/* cdfcat.c -- This program concatenates selected variable slices
 * together to make a single variable slice.  Selection is made
 * on the basis of a scalar field that increases monotonically
 * through the input file.  All variable fields have their dimension
 * increased by one, with the name of the selection field becoming
 * the name of the dimension.  This dimension becomes the first
 * dimension.  The selection field thus becomes an index field
 * for that dimension.
 */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "cdfhdr.h"

static char hbufin[HBMAX][LINE],hbufout[HBMAX][LINE],testvar[LINE];
static char com[HBMAX][LINE],cline[LINE],pname[LINE],pval[LINE];
static char type;
static int nvf,loop,goop,nclines,testpos;
static long nstatic,nvaria,testptr,max,count,poop;
static float start,finish,tv1,tv2,dtv;
static float *sbuff,*vbuff,**abuff;
static struct field *fp;
static long offset[HBMAX],length[HBMAX];

main(argc,argv)
int argc;
char *argv[];
{
/* check arguments */
  if (argc != 5) {
    fprintf (stderr,"Usage: cdfcat testvar start end maxslices\n");
    exit(1);
  }
  strcpy(testvar,argv[1]);
  start = atof(argv[2]);
  finish = atof(argv[3]);
  max = atoi(argv[4]);
/* read input header from standard input */
  gethdr(stdin,hbufin,HBMAX);
/* check format and die if not float */
  if (getfmt(hbufin,HBMAX) != 'f') {
    fprintf(stderr,"cdfcat: format not float\n");
    exit(1);
  }
/* see if testvar is here */
  if ((fp = seekfld(hbufin,HBMAX,'v',testvar)) == NULL) {
    fprintf(stderr,"cdfcat: test field %s doesn't exist\n",testvar);
    exit(1);
  }
  if (fp->dim != 0) {
    fprintf(stderr,"cdfcat: test field %s isn't scalar\n",testvar);
    exit(1);
  }
  testptr = fp->ptr;
/* get static slice */
  nstatic = elemcnt(hbufin,HBMAX,'s');
  if ((sbuff = (float *) malloc(sizeof(float)*nstatic)) ==
      (float *)NULL) {
    fprintf(stderr,"cdfcat: can't allocate enough memory\n");
    exit(1);
  }
  if (getslice(stdin,nstatic,sbuff) == EOF) {
    fprintf(stderr,"cdfcat: unexpected EOF reading static slice\n");
    exit(1);
  }
/* evaluate size of buffers needed */
  nvf = 0;
  nvaria = 0;
  while ((fp = getfld(hbufin,HBMAX,'v',nvf)) != NULL) {
    if (fp->dim > 3) {
      fprintf(stderr,"cdfcat: %s would have dimension > 4\n",fp->fname);
      exit(1);
    }
/* incidentally, find position of test variable */
    if (strcmp(fp->fname,testvar) == 0) testpos = nvf;
    nvaria += fp->tsize;
    nvf++;
  }
  if (nvaria*max > MAXSLICE) {
    fprintf(stderr,"cdfcat: proposed variable slice is too big\n");
    fprintf(stderr,"        slice: %d, maxslice: %d\n",nvaria*max,MAXSLICE);
    exit(1);
  }
/* allocate input buffer */
  if ((vbuff = (float *) malloc(sizeof(float)*nvaria)) ==
      (float *)NULL) {
    fprintf(stderr,"cdfcat: can't get enough memory\n");
    exit(1);
  }
/* allocate pointers to output buffers for individual fields */
  if ((abuff = (float **) malloc(sizeof(float *)*nvf)) ==
      (float **)NULL) {
    fprintf(stderr,"cdfcat: can't get enough memory\n");
    exit(1);
  }
/* allocate buffers for construction of concatenated fields */
  for (loop = 0; loop < nvf; loop++) {
    fp = getfld(hbufin,HBMAX,'v',loop);
    if ((abuff[loop] = (float *)malloc(sizeof(float)*fp->tsize*max)) ==
	(float *)NULL) {
      fprintf(stderr,"cdfcat: can't get enough memory\n");
      exit(1);
    }
/* fill length and offset tables as well */
    length[loop] = fp->tsize;
    offset[loop] = fp->ptr;
  }
/* read slices */
  count = 0;
  while (getslice(stdin,nvaria,vbuff) != EOF) {
/* test for test variable in range */
    if (vbuff[testptr] < start) continue;
    if (vbuff[testptr] > finish) break;
    if (count == 0) tv1 = vbuff[testptr];
    tv2 = vbuff[testptr];
    if (count >= max) {
      fprintf(stderr,"cdfcat: more slices than anticipated\n");
      exit(1);
    }
/* loop on fields */
    for (loop = 0; loop < nvf; loop++) {
/* do transfer */
      for (poop = 0; poop < length[loop]; poop++) {
        *(abuff[loop] + (length[loop])*count + poop) =
             vbuff[offset[loop] + poop];
      }
    }
/* increment slice count */
    count++;
  }
/* check for trivial case */
  if (count <= 0) {
    fprintf(stderr,"cdfcat: only %ld variable slices\n",count);
    exit(1);
  }
/* create new header */
  nullhdr(hbufout,HBMAX,"float");
/* transfer comments */
  nclines = getcmt(hbufin,HBMAX,com);
  for (loop = 0; loop < nclines; loop++) {
    addcline(hbufout,HBMAX,com[loop]);
  }
/* add one of our own */
  sprintf(cline,"cdfcat: %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  addcline(hbufout,HBMAX,cline);
/* transfer parameters */
  loop = 0;
  while (getpar(hbufin,HBMAX,loop++,pname,pval) == OK) {
    addpar(hbufout,HBMAX,pname,pval);
  }
/* add parameters to help cdfplot */
  if (count <= 1) dtv = 0.;
  else dtv = (tv2 - tv1)/(count - 1);
  sprintf(pval,"%g",dtv);
  sprintf(pname,"d%s",testvar);
  addpar(hbufout,HBMAX,pname,pval);
  sprintf(pval,"%g",tv1);
  sprintf(pname,"%s0",testvar);
  addpar(hbufout,HBMAX,pname,pval);
/* variable fields */
  loop = 0;
  while ((fp = getfld(hbufin,HBMAX,'v',loop++)) != NULL) {
    fp->dsize4 = fp->dsize3;
    fp->dsize3 = fp->dsize2;
    fp->dsize2 = fp->dsize1;
    fp->dsize1 = count;
    strcpy(fp->dname4,fp->dname3);
    strcpy(fp->dname3,fp->dname2);
    strcpy(fp->dname2,fp->dname1);
    strcpy(fp->dname1,testvar);
    fp->dim += 1;
/* special case -- test field goes in static slice */
    if (strcmp(fp->fname,testvar) == 0) type = 's';
    else type = 'v';
    addfld(hbufout,HBMAX,type,fp->fname,fp->smul,fp->sadd,fp->bytes,
      fp->dim,fp->dname1,fp->dsize1,fp->dname2,fp->dsize2,
      fp->dname3,fp->dsize3,fp->dname4,fp->dsize4);
  }
/* static fields */
  loop = 0;
  while ((fp = getfld(hbufin,HBMAX,'s',loop++)) != NULL) {
    addfld(hbufout,HBMAX,'s',fp->fname,fp->smul,fp->sadd,fp->bytes,
      fp->dim,fp->dname1,fp->dsize1,fp->dname2,fp->dsize2,
      fp->dname3,fp->dsize3);
  }
/* write new header */
  puthdr(stdout,hbufout,HBMAX);
/* write static slice by hand */
  printf("@%15ld",nstatic + count);
  if (FLOATSWAP) floatswap(count,abuff[testpos]);
  fwrite(abuff[testpos],sizeof(float),count,stdout);
  if (FLOATSWAP) floatswap(nstatic,sbuff);
  fwrite(sbuff,sizeof(float),nstatic,stdout);
/* variable slice must also be written by hand */
  nvaria = elemcnt(hbufout,HBMAX,'v');
  printf("@%15ld",nvaria);
  goop = loop = 0;
  while ((fp = getfld(hbufout,HBMAX,'v',loop)) != NULL) {
/* skip over test variable, which has been moved to static slice */
    if (loop == testpos) goop++;
    if (FLOATSWAP) floatswap(fp->tsize,abuff[goop]);
    fwrite(abuff[goop],sizeof(float),fp->tsize,stdout);
    loop++;
    goop++;
  }
  exit(0);
}
