/* cdfcenter -- Interpolate grid edge values of a field to the grid
 * center.  This is special for unstaggering data in numerical models
 * with staggered grids.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

static char hbuff[HBMAX][LINE];            /* input header buffer */
static float *sbuff,*vbuff;                /* slice buffers */
static long nsbuff,nvbuffin,nvbuffout;     /* slice buffer sizes */
static struct field *fp;                   /* field pointer temp */
static long pos,loop,instance;             /* looping indices */
static char comment[LINE];                 /* new comment line */
static char *oldindex,*newindex;           /* names of index fields */
static char *suffix;                       /* suffix added to old name */
static int nchange;                        /* number of fields to be changed */
static char **oldnames;                    /* names of fields to be changed */
static char **newnames;                    /* names of changed fields */
static struct field **oldfp;               /* original field struct pointers */
static struct field **newfp;               /* changed field struct pointers */
static float **olddata,**newdata;          /* data pointers to fields */
static float *tmpdata;                     /* temporary buffer for data */
static long start,incr,size;               /* stuff for subspace1 */
static long oldisize,newisize;             /* sizes of index fields */

main(int argc, char *argv[])
{

  /* check args */
  if (argc < 5) {
    fprintf(stderr,"Usage: cdfcenter oldindex newindex suffix field1 ...\n");
    exit(1);
  }
  oldindex = argv[1];
  newindex = argv[2];
  suffix = argv[3];
  nchange = argc - 4;

  /* get some memory */
  if ((oldnames = (char**)malloc(nchange*sizeof(char*))) == (char**)NULL) {
    fprintf(stderr,"cdfcenter: can't get memory!\n");
    exit(1);
  }
  if ((newnames = (char**)malloc(nchange*sizeof(char*))) == (char**)NULL) {
    fprintf(stderr,"cdfcenter: can't get memory!\n");
    exit(1);
  }
  if ((oldfp = (struct field**)malloc(nchange*sizeof(struct field*))) ==
      (struct field**)NULL) {
    fprintf(stderr,"cdfcenter: can't get memory!\n");
    exit(1);
  }
  if ((newfp = (struct field**)malloc(nchange*sizeof(struct field*))) ==
      (struct field**)NULL) {
    fprintf(stderr,"cdfcenter: can't get memory!\n");
    exit(1);
  }
  if ((olddata = (float**)malloc(nchange*sizeof(float*))) == (float**)NULL) {
    fprintf(stderr,"cdfcenter: can't get memory!\n");
    exit(1);
  }
  if ((newdata = (float**)malloc(nchange*sizeof(float*))) == (float**)NULL) {
    fprintf(stderr,"cdfcenter: can't get memory!\n");
    exit(1);
  }

  /* fill in field names */
  for (pos = 0; pos < nchange; pos++) {
    oldnames[pos] = argv[pos + 4];
    if ((newnames[pos] = (char*)malloc(LINE*sizeof(char))) == (char*)NULL) {
      fprintf(stderr,"cdfcenter: can't get memory!\n");
      exit(1);
    }
    sprintf(newnames[pos],"%s%s",oldnames[pos],suffix);
  }

  /* get a header from standard input */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfcenter: float format expected\n");
    exit(1);
  }

  /* add a comment */
  sprintf(comment,"cdfcenter: %s %s %s\n",oldindex,newindex,suffix);
  addcline(hbuff,HBMAX,comment);
  for (pos = 0; pos < nchange; pos++) {
    sprintf(comment,"  %s\n",oldnames[pos]);
    addcline(hbuff,HBMAX,comment);
  }

  /* check to see if index fields exist and are the right size */
  fp = seekfld(hbuff,HBMAX,'s',oldindex);
  oldisize = fp->dsize1;
  if ((fp->dim != 1) || (strcmp(oldindex,fp->dname1) != 0)) {
    fprintf(stderr,"cdfcenter: field %s is not an index field\n",oldindex);
    exit(1);
  }
  fp = seekfld(hbuff,HBMAX,'s',newindex);
  newisize = fp->dsize1;
  if ((fp->dim != 1) || (strcmp(newindex,fp->dname1) != 0)) {
    fprintf(stderr,"cdfcenter: field %s is not an index field\n",newindex);
    exit(1);
  }
  if (newisize + 1 != oldisize) {
    fprintf(stderr,"cdfcenter: size of newindex != oldindex - 1\n");
    exit(1);
  }

  /* compute input buffer sizes */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbuffin = elemcnt(hbuff,HBMAX,'v');

  /* add new centered fields -- assume corner fields in variable slice */
  for (pos = 0; pos < nchange; pos++) {

    /* first, find the field of interest */
    fp = seekfld(hbuff,HBMAX,'v',oldnames[pos]);

    /* see if it has the old index field and create new one */
    if (fp->dim >= 1 && strcmp(oldindex,fp->dname1) == 0) {
      addfld(hbuff,HBMAX,'v',newnames[pos],fp->smul,fp->sadd,fp->bytes,fp->dim,
	     newindex,newisize,
	     fp->dname2,fp->dsize2,
	     fp->dname3,fp->dsize3,
	     fp->dname4,fp->dsize4);
    }
    else if (fp->dim >= 2 && strcmp(oldindex,fp->dname2) == 0) {
      addfld(hbuff,HBMAX,'v',newnames[pos],fp->smul,fp->sadd,fp->bytes,fp->dim,
	     fp->dname1,fp->dsize1,
	     newindex,newisize,
	     fp->dname3,fp->dsize3,
	     fp->dname4,fp->dsize4);
    }
    else if (fp->dim >= 3 && strcmp(oldindex,fp->dname3) == 0) {
      addfld(hbuff,HBMAX,'v',newnames[pos],fp->smul,fp->sadd,fp->bytes,fp->dim,
	     fp->dname1,fp->dsize1,
	     fp->dname2,fp->dsize2,
	     newindex,newisize,
	     fp->dname4,fp->dsize4);
    }
    else if (fp->dim >= 4 && strcmp(oldindex,fp->dname4) == 0) {
      addfld(hbuff,HBMAX,'v',newnames[pos],fp->smul,fp->sadd,fp->bytes,fp->dim,
	     fp->dname1,fp->dsize1,
	     fp->dname2,fp->dsize2,
	     fp->dname3,fp->dsize3,
	     newindex,newisize);
    }
    else {
      fprintf(stderr,"cdfcenter: field %s doesn't have index field %s\n",
	      oldnames[pos],oldindex);
      exit(1);
    }
  }

  /* compute the output buffer size and make the buffers */
  nvbuffout = elemcnt(hbuff,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuffout);

  /* get field and data pointers */
  for (pos = 0; pos < nchange; pos++) {
    olddata[pos] = getptr2(hbuff,HBMAX,vbuff,'v','d',
			   oldnames[pos],&oldfp[pos]);
    newdata[pos] = getptr2(hbuff,HBMAX,vbuff,'v','d',
			   newnames[pos],&newfp[pos]);
  }

  /* write the header */
  puthdr(stdout,hbuff,HBMAX);

  /* read in and write the static slice */
  if (getslice(stdin,nsbuff,sbuff) == EOF) {
    fprintf(stderr,"cdfcenter: unexpected eof reading static slice\n");
    exit(1);
  }
  putslice(stdout,nsbuff,sbuff);

  /* allocate temporary data space */
  if ((tmpdata = (float*)malloc(oldisize*sizeof(float))) == (float*)NULL) {
    fprintf(stderr,"cdfcenter: can't get memory!\n");
    exit(1);
  }

  /* loop on variable slices */
  while (getslice(stdin,nvbuffin,vbuff) != EOF) {

    /* loop on variables of interest */
    for (pos = 0; pos < nchange; pos++) {
      
      /* loop on 1-D rays through each variable */
      instance = 0;
      while (subspace1(oldfp[pos],oldindex,instance,
		       &start,&incr,&size) != FAIL) {

	/* transfer ray to temporary space */
	for (loop = 0; loop < size; loop++) {
	  tmpdata[loop] = olddata[pos][start + incr*loop];
	}

	/* do the centering calculation and transfer data to new field */
	if (subspace1(newfp[pos],newindex,instance,
		      &start,&incr,&size) == FAIL) {
	  fprintf(stderr,"cdfcenter: unexpected crash -- call NCSB!\n");
	  exit(1);
	}
	for (loop = 0; loop < size; loop++) {
	  newdata[pos][start + incr*loop] = 0.5*(tmpdata[loop] +
						 tmpdata[loop + 1]);
	}

	/* go to next ray */
	instance++;
      }
    }

    /* output variable slice */
    putslice(stdout,nvbuffout,vbuff);
  }

  /* done! */
  exit(0);
}
