/* cdfentform -- compute fields needed for an entropy-p chart */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
/* constants (Iribarne and Godson) */
#define CTOK 273.15               /* Celsius to Kelvin conversion */
#define RD 287.05                 /* dry air gas constant */
#define EPS 0.622                 /* vapor to dry air mol weight ratio */
#define RV (RD/EPS)               /* water vapor gas constant */
#define CPD 1005.                 /* specific heat const pres dry air */
#define CPV 1850.                 /* specific heat const pres water vapor */
#define CL 4218.                  /* specific heat liquid water */
#define CIMEAN 1959.              /* specific heat of ice at -20 C */
#define LR 2.5008e6               /* heat of vaporization at 0 C */
#define LFR 3.337e5               /* heat of fusion at 0 C */
#define DLFDT 2.45e3              /* slope of heat of fusion curve */
#define TR 273.15                 /* reference temperature (K) */
#define PR 100000.                /* reference pressure (Pa) */
#define TMIN 190.                 /* minimum reasonable atmospheric temp */
#define GEE 9.807                 /* acceleration of gravity */
#define DELTA 0.606               /* virtual temperature factor */

/* needed constants */
#define P0 1000.
#define KAPPA 0.286

float sentcalc();
float escalc();
float *floatbuff();
float thcalc();

main(argc,argv)
int argc;
char *argv[];
{
  float p1,p2,ent1,ent2,delp,delent;
  float *p,*ent,*theta;
  long ip,ient,np,nent;
  int handle,dhent,dhp;
  float *floatbuff();

/* do command line */
  if (argc != 7) {
    fprintf(stderr,"Usage: cdfentform p1 p2 np ent1 ent2 nent\n");
    exit(1);
  }
  p1 = atof(argv[1]);
  p2 = atof(argv[2]);
  np = atoi(argv[3]);
  if (np <= 1) {
    fprintf(stderr,"cdfentform: np = %d must be > 1\n",np);
    exit(1);
  }
  ent1 = atof(argv[4]);
  ent2 = atof(argv[5]);
  nent = atoi(argv[6]);
  if (nent <= 1) {
    fprintf(stderr,"cdfentform: nent = %d must be > 1\n",nent);
    exit(1);
  }

/* allocate space */
  p = floatbuff(np);
  ent = floatbuff(nent);
  theta = floatbuff(np*nent);

/* fill index fields */
  delp = (p2 - p1)/(np - 1);
  for (ip = 0; ip < np; ip++) {
    p[ip] = p1 + ip*delp;
  }
  delent = (ent2 - ent1)/(nent - 1);
  for (ient = 0; ient < nent; ient++) {
    ent[ient] = ent1 + ient*delent;
  }

/* loop on grid */
  for (ip = 0; ip < np; ip++) {
    for (ient = 0; ient < nent; ient++) {
      theta[ip + np*ient] = thcalc(p[ip],ent[ient]);
    }
  }

/* put it out in a candis file on the standard output */
  handle = candiscreate("-",0);
  candiscomment(handle,"cdfentform:");
  dhent = candisidim(handle,"entropy",nent,ent);
  dhp = candisidim(handle,"pressure",np,p);
  candisfld(handle,"theta",2,dhent,dhp,0,0,theta);
  candiswrite(handle);

/* done! */
  exit(0);
}

/* calculate theta, given the pressure (mb) and entropy (J/kg/K) */
float thcalc(p,ent)
float p,ent;
{
  float th,dentdth,t2,t1,es,ent2,ent1;
  int i;
  float entcalc();

/* initialize iteration loop */
  th = 300.;

/* loop till we poop */
  for (i = 0; i < 8; i++) {
    t2 = th*pow(p/P0,KAPPA);
    t1 = (th - 1.)*pow(p/P0,KAPPA);
    ent2 = sentcalc(p*100.,t2);
    ent1 = sentcalc(p*100.,t1);
    dentdth = ent2 - ent1;
    th += (ent - ent2)/dentdth;
  }

/* return result */
  return(th);
}

/* thermo -- This is a thermodynamics package that tries to be as
 * accurate as warranted by real atmospheric data for all realistic
 * conditions.
 */

#define A 17.269                  /* constants for Tetens' formula */
#define B 35.86
#define C 610.78
#define D 273.16
#define LOOPOUT 200               /* max number of iteration loops */
#define QVMIN 1.e-8               /* smallest allowed mixing ratio */

/* compute saturated entropy */
float sentcalc(p,t)
float p;                         /* total pressure (Pa) */
float t;                         /* temperature (K) */
{
  float qs,qv,qt,pvs,pd,s;
  pvs = escalc(t);     /* saturation vapor pressure */
  pd = p - pvs;    /* partial pressure of dry air at saturation */
  qs = qv = qt = EPS*pvs/pd;     /* saturation mixing ratio */
  pd = p*(1. - qv/(EPS + qv));   /* partial pressure of dry air */
/* this formula is consistent both with Iribarne and Godson and with
 * Kerry Emanuel's reversible theta-e -- I also derived it myself.
 */
  s = (CPD + CL*qt)*log(t/TR) - RD*log(pd/PR) +
       (LR + (CPV - CL)*(t - TR))*qv/t;
  if (qv > 0.) s -= qv*RV*log(qv/qs);
  return(s);
}

/* Tetens' formula for water saturation vapor pressure */
/*
float escalc(t)
float t;
{
  return(C*exp(A*(t - D)/(t - B)));
}
*/

/* Formula for water saturation vapor pressure consistent with
 * Clausius-Clapyron equation.
 */
float escalc(t)
float t;
{
  float l,lr;
  lr = LR;
/* uncomment next line to use ice adiabat below 0 C */
/*
  if (t < 273.15) lr += LFR;
*/
  l = lr - (CL - CPV)*(t - TR);
  return(C*pow(TR/t,(CL - CPV)/RV)*exp((lr/TR - l/t)/RV));
}

/* compute mixing ratio, given pressure and dewpoint */
float mratio(p,td)
float p;                          /* pressure (Pa) */
float td;                         /* dewpoint temperature (K) */
{
  float temp;
  temp = escalc(td);
  return(EPS*temp/(p - temp));
}

/* floatbuff -- This routine uses malloc to allocate float buffer
 * space.  The argument nelem defines the number of floating
 * point elements.  It returns a float pointer to the space.  If malloc
 * fails, getbuff dies with an error message.
 */
float *floatbuff(nelem)
long nelem;
{
  float *point;
  if ((point = (float*)malloc((unsigned)(sizeof(float)*nelem)))
       == (float*)NULL) {
    fprintf(stderr,"floatbuff: can't get %d words\n",nelem);
    exit(1);
  }
  return(point);
}
