/* cdfentropy -- Compute Raymond and Blyth's entropy */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

/* constants (Iribarne and Godson) */
#define CTOK 273.15               /* Celsius to Kelvin conversion */
#define RD 287.05                 /* dry air gas constant */
#define EPS 0.622                 /* vapor to dry air mol weight ratio */
#define RV (RD/EPS)               /* water vapor gas constant */
#define CPD 1005.                 /* specific heat const pres dry air */
#define CPV 1850.                 /* specific heat const pres water vapor */
#define CL 4218.                  /* specific heat liquid water */
#define CIMEAN 1959.              /* specific heat of ice at -20 C */
#define LR 2.5008e6               /* heat of vaporization at 0 C */
#define LFR 3.337e5               /* heat of fusion at 0 C */
#define DLFDT 2.45e3              /* slope of heat of fusion curve */
#define TR 273.15                 /* reference temperature (K) */
#define PR 100000.                /* reference pressure (Pa) */
#define TMIN 190.                 /* minimum reasonable atmospheric temp */
#define GEE 9.807                 /* acceleration of gravity */
#define DELTA 0.606               /* virtual temperature factor */

/* thermodynamic function declarations */
float escalc();                   /* saturation vapor pressure */
float ent();                      /* entropy */
float mratio();                   /* mixing ratio */

char hbuff[HBMAX][LINE];
long nsbuff,nvbuffold,nvbuff;
float *sbuff,*vbuff;
float *pres;                         /* pressure (mb) */
float *temp;                         /* temperature (C) */
float *moisture;                     /* mixing ratio (g/kg) or dewpt (C) */
float *entropy;                      /* specific (J/K/kg) */
float *satmr;                        /* saturation mixing ratio (g/kg) */
float *rhol;                         /* liquid water density (g/m^3) */
char cline[LINE];
float r,p,t,es,l;
float bad,badlim;
long size;
long loop;
int dpflag,lflag;
struct field *fp;

main(argc,argv)
int argc;
char *argv[];
{

/* command line arguments */
  if ((argc != 7 && argc != 8) || (argc > 1 && argv[1][0] != '-')) {
    fprintf(stderr,
         "Usage: cdfentropy -d|q pres temp dwpt|mixrat entropy satmr [rhol]\n");
    exit(1);
  }

/* set lflag if liquid water to be included */
  lflag = (argc == 8);

/* check to see whether to expect dewpoint or mixing ratio */
  if (strcmp(argv[1],"-d") == 0) dpflag = 1;
  else if (strcmp(argv[1],"-q") == 0) dpflag = 0;
  else {
    fprintf(stderr,"cdfentropy: unknown option %s\n",argv[1]);
    exit(1);
  }

/* get header */
  gethdr(stdin,hbuff,HBMAX);

/* check format */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfentropy: float format expected\n");
    exit(1);
  }

/* get bad data values */
  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;

/* add comment */
  sprintf(cline,"cdfentropy: %s %s %s\n",argv[1],argv[2],argv[3]);
  addcline(hbuff,HBMAX,cline);
  if (lflag) sprintf(cline,"  %s %s %s %s\n",argv[4],argv[5],argv[6],argv[7]);
  else sprintf(cline,"  %s %s %s\n",argv[4],argv[5],argv[6]);
  addcline(hbuff,HBMAX,cline);

/* get size of incoming variable slice */
  nvbuffold = elemcnt(hbuff,HBMAX,'v');

/* add new fields if needed -- assume same size and shape as pressure field */
  if((fp = seekfld(hbuff,HBMAX,'v',argv[2])) == NULL) {
    fprintf(stderr,"cdfentropy: field %s doesn't exist\n",argv[2]);
    exit(1);
  }
  size = fp->tsize;

/* entropy */
  if (seekfld(hbuff,HBMAX,'v',argv[5]) == NULL) {
    fp = seekfld(hbuff,HBMAX,'v',argv[2]);
    addfld(hbuff,HBMAX,'v',argv[5],10.,0.,'s',fp->dim,
         fp->dname1,fp->dsize1,
         fp->dname2,fp->dsize2,
         fp->dname3,fp->dsize3,
         fp->dname4,fp->dsize4);
  }

/* saturation mixing ratio */
  if (seekfld(hbuff,HBMAX,'v',argv[6]) == NULL) {
    fp = seekfld(hbuff,HBMAX,'v',argv[2]);
    addfld(hbuff,HBMAX,'v',argv[6],10.,0.,'s',fp->dim,
         fp->dname1,fp->dsize1,
         fp->dname2,fp->dsize2,
         fp->dname3,fp->dsize3,
         fp->dname4,fp->dsize4);
  }

/* create buffers */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuff);

/* get pointers */
  pres = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[2],&fp);
  temp = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[3],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfentropy: different sizes for %s and %s\n",
         argv[2],argv[3]);
    exit(1);
  }
  moisture = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[4],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfentropy: different sizes for %s and %s\n",
         argv[2],argv[4]);
    exit(1);
  }
  entropy = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[5],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfentropy: different sizes for %s and %s\n",
         argv[2],argv[5]);
    exit(1);
  }
  satmr = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[6],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfentropy: different sizes for %s and %s\n",
         argv[2],argv[6]);
    exit(1);
  }
  if (lflag) {
    rhol = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[7],&fp);
    if (fp->tsize != size) {
      fprintf(stderr,"cdfentropy: different sizes for %s and %s\n",
           argv[2],argv[7]);
      exit(1);
    }
  }

/* write header buffer */
  puthdr(stdout,hbuff,HBMAX);

/* transfer static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

/* loop on variable slices */
  while (getslice(stdin,nvbuffold,vbuff) != EOF) {

/* loop on elements */
    for (loop = 0; loop < size; loop++) {

/* get and convert data -- skip if data bad */
      t = temp[loop];
      p = pres[loop];
      r = moisture[loop];
      if (lflag) l = rhol[loop];
      else l = 0.;
      if (fabs(t) > badlim || fabs(p) > badlim || fabs(r) > badlim ||
           fabs(l) > badlim) {
        entropy[loop] = bad;
        satmr[loop] = bad;
      }
      else {
        t += CTOK;
	p *= 100.;

/* r is initially dewpoint temperature, convert to mixing ratio */
        if (dpflag) {
          r += CTOK;

/* if dewpoint exceeds temperature, make temperature = dewpoint */
          if (r > t) t = r;
	  r = mratio(p,r);
        }
        else {

/* r is initially mixing ratio in g/kg -- convert to g/g */
          r /= 1000.;
        }

/* get the liquid water mixing ratio */
        if (lflag) l *= (1.e-5*RD*t/p);

/* compute entropy */
        entropy[loop] = ent(p,t,r + l);

/* compute saturation mixing ratio in g/kg */
	satmr[loop] = 1000.*mratio(p,t);
      }
    }

/* write variable slice */
    putslice(stdout,nvbuff,vbuff);
  }
}

/* thermo -- This is a thermodynamics package that tries to be as
 * accurate as warranted by real atmospheric data for all realistic
 * conditions.
 */

#define A 17.269                  /* constants for Tetens' formula */
#define B 35.86
#define C 610.78
#define D 273.16
#define LOOPOUT 200               /* max number of iteration loops */
#define QVMIN 1.e-8               /* smallest allowed mixing ratio */

/* compute entropy */
float ent(p,t,qt)
float p;                         /* total pressure (Pa) */
float t;                         /* temperature (K) */
float qt;                        /* total water mixing ratio (g/g) */
{
  float qs,qv,pvs,pd,s;
  pvs = escalc(t);     /* saturation vapor pressure */
  pd = p - pvs;    /* partial pressure of dry air at saturation */
  qs = qv = EPS*pvs/pd;     /* saturation mixing ratio */
  if (qt < qv) qv = qt;     /* check for saturation */
  pd = p*(1. - qv/(EPS + qv));   /* partial pressure of dry air */
/* this formula is consistent both with Iribarne and Godson and with
 * Kerry Emanuel's reversible theta-e -- I also derived it myself.
 */
  s = (CPD + CL*qt)*log(t/TR) - RD*log(pd/PR) +
       (LR + (CPV - CL)*(t - TR))*qv/t;
  if (qv > 0.) s -= qv*RV*log(qv/qs);
  return(s);
}

/* Tetens' formula for water saturation vapor pressure */
/*
float escalc(t)
float t;
{
  return(C*exp(A*(t - D)/(t - B)));
}
*/

/* Formula for water saturation vapor pressure consistent with
 * Clausius-Clapyron equation.
 */
float escalc(t)
float t;
{
  float l,lr;
  lr = LR;
/* uncomment next line to use ice adiabat below 0 C */
/*
  if (t < 273.15) lr += LFR;
*/
  l = lr - (CL - CPV)*(t - TR);
  return(C*pow(TR/t,(CL - CPV)/RV)*exp((lr/TR - l/t)/RV));
}

/* compute mixing ratio, given pressure and dewpoint */
float mratio(p,td)
float p;                          /* pressure (Pa) */
float td;                         /* dewpoint temperature (K) */
{
  float temp;
  temp = escalc(td);
  return(EPS*temp/(p - temp));
}
