/* cdffilter -- Do low and high pass filtering on fields */
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "cdfhdr.h"

extern int optind,optopt,opterr;
extern char *optarg;

static char *hbuff[HBMAX][LINE];       /* header buffer */
static char cline[LINE];               /* character line buffer */
static long nsbuff,nvbin,nvbout;       /* slice buffer sizes */
static float *sbuff,*vbuff;            /* slice buffer pointers */
static float *ifield;                  /* pointer to specified index field */
static float *sfield;                  /* pointer to specified field */
static float *dfield;                  /* pointer to filtered field */
static float *tfield;                  /* pointer to temporary field */
static long instance;                  /* looping var over instances */
static long loop,xloop,ploop;          /* the other looping variables */
static struct field *fp;               /* field structure temporary pointer */
static struct field *fps;              /* field structure pointer for
					   specified field */
static float bad,badlim;               /* bad data values */
static long start,incr,size,csize;     /* loop variables */
static int highpass;                   /* 1 = high pass, 0 = low pass */
static int cyclic;                     /* 0 = non-cyclic, 1 = cyclic */
static int rcyclic;                    /* 0 = non-rcyclic, 1 = rcyclic */
static long index,pindex;              /* field looping variables */
static double sum;                     /* integration temporary */
static float lambda;                   /* filter smoothing length */
static float dx;                       /* step in index field */
static float alpha;                    /* smoothing parameter */


main(argc,argv)
int argc;
char *argv[];
{
  char xxx;

  /* check arguments */
  if (argc < 6) {
    fprintf(stderr,
         "Usage: cdffilter -[lh] [-c|r] smooth_len filtered_fld field dim\n");
    exit(1);
  }

  /* get options */
  opterr = 0;
  highpass = -1;
  cyclic = rcyclic = 0;
  while ((xxx = getopt(argc,argv,"+lhcr")) != -1) {
    if (xxx == '?') {
      fprintf(stderr,"cdffilter: unknown option %c\n",optopt);
      exit(1);
    }
    if (xxx == 'h') highpass = 1;
    if (xxx == 'l') highpass = 0;
    if (xxx == 'c') cyclic = 1;
    if (xxx == 'r') rcyclic = 1;
  }
  if (highpass == -1) {
    fprintf(stderr,"cdffilter: must have either -h or -l option\n");
    exit(1);
  }
  if (cyclic == 1 && rcyclic == 1) {
    fprintf(stderr,"cdffilter: either -r or -c, not both\n");
    exit(1);
  }
  if (argc - optind != 4) {
    fprintf(stderr,
         "Usage: cdffilter -[lh] [-c|r] [--] smooth_len filtered_fld field dim\n");
    exit(1);
  }

  /* get header */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdffilter: float format expected\n");
    exit(1);
  }

  /* get lambda */
  lambda = atof(argv[optind]);


  if (lambda == 0.) {
    fprintf(stderr,"cdffilter: inappropriate smoothing length: %g\n",lambda);
    exit(1);
  }

  /* get bad data values */
  bad = (seekpar(hbuff,HBMAX,"bad",cline) == OK) ? atof(cline) : BAD;
  badlim = (seekpar(hbuff,HBMAX,"badlim",cline) == OK) ? atof(cline) : BADLIM;

  /* compute input buffer lengths */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');

  /* add stuff to header, comment first */
  if (highpass) {
    if (cyclic) sprintf(cline,"cdffilter: -hc\n");
    else if (rcyclic) sprintf(cline,"cdffilter: -hr\n");
    else sprintf(cline,"cdffilter: -h\n");
  }
  else {
    if (cyclic) sprintf(cline,"cdffilter: -lc\n");
    else if (rcyclic) sprintf(cline,"cdffilter: -lr\n");
    else sprintf(cline,"cdffilter: -l\n");
  }
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s %s %s\n",argv[optind],argv[optind + 1]
	  ,argv[optind + 2],argv[optind + 3]);
  addcline(hbuff,HBMAX,cline);

  /* unfiltered field next -- same dimensions for filtered field */
  if ((fp = seekfld(hbuff,HBMAX,'v',argv[optind + 2])) == NULL) {
    fprintf(stderr,"cdffilter: %s field not found in variable slice\n",
	    argv[optind + 2]);
    exit(1);
  }
  addfld(hbuff,HBMAX,'v',argv[optind + 1],fp->smul,fp->sadd,fp->bytes,fp->dim,
       fp->dname1,fp->dsize1,fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,fp->dname4,fp->dsize4);

  /* allocate buffers */
  nvbout = elemcnt(hbuff,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbout);

  /* assign field and structure pointers */
  sfield = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[optind + 2],&fps);
  dfield = getptr(hbuff,HBMAX,vbuff,'v','d',argv[optind + 1]);
  if ((ifield = getptr2(hbuff,HBMAX,sbuff,'s','r',argv[optind + 3],&fp))
      == NULL) {
    fprintf(stderr,"cdffilter: %s field doesn't exist\n",argv[optind + 3]);
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->fname,fp->dname1) != 0)) {
    fprintf(stderr,"cdffilter: %s field is not an index field\n",
	    argv[optind + 3]);
    exit(1);
  }

  /* write header */
  puthdr(stdout,hbuff,HBMAX);

  /* transfer static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

  /* loop on variable slices */
  while (getslice(stdin,nvbin,vbuff) != EOF) {

    /* loop on subspace instances */
    instance = 0;
    while (subspace1(fps,argv[optind + 3],instance,&start,&incr,&size)
	   != FAIL) {
      tfield = getbuff(size);

      /* do a non-cyclic low pass filtering job -- forward filtering first */
      if (cyclic == 0 && rcyclic == 0) {
	for (loop = 0; loop < size; loop++) {
	  index = start + incr*loop;
	  if (loop == 0 || fabs(sfield[index]) > badlim ||
	      fabs(sfield[index - incr]) > badlim) dfield[index]
						     = sfield[index];
	  else {
	    dx = ifield[loop] - ifield[loop - 1];
	    alpha = dx/lambda;
	    if (alpha < 0.) {
	      alpha = -alpha;
	    }
	    else if (alpha >= 1.) {
	      fprintf(stderr,"cdffilter: smoothing length < grid size\n");
	      exit(1);
	    }
	    dfield[index] = alpha*sfield[index] +
	      (1. - alpha)*dfield[index - incr];
	  }
	  tfield[loop] = dfield[index];
	}
	
	/* backward filtering for symmetry */
	for (loop = size - 1; loop >= 0; loop--) {
	  index = start + incr*loop;
	  if (loop == size - 1 || fabs(sfield[index]) > badlim ||
	      fabs(sfield[index + incr]) > badlim)
	    dfield[index] = sfield[index];
	  else {
	    dx = ifield[loop + 1] - ifield[loop];
	    alpha = dx/lambda;
	    if (alpha < 0.) {
	      alpha = -alpha;
	    }
	    else if (alpha >= 1.) {
	      fprintf(stderr,"cdffilter: smoothing length < grid size\n");
	      exit(1);
	    }
	    dfield[index] = alpha*tfield[loop] +
	      (1. - alpha)*dfield[index + incr];
	  }
	}
      }

      /* do a cyclic low pass filtering job -- forward filtering first */
      else {

	/* if rcyclic, ignore last element for now */
	csize = rcyclic ? size - 1 : size;
	if (csize < 1) {
	  fprintf(stderr,"cdffilter: no repeated cyclic for dim size < 2\n");
	  exit(1);
	}

	/* for cyclic we must assume that dx is constant */
	dx = (ifield[size - 1] - ifield[0])/size;
	alpha = dx/lambda;
	if (alpha < 0.) {
	  alpha = -alpha;
	}
	else if (alpha >= 1.) {
	  fprintf(stderr,"cdffilter: smoothing length < grid size\n");
	  exit(1);
	}

	/* go around twice to minimize starting transient */
	for (xloop = -csize + 1; xloop < csize; xloop++) {
	  loop = xloop < 0 ? xloop + csize : xloop;
	  ploop = xloop - 1 < 0 ? xloop - 1 + csize : xloop - 1;
	  index = start + incr*loop;
	  pindex = start + incr*ploop;

	  /* no bad data allowed in cyclic */
	  if (fabs(sfield[index]) > badlim || fabs(sfield[pindex]) > badlim) {
	    fprintf(stderr,"cdffilter: bad data not allowed in cyclic\n");
	  }
	  if (xloop == -csize + 1) dfield[index] = sfield[index];
	  else {
	    dfield[index] = alpha*sfield[index] + (1. - alpha)*dfield[pindex];
	  }
	  tfield[loop] = dfield[index];
	}
	
	/* backward filtering for symmetry */
	for (xloop = 2*csize - 2; xloop >= 0; xloop--) {
	  loop = xloop > csize - 1 ? xloop - csize : xloop;
	  ploop = xloop + 1 > csize - 1 ? xloop + 1 - csize : xloop + 1;
	  index = start + incr*loop;
	  pindex = start + incr*ploop;
	  if (xloop == 2*csize - 2) dfield[index] = tfield[loop];
	  else {
	    dfield[index] = alpha*tfield[loop] + (1. - alpha)*dfield[pindex];
	  }
	}
      }

      /* fill in repeated element if necessary */
      if (rcyclic) dfield[start + incr*csize] = dfield[start];

      /* if high pass filter is desired, take the complement */
      if (highpass) {
        for (loop = 0; loop < size; loop++) {
          index = start + incr*loop;
          if (fabs(sfield[index]) < badlim) {
            dfield[index] = sfield[index] - dfield[index];
          }
        }
      }

      /* increment instance */
      instance++;
      free(tfield);
    }
    /* write variable slice */
    putslice(stdout,nvbout,vbuff);
  }
  exit(0);
}
