/* cdfgriv.c -- Reorganize 2-D grids into 1-D lists with
 * 1-D coordinate locations.
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "cdfhdr.h"

/* Candis definitions */
static char hbuff[HBMAX][LINE];
static char cline[LINE];
static long nsbin,nsbout,nvbin,nvbout;
static float *sbuff,*vbuff;
static float bad,badlim;
static char badtxt[LINE], badlimtxt[LINE];

/* other definitions */
static char *xname,*yname;       /* names of input index fields */
static char iname[LINE];         /* name of output index field */
static char nxname[LINE],nyname[LINE];  /* new x and y names */
static long xthin,ythin;         /* amount to thin each index field */
static char *varnames[HBMAX];    /* names of variables to be serialized */
static char nvarnames[HBMAX][LINE];  /* new names of these variables */
static long nvars;               /* number of these variables */
static long nindex;              /* size of new index field */
static long xsize,ysize;         /* dimension sizes */
struct field *fp;                /* pointer to field structure */
static float *indexptr,*xptr,*yptr,*xlptr,*ylptr,*oldptr[HBMAX],
  *newptr[HBMAX];                /* ptrs to flds */

main(argc,argv)
     int argc;
     char *argv[];
{
  long i,j,k,l;

  /* check arguments */
  if (argc < 6) {
    fprintf(stderr,"Usage: cdfgriv xname xthin yname ythin var1 ...\n");
    exit(1);
  }
  nvars = argc - 5;
  if (nvars > HBMAX) {
    fprintf(stderr,"cdfgriv: too many variables!\n");
    exit(1);
  }
  xname = argv[1];
  yname = argv[3];
  xthin = atof(argv[2]);
  if (xthin < 1) xthin = 1;
  ythin = atof(argv[4]);
  if (ythin < 1) ythin = 1;
  for (i = 0; i < nvars; i++) varnames[i] = argv[i + 5];

  /* get header */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfgriv: float format expected\n");
    exit(1);
  }

  /* get bad and badlim if they exist, otherwise defaults */
  bad = BAD;
  badlim = BADLIM;
  if (seekpar(hbuff,HBMAX,"bad",badtxt) == OK &&
      seekpar(hbuff,HBMAX,"badlim",badlimtxt) == OK) {
    bad = atoi(badtxt);
    badlim = atoi(badlimtxt);
  }

  /* loop on input variable fields to pick out ones of interest and
     check to see that they all have the right shape and size */
  for (i = 0; i < nvars; i++) {
    if ((fp = seekfld(hbuff,HBMAX,'v',varnames[i])) == NULL) {
      fprintf(stderr,"cdfgriv: variable field %s does not exist\n",
	      varnames[i]);
      exit(1);
    }
    if (fp->dim != 2) {
      fprintf(stderr,"cdfgriv: field %s is not 2-D\n",varnames[i]);
      exit(1);
    }
    if (strcmp(xname,fp->dname1) != 0 || strcmp(yname,fp->dname2) != 0) {
      fprintf(stderr,"cdfgriv: nonexistent or badly ordered dimensions\n",
	      xname,yname);
      exit(1);
    }
    xsize = fp->dsize1;
    ysize = fp->dsize2;
  }

  /* given the dimension sizes and skip intervals, compute new
     index field size */
  nindex = 0;
  for (i = xthin/2; i < xsize; i += xthin) {
    for (j = ythin/2; j < ysize; j += ythin) {
      nindex++;
    }
  }

  /* quit if nindex = 0 */
  if (nindex == 0) {
    fprintf(stderr,"cdfgriv: xthin or ythin too big\n");
    exit(1);
  }

  /* get the old buffer sizes */
  nsbin = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');

  /* find unused variable names */
  for (i = 1; i <= 9; i++) {

    /* construct new index name */
    sprintf(iname,"index%d",i);

    /* construct the new variable names */
    sprintf(nxname,"%s.l%d",xname,i);
    sprintf(nyname,"%s.l%d",yname,i);
    for (j = 0; j < nvars; j++) {
      sprintf(nvarnames[j],"%s.l%d",varnames[j],i);
    }

    /* if index name is available, break out of loop */
    /* if index is available, assume others are as well */
    if (seekfld(hbuff,HBMAX,'s',iname) == NULL) {
      break;
    }

    /* can't find any unused names */
    if (i > 8) {
      fprintf(stderr,"cdfgriv: no available index field name\n");
      exit(1);
    }
  }

  /* add index var and other vars to header */
  addfld(hbuff,HBMAX,'s',iname,1.,0.,'l',1,iname,nindex);
  addfld(hbuff,HBMAX,'v',nxname,1.,0.,'l',1,iname,nindex);
  addfld(hbuff,HBMAX,'v',nyname,1.,0.,'l',1,iname,nindex);
  for (i = 0; i < nvars; i++) {
    addfld(hbuff,HBMAX,'v',nvarnames[i],1.,0.,'l',1,iname,nindex);
  }

  /* add other stuff to the header */
  addpar(hbuff,HBMAX,"index0","0.");
  addpar(hbuff,HBMAX,"dindex","1.");
  sprintf(cline,"cdfgriv: %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  addcline(hbuff,HBMAX,cline);
  i = 0;
  strcpy(cline,"  ");
  while (i < nvars) {
    if (strlen(cline) + strlen(varnames[i]) > 75) {
      strcat(cline,"\n");
      addcline(hbuff,HBMAX,cline);
      strcpy(cline,"  ");
    }
    strcat(cline," ");
    strcat(cline,varnames[i]);
    i++;
  }
  strcat(cline,"\n");
  addcline(hbuff,HBMAX,cline);

  /* put the header */
  puthdr(stdout,hbuff,HBMAX);

  /* get the new buffer sizes and allocate space */
  nsbout = elemcnt(hbuff,HBMAX,'s');
  nvbout = elemcnt(hbuff,HBMAX,'v');
  sbuff = getbuff(nsbout);
  vbuff = getbuff(nvbout);

  /* read in static slice */
  getslice(stdin,nsbin,sbuff);

  /* get pointer to index field named index as well as x and y ptrs */
  indexptr = getptr(hbuff,HBMAX,sbuff,'s','d',iname);
  xptr = getptr(hbuff,HBMAX,sbuff,'s','d',xname);
  yptr = getptr(hbuff,HBMAX,sbuff,'s','d',yname);

  /* fill the index field index */
  for (i = 0; i < nindex; i++) indexptr[i] = (float)i;

  /* write the static slice */
  putslice(stdout,nsbout,sbuff);

  /* get the variable field pointers */
  xlptr = getptr(hbuff,HBMAX,vbuff,'v','d',nxname);
  ylptr = getptr(hbuff,HBMAX,vbuff,'v','d',nyname);
  for (i = 0; i < nvars; i++) {
    oldptr[i] = getptr(hbuff,HBMAX,vbuff,'v','d',varnames[i]);
    newptr[i] = getptr(hbuff,HBMAX,vbuff,'v','d',nvarnames[i]);
  }

  /* loop on variable slices */
  while (getslice(stdin,nvbin,vbuff) != EOF) {
    
    /* transfer the data */
    k = 0;
    for (i = xthin/2; i < xsize; i += xthin) {
      for (j = ythin/2; j < ysize; j += ythin) {
	xlptr[k] = xptr[i];
	ylptr[k] = yptr[j];
	for (l = 0; l < nvars; l++) {
	  newptr[l][k] = oldptr[l][j + ysize*i];
	}
	k++;
      }
    }
    
    /* write the variable slices */
    putslice(stdout,nvbout,vbuff);
  }

  /* done! */
}
