/* cdfinvindex -- Replace an index field with another 1-D field which
 * is a function only of the initial index field.
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "cdfhdr.h"

static char hbuffin[HBMAX][LINE];          /* input header buffer */
static char hbuffout[HBMAX][LINE];         /* output header buffer */
static float *sbuff,*vbuff;                /* slice buffers */
static long nsbuff,nvbuff;                 /* slice buffer sizes */
static struct field *fp;                   /* field pointer temp */
static long pos,loop;                      /* looping indices */
static char *oldname,*newname;             /* index field names */
static char comment[LINE];                 /* new comment line */
static char pname[LINE],pvalue[LINE];      /* parameter name and value */
static char indexname0[LINE];              /* index parameter base name */
static char dindexname[LINE];              /* index parameter increment name */
static long isize;                         /* size of index field */
static char precision;                     /* field precision */
static char fname[LINE];                   /* field name */
static float smul,sadd;                    /* conversion parameters */
static int dim,n1,n2,n3,n4;                /* num dims and dim sizes */
static char d1[LINE],d2[LINE],d3[LINE],d4[LINE];  /* dimension names */
static float *oldptr,*newptr;              /* pointers to old and new flds */
static float *oldtmp;                      /* storage for old index field */
static char newpval0[LINE],newdpval[LINE]; /* new index parameter values */
static char newpname0[LINE],newdpname[LINE];  /* new index parameter names */

static void swapnames();

main(int argc, char *argv[])
{

  /* check args */
  if (argc != 3) {
    fprintf(stderr,"Usage: cdfinvindex oldindex newindex\n");
    exit(1);
  }
  oldname = argv[1];
  newname = argv[2];

  /* get a header from standard input */
  gethdr(stdin,hbuffin,HBMAX);
  if (getfmt(hbuffin,HBMAX) != 'f') {
    fprintf(stderr,"cdfinvindex: float format expected\n");
    exit(1);
  }

  /* create a null output header */
  nullhdr(hbuffout,HBMAX,"float");

  /* copy over comments and add a new comment */
  copycmt(hbuffout,HBMAX,hbuffin,HBMAX);
  sprintf(comment,"cdfinvindex %s %s\n",oldname,newname);
  addcline(hbuffout,HBMAX,comment);

  /* modify parameters, eliminating reference to old index field */
  pos = -1;
  sprintf(indexname0,"%s0",oldname);
  sprintf(dindexname,"d%s",oldname);
  while (getpar(hbuffin,HBMAX,++pos,pname,pvalue) != FAIL) {
    if (strcmp(pname,indexname0) == 0) continue;
    if (strcmp(pname,dindexname) == 0) continue;
    addpar(hbuffout,HBMAX,pname,pvalue);
  }

  /* look for old index field in static slice */
  if ((fp = seekfld(hbuffin,HBMAX,'s',oldname)) == NULL) {
    fprintf(stderr,"cdfinvindex: %s not found in static slice\n");
    exit(1);
  }
  if (fp->dim != 1) {
    fprintf(stderr,"cdfinvindex: field %s has dimension %d != 1\n",
	    fp->dim);
    exit(1);
  }
  if (strcmp(oldname,fp->dname1) != 0) {
    fprintf(stderr,"cdfinvindex: %s is not an index field -- %s != %s\n",
	    oldname,fp->dname1);
    exit(1);
  }
  isize = fp->dsize1;

  /* look for new field in either static or variable slice */
  if ((fp = seekfld(hbuffin,HBMAX,'s',newname)) == NULL) {
    if ((fp = seekfld(hbuffin,HBMAX,'v',newname)) == NULL) {
      fprintf(stderr,"cdfinvindex: field %s not found\n",newname);
      exit(1);
    }
  }
  if (fp->dim != 1) {
    fprintf(stderr,"cdfinvindex: %s must have one dimension\n",newname);
    exit(1);
  }
  if (strcmp(oldname,fp->dname1) != 0) {
    fprintf(stderr,"cdfinvindex: %s must be a function of %s\n",
	    newname,oldname);
    exit(1);
  }

  /* interchange oldname and newname in all fields and dimension names --
  * static slice first... */
  pos = 0;
  while ((fp = getfld(hbuffin,HBMAX,'s',pos)) != NULL) {
    strcpy(fname,fp->fname);
    smul = fp->smul;
    sadd = fp->sadd;
    precision = fp->bytes;
    dim = fp->dim;
    strcpy(d1,fp->dname1);
    strcpy(d2,fp->dname2);
    strcpy(d3,fp->dname3);
    strcpy(d4,fp->dname4);
    n1 = fp->dsize1;
    n2 = fp->dsize2;
    n3 = fp->dsize3;
    n4 = fp->dsize4;
    swapnames(fname,oldname,newname);
    if (dim >= 1) swapnames(d1,oldname,newname);
    if (dim >= 2) swapnames(d2,oldname,newname);
    if (dim >= 3) swapnames(d3,oldname,newname);
    if (dim >= 4) swapnames(d4,oldname,newname);
    addfld(hbuffout,HBMAX,'s',fname,smul,sadd,precision,dim,
	   d1,n1,d2,n2,d3,n3,d4,n4);
    pos++;
  }

  /* ... then variable slice */
  pos = 0;
  while ((fp = getfld(hbuffin,HBMAX,'v',pos)) != NULL) {
    strcpy(fname,fp->fname);
    smul = fp->smul;
    sadd = fp->sadd;
    precision = fp->bytes;
    dim = fp->dim;
    strcpy(d1,fp->dname1);
    strcpy(d2,fp->dname2);
    strcpy(d3,fp->dname3);
    strcpy(d4,fp->dname4);
    n1 = fp->dsize1;
    n2 = fp->dsize2;
    n3 = fp->dsize3;
    n4 = fp->dsize4;
    swapnames(fname,oldname,newname);
    if (dim >= 1) swapnames(d1,oldname,newname);
    if (dim >= 2) swapnames(d2,oldname,newname);
    if (dim >= 3) swapnames(d3,oldname,newname);
    if (dim >= 4) swapnames(d4,oldname,newname);
    addfld(hbuffout,HBMAX,'v',fname,smul,sadd,precision,dim,
	   d1,n1,d2,n2,d3,n3,d4,n4);
    pos++;
  }

  /* allocate static and variable slice buffers */
  nsbuff = elemcnt(hbuffin,HBMAX,'s');
  nvbuff = elemcnt(hbuffin,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuff);

  /* get static slice and first variable slice */
  if (getslice(stdin,nsbuff,sbuff) == EOF) {
    fprintf(stderr,"cdfinvindex: unexpected eof reading static slice\n");
    exit(1);
  }
  if (getslice(stdin,nvbuff,vbuff) == EOF) {
    fprintf(stderr,
	    "cdfinvindex: unexpected eof reading first variable slice\n");
    exit(1);
  }

  /* get pointers to old and new index fields */
  oldptr = getptr(hbuffin,HBMAX,sbuff,'s','d',oldname);
  if ((newptr = getptr(hbuffin,HBMAX,sbuff,'s','r',newname)) == NULL) {
    newptr = getptr(hbuffin,HBMAX,vbuff,'v','d',newname);
  }

  /* add parameters for new field */
  sprintf(newpval0,"%g",newptr[0]);
  sprintf(newdpval,"%g",newptr[1] - newptr[0]);
  sprintf(newpname0,"%s0",newname);
  sprintf(newdpname,"d%s",newname);
  addpar(hbuffout,HBMAX,newpname0,newpval0);
  addpar(hbuffout,HBMAX,newdpname,newdpval);

  /* retain the old index field data and replace with the new */
  oldtmp = getbuff(isize);
  for (loop = 0; loop < isize; loop++) {
    oldtmp[loop] = oldptr[loop];
    oldptr[loop] = newptr[loop];
    newptr[loop] = oldtmp[loop];
  }

  /* write the new header */
  puthdr(stdout,hbuffout,HBMAX);

  /* write the static slice and the first variable slice */
  putslice(stdout,nsbuff,sbuff);
  putslice(stdout,nvbuff,vbuff);

  /* transfer any subsequent variable slices, writing contents of
   * old index field to the new location -- if it is in the static
   * slice, it doesn't matter
   */
  while (getslice(stdin,nvbuff,vbuff) != EOF) {
    for (loop = 0; loop < isize; loop++) {
      newptr[loop] = oldtmp[loop];
    }
    putslice(stdout,nvbuff,vbuff);
  }

  /* done! */
  exit(0);
}

/* swapnames -- interchange names */
static void swapnames(char *name, char *oldname, char *newname)
{
  char tmpname[LINE];
  strcpy(tmpname,name);
  if (strcmp(tmpname,oldname) == 0) strcpy(name,newname);
  if (strcmp(tmpname,newname) == 0) strcpy(name,oldname);
}
