/* cdfmath -- This program executes general math functions on common
 * data format files.  The math expression is assumed to be in reverse
 * Polish, with fields, operators, and constants.
 */
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "cdfhdr.h"
#define TINY 1.e-30
#define DEGTORAD 57.29577951

/* Candis definitions */
static char hbuff[HBMAX][LINE];
static char cline[LINE];
static long nsbuff,nvbin,nvbout;
static float *sbuff,*vbuff;
static float bad,badlim;

/* parse table structure */
#define VAR 1
#define CONST 2
#define FN 3
#define OP 4
#define PLUS 1
#define MINUS 2
#define TIMES 3
#define DIV 4
#define GT 5
#define LT 6
#define QUESTION 7
#define ASSIGN 8
#define EXP 1
#define LOG 2
#define LOG10 3
#define POW 4
#define SIN 5
#define COS 6
#define ATAN2 7
#define MOD 8
#define ABS 9

static struct parsetable {
  int etype,optype,fntype;
  float *ptr;
  char fname[LINE];
  long a[4];
  int newfldflag;
} parsetable;

/* parse table info */
#define MAXENTRIES 50
static struct parsetable pt[MAXENTRIES];
static int numentries;
static int entrypos;

/* index field information */
static int numifields;
static long size[4],newsize[4];
static char name[4][LINE],newname[4][LINE];

/* computational stack */
static float stack[MAXENTRIES];
static float *temptr;
static int stackptr;

/* other stuff */
static int poop,loop;
static long ii1,ii2,ii3,ii4;
static int lastetype;
static long indexfld;
static long skip;
static int acum[4];
static int newdim;
static struct field *fp;
static char *gettoken();
static void syntax();
static struct parsetable tokentype();
static void parse();
static void ordererror();

/******************* main program *************************/

main(argc,argv)
int argc;
char *argv[];
{
  double temp1;

/* check arguments */

  if (argc != 2) {
    fprintf(stderr,"Usage: cdfmath 'expression'\n");
    fprintf(stderr,"       expression = element element ...\n");
    fprintf(stderr,
         "       element = field or constant or operator or function\n");
    fprintf(stderr,"       field = variable field or static field\n");
    fprintf(stderr,"       operator = + or - or * or / or > or < or ? or =\n");
    fprintf(stderr,
         "       function = exp or sin or cos or log or log10 or pow\n");
    fprintf(stderr,"            or atan2 or mod or abs\n");
    fprintf(stderr,"       sin and cos arguments in degrees\n");
    fprintf(stderr,"       atan2 returns angle in degrees\n");
    fprintf(stderr,"       reverse polish notation used\n");
    exit(1);
  }

/* get header */

  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfmath: float format expected\n");
    exit(1);
  }

/* get bad data parameters */

  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;
  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;

/* add comment */

  sprintf(cline,"cdfmath: %s\n",argv[1]);
  addcline(hbuff,HBMAX,cline);

/* get input slice sizes */

  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');

/* set default dimension sizes */

  size[0] = size[1] = size[2] = size[3] = 1;

/* survey index fields */

  numifields = 0;
  loop = 0;
  while ((fp = getfld(hbuff,HBMAX,'s',loop++)) != NULL) {
    if ((fp->dim == 1) && (strcmp(fp->fname,fp->dname1) == 0)) {
      numifields++;
      if (numifields > 4) {
        fprintf(stderr,"cdfmath: too many index fields\n");
        exit(1);
      }
      if (numifields == 1) {
        strcpy(name[0],fp->dname1);
        size[0] = fp->dsize1;
      }
      else if (numifields == 2) {
        strcpy(name[1],fp->dname1);
        size[1] = fp->dsize1;
      }
      else if (numifields == 3) {
        strcpy(name[2],fp->dname1);
        size[2] = fp->dsize1;
      }
      else if (numifields == 4) {
        strcpy(name[3],fp->dname1);
        size[3] = fp->dsize1;
      }
    }
  }

/* parse the expression */

  parse(argv[1]);

/* check for common errors */
  if (numentries < 3) {
    fprintf(stderr,"cdfmath: syntax error\n");
    exit(1);
  }
  if (pt[numentries - 1].etype != OP || pt[numentries - 1].optype != ASSIGN) {
    fprintf(stderr,"cdfmath: the command must end with '='\n");
    exit(1);
  }
  for (loop = 0; loop < numentries - 1; loop++) {
    if (pt[loop].etype == VAR && pt[loop].newfldflag &&
         !(pt[loop + 1].etype == OP && pt[loop + 1].optype == ASSIGN)) {
      fprintf(stderr,"cdfmath: field %s doesn't exist\n",pt[loop].fname);
      exit(1);
    }
  }

/* write the header */

  puthdr(stdout,hbuff,HBMAX);

/* get output variable slice buffer size */

  nvbout = elemcnt(hbuff,HBMAX,'v');

/* allocate buffers */

  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbout);

/* assign field pointers */

  for (loop = 0; loop < numentries; loop++) {
    if (pt[loop].etype == VAR) {
      if ((pt[loop].ptr = getptr(hbuff,HBMAX,vbuff,'v','r',pt[loop].fname))
           == NULL) {
        pt[loop].ptr = getptr(hbuff,HBMAX,sbuff,'s','d',pt[loop].fname);
      }
    }
  }

/* transfer the static field */

  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

/* find minimum looping that will do the job */

  for (poop = 0; poop < 4; poop++) newsize[poop] = 1;
  for (loop = 0; loop < numentries; loop++) {
    if (pt[loop].etype == VAR) {
      for (poop = 0; poop < 4; poop++) {
        if (pt[loop].a[poop]) newsize[poop] = size[poop];
      }
    }
  }

/* loop on variable slices */

  while (getslice(stdin,nvbin,vbuff) != EOF) {

/* loop on indices */

    for (ii1 = 0; ii1 < newsize[0]; ii1++) {
      for (ii2 = 0; ii2 < newsize[1]; ii2++) {
        for (ii3 = 0; ii3 < newsize[2]; ii3++) {
          for (ii4 = 0; ii4 < newsize[3]; ii4++) {

/* loop on parse table */

            lastetype = stackptr = 0;
            for (loop = 0; loop < numentries; loop++) {

/* check for element type */

              switch(pt[loop].etype) {

/* field type */

              case VAR :
                indexfld = pt[loop].a[0]*ii1 + pt[loop].a[1]*ii2 +
                     pt[loop].a[2]*ii3 + pt[loop].a[3]*ii4;
                if (stackptr >= MAXENTRIES - 1) {
                  fprintf(stderr,"cdfmath: expression too complex\n");
                  exit(1);
                }
                stack[stackptr++] = *(pt[loop].ptr + indexfld);
                break;

/* constant type */

              case CONST :
                if (stackptr >= MAXENTRIES - 1) {
                  fprintf(stderr,"cdfmath: expression too complex\n");
                  exit(1);
                }
                stack[stackptr++] = *pt[loop].ptr;
                break;

/* operator type */

              case OP :

/* which operator? */

                switch(pt[loop].optype) {

                case PLUS :
                  if (stackptr < 2) syntax("+");
                  temptr = stack + stackptr - 2;
                  if ((fabs(*temptr) > badlim) ||
                       (fabs(*(temptr + 1)) > badlim)) {
                    *temptr = bad;
                    stackptr--;
                    break;
                  }
                  *temptr += *(temptr + 1);
                  stackptr--;
                  break;
                case MINUS :
                  if (stackptr < 2) syntax("-");
                  temptr = stack + stackptr - 2;
                  if ((fabs(*temptr) > badlim) ||
                       (fabs(*(temptr + 1)) > badlim)) {
                    *temptr = bad;
                    stackptr--;
                    break;
                  }
                  *temptr -= *(temptr + 1);
                  stackptr--;
                  break;
                case TIMES :
                  if (stackptr < 2) syntax("*");
                  temptr = stack + stackptr - 2;
                  if ((fabs(*temptr) > badlim) ||
                       (fabs(*(temptr + 1)) > badlim)) {
                    *temptr = bad;
                    stackptr--;
                    break;
                  }
                  *temptr *= *(temptr + 1);
                  stackptr--;
                  break;
                case DIV :
                  if (stackptr < 2) syntax("/");
                  temptr = stack + stackptr - 2;
                  if ((fabs(*temptr) > badlim) ||
                       (fabs(*(temptr + 1)) > badlim) ||
                       (fabs(*(temptr + 1)) < TINY)) {
                    *temptr = bad;
                    stackptr--;
                    break;
                  }
                  *temptr /= *(temptr + 1);
                  stackptr--;
                  break;
                case GT :
                  if (stackptr < 2) syntax(">");
                  temptr = stack + stackptr - 2;
                  if ((fabs(*temptr) > badlim) ||
                       (fabs(*(temptr + 1)) > badlim)) {
                    *temptr = bad;
                    stackptr--;
                    break;
                  }
                  if (*temptr <= *(temptr + 1)) *temptr = bad;
                  stackptr--;
                  break;
                case LT :
                  if (stackptr < 2) syntax("<");
                  temptr = stack + stackptr - 2;
                  if ((fabs(*temptr) > badlim) ||
                       (fabs(*(temptr + 1)) > badlim)) {
                    *temptr = bad;
                    stackptr--;
                    break;
                  }
                  if (*temptr >= *(temptr + 1)) *temptr = bad;
                  stackptr--;
                  break;
                case QUESTION :
                  if (stackptr < 3) syntax("?");
                  temptr = stack + stackptr - 3;
                  if (fabs(*temptr) < badlim) {
                    *temptr = *(temptr + 1);
                  }
                  else {
                    *temptr = *(temptr + 2);
                  }
                  stackptr -= 2;
                  break;
                case ASSIGN :
                  poop = loop - 1;
                  if (lastetype != VAR) syntax("=");
                  if (stackptr != 2) syntax("=");
                  if ((pt[poop].a[0] == 0) && (ii1 != 0)) break;
                  if ((pt[poop].a[1] == 0) && (ii2 != 0)) break;
                  if ((pt[poop].a[2] == 0) && (ii3 != 0)) break;
                  if ((pt[poop].a[3] == 0) && (ii4 != 0)) break;
                  indexfld = pt[poop].a[0]*ii1 + pt[poop].a[1]*ii2 +
                       pt[poop].a[2]*ii3 + pt[poop].a[3]*ii4;
                  stackptr -= 2;
                  *(pt[poop].ptr + indexfld) = stack[stackptr];
                  break;
                }
                break;

/* function type */

              case FN :

/* which function? */

                switch(pt[loop].fntype) {
                case EXP :
                  if (stackptr < 1) syntax("exp");
                  temptr = stack + stackptr - 1;
                  if (fabs(*temptr) > badlim) break;
                  *temptr = exp(*temptr);
                  break;
                case SIN :
                  if (stackptr < 1) syntax("sin");
                  temptr = stack + stackptr - 1;
                  if (fabs(*temptr) > badlim) break;
                  *temptr = sin(*temptr/DEGTORAD);
                  break;
                case COS :
                  if (stackptr < 1) syntax("cos");
                  temptr = stack + stackptr - 1;
                  if (fabs(*temptr) > badlim) break;
                  *temptr = cos(*temptr/DEGTORAD);
                  break;
                case LOG :
                  if (stackptr < 1) syntax("log");
                  temptr = stack + stackptr - 1;
                  if (fabs(*temptr) > badlim) break;
                  if (*temptr > TINY)
                    *temptr = log(*temptr);
                  else *temptr = bad;
                  break;
                case LOG10 :
                  if (stackptr < 1) syntax("log10");
                  temptr = stack + stackptr - 1;
                  if (fabs(*temptr) > badlim) break;
                  if (*temptr > TINY)
                    *temptr = log10(*temptr);
                  else *temptr = bad;
                  break;
                case ABS :
                  if (stackptr < 1) syntax("abs");
                  temptr = stack + stackptr - 1;
                  if (fabs(*temptr) > badlim) break;
                  *temptr = fabs(*temptr);
                  break;
                case POW :
                  if (stackptr < 2) syntax("pow");
                  temptr = stack + stackptr - 2;
                  if (fabs(*temptr) > badlim || fabs(*(temptr + 1)) > badlim
		      || *temptr < TINY) {
		    *temptr = bad;
                    stackptr--;
                    break;
                  }
		  *temptr = pow(*temptr,*(temptr + 1));
                  stackptr--;
                  break;
                case ATAN2 :
                  if (stackptr < 2) syntax("atan2");
                  temptr = stack + stackptr - 2;
                  if (fabs(*temptr) > badlim || fabs(*(temptr + 1)) > badlim ||
                       (fabs(*temptr) < TINY && fabs(*(temptr + 1)) < TINY)) {
                    *temptr = bad;
                  }
                  else {
                    *temptr = DEGTORAD*atan2(*temptr,*(temptr + 1));
                    if (*temptr < 0.) *temptr += 360.;
                    else if (*temptr > 360.) *temptr -= 360.;
                  }
                  stackptr--;
                  break;
                case MOD :
                  if (stackptr < 2) syntax("mod");
                  temptr = stack + stackptr - 2;
                  if (fabs(*temptr) > badlim || *(temptr + 1) < TINY) {
                    *temptr = bad;
                  }
                  else {
                    temp1 = (double)*temptr/(double)(*(temptr + 1));
                    temp1 += (int)temp1 + 2;
                    temp1 -= (int)temp1;
                    *temptr = temp1*(*(temptr + 1));
                  }
                  stackptr--;
                  break;
                }
                break;
              }

/* send etype to lastetype */

              lastetype = pt[loop].etype;
            }
          }
        }
      }
    }

/* put variable slice */

    putslice(stdout,nvbout,vbuff);
  }
  exit(0);
}

/********************* report a syntax error *******************/

static void syntax(string)
char *string;	
{
  fprintf(stderr,"cdfmath: syntax error near %s\n",string);
  exit(1);
}

/*********************** get a token ***************************/

static char *gettoken(line)
char *line;
{
  int tokenpos;
  static char tline[LINE];
  char c;

/* get an entry surrounded by spaces or ended by a null */

  tokenpos = 0;
  while ((c = line[entrypos]) != '\0') {
    entrypos++;

/* ignore leading spaces */

    if (c == ' ') {
      if (tokenpos > 0) break;
    }

/* if not a space, incorporate */

    else tline[tokenpos++] = c;
  }

/* if nothing incorporated, return a null */

  if (tokenpos == 0) {
    return(NULL);
  }

/* otherwise null terminate and return the token */

  tline[tokenpos] = '\0';
  return(tline);
}

/************************** determine token type ****************/

static struct parsetable tokentype(token)
char *token;
{
  static struct parsetable cur;
  cur.newfldflag = 0;

/* operators */
 
  if (strlen(token) == 1) {
    cur.etype = OP;
    if (token[0] == '+') {
      cur.optype = PLUS;
      return(cur);
    }
    if (token[0] == '-') {
      cur.optype = MINUS;
      return(cur);
    }
    if (token[0] == '*') {
      cur.optype = TIMES;
      return(cur);
    }
    if (token[0] == '/') {
      cur.optype = DIV;
      return(cur);
    }
    if (token[0] == '>') {
      cur.optype = GT;
      return(cur);
    }
    if (token[0] == '<') {
      cur.optype = LT;
      return(cur);
    }
    if (token[0] == '?') {
      cur.optype = QUESTION;
      return(cur);
    }
    if (token[0] == '=') {
      cur.optype = ASSIGN;
      return(cur);
    }
  }
      
/* functions */

  cur.etype = FN;
  if (strcmp(token,"exp") == 0) {
    cur.fntype = EXP;
    return(cur);
  }
  if (strcmp(token,"sin") == 0) {
    cur.fntype = SIN;
    return(cur);
  }
  if (strcmp(token,"cos") == 0) {
    cur.fntype = COS;
    return(cur);
  }
  if (strcmp(token,"log") == 0) {
    cur.fntype = LOG;
    return(cur);
  }
  if (strcmp(token,"log10") == 0) {
    cur.fntype = LOG10;
    return(cur);
  }
  if (strcmp(token,"pow") == 0) {
    cur.fntype = POW;
    return(cur);
  }
  if (strcmp(token,"atan2") == 0) {
    cur.fntype = ATAN2;
    return(cur);
  }
  if (strcmp(token,"mod") == 0) {
    cur.fntype = MOD;
    return(cur);
  }
  if (strcmp(token,"abs") == 0) {
    cur.fntype = ABS;
    return(cur);
  }

/* fields */

  cur.etype = VAR;
  if (isalpha(token[0])) {
    strcpy(cur.fname,token);
    if ((fp = seekfld(hbuff,HBMAX,'v',token)) == NULL) {
      if ((fp = seekfld(hbuff,HBMAX,'s',token)) == NULL) {

/* field doesn't exist -- create it */

        cur.newfldflag = 1;
        for (poop = 0; poop < 4; poop++) acum[poop] = 0;
        for (loop = 0; loop < numentries; loop++) {
          for (poop = 0; poop < 4; poop++) {
            if (pt[loop].etype == VAR && pt[loop].a[poop]) acum[poop] = 1;
          }
        }
        newdim = 0;
        for (poop = 0; poop < 4; poop++) {
          if (acum[poop]) {
            newsize[newdim] = size[poop];
            strcpy(newname[newdim++],name[poop]);
          }
        }
        addfld(hbuff,HBMAX,'v',token,1.,0.,'l',newdim,
             newname[0],newsize[0],
             newname[1],newsize[1],
             newname[2],newsize[2],
             newname[3],newsize[3]);
        fp = seekfld(hbuff,HBMAX,'v',token);
      }
    }

/* get the a's */

    for (loop = 0; loop < 4; loop++) cur.a[loop] = 0;
    skip = 1;
    loop = -1;
    if (fp->dim > 0) {
      while (++loop < numifields && strcmp(name[loop],fp->dname1) != 0);
      if (loop >= numifields) ordererror(fp->dname1);
      cur.a[loop] = 1;
    }
    if (fp->dim > 1) {
      while (++loop < numifields && strcmp(name[loop],fp->dname2) != 0);
      if (loop >= numifields) ordererror(fp->dname2);
      cur.a[loop] = 1;
    }
    if (fp->dim > 2) {
      while (++loop < numifields && strcmp(name[loop],fp->dname3) != 0);
      if (loop >= numifields) ordererror(fp->dname3);
      cur.a[loop] = 1;
    }
    if (fp->dim > 3) {
      while (++loop < numifields && strcmp(name[loop],fp->dname4) != 0);
      if (loop >= numifields) ordererror(fp->dname4);
      cur.a[loop] = 1;
    }
    if (cur.a[3] != 0) skip *= size[3];
    cur.a[2] *= skip;
    if (cur.a[2] != 0) skip *= size[2];
    cur.a[1] *= skip;
    if (cur.a[1] != 0) skip *= size[1];
    cur.a[0] *= skip;

    return(cur);
  }

/* must be a number */

  cur.etype = CONST;
  cur.ptr = getbuff(1);
  *cur.ptr = atof(token);
  cur.a[0] = cur.a[1] = cur.a[2] = cur.a[3] = 0;
  return(cur);
}

/*********************** parse the command line ******************/

static void parse(comline)
char *comline;
{
  char *token;

/* initialize things */

  numentries = entrypos = 0;

/* loop on tokens */

  while ((token = gettoken(comline)) != NULL) {

/* limit check */

    if (numentries >= MAXENTRIES) {
      fprintf(stderr,"cdfmath: expression too complex -- simplify\n");
      exit(1);
    }

/* fill in parse table */

    pt[numentries] = tokentype(token);
    numentries++;
  }
}

/* error exit if dimensions out of order */
static void ordererror(dname)
char *dname;
{
  fprintf(stderr,"cdfmath: index field %s nonexistent or out of order\n",dname);
  exit(1);
}
