/* cdfmr -- Compute mixing ratio from temperature and dewpoint */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

/* constants (Iribarne and Godson) */
#define CTOK 273.15               /* Celsius to Kelvin conversion */
#define RD 287.05                 /* dry air gas constant */
#define EPS 0.622                 /* vapor to dry air mol weight ratio */
#define RV (RD/EPS)               /* water vapor gas constant */
#define CPD 1005.                 /* specific heat const pres dry air */
#define CPV 1850.                 /* specific heat const pres water vapor */
#define CL 4218.                  /* specific heat liquid water */
#define CIMEAN 1959.              /* specific heat of ice at -20 C */
#define LR 2.5008e6               /* heat of vaporization at 0 C */
#define LFR 3.337e5               /* heat of fusion at 0 C */
#define DLFDT 2.45e3              /* slope of heat of fusion curve */
#define TR 273.15                 /* reference temperature (K) */
#define PR 100000.                /* reference pressure (Pa) */
#define TMIN 190.                 /* minimum reasonable atmospheric temp */
#define GEE 9.807                 /* acceleration of gravity */
#define DELTA 0.606               /* virtual temperature factor */

/* thermodynamic function declarations */
float escalc();                   /* saturation vapor pressure */
float ent();                      /* entropy */
float mratio();                   /* mixing ratio */

char hbuff[HBMAX][LINE];
long nsbuff,nvbuffold,nvbuff;
float *sbuff,*vbuff;
float *pres;                         /* pressure (mb) */
float *temp;                         /* temperature (C) */
float *dewpt;                        /* dewpt (C) */
float *mr;                           /* mixing ratio (g/kg) */
char cline[LINE];
float r,p,t,es;
float bad,badlim;
long size;
long loop;
static struct field *fp;

main(argc,argv)
int argc;
char *argv[];
{

/* command line arguments */
  if (argc != 5) {
    fprintf(stderr,"Usage: cdfmr pres temp dwpt mr\n");
    exit(1);
  }

/* get header */
  gethdr(stdin,hbuff,HBMAX);

/* check format */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfmr: float format expected\n");
    exit(1);
  }

/* get bad data values */
  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;

/* add comment */
  sprintf(cline,"cdfmr: %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  addcline(hbuff,HBMAX,cline);

/* get size of incoming variable slice */
  nvbuffold = elemcnt(hbuff,HBMAX,'v');

/* add new fields if needed -- assume same size and shape as pressure field */
  if((fp = seekfld(hbuff,HBMAX,'v',argv[1])) == NULL) {
    fprintf(stderr,"cdfmr: field %s doesn't exist\n",argv[1]);
    exit(1);
  }
  size = fp->tsize;

/* mr */
  if (seekfld(hbuff,HBMAX,'v',argv[4]) == NULL) {
    fp = seekfld(hbuff,HBMAX,'v',argv[1]);
    addfld(hbuff,HBMAX,'v',argv[4],10.,0.,'s',fp->dim,
         fp->dname1,fp->dsize1,
         fp->dname2,fp->dsize2,
         fp->dname3,fp->dsize3,
         fp->dname4,fp->dsize4);
  }

/* create buffers */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuff);

/* get pointers */
  pres = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[1],&fp);
  temp = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[2],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfmr: different sizes for %s and %s\n",
         argv[1],argv[2]);
    exit(1);
  }
  dewpt = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[3],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfmr: different sizes for %s and %s\n",
         argv[1],argv[3]);
    exit(1);
  }
  mr = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[4],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfmr: different sizes for %s and %s\n",
         argv[1],argv[4]);
    exit(1);
  }

/* write header buffer */
  puthdr(stdout,hbuff,HBMAX);

/* transfer static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

/* loop on variable slices */
  while (getslice(stdin,nvbuffold,vbuff) != EOF) {

/* loop on elements */
    for (loop = 0; loop < size; loop++) {

/* get and convert data -- skip if data bad */
      t = temp[loop];
      p = pres[loop];
      r = dewpt[loop];
      if (fabs(t) > badlim || fabs(p) > badlim || fabs(r) > badlim) {
        mr[loop] = bad;
      }
      else {
        t += CTOK;
	r += CTOK;
	p *= 100.;

/* if dewpoint exceeds temperature, make temperature = dewpoint */
	if (r > t) t = r;

/* compute mixing ratio (g/kg) */
	mr[loop] = 1000.*mratio(p,r);
      }
    }

/* write variable slice */
    putslice(stdout,nvbuff,vbuff);
  }
}

/* thermo -- This is a thermodynamics package that tries to be as
 * accurate as warranted by real atmospheric data for all realistic
 * conditions.
 */

#define A 17.269                  /* constants for Tetens' formula */
#define B 35.86
#define C 610.78
#define D 273.16
#define LOOPOUT 200               /* max number of iteration loops */
#define QVMIN 1.e-8               /* smallest allowed mixing ratio */

/* compute entropy */
float ent(p,t,qt)
float p;                         /* total pressure (Pa) */
float t;                         /* temperature (K) */
float qt;                        /* total water mixing ratio (g/g) */
{
  float qs,qv,pvs,pd,s;
  pvs = escalc(t);     /* saturation vapor pressure */
  pd = p - pvs;    /* partial pressure of dry air at saturation */
  qs = qv = EPS*pvs/pd;     /* saturation mixing ratio */
  if (qt < qv) qv = qt;     /* check for saturation */
  pd = p*(1. - qv/(EPS + qv));   /* partial pressure of dry air */
/* this formula is consistent both with Iribarne and Godson and with
 * Kerry Emanuel's reversible theta-e -- I also derived it myself.
 */
  s = (CPD + CL*qt)*log(t/TR) - RD*log(pd/PR) +
       (LR + (CPV - CL)*(t - TR))*qv/t;
  if (qv > 0.) s -= qv*RV*log(qv/qs);
  return(s);
}

/* Tetens' formula for water saturation vapor pressure */
/*
float escalc(t)
float t;
{
  return(C*exp(A*(t - D)/(t - B)));
}
*/

/* Formula for water saturation vapor pressure consistent with
 * Clausius-Clapyron equation.
 */
float escalc(t)
float t;
{
  float l,lr;
  lr = LR;
/* uncomment next line to use ice adiabat below 0 C */
/*
  if (t < 273.15) lr += LFR;
*/
  l = lr - (CL - CPV)*(t - TR);
  return(C*pow(TR/t,(CL - CPV)/RV)*exp((lr/TR - l/t)/RV));
}

/* compute mixing ratio, given pressure and dewpoint */
float mratio(p,td)
float p;                          /* pressure (Pa) */
float td;                         /* dewpoint temperature (K) */
{
  float temp;
  temp = escalc(td);
  return(EPS*temp/(p - temp));
}
