/* cdfnewent -- Compute Lopez-Carrillo's entropy */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

/* constants (Iribarne and Godson) */
#define RD 287.05                 /* dry air gas constant */
#define EPS 0.622                 /* vapor to dry air mol weight ratio */
#define RV (RD/EPS)               /* water vapor gas constant */
#define CPD 1005.                 /* specific heat const pres dry air */
#define CPV 1850.                 /* specific heat const pres water vapor */
#define CL 4218.                  /* specific heat liquid water */
#define CIMEAN 1959.              /* specific heat of ice at -20 C */
#define LR 2.5008e6               /* heat of vaporization at 0 C */
#define LFR 3.337e5               /* heat of fusion at 0 C */
#define DLFDT 2.45e3              /* slope of heat of fusion curve */
#define TR 273.15                 /* reference temperature (K) */
#define PTP 610.78                /* vapor pressure at triple pt (Pa) */
#define PR 100000.                /* reference pressure (Pa) */
#define TMIN 190.                 /* minimum reasonable atmospheric temp */
#define GEE 9.807                 /* acceleration of gravity */
#define DELTA 0.606               /* virtual temperature factor */

/* thermodynamic function declarations */
float escalc();                   /* saturation vapor pressure */
float ent();                      /* entropy */

char hbuff[HBMAX][LINE];
long nsbuff,nvbuffold,nvbuff;
float *sbuff,*vbuff;
float *pres;                         /* pressure (mb) */
float *temp;                         /* temperature (C) */
float *moisture;                     /* mixing ratio (g/kg) or dewpt (C) */
float *entropy;                      /* specific entropy (J/K/kg) */
float *satent;                       /* saturation entropy (J/K/kg) */

char cline[LINE];
float r,rsat,rc,rv,p,t,pvs;
float bad,badlim;
long size;
long loop;
struct field *fp;

main(argc,argv)
int argc;
char *argv[];
{

  /* command line arguments */
  if (argc != 6) {
    fprintf(stderr,"Usage: cdfnewent pres temp mixrat entropy satent\n");
    exit(1);
  }

  /* get header */
  gethdr(stdin,hbuff,HBMAX);

  /* check format */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfnewent: float format expected\n");
    exit(1);
  }

  /* get bad data values */
  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;

  /* add comment */
  sprintf(cline,"cdfnewent: %s %s %s\n",argv[1],argv[2],argv[3]);
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s\n",argv[4],argv[5]);
  addcline(hbuff,HBMAX,cline);

  /* get size of incoming variable slice */
  nvbuffold = elemcnt(hbuff,HBMAX,'v');

  /* add new fields if needed -- assume same size and shape as pressure */
  if((fp = seekfld(hbuff,HBMAX,'v',argv[1])) == NULL) {
    fprintf(stderr,"cdfnewent: field %s doesn't exist\n",argv[1]);
    exit(1);
  }
  size = fp->tsize;

  /* entropy */
  if (seekfld(hbuff,HBMAX,'v',argv[4]) == NULL) {
    fp = seekfld(hbuff,HBMAX,'v',argv[1]);
    addfld(hbuff,HBMAX,'v',argv[4],10.,0.,'s',fp->dim,
         fp->dname1,fp->dsize1,
         fp->dname2,fp->dsize2,
         fp->dname3,fp->dsize3,
         fp->dname4,fp->dsize4);
  }

  /* satent */
  if (seekfld(hbuff,HBMAX,'v',argv[5]) == NULL) {
    fp = seekfld(hbuff,HBMAX,'v',argv[1]);
    addfld(hbuff,HBMAX,'v',argv[5],10.,0.,'s',fp->dim,
         fp->dname1,fp->dsize1,
         fp->dname2,fp->dsize2,
         fp->dname3,fp->dsize3,
         fp->dname4,fp->dsize4);
  }

  /* create buffers */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuff);

  /* get pointers */
  pres = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[1],&fp);
  temp = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[2],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfnewent: different sizes for %s and %s\n",
         argv[1],argv[2]);
    exit(1);
  }
  moisture = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[3],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfnewent: different sizes for %s and %s\n",
         argv[1],argv[3]);
    exit(1);
  }
  entropy = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[4],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfnewent: different sizes for %s and %s\n",
         argv[1],argv[4]);
    exit(1);
  }
  satent = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[5],&fp);
  if (fp->tsize != size) {
    fprintf(stderr,"cdfnewent: different sizes for %s and %s\n",
         argv[1],argv[5]);
    exit(1);
  }

  /* write header buffer */
  puthdr(stdout,hbuff,HBMAX);

  /* transfer static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

  /* loop on variable slices */
  while (getslice(stdin,nvbuffold,vbuff) != EOF) {

    /* loop on elements */
    for (loop = 0; loop < size; loop++) {

      /* get and convert data -- skip if data bad */
      t = temp[loop];
      p = pres[loop];
      r = moisture[loop];
      if (fabs(t) > badlim || fabs(p) > badlim || fabs(r) > badlim) {
        entropy[loop] = bad;
        satent[loop] = bad;
      }
      else {
        t += TR;
	p *= 100.;

	/* r is initially mixing ratio in g/kg -- convert to g/g */
	r /= 1000.;

	/* compute saturation mixing ratio */
	pvs = escalc(t);
	rsat = EPS*pvs/(p - pvs);

	/* compute vapor and condensed parts */
	if (r > rsat) {
	  rv = rsat;
	  rc = r - rv;
	}
	else {
	  rv = r;
	  rc = 0.;
	}

	/* compute entropy */
        entropy[loop] = ent(p,t,rv,rc);

	/* compute saturated entropy */
	satent[loop] = ent(p,t,rsat,0.);
      }
    }

    /* write variable slice */
    putslice(stdout,nvbuff,vbuff);
  }
}

/* thermo -- This is a thermodynamics package that tries to be as
 * accurate as warranted by real atmospheric data for all realistic
 * conditions.
 */

/* compute Lopez-Carrillo's entropy */
float ent(p,t,rv,rc)
float p;                         /* total pressure (Pa) */
float t;                         /* temperature (K) */
float rv;                        /* vapor mixing ratio (g/g) */
float rc;                        /* condensate mixing ratio (g/g) */
{
  float rl,ri,pv,pd,s;

  /* decide whether we have liquid or ice */
  if (t > TR) {
    rl = rc;
    ri = 0.;
  }
  else {
    rl = 0.;
    ri = rc;
  }

  /* compute vapor pressure and pressure of dry air */
  pv = (p*rv)/(EPS + rv);
  pd = p - pv;

  /* compute entropy */
  s = (CPD + rv*CPV + rl*CL + ri*CIMEAN)*log(t/TR) - RD*log(pd/PR)
    - rv*RV*log(pv/PTP) + (LR*rv - LFR*ri)/TR;
  return(s);
}

/* Formula for water/ice saturation vapor pressure consistent with
 * Clausius-Clapyron equation.
 */
float escalc(t)
float t;
{
  float l,lr,cc;

  /* decide whether above freezing */
  if (t > TR) {
    lr = LR;
    cc = CL;
  }
  else {
    lr = LR + LFR;
    cc = CIMEAN;
  }

  /* compute the saturation vapor pressure over water/ice */
  l = lr - (cc - CPV)*(t - TR);
  return(PTP*pow(TR/t,(cc - CPV)/RV)*exp((lr/TR - l/t)/RV));
}
