/* cdfrotate -- Rotates data from a cdf file on a plane of two index fields x
                and y around a point in the plane by some angle theta, linearly
                interpolating data over a given rectangular domain in the
                rotated x-prime/y-prime plane.  Optionally adjusts vector
                fields to match the new axes.  Reduces dimensionality of data
                if either dimension in the rotated plane is of size 1.

   cdfocut  --  Degenerate version of cdfrotate.  Interpolates over a section
                of the x-prime axis only, reducing dimensionality of data.
                Optionally rotates one or more pairs of vector fields.
                This version has a different usage than cdfrotate, as it is
                intended to replace a previous faulty cdfocut utility.

   To switch between the two versions of this file, remove the comment
   delimiters around one setup section below, and place them similarly
   around the other setup section.  All other functions are identical.
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "cdfhdr.h"

#define MAXVECTORS 20
#define SET 0
#define ALL 1
#define ALLDEF 2
#define DEF ALLDEF

static char myname[80];
static float xval, yval;                   /* new axis origin/rotation point */
static double theta;                       /* angle of rotation */
static char xname[LINE],yname[LINE];       /* original index field names */
static float step;                         /* smaller of dx and dy */
static int xsize, ysize;                   /* sizes of index dimensions */
static float bad=BAD, badlim=BADLIM;       /* bad data limits */

static float xp_xy(), yp_xy(), x_xpyp(), y_xpyp();   /* internal functions */
static float interp(), *pointer();                   /* internal functions */
static long zsize(), wsize();                        /* internal functions */
static void cpyfld();                                /* internal functions */


main(argc,argv)
int argc;
char *argv[];
{
  char uname[MAXVECTORS][LINE], vname[MAXVECTORS][LINE];
                                          /* original vector field names */
  int numvectors = 0;                     /* number of vectors to adjust */
  float xplow=0,xphigh=0,yplow=0,yphigh=0;/* bounds on x-prime and y-prime */
  float xlow, xhigh, ylow, yhigh;         /* limits on ALL range type */
  int xlimtype=DEF, ylimtype=DEF;         /* flags for limit types requested */
  int i,j,k;                              /* loop variables */
  char inheader[HBMAX][LINE];             /* buffer for old header */
  char outheader[HBMAX][LINE];            /* buffer for new header */
  int xpsize, ypsize;                     /* sizes of new dimensions */
  struct field *fp;                       /* field description pointer */
  struct field fpin, fpout;               /* input/output field descriptions */
  struct field fpu, fpv;                  /* u/v vector field descriptions */
  float x0, y0;                           /* low values of index fields */
  float xp0, yp0;                         /* low values of new dimensions */
  float dx, dy;                           /* steps in index fields */
  char string1[LINE*2], string2[LINE*2],  /* general purpose strings */
       string3[LINE*2], string4[LINE*2];
  char pname[LINE], pvalue[LINE];         /* parameter name and value */
  char type;                              /* slice type ('s' or 'v') */
  long buffersizein, buffersizeout;
           /* larger of static and variable slice sizes for input and output */
  float *inbuffer, *outbuffer;            /* input/output slice buffers */
  int flag;                               /* flags presence of dimensions */
  long xloop, yloop, zloop, wloop;        /* loop variables for dimensions */
  int isuvector, isvvector;               /* flags marking vector fields */
  float x, y, xp, yp;                     /* coordinate values at a point */
  float xindex, yindex;                   /* indices into x and y dimensions */
  float u, v;                             /* u and v vector components */

  char fname[LINE],dname1[LINE],dname2[LINE],dname3[LINE],dname4[LINE];
  long dsize1, dsize2, dsize3, dsize4;   
  int dim;
                     /* temporary storage for altered field description data */


/********************** SETUP -- CDFROTATE VERSION ***************************/

/*

  strcpy(myname,argv[0]);
  if(argc < 6)
  {
    fprintf(stderr,"%s: Invalid number of arguments.\n",myname);
    fprintf(stderr,"%s: Usage:\n",myname);
    fprintf(stderr,"%s:   %s x_name y_name x_value y_value theta\n",
        myname,myname);
    fprintf(stderr,"%s:     [-xprange xplow xphigh] [-yprange yplow yphigh]\n"
        ,myname);
    fprintf(stderr,"%s:     [-xp0] [-yp0] [-all]\n",myname);
    fprintf(stderr,"%s:     [-allrange xlow ylow xhigh yhigh]\n",myname);
    fprintf(stderr,"%s:     [-vector u_name v_name] [-vector . . . .]\n"
        ,myname);
    exit(1);
  }

  strcpy(xname,argv[1]);
  strcpy(yname,argv[2]);
  xval = atof(argv[3]);
  yval = atof(argv[4]);
  theta = (double)atof(argv[5]) * (double)4 * atan((double)1) / (double)180;

  i = 6;
  while(i < argc)
  {
    if(!strcmp(argv[i],"-xprange"))
    {
      if(argc < i+3)
      {
        fprintf(stderr,"%s: Invalid number of -xprange arguments\n",myname);
        exit(1);
      }
      xplow = atof(argv[++i]);
      xphigh = atof(argv[++i]);
      xlimtype = SET;
    }
    else if(!strcmp(argv[i],"-yprange"))
    {
      if(argc < i+3)
      {
        fprintf(stderr,"%s: Invalid number of -yprange arguments\n",myname);
        exit(1);
      }
      yplow = atof(argv[++i]);
      yphigh = atof(argv[++i]);
      ylimtype = SET;
    }
    else if(!strcmp(argv[i],"-all"))
        xlimtype = ylimtype = ALLDEF;
    else if(!strcmp(argv[i],"-allrange"))
    {
      if(argc < i+5)
      {
        fprintf(stderr,"%s: Invalid number of -allrange arguments\n",myname);
        exit(1);
      }
      xlow = atof(argv[++i]);
      ylow = atof(argv[++i]);
      xhigh = atof(argv[++i]);
      yhigh = atof(argv[++i]);
      xlimtype = ylimtype = ALL;
    }
    else if(!strcmp(argv[i],"-xp0"))
    {
      xplow = 0;
      xphigh = 0;
      xlimtype = SET;
    }
    else if(!strcmp(argv[i],"-yp0"))
    {
      yplow = 0;
      yphigh = 0;
      ylimtype = SET;
    }
    else if(!strcmp(argv[i],"-vector"))
    {
      if(argc < i+3)
      {
        fprintf(stderr,"%s: Invalid number of -vector arguments\n",myname);
        exit(1);
      }
      strcpy(uname[numvectors],argv[++i]);
      strcpy(vname[numvectors],argv[++i]);
      numvectors++;
    }
    else
    {
      fprintf(stderr,"%s: Invalid argument: %s\n",myname,argv[i]);
      exit(1);
    }
    i++;
  }

*/

/*****************************************************************************/





/********************** SETUP -- CDFOCUT VERSION *****************************/
  strcpy(myname,argv[0]);
  if(argc < 8)
  {
    fprintf(stderr,"%s: Invalid number of arguments.\n",myname);
    fprintf(stderr,"%s: Usage:\n",myname);
    fprintf(stderr,"%s:   %s x_name y_name x_value y_value theta xplow xphigh\n"
        ,myname,myname);
    fprintf(stderr,"%s:     [u_name1 v_name1] [u_name2 v_name2] [. . . .]\n"
        ,myname);
    exit(1);
  }

  strcpy(xname,argv[1]);
  strcpy(yname,argv[2]);
  xval = atof(argv[3]);
  yval = atof(argv[4]);
  theta = (double)atof(argv[5]) * (double)4 * atan((double)1) / (double)180;
  xplow = atof(argv[6]);
  xphigh = atof(argv[7]);

  xlimtype = SET;
  yplow = yphigh = 0;
  ylimtype = SET;

  i = 8;
  while(i < argc)
  {
    strcpy(uname[numvectors],argv[i++]);
    if(i >= argc)
    {
      fprintf(stderr,
              "%s: There is no v_name%d argument after u_name%d argument %s\n",
              myname, numvectors+1, numvectors+1, argv[i-1]);
      exit(1);
    }
    strcpy(vname[numvectors++],argv[i++]);
  }
/*****************************************************************************/






  /* check that low and high limits are low and high */
  if(xphigh < xplow)
  {
    fprintf(stderr,"%s: xphigh cannot be less than xplow\n",myname);
    exit(1);
  }
  if(yphigh < yplow)
  {
    fprintf(stderr,"%s: yphigh cannot be less than yplow\n",myname);
    exit(1);
  }
  if(xhigh < xlow)
  {
    fprintf(stderr,"%s: xhigh cannot be less than xlow\n",myname);
    exit(1);
  }
  if(yhigh < ylow)
  {
    fprintf(stderr,"%s: yhigh cannot be less than ylow\n",myname);
    exit(1);
  }


  /* get the header from the standard input */
  gethdr(stdin,inheader,HBMAX);

  /* make sure of float format */
  if (getfmt(inheader,HBMAX) != 'f')
  {
    fprintf(stderr,"%s: Input must be Float format\n",myname);
    exit(1);
  }

  /* check for x and y index fields; get field sizes */
  if ((fp = seekfld(inheader,HBMAX,'s',xname)) == NULL) {
    fprintf(stderr,"%s: no static field named %s\n",myname,xname);
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->dname1,xname) != 0)) {
    fprintf(stderr,"%s: %s is not an index field\n",myname,xname);
    exit(1);
  }
  xsize = fp->dsize1;
  if ((fp = seekfld(inheader,HBMAX,'s',yname)) == NULL) {
    fprintf(stderr,"%s: no static field named %s\n",myname,xname);
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->dname1,yname) != 0)) {
    fprintf(stderr,"%s: %s is not an index field\n",myname,xname);
    exit(1);
  }
  ysize = fp->dsize1;

  /* check for vector fields */
  for(i = 0; i < numvectors; i++)
  {
    if ((fp = seekfld(inheader,HBMAX,'v',uname[i])) == NULL)
    {
      fprintf(stderr,"%s: %s field not found\n",myname,uname[i]);
      exit(1);
    }
    cpyfld(&fpu,fp);
    if ((fp = seekfld(inheader,HBMAX,'v',vname[i])) == NULL)
    {
      fprintf(stderr,"%s: %s field not found\n",myname,vname[i]);
      exit(1);
    }
    cpyfld(&fpv,fp);
   
    /* make sure u and v fields have same structure */
    if(fpu.tsize != fpv.tsize ||
       fpu.dim != fpv.dim ||
       fpu.dsize1 != fpv.dsize1 ||
       fpu.dsize2 != fpv.dsize2 ||
       fpu.dsize3 != fpv.dsize3 ||
       fpu.dsize4 != fpv.dsize4 ||
       strcmp(fpu.dname1,fpv.dname1) ||
       strcmp(fpu.dname2,fpv.dname2) ||
       strcmp(fpu.dname3,fpv.dname3) ||
       strcmp(fpu.dname4,fpv.dname4) )
    {
      fprintf(stderr,"%s: vector fields %s and %s have different structures\n",
              myname,fpu.fname,fpv.fname);
      exit(1);
    }
  }

  /* get dx and dy; if not present, default to 1 */
  sprintf(pname,"d%s",xname);
  if(seekpar(inheader,HBMAX,pname,pvalue) == FAIL)
    dx = 1;
  else
    dx=atof(pvalue);
  sprintf(pname,"d%s",yname);
  if(seekpar(inheader,HBMAX,pname,pvalue) == FAIL)
    dy = 1;
  else
    dy=atof(pvalue);

  /* get bad data limits */
  if (seekpar(inheader,HBMAX,"bad",pvalue) != FAIL)
    bad=atof(pvalue);
  if (seekpar(inheader,HBMAX,"badlim",pvalue) != FAIL)
    badlim=atof(pvalue);

/* find xp and yp ranges */
  /* get x0 and y0 from parameter list or default to 0 if not present */
  sprintf(pname,"%s0",xname);
  if (seekpar(inheader,HBMAX,pname,pvalue) == FAIL)
    x0 = 0.0;
  else
    x0 = atof(pvalue);
  sprintf(pname,"%s0",yname);
  if (seekpar(inheader,HBMAX,pname,pvalue) == FAIL)
    y0 = 0.0;
  else
    y0 = atof(pvalue);
  
  /* if dx and/or dy are negative, make them positive and put x0/y0 at the
     other end of the data */
/*
  if(dx < 0)
  {
    x0 = x0+dx*(xsize-1);
    dx = -dx;
  }
  if(dy < 0)
  {
    y0 = y0+dy*(ysize-1);
    dy = -dy;
  }
*/

  /* set step value for xp/yp grid */
  step = fabs(dx) < fabs(dy) ? fabs(dx) : fabs(dy);

  /* convert ALLDEF range type to ALL with default arguments */
  if(xlimtype == ALLDEF || ylimtype == ALLDEF)
  {
    xlow = (dx >= 0) ? x0 : x0+dx*(xsize-1);
    xhigh = (dx >= 0) ? x0+dx*(xsize-1) : x0;
    ylow = (dy >= 0) ? y0 : y0+dy*(ysize-1);
    yhigh = (dy >= 0) ? y0+dy*(ysize-1) : y0;
  }
  if(xlimtype == ALLDEF)
    xlimtype = ALL;
  if(ylimtype == ALLDEF)
    ylimtype = ALL;
  

  /* on SET limit type, just use the xp limits the command line requested */ 
  if(xlimtype == SET)
  {
    xp0 = xplow;
    xpsize = ceil((xphigh-xplow)/step)+1;
  }

  /* on ALL limit type, use the min/max xp values of the domain corners */
  else if(xlimtype == ALL)
  {
    xplow = xp_xy(xlow,ylow);
    xplow = xplow > xp_xy(xlow,yhigh) ? xp_xy(xlow,yhigh) : xplow;
    xplow = xplow > xp_xy(xhigh,yhigh) ? xp_xy(xhigh,yhigh) : xplow;
    xplow = xplow > xp_xy(xhigh,ylow) ? xp_xy(xhigh,ylow) : xplow;

    xphigh = xp_xy(xlow,ylow);
    xphigh = xphigh < xp_xy(xlow,yhigh) ? xp_xy(xlow,yhigh) : xphigh;
    xphigh = xphigh < xp_xy(xhigh,yhigh) ? xp_xy(xhigh,yhigh) : xphigh;
    xphigh = xphigh < xp_xy(xhigh,ylow) ? xp_xy(xhigh,ylow) : xphigh;

    /* make xp grid points line up with origin */
    xplow = 0-ceil((0-xplow)/step)*step;

    xp0 = xplow;
    xpsize = ceil((xphigh-xplow)/step)+1;
  }

  /* on SET limit type, just use the yp limits the command line requested */
  if(ylimtype == SET)
  {
    yp0 = yplow;
    ypsize = ceil((yphigh-yplow)/step)+1;
  }

  /* on ALL limit type, use the min/max yp values of the domain corners */
  else if(ylimtype == ALL)
  {
    yplow = yp_xy(xlow,ylow);
    yplow = yplow > yp_xy(xlow,yhigh) ? yp_xy(xlow,yhigh) : yplow;
    yplow = yplow > yp_xy(xhigh,yhigh) ? yp_xy(xhigh,yhigh) : yplow;
    yplow = yplow > yp_xy(xhigh,ylow) ? yp_xy(xhigh,ylow) : yplow;

    yphigh = yp_xy(xlow,ylow);
    yphigh = yphigh < yp_xy(xlow,yhigh) ? yp_xy(xlow,yhigh) : yphigh;
    yphigh = yphigh < yp_xy(xhigh,yhigh) ? yp_xy(xhigh,yhigh) : yphigh;
    yphigh = yphigh < yp_xy(xhigh,ylow) ? yp_xy(xhigh,ylow) : yphigh;

    /* make yp grid points line up with origin */
    yplow = 0-ceil((0-yplow)/step)*step;

    yp0 = yplow;
    ypsize = ceil((yphigh-yplow)/step)+1;
  }


/* Build output header */
  nullhdr(outheader,HBMAX,"float");

  /* transfer comment section */
  copycmt(outheader,HBMAX,inheader,HBMAX);

  /* add comment */
  sprintf(string1,"%s: %s %s %s %s %s\n",argv[0],argv[1],argv[2],argv[3],
          argv[4],argv[5]);
  addcline(outheader,HBMAX,string1);
  for(i = 1; i <= (argc-1)/6; i++)
  {
    strcpy(string1," ");  
    for(j = i*6; j < argc && j < (i+1)*6; j++)
    {
      strcat(string1," ");
      strcat(string1,argv[j]);
    }
    strcat(string1,"\n");
    addcline(outheader,HBMAX,string1);
  }

  /* put xp0, yp0, dxp, and dyp parameters into output header if the dimension
     is not size 1 (and so will be removed) */
  if(xpsize > 1)
  {
    sprintf(pname,"%sp0",xname);
    sprintf(pvalue,"%f",xp0);
    addpar(outheader,HBMAX,pname,pvalue);
    sprintf(pname,"d%sp",xname);
    sprintf(pvalue,"%f",step);
    addpar(outheader,HBMAX,pname,pvalue);
  }
  if(ypsize > 1)
  {
    sprintf(pname,"%sp0",yname);
    sprintf(pvalue,"%f",yp0);
    addpar(outheader,HBMAX,pname,pvalue);
    sprintf(pname,"d%sp",yname);
    sprintf(pvalue,"%f",step);
    addpar(outheader,HBMAX,pname,pvalue);
  }

  /* pass all other parameters */
  sprintf(string1,"%s0",xname);
  sprintf(string2,"%s0",yname);
  sprintf(string3,"d%s",xname);
  sprintf(string4,"d%s",yname);
  i = 0;
  while(getpar(inheader,HBMAX,i++,pname,pvalue) == OK)
    if(strcmp(pname,string1) && strcmp(pname,string2) && 
              strcmp(pname,string3) && strcmp(pname,string4))
      addpar(outheader,HBMAX,pname,pvalue);



  /* sift through the field descriptions, replacing x and y with xp and yp
     and the corresponding dimension sizes throughout, (or removing them if
     a given dimension is size 1) and replacing vector fields u and v with up
     and vp; loop through static and variable sections */
  for(i = 0; i <= 1; i++)
  {
    type = i == 0 ? 's' : 'v';
    j = 0;
    while((fp = getfld(inheader,HBMAX,type,j++)) != NULL)
    {
      /* get current values */
      fname[0] = dname1[0] = dname2[0] = dname3[0] = dname4[0] = '\0';
      strcpy(fname,fp->fname);
      strcpy(dname1,fp->dname1);
      strcpy(dname2,fp->dname2);
      strcpy(dname3,fp->dname3);
      strcpy(dname4,fp->dname4);
      dsize1 = fp->dsize1;
      dsize2 = fp->dsize2;
      dsize3 = fp->dsize3;
      dsize4 = fp->dsize4;
      dim = fp->dim;

      /* replace dimension names and sizes */
      flag = 0;
      if(fp->dim >= 1 && !strcmp(dname1,xname))
      {
        sprintf(dname1,"%sp",xname);
        dsize1 = xpsize;
        flag++;
      }
      else if(fp->dim >= 1 && !strcmp(dname1,yname))
      {
        sprintf(dname1,"%sp",yname);
        dsize1 = ypsize;
        flag += 8;
      }
      if(fp->dim >= 2 && !strcmp(dname2,xname))
      {
        sprintf(dname2,"%sp",xname);
        dsize2 = xpsize;
        flag++;
      }
      else if(fp->dim >= 2 && !strcmp(dname2,yname))
      {
        sprintf(dname2,"%sp",yname);
        dsize2 = ypsize;
        flag += 8;
      }
      if(fp->dim >= 3 && !strcmp(dname3,xname))
      {
        sprintf(dname3,"%sp",xname);
        dsize3 = xpsize;
        flag++;
      }
      else if(fp->dim >= 3 && !strcmp(dname3,yname))
      {
        sprintf(dname3,"%sp",yname);
        dsize3 = ypsize;
        flag += 8;
      }
      if(fp->dim >= 4 && !strcmp(dname4,xname))
      {
        sprintf(dname4,"%sp",xname);
        dsize4 = xpsize;
        flag++;
      }
      else if(fp->dim >= 4 && !strcmp(dname4,yname))
      {
        sprintf(dname4,"%sp",yname);
        dsize4 = ypsize;
        flag += 8;
      }

      /* replace index field names */
      if(!strcmp(fname,xname))
      {
        sprintf(fname,"%sp",xname);
        flag = 16;
      }
      if(!strcmp(fname,yname))
      {
        sprintf(fname,"%sp",yname);
        flag = 32;
      }

      /* replace vector field names */
      for(k = 0; k < numvectors; k++)
      {
        if(!strcmp(fname,uname[k]))
          sprintf(fname,"%sp",uname[k]);
        if(!strcmp(fname,vname[k]))
          sprintf(fname,"%sp",vname[k]); 
      }

      /* remove size 1 xp or yp dimensions */
      if(xpsize == 1)
      {
        sprintf(string1,"%sp",xname);
        if(!strcmp(string1,dname1))
        {
          strcpy(dname1,dname2);
          strcpy(dname2,dname3);
          strcpy(dname3,dname4);
          dname4[0] = '\0';
          dsize1 = dsize2;
          dsize2 = dsize3;
          dsize3 = dsize4;
          dim--;
        }
        else if(!strcmp(string1,dname2))
        {
          strcpy(dname2,dname3);
          strcpy(dname3,dname4);
          dname4[0] = '\0';
          dsize2 = dsize3;
          dsize3 = dsize4;
          dim--;
        }
        else if(!strcmp(string1,dname3))
        {
          strcpy(dname3,dname4);
          dname4[0] = '\0';
          dsize3 = dsize4;
          dim--;
        }
        else if(!strcmp(string1,dname3))
        {
          dname4[0] = '\0';
          dim--;
        }
      }
      if(ypsize == 1)
      {
        sprintf(string1,"%sp",yname);
        if(!strcmp(string1,dname1))
        {
          strcpy(dname1,dname2);
          strcpy(dname2,dname3);
          strcpy(dname3,dname4);
          dname4[0] = '\0';
          dsize1 = dsize2;
          dsize2 = dsize3;
          dsize3 = dsize4;
          dim--;
        }
        else if(!strcmp(string1,dname2))
        {
          strcpy(dname2,dname3);
          strcpy(dname3,dname4);
          dname4[0] = '\0';
          dsize2 = dsize3;
          dsize3 = dsize4;
          dim--;
        }
        else if(!strcmp(string1,dname3))
        {
          strcpy(dname3,dname4);
          dname4[0] = '\0';
          dsize3 = dsize4;
          dim--;
        }
        else if(!strcmp(string1,dname3))
        {
          dname4[0] = '\0';
          dim--;
        }
      }

      /* die if a field had x but not y or vice versa */
      if(flag != 0 && flag != 9 && flag != 16 && flag != 32)
      {
        fprintf(stderr,"%s: field %s has %s or %s dimensions but not both!\n",
                myname,fname,xname,yname);
        exit(1);
      }

      /* write the new field unless it is a size 1 index field */
      if(!(flag == 16 && xpsize == 1) && !(flag == 32 && ypsize == 1))
        addfld(outheader,HBMAX,type,fname,fp->smul,fp->sadd,fp->bytes,
             dim,dname1,dsize1,dname2,dsize2,dname3,dsize3,dname4,dsize4);

    }
  }
  /* write the new header to the standard output */
  puthdr(stdout,outheader,HBMAX);


  /* prepare buffers */
  buffersizein = elemcnt(inheader,HBMAX,'s') > elemcnt(inheader,HBMAX,'v') ? 
                 elemcnt(inheader,HBMAX,'s') : elemcnt(inheader,HBMAX,'v');
  buffersizeout = elemcnt(outheader,HBMAX,'s') > elemcnt(outheader,HBMAX,'v') ? 
                  elemcnt(outheader,HBMAX,'s') : elemcnt(outheader,HBMAX,'v');
  inbuffer = getbuff(buffersizein);
  outbuffer = getbuff(buffersizeout);
  
 

/* Process data */
  /* loop through slices */
  type = 's';
  while(getslice(stdin,elemcnt(inheader,HBMAX,type),inbuffer) != EOF)
  {
    /* clear output slice buffer */
    for(i = 0; i < buffersizeout; i++)
      outbuffer[i] = bad;

    /* loop through fields */
    i = 0;
    while((fp = getfld(inheader,HBMAX,type,i++)) != NULL)
    {
      cpyfld(&fpin,fp);

      /* skip if the field has been deleted (size 1 index field) */
      sprintf(string1,"%sp",fpin.fname);
      if((fp = seekfld(outheader,HBMAX,type,fpin.fname)) != NULL || 
         (fp = seekfld(outheader,HBMAX,type,string1)) != NULL)
      {
        /* use the ouput field description that was just found != NULL */
        cpyfld(&fpout,fp);

        /* if it's an index field, fill it with the proper values */
        if(!strcmp(fpin.fname,xname))
          for(j = 0; j < xpsize; j++)
            outbuffer[fpout.ptr + j] = xp0 + step * j;
        else if(!strcmp(fpin.fname,yname))
          for(j = 0; j < ypsize; j++)
            outbuffer[fpout.ptr + j] = yp0 + step * j;
  
        /* if the x and y dimensions are not present, just copy the data */
        else if(strcmp(xname,fpin.dname1) && strcmp(xname,fpin.dname2) &&
                strcmp(xname,fpin.dname3) && strcmp(xname,fpin.dname4) ) 
        {
          for(j = 0; j < fpin.tsize; j++)
            outbuffer[fpout.ptr + j] = inbuffer[fpin.ptr + j];
        }
  
        /* otherwise, rotate and interpolate */
        else
        {
          /* loop through last (w) dimension */
          for(wloop = 0; wloop < wsize(&fpin); wloop++)
          {
            /* loop through next-to-last (z) dimension */
            for(zloop = 0; zloop < zsize(&fpin); zloop++)
            {
              /* loop through yp dimension */
              for(yloop = 0; yloop < ypsize; yloop++)
              {
                /* loop through xp dimension */
                for(xloop = 0; xloop < xpsize; xloop++)
                {
                  /* calculate coordinates */
                  xp = xp0 + xloop * step;
                  yp = yp0 + yloop * step;
                  x = x_xpyp(xp,yp);
                  y = y_xpyp(xp,yp);
                  xindex = (x-x0)/dx;
                  yindex = (y-y0)/dy;
  
                  /* interpolate */
                  *pointer(&fpout,outbuffer,xloop,yloop,zloop,wloop) =
                      interp(&fpin,inbuffer,xindex,yindex,zloop,wloop);
                }
              }
            }
          }
        }
      }
    }

    /* rotate vector fields in variable slice only */
    if(type == 'v')
    {
      for(i = 0; i < numvectors; i++)
      {
        /* get u/v field descriptions */
        sprintf(string1,"%sp",uname[i]);
        fp = seekfld(outheader,HBMAX,type,string1);
        cpyfld(&fpu,fp);
        sprintf(string1,"%sp",vname[i]);
        fp = seekfld(outheader,HBMAX,type,string1);
        cpyfld(&fpv,fp);

        /* loop through all data points */
        for(j = 0; j < fpu.tsize; j++)
        {
          /* get old (u and v) components */
          u = outbuffer[fpu.ptr + j];
          v = outbuffer[fpv.ptr + j];
        
          /* put new (up and vp) cpmponents */
          outbuffer[fpu.ptr + j] = (fabs(u)>badlim || fabs(v)>badlim) ? bad : 
                                   u*cos(theta)+v*sin(theta);
          outbuffer[fpv.ptr + j] = (fabs(u)>badlim || fabs(v)>badlim) ? bad :
                                   v*cos(theta)-u*sin(theta);
        }
      }
    }
    /* write slice */
    putslice(stdout,elemcnt(outheader,HBMAX,type),outbuffer);
    type = 'v';
  }
}

/* interpolation routine */
static float interp(fp,buffer,xindex,yindex,z,w)
struct field *fp;
float *buffer;
float xindex,yindex;
int w,z;
{
  int x1,x2,y1,y2;               /* borders of interpolation area */
  float bl, tl, tr, br;          /* values at corners (t=top, l=left, etc)*/
  float left, right, mid;        /* intermediate interp. results */

  /* find corners of interpolation area; return bad value if out of range */
  if((x1 = floor(xindex)) < 0 || (y1 = floor(yindex)) < 0 || 
          (x2 = ceil(xindex)) > xsize-1 || (y2 = ceil(yindex)) > ysize-1)
    return(bad);

  /* find values at corners */
  bl = *pointer(fp,buffer,x1,y1,z,w);
  tl = *pointer(fp,buffer,x1,y2,z,w);
  tr = *pointer(fp,buffer,x2,y2,z,w);
  br = *pointer(fp,buffer,x2,y1,z,w);

  /* return bad value if any of corners are bad */
  if(fabs(bl) >= badlim || fabs(tl) >= badlim ||
     fabs(br) >= badlim || fabs(tr) >= badlim)
    return(bad);

  /* interpolate along right and left sides */
  if(y2 == y1)
  {
    right = br;
    left = bl;
  }
  else
  {
    right = (yindex-y1)/(y2-y1)*(tr-br)+br;
    left  = (yindex-y1)/(y2-y1)*(tl-bl)+bl;
  }

  /* interpolate to the center */
  if(x2 == x1)
    mid = left;
  else
    mid = (xindex-x1)/(x2-x1)*(right-left)+left;

  return(mid);
}

/* sizes of other dimensions in a field */
static long zsize(fp)
struct field *fp;
{
  if(fp->dim == 2) return(1);
  if(strcmp(xname,fp->dname1) && strcmp(yname,fp->dname1)) return(fp->dsize1);
  if(strcmp(xname,fp->dname2) && strcmp(yname,fp->dname2)) return(fp->dsize2);
  if(strcmp(xname,fp->dname3) && strcmp(yname,fp->dname3)) return(fp->dsize3);
  fprintf(stderr,"Error in routine: zsize\n");
  exit(1);
}

static long wsize(fp)
struct field *fp;
{
  int blanks = 0;
  if(fp->dim < 4) return(1);
  if(strcmp(xname,fp->dname1) && strcmp(yname,fp->dname1)) blanks++;
  if(blanks = 2) return(fp->dsize1);
  if(strcmp(xname,fp->dname2) && strcmp(yname,fp->dname2)) blanks++;
  if(blanks = 2) return(fp->dsize2);
  if(strcmp(xname,fp->dname3) && strcmp(yname,fp->dname3)) blanks++;
  if(blanks = 2) return(fp->dsize3);
  if(strcmp(xname,fp->dname4) && strcmp(yname,fp->dname4)) blanks++;
  if(blanks = 2) return(fp->dsize4);
  fprintf(stderr,"Error in routine: wsize\n");
  exit(1);
}



/* Pointer into slice buffers through dimensions */
static float *pointer(fp,buffer,x,y,z,w)
struct field *fp;
float *buffer;
int x,y,z,w;
{
  char xpname[LINE], ypname[LINE];
  int xd = 3, yd = 3, zd = 0, wd = 0;               /* dimension numbers */
  long offset = 0;
  long step[4];
  long index[4];
  int i;

  step[0] = step[1] = step[2] = step[3] = 1;

  sprintf(xpname,"%sp",xname);
  sprintf(ypname,"%sp",yname);

  /* locate dimensions */
  if(!strcmp(fp->dname1,xname) || !strcmp(fp->dname1,xpname))
    xd = 0;
  if(!strcmp(fp->dname2,xname) || !strcmp(fp->dname2,xpname))
    xd = 1;
  if(!strcmp(fp->dname3,xname) || !strcmp(fp->dname3,xpname))
    xd = 2;
  if(!strcmp(fp->dname4,xname) || !strcmp(fp->dname4,xpname))
    xd = 3;
  if(!strcmp(fp->dname1,yname) || !strcmp(fp->dname1,ypname))
    yd = 0;
  if(!strcmp(fp->dname2,yname) || !strcmp(fp->dname2,ypname))
    yd = 1;
  if(!strcmp(fp->dname3,yname) || !strcmp(fp->dname3,ypname))
    yd = 2;
  if(!strcmp(fp->dname4,yname) || !strcmp(fp->dname4,ypname))
    yd = 3;

  while(zd == xd || zd == yd) zd++;
  while(wd == xd || wd == yd || wd == zd) wd++;

  index[xd] = x;
  index[yd] = y;
  index[zd] = z;
  index[wd] = w;

  if(fp->dim == 4)
  { 
    step[3] = 1;
    step[2] = fp->dsize4;
    step[1] = fp->dsize3 * step[2];
    step[0] = fp->dsize2 * step[1];
  }
  else if(fp->dim == 3)
  {
    step[2] = 1;
    step[1] = fp->dsize3;
    step[0] = fp->dsize2 * step[1];
  }
  else if(fp->dim == 2)
  {
    step[1] = 1;
    step[0] = fp->dsize2;
  }
  else
    step[0] = 1;

  for(i = 0; i < fp->dim; i++)
  {
    offset += index[i] * step[i];
  }

  return (&(buffer[fp->ptr + offset]));
}


/* Coordinate transformations */

static float xp_xy(x,y)
float x,y;
{
  return((x-xval)*cos(theta)+(y-yval)*sin(theta));
}

static float yp_xy(x,y)
float x,y;
{
  return(-(x-xval)*sin(theta)+(y-yval)*cos(theta));
}

static float x_xpyp(xp,yp)
float xp,yp;
{
  return(xp*cos(theta)-yp*sin(theta)+xval);
}

static float y_xpyp(xp,yp)
float xp,yp;
{
  return(xp*sin(theta)+yp*cos(theta)+yval);
}


/* field description copying utility */
static void cpyfld(fp1,fp2)
struct field *fp1, *fp2;
{
  /* copy values */
  strcpy(fp1->fname,fp2->fname);
  fp1->smul = fp2->smul;
  fp1->sadd = fp2->sadd;
  fp1->bytes = fp2->bytes;
  fp1->dim = fp2->dim;
  strcpy(fp1->dname1,fp2->dname1);
  strcpy(fp1->dname2,fp2->dname2);
  strcpy(fp1->dname3,fp2->dname3);
  strcpy(fp1->dname4,fp2->dname4);
  fp1->dsize1 = fp2->dsize1;
  fp1->dsize2 = fp2->dsize2;
  fp1->dsize3 = fp2->dsize3;
  fp1->dsize4 = fp2->dsize4;
  fp1->tsize = fp2->tsize;
  fp1->ptr = fp2->ptr;

  /* clear excess dimension names */
  if(fp1->dim < 4)
    fp1->dname4[0] = '\0';
  if(fp1->dim < 3)
    fp1->dname3[0] = '\0';
  if(fp1->dim < 2)
    fp1->dname2[0] = '\0';
  if(fp1->dim < 1)
    fp1->dname1[0] = '\0';
}
