/* cdfroll -- This program cyclically shifts values of variable slice
 * fields along a specified dimension in order to make a Galilean
 * transformation.
 */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "cdfhdr.h"

static char hbuff[HBMAX][LINE];   /* header buffer */
static float *sbuffs,*sbuffv;   /* slice buffers  */

main(argc,argv)
int argc;
char *argv[];
{
  long loop,poop,vpos,nshift; /* used in looping */
  char *rolldim,*timevar,*rollrate;    /* for command line arguments */
  float *dimptr,*timeptr;     /* pointers to dimension and time fields */
  float *fldptr,*tmpptr;      /* pointers to var fields and temporary */
  float rollratevar;          /* float value of rollrate */
  int repeat;                 /* 1 if last element repeats first */
  long dimsize;               /* effective size of dimension to be rolled */
  long actualdsize;           /* actual size of dimension to be rolled */
  float dstep;                /* grid size of selected dimension */
  float shift;               /* used in shift calculation */
  char cl[LINE],pvalue[LINE]; /* comment and parameter line data */
  long inst,inc,siz,where;    /* used with the subspace routine */
  long nsels,nvels;           /* number of static and variable elements */
  struct field *fp;           /* pointer to the field structure */

  /* check command line arguments */
  if (argc != 5) {
    fprintf(stderr,"Usage: cdfroll -[c|r] dimension timevar rate\n");
    exit(1);
  }
  if (strcmp(argv[1],"-c") == 0) repeat = 0;
  else if (strcmp(argv[1],"-r") == 0) repeat = 1;
  else {
    fprintf(stderr,"cdfroll: unrecognized flag %s\n",argv[1]);
    exit(1);
  }
  rolldim = argv[2];
  timevar = argv[3];
  rollrate = argv[4];
  rollratevar = atof(rollrate);

  /* get header from stdin */
  fprintf(stderr,"just before gethdr\n");
  gethdr(stdin,hbuff,HBMAX);
  fprintf(stderr,"just after gethdr\n");

  /* check format */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfroll: float format expected\n");
    exit(1);
  }

  /* check to see if dimension exists */
  if ((fp = seekfld(hbuff,HBMAX,'s',rolldim)) == NULL) {
    fprintf(stderr,"cdfroll: dimension %s doesn't exist\n",rolldim);
    exit(1);
  }
  else if ((fp->dim != 1) || (strcmp(fp->dname1,rolldim) != 0)) {
    fprintf(stderr,"cdfroll: %s is not a dimension\n",rolldim);
  }
  else if (repeat) {
    dimsize = fp->dsize1 - 1;
  }
  else {
    dimsize = fp->dsize1;
  }
  if (dimsize < 2) {
    fprintf(stderr,"cdfroll: effective dimension size < 2\n");
    exit(1);
  }
  actualdsize = fp->dsize1;

  /* make a temporary buffer for doing roll */
  tmpptr = getbuff(dimsize);

  /* check to see if timevar exists and is zero dimensional */
  if ((fp = seekfld(hbuff,HBMAX,'v',timevar)) == NULL) {
    fprintf(stderr,"cdfroll: field %s does not exist\n",timevar);
    exit(1);
  }
  if (fp->dim != 0) {
    fprintf(stderr,"cdfroll: time field %s must have zero dimensions\n",
	    timevar);
    exit(1);
  }

  /* add comment line */
  sprintf(cl,"cdfsmooth: %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  addcline(hbuff,HBMAX,cl);

  /* write header */
  puthdr(stdout,hbuff,HBMAX);

  /* get memory for slices */
  nsels = elemcnt(hbuff,HBMAX,'s');
  sbuffs = getbuff(nsels);
  nvels = elemcnt(hbuff,HBMAX,'v');
  sbuffv = getbuff(nvels);

  /* transfer static slice */
  getslice(stdin,nsels,sbuffs);
  putslice(stdout,nsels,sbuffs);

  /* get pointer to time and dimension fields and compute the grid step */
  timeptr = getptr(hbuff,HBMAX,sbuffv,'v','d',timevar);
  dimptr = getptr(hbuff,HBMAX,sbuffs,'s','d',rolldim);
  dstep = dimptr[1] - dimptr[0];   /* assume uniform grid */

  /* loop through the variable slices */
  while (getslice(stdin,nvels,sbuffv) != EOF) {

    /* compute the number of cells to shift data */
    shift = timeptr[0]*rollratevar/dstep;
    while (shift < 0) shift += (float)dimsize;
    nshift = (long)(shift + 0.5);   /* do integer conversion on pos float */
    while (nshift >= dimsize) nshift -= dimsize;

    /* loop through the variable fields */
    vpos = 0;
    while ((fp = getfld(hbuff,HBMAX,'v',vpos++)) != NULL) {
      fldptr = getptr(hbuff,HBMAX,sbuffv,'v','d',fp->fname);

      /* loop through instances of specified dimension in field */
      inst = 0;
      while (subspace1(fp,rolldim,inst++,&where,&inc,&siz) != FAIL) {

	/* decrement size by 1 if repeat */
	if (repeat) siz--;
	if (siz != dimsize) {
	  fprintf(stderr,"cdfroll: siz %d != dimsize %d -- panic!\n",
		  siz,dimsize);
	  exit(1);
	}
	
	/* shift data by specified amount */
	for (loop = 0; loop < siz; loop++) {
	  poop = loop - nshift;
	  if (poop < 0) poop += siz;
	  tmpptr[poop] = fldptr[where + inc*loop];
	}
	for (loop = 0; loop < siz; loop++) {
	  fldptr[where + inc*loop] = tmpptr[loop];
	}

	/* set repeated value if needed */
	if (repeat) {
	  fldptr[where + inc*(actualdsize - 1)] = fldptr[where];
	}
      }
    }

    /* output variable slice */
    putslice(stdout,nvels,sbuffv);
  }
  exit(0);
}
