#include <stdio.h>
#include "cdfhdr.h"

/*  Program SST -- To be used to collect Reynolds weekly sst data and make  */
/*  into a candis file with one week's data per variable slice.  The data   */
/*  include a start date, end date, number of days (7) and an index.  The   */
/*  temperature values are in a two dimensional variable called sst which   */
/*  depends on latitude and longitude.                                      */
/*                                                                          */
/*  First Version:  May 12, 1998 by James Fox                               */
/*                                                                          */
/*  Revised:                                                                */
/*  13 May 1998: Added VAX and FLOATSWAP tests for byte swapping routines   */
/*  so that this program will work on big endian machines.  Also made       */
/*  some minor cosmetic changes. -- D J Raymond                             */
/*  18 August 1998: Several changes: 1) Changed the ordering of lat and     */
/*  lon in the output so it plots right without the need for cdforder;      */
/*  2) Defined index fields lon and lat; 3) Fixed the output field names    */
/*  so that they aren't needed on the command line -- also made the         */
/*  name of the mask file mandatory; 4) Added robustness here and there;    */
/*  Add a Julian Day calculation relative to a specified date.              */

#define NLON 360  /* number of longitude points */
#define NLAT 180  /* number of latitude points */
#define LATLON (NLON*NLAT)  /* the number of coordinate points */
#define LON0 -179.5  /* beginning longitude */
#define LAT0 -89.5   /* beginning latitude */
#define DLON 1.  /* longitude step */
#define DLAT 1.  /* latitude step */
#define CLON0 "-179.5"  /* ascii beginning longitude */
#define CLAT0 "-89.5"  /* ascii beginning latitude */
#define CDLON "1."  /* ascii longitude step */
#define CDLAT "1."  /* ascii latitude step */

static long junk[1];        /* this is for throwing out xtra fortran bytes */
static long ln[8];          /* read the long data here */
static short isst[LATLON];   /* read the temp values here */

static char hbuff[HBMAX][LINE];    /* Candis header */

static char *startyear, *startmonth, *startday;
static char *endyear, *endmonth, *endday, *numdays, *jday;
static char *longitude, *latitude;
static char *temperature, *maskfile;  /* pointers to command line args */

static long nsbuff, nvbuff; 
static float *sbuff, *vbuff;       /* buffer stuff */

static float *yrst, *mst, *dst, *yrref, *ls;
static float *mref, *dref, *ndays;
static float *julday, *sst, *lon, *lat;           /* field pointers */  

long julian_date();

main(argc, argv)
int argc;
char *argv[];

{

  int i, j;
  int refyear, refmonth, refday;
  float lyear;
  float fsst[LATLON];              /* put the floating point temps here */
  char cline[LINE];               /* for the comment line */
  char a[2];
  float smul = 100.0;       
  float sadd = 0.0;           /* Part of the description of Candis fields */

  FILE *mask;   /* herein is contained the land-sea mask data */

  /* check arguments */

  if(argc != 5){
    fprintf(stderr,
	    "Usage: cdfsst refyear(4 digits) refmonth refday maskfile\n");
    exit(1);
  }
  refyear = atoi(argv[1]);
  refmonth = atoi(argv[2]);
  refday = atoi(argv[3]);
  maskfile = argv[4];

  /* prepare all the Candis related variables and structures to receive */
  /* the data */

  nullhdr(hbuff, HBMAX, "float");

  addpar(hbuff, HBMAX, "bad", "99999.0");
  addpar(hbuff, HBMAX, "badlim", "99998.0");

  addcline(hbuff, HBMAX, "cdfsst\n");
    
  addfld(hbuff, HBMAX, 's', "lsmask", smul, sadd, 'l', 2, "lon",
	 NLON, "lat", NLAT);
  addfld(hbuff, HBMAX, 's', "lon", smul, sadd, 'l', 1, "lon", NLON);
  addfld(hbuff, HBMAX, 's', "lat", smul, sadd, 'l', 1, "lat", NLAT);
  addfld(hbuff, HBMAX, 'v', "year1", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "month1", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "day1", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "refyear", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "refmonth", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "refday", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "numdays", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "jday", smul, sadd, 'l', 0);
  addfld(hbuff, HBMAX, 'v', "sst", smul, sadd, 'l', 2, "lon",
	 NLON, "lat", NLAT);

  nsbuff = elemcnt(hbuff, HBMAX, 's');
  nvbuff = elemcnt(hbuff, HBMAX, 'v');

  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuff);

  ls = getptr(hbuff, HBMAX, sbuff, 's', 'd', "lsmask");
  lon = getptr(hbuff, HBMAX, sbuff, 's', 'd', "lon");
  lat = getptr(hbuff, HBMAX, sbuff, 's', 'd', "lat");
  yrst = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "year1");
  mst = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "month1");
  dst = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "day1");
  yrref = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "refyear");
  mref = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "refmonth");
  dref = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "refday");
  ndays = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "numdays");
  julday = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "jday");
  sst = getptr(hbuff, HBMAX, vbuff, 'v', 'd', "sst");

  addpar(hbuff, HBMAX, "lat0", CLAT0);
  addpar(hbuff, HBMAX, "dlat", CDLAT);
  addpar(hbuff, HBMAX, "lon0", CLON0);
  addpar(hbuff, HBMAX, "dlon", CDLON); 

  for (i = 0; i < NLON; i++) lon[i] = LON0 + i*DLON;
  for (i = 0; i < NLAT; i++) lat[i] = LAT0 + i*DLAT;

  mask = fopen(maskfile, "r"); /* now read in the land-sea mask data */
  
  if (mask == NULL){
    fprintf(stderr, "cdfsst: Maskfile %s not found.\n", maskfile);
    exit(1);
  }
  
  a[1] = '\0';
  for(i = 0; i < NLAT; i++){
    for (j = 0; j < NLON; j++){
      if ((a[0] = fgetc(mask)) == EOF){
	fprintf(stderr,"cdfsst: Mask %s file is incomplete!\n", maskfile);
	exit(1);
      }
      ls[i + NLAT*j] = atoi(a);
    }
  }
 
  fclose(mask);

  puthdr(stdout, hbuff, HBMAX);
  putslice(stdout, nsbuff, sbuff);

  /* this is the big data read loop */

  while(fread(junk, 4, 1, stdin) != 0){

    /* read the important stuff */

    if (fread(ln, 4, 8, stdin) == 0){   /* this is the date-time stuff */
      fprintf(stderr, "cdfsst: File read error or unexpected end of file. ");
      fprintf(stderr, "This file probably has missing bytes.\n");  
      exit(1);
    }                                 
    
    if (FLOATSWAP) floatswap(8,(float*)ln); /* its big-endian */
    yrst[0] = ln[0];
    mst[0] = ln[1];
    dst[0] = ln[2];
    yrref[0] = refyear;
    mref[0] = refmonth;
    dref[0] = refday;
    ndays[0] = ln[6];

    /* patch for millenium problem */
    lyear = yrst[0] < 57. ? 2000. + yrst[0] : 1900. + yrst[0];
    julday[0] = julian_date((long)lyear, (long)mst[0], (long)dst[0]) -
      julian_date(refyear, refmonth, refday);
    
    if (fread(junk, 4, 1, stdin) == 0){  /* more formatting bytes */  
      fprintf(stderr, "cdfsst: File read error or unexpected end of file. ");
      fprintf(stderr, "This file probably has missing bytes.\n");
      exit(1);
    }
    
    if (fread(junk, 4, 1, stdin) == 0){  /* throw out the formatting */
      fprintf(stderr, "cdfsst: File read error or unexpected end of file. ");
      fprintf(stderr, "This file probably has missing bytes.\n");       
      exit(1);
    }
      
    if (fread(isst, 2, LATLON, stdin) == 0){  /* read the temperatures */
      fprintf(stderr, "cdfsst: File read error or unexpected end of file. ");
      fprintf(stderr, "This file probably has missing bytes.\n");
      exit(1);
    }

    if (VAX) shortswap(LATLON, isst);          /* they are big-endian */
    
    for(i = 0; i < LATLON; i++){
      fsst[i] = 0.01*(float)isst[i]; /* convert to float temperature */
    }
      
      
    for(i = 0; i < NLAT; i++){
      for(j = 0; j < NLON; j++){
	sst[i + NLAT*j] = fsst[NLON*i + j];
      }
    }
      
    putslice(stdout, nvbuff, vbuff);
    
    fread(junk, 4, 1, stdin);          /* yet more fortran garbage */
  }                                    /* bad fortran! bad! No biscuit! */

}
