/* thermo -- This is a thermodynamics package that tries to be as
 * accurate as warranted by real atmospheric data for all realistic
 * conditions.
$Id: thermo.c,v 1.2 1992/02/07 03:03:23 raymond Exp raymond $
 */
#include <math.h>
#include <stdio.h>
#include "thermo.h"

#define A 17.269                  /* constants for Tetens' formula */
#define B 35.86
#define C 610.78
#define D 273.16
#define LOOPOUT 200               /* max number of iteration loops */
#define QVMIN 1.e-8               /* smallest allowed mixing ratio */

/* invert the entropy equation -- use Newton's method
 */
void invent(p,s,qt,t,qv)

/* input */
float p;                         /* total pressure (Pa) */
float s;                         /* entropy (J/K) */
float qt;                        /* total water mixing ratio (g/g) */

/* output */
float *t;                        /* temperature (K) */
float *qv;                       /* vapor mixing ratio (g/g) */
{
  int i;
  float fs1,fs2,ts,dts,qvsat;
  ts = 200. + 1.e-3*p;
  for (i = 0; i < LOOPOUT; i++) {
    fs1 = ent(p,ts,qt) - s;
    fs2 = ent(p,ts - 1.,qt) - s;
    dts = fs1/(fs2 - fs1);
    ts += dts;
    if (fabs(dts) < .1) break;
    if (i == LOOPOUT - 1) {
      fprintf(stderr,"invent: loopout error: p, s, qt = %g %g %g\n",p,s,qt);
      exit(1);
    }
  }
  qvsat = mratio(p,ts);       /* saturation mixing ratio at computed temp */
  if (qvsat > qt) *qv = qt;       /* check for saturation */
  else *qv = qvsat;
  *t = ts;
}

/* compute entropy */
float ent(p,t,qt)
float p;                         /* total pressure (Pa) */
float t;                         /* temperature (K) */
float qt;                        /* total water mixing ratio (g/g) */
{
  float qs,qv,pvs,pd,s;
  pvs = es(t);     /* saturation vapor pressure */
  pd = p - pvs;    /* partial pressure of dry air at saturation */
  qs = qv = EPS*pvs/pd;     /* saturation mixing ratio */
  if (qt < qv) qv = qt;     /* check for saturation */
  pd = p*(1. - qv/(EPS + qv));   /* partial pressure of dry air */
/* this formula is consistent both with Iribarne and Godson and with
 * Kerry Emanuel's reversible theta-e -- I also derived it myself.
 */
  s = (CPD + CL*qt)*log(t/TR) - RD*log(pd/PR) +
       (LR + (CPV - CL)*(t - TR))*qv/t;
  if (qv > 0.) s -= qv*RV*log(qv/qs);
  return(s);
}

/* Tetens' formula for water saturation vapor pressure */
/*
float es(t)
float t;
{
  return(C*exp(A*(t - D)/(t - B)));
}
*/

/* Formula for water saturation vapor pressure consistent with
 * Clausius-Clapyron equation.
 */
float es(t)
float t;
{
  float l,lr;
  lr = LR;
/* uncomment next line to use ice adiabat below 0 C */
/*
  if (t < 273.15) lr += LFR;
*/
  l = lr - (CL - CPV)*(t - TR);
  return(C*pow(TR/t,(CL - CPV)/RV)*exp((lr/TR - l/t)/RV));
}

/* compute mixing ratio, given pressure and dewpoint */
float mratio(p,td)
float p;                          /* pressure (Pa) */
float td;                         /* dewpoint temperature (K) */
{
  float temp;
  temp = es(td);
  return(EPS*temp/(p - temp));
}
