#! /usr/bin/python
#
#
#------------------------------------------------------------------
HELP =    """
Script:ng3dvar_PrepModel.py \t

 usage: ng3dvar_PrepModel.py [help] -g GridFile -m ModelFile > stdout \t

Description: This script takes a file containing model velocity data
and re-navigates the horizontal velocity to the analysis 
grid define by the GridFile. The velocities are Earth relative, which is
what is expected by ng3dvar. For convenience, we also include these
velocities with respect to the storm reference frame.


Arguments:
-g GridFile is the file with the observations to be analyzed with
   ng3dvar. This file is typically created by ngradcart or nagauxcart, 
   and as such, it contains the analysis grid information.

-m ModelFile is the file with model data that we wish to use as a 
   background during the ng3dvar analysis of the GridFile data. 
   So far only GFS final analysis has been implemented.

Output:
It is sent to the standard output and it is ready for 
ng3dvar (option --pbg). It contains horizontal velocities defined
on the analysis grid, but only at locations where there are model 
data with in any given analysis grid cell. The vertical location is
found by converting the pressure level to height using the standard
atmosphere (see the function 'FixDomain' below). The velocities
are named vx and vy and they Earth relative. For convenience,
we have add these velocities with respect to the storm reference frame 
with the names vxs and vys.

Note that the ouput could be send to cdffill to fill out the horizontal
analysis grid with a linear interpolation. However, linear interpolation
in the vertical should be use with caution, because the spacing of
the levels is non-linear, as they come from the conversion of pressure
to height.

Carlos Lopez Carrillo\t
GRC,NMT\t
    """
#--------------------------------------------------------------------
# Modules
import os,sys, string
#import argparse
from numpy import *
from pycandis import *
from cdfclass import *
#--------------------------------------------------------------------
def Main(ComArg) :#{
  """ 
    Assumes lonlat grid and model data to be good data.
  """

  [Gfile,Mfile] = ProArg(ComArg); #Get Command Line Arguments

  G = GetCandis(Gfile);  #Get Analysis grid data

  F = FixDomain(G,Mfile);#Get Model data on the same Domain as grid data.

  ReNavigate(G,F);       #Model data to the Analysis Grid.

    
#}
#--------------------------------------------------------------------
def ProArg(ComArg) : #{
    """
    Quick and dirty check of command line parameters.
    I'd rather use argparse, but it is available starting with python 7.
    """

    USAGE = string.split(HELP,'\t')[1].lstrip('\n');

    if len(ComArg) != 5 : print " %s" % (USAGE); quit();
    reqopt = 0;
    n = len(ComArg);
    for i in range(1,n):#{
        if (ComArg[i] == '-g') : Gfile = ComArg[i+1]; reqopt = reqopt+1;
        if (ComArg[i] == '-m') : Mfile = ComArg[i+1]; reqopt = reqopt+1;

    if reqopt != 2 : print " %s" % (USAGE); quit();
            
    return [Gfile, Mfile];
#}
#--------------------------------------------------------------------
def GetCandis(fname) : #{ Read in candis file
    x = Candis();
    x.read(fname);
    x.close()

    return(x);
#}
#--------------------------------------------------------------------
def FixDomain(G,Mfile): #{

    Wfile = "ng3dvar_PrepModel-00.aux";

    PMSL  = 1000.0;   # sea level pressure (hPa) */
    TMSL  = 298.0;    # sea level temperature (K) */
    GAMMA = 5.0;      # temperature lapse rate (K/km) */
    GEE   = 9.8;      # acceleration of gravity (m/s^2) */
    RGAS  = 287.6;    # gas constant for air (J/kg/K) */
    z1    = TMSL/GAMMA;
    rmu   = (RGAS*GAMMA*0.001)/GEE;

    x0 = G.params['lon0'];
    dx = G.params['dlon'];
    Nx = len(G.sfields['lon'].data);
    xf = x0 + dx * (Nx -1)

    y0 = G.params['lat0'];
    dy = G.params['dlat'];
    Ny = len(G.sfields['lat'].data);
    yf = y0 + dy * (Ny -1)


    C = CDF();

    #Build alt = z1 * [1 - (P/Ps)^rmu]
    C.i("cdfmath '%f 1 pres %f / %f pow - * alt =' < %s" % 
        (z1,PMSL,rmu,Mfile));
    C.f("cdfinvindex %s %s" % ("pres","alt"));
    C.f("cdfextr -rs pres");
    C.f("cdfwindow lat %f %f" % (y0,yf));

    if x0 < 0 : # (+) lon to the East and (-) to the West
        ax0 = 360 + x0; 
        axf = 360 + xf; 
        C.f("cdfwindow lon %f %f" % (ax0,axf));
        C.f("cdfmath 'lon 360 - alon ='");
        C.f("cdfinvindex %s %s" % ("lon","alon"));
        C.f("cdfmath 'alon  lon ='");
        C.f("cdfinvindex %s %s" % ("alon","lon"));
        C.f("cdfextr -rs alon");
    else : # MonoIncreasing to the East
        C.f("cdfwindow lon %f %f" % (x0,xf));

    C.o("cdfextr jday hour vx vy > %s" % (Wfile));

    C.rs();
    #C.r();
    #C.d();
    
    F = GetCandis(Wfile);

    COMM = "rm -f %s" % (Wfile);
    run(COMM);

    return(F);

#}
#--------------------------------------------------------------------
def ReNavigate(G,F): #{
#
  EARTH_R = 6370.0;
  D2R = pi/180;

  # Parameters from Data file (On analysis grid)
  BAD     = G.params['bad'];
  BADLIM  = G.params['badlim'];
  reftime = G.params['reftime'];
  svx = G.params['svx'];
  svy = G.params['svy'];
  xg0 = G.params['lon0'];
  dgx = G.params['dlon'];
  yg0 = G.params['lat0'];
  dgy = G.params['dlat'];
  dgz = G.params['dz'];
  Ngx = len(G.sfields['lon'].data);
  Ngy = len(G.sfields['lat'].data);
  Ngz = len(G.sfields['z'].data);

  # Model Data
  xm0 = F.params['lon0'];
  ym0 = F.params['lat0'];
  xm  = F.sfields['lon'].data;
  ym  = F.sfields['lat'].data;
  zm  = F.sfields['alt'].data;
  jday = F.vfields['jday'].data;
  hour = F.vfields['hour'].data;
  vmx  = F.vfields['vx'].data;
  vmy  = F.vfields['vy'].data;
  Nmx = len(xm);
  Nmy = len(ym);
  Nmz = len(zm);
  monotime = 3.6*hour; #ks

  #navigate the location --all model values have the same time
  deltatime = monotime - reftime;
  deltax = svx*deltatime;
  deltay = svy*deltatime;
  deltaxRad = deltax/(EARTH_R*D2R);
  deltayRad = deltax/(EARTH_R*D2R);

  printit = 0; #4 Debugging
  if (printit == 1) :
     print "svx=%f" % (svx);
     print "svy=%f" % (svy);
     print "xg0=%f" % (xg0);
     print "yg0=%f" % (yg0);
     print "reftime=%f" % (reftime);
     print "-----";
     print "monotime=%f" % (monotime);
     print "jday=%f" % (jday);
     print "hour=%f" % (hour);
     print "xm0=%f" % (xm0);
     print "ym0=%f" % (ym0);
     print "Nmx=%f" % (Nmx);
     print "Nmy=%f" % (Nmy);
     print "Nmz=%f" % (Nmz);
     print "-----";
     print "deltatime=%f" % (deltatime);
     print "deltax=%f" % (deltax);
     print "deltay=%f" % (deltay);

  #quit();

  # Prepare Candis Output
  o = Candis();
  o.comments.append("ng3dvar_PrepModel.py");
  o.params  = G.params
  o.sfields = G.sfields;
  o.vfields['vx']   = Field(['lon', 'lat', 'z'], zeros([Ngx,Ngy,Ngz]))
  o.vfields['vy']   = Field(['lon', 'lat', 'z'], zeros([Ngx,Ngy,Ngz]))
  #
  #For convinience, we also include the model velocities with respect
  #To the reference frame moving with the storm.
  #
  o.vfields['vxs'] = Field(['lon', 'lat', 'z'], zeros([Ngx,Ngy,Ngz]))
  o.vfields['vys'] = Field(['lon', 'lat', 'z'], zeros([Ngx,Ngy,Ngz]))

  vx  = o.vfields['vx'].data;
  vy  = o.vfields['vy'].data;
  vxs  = o.vfields['vxs'].data;
  vys  = o.vfields['vys'].data;

  gnum = len(vx.flat);# = Ngx*Ngy*Ngz;
  sigma = zeros(gnum);

  for im in range(0,Nmx) :#{
    for jm in range(0,Nmy) :#{

      #Find location on the Analysis grid
      xloc = xm[im] - xg0 - deltaxRad/cos(ym[jm]*D2R);
      yloc = ym[jm] - yg0 - deltayRad;

      #avoid possible negative value truncation problems 
      ix = int(xloc/dgx + 1.5) -1;
      iy = int(yloc/dgy + 1.5) -1;
      
      #skip if out of space range or on very edge
      if (ix > 0 and ix < (Ngx -1) and iy > 0 and iy < (Ngy - 1)) :#{
        for km in range(0,Nmz) :#{

            zloc = zm[km];
            iz = int(zloc/dgz + 1.5) -1;

            if (iz > 0 and iz < (Ngz -1)) : #{

              #good location --increment values for average calculation
              #Assumes good data
              pg = iz + Ngz *(iy + Ngy * ix); 
              pm = km + Nmz *(jm + Nmy * im);

              #print "(Ngx,Ngy,Ngz) =  (%i,%i,%i)" % (Ngx,Ngy,Ngz);
              #print "(Nmx,Nmy,Nmz) =  (%i,%i,%i)" % (Nmx,Nmy,Nmz);
              #print "(ix,iy,iz,pg) =  (%i,%i,%i,%i)" % (ix,iy,iz,pg);
              #print "(im,jm,km,pm) =  (%i,%i,%i,%i)" % (im,jm,km,pm);

              sigma.flat[pg] += 1.;
              vx.flat[pg] += vmx.flat[pm];
              vy.flat[pg] += vmy.flat[pm];
            #}
        #} 
      #}
    #}
  #}


  #take average where data are good, set things to bad otherwise */
  for i in range(0,gnum) :#{
      if (sigma[i] > 0 ) :#{
        vx.flat[i] = vx.flat[i] / sigma.flat[i];
        vy.flat[i] = vy.flat[i] / sigma.flat[i];

        vxs.flat[i] = vx.flat[i] - svx;
        vys.flat[i] = vy.flat[i] - svy;
      else :
        vx.flat[i] = BAD;
        vy.flat[i] = BAD;
     #}
  #}

  o.write();
#}
#
#--------------------------------------------------------------------
#
# Shell interaction
#
if __name__ == '__main__' : #{
    na = len(sys.argv);
    if na == 2 and sys.argv[1] == 'help':
        print "%s" % (HELP);
    else :
        Main(sys.argv);
    #}
#}
