// 
// main -- Implements a Constrained Minimization (Penalty Method) 
//         using the unconstrained CG Algorithm of Stephen Nash.
// 

#include "ng3dvar.h"

//
// external Functions
//
double *Dgetbuff(long nelem);
int gpe(int Nx, int Ny, int Nz, int c, int i, int j, int k);
void CheckCLP(struct clpInfo *clp);
void ScaleTF(struct funInfo *finfo,int N,double *x,double *f,double *g);
double mc_L1norm(struct funInfo *finfo,double *x);
double scmc_L1norm(struct funInfo *finfo,double *x);
double RadMissFit_L2norm(struct funInfo *finfo,int N,double *x,double *g);
double BGFit_L2norm(struct funInfo *finfo,int N,double *x,double *g);
void fun_ws2f(struct clpInfo *clp,struct funInfo *finfo,double *vel,char *Ofile);
void SetInitalSol(struct funInfo *finfo,double *xx);
struct clpInfo *fun_cgeta(int argc,char **argv);
struct RadarDataInfo *SetRDInfo();
struct BGFdata *SetBGFdata(struct clpInfo *clp, struct gridInfo *gridi);

int cg_CONMINc(int N,double *X,double *F,double *G, struct funInfo *finfo,
	       int (*CALCFG)(void *,int,void *,void *,void *));

int calcfg(struct funInfo *finfo,int N, double *x, double *f, double *g);
void fun_writeModel(struct gridInfo *gridi, struct BGFdata *bgf);

struct funInfo *SetfunInfo(
			   struct clpInfo *clp, 
			   struct RadarDataInfo *rdi,
			   struct BGFdata *bgf,
			   struct gridInfo *gridi, 
			   struct cgInfo *cgi,
			   double PRA,double PMC,double PBC,double PHS,
			   double PVS,double PBG,double PXC);


//
//  
//
main(int argc,char **argv) {

  //
  // Define Variables and Set Default Values
  //
  float MCtol = 0.001; // MassContinuity Tolerance on the Main loop.
  int NGD =  3;        // Number of Grid Dimensions;
  int NFC =  3;        // Number of Field Components;

  // 
  // A positive value includes the associated term to the TargetFunction
  //
  double PRA = 1.0; // Scale for RadarMisffit --always one.
  double PMC = 1.0; // Scale for MassContinuity Term
  double PHS = 1.0; // Horizontal Smoothing Term
  double PVS = 0.0; // Vertical Smoothing Term
  double PBC = 0.0; // Scale for BoundaryConditions Term
  double PXC = 0.0; // ExtraConstrain
  double PBG = 0.0; // Background Term

  //double PP;          // PenaltyParameter
  //double PP0 = 100.0; //1e9; // PenaltyParameter 
  double GRP = 1.5 ;  //GrowRate of Penalty 
  int Nitera = 100;    //10; //1; Outer Loop Iterations.
  int ol_out = -1;    //No intermediate output
  int ini_out = -1;    //No output of initial solution
  int scale  = -1;    //Do not scale before entering outer loop.
  int cg_nfe = 5000;  //Max number of function evaluations in conmin

  double CC1;            //Convergence Criteria.
  double RadMissFit;     //RadarMissfit
  double BGFit;     //BackgroundFit
  double MCMax;          //MaxDivergence

  int Nxg,Nyg,Nzg,i,j,k,nn; // looping variables

  //Solution
  int Nx;            
  double *x0,*xx,*xg,xf; 
  double DX,DXQ,XSQ;
  char Ofile[MAXCHAR];

  // 
  // Structures to move information around
  //
  struct clpInfo *clp;       //command line parameters
  struct RadarDataInfo *rdi; //radar data
  struct BGFdata *bgf;       //Background field data
  struct funInfo *ifg;       //additional info required to evaluate the target fun
  struct gridInfo *gridi;    //Grid information
  struct cgInfo *cgi;        //information for cg

  int ifun,iter,nflag;  
  //double *TWA;

  //
  // Get Command Line Arguments
  //
  clp = fun_cgeta(argc,argv); //CheckCLP(clp);

  // 
  // Check if user wants to override defaults.
  // Honor User request, if any.
  //
  if ( (*clp).cg_nfe > 0 ) cg_nfe = (*clp).cg_nfe; 

  if ( (*clp).phs >= 0 )  PHS =  (*clp).phs;
  if ( (*clp).pvs >= 0 )  PVS =  (*clp).pvs;
  if ( (*clp).pbg >= 0 )  PBG =  (*clp).pbg;

  //The following test is to implement the un-enforcement of
  //the mass continuity. when PMC is negative the mass continuity
  //term is excluded from the target function. Nitera =1 forces
  //the main loop to quit regardless of the size of the mass continuity
  //term.
  if ( fabs((*clp).ol_p0)  < 0.000001 ) { PMC  = -1.0; Nitera = 1;}

  if ( (*clp).ol_p0  > 0.000001 ) PMC    = (*clp).ol_p0;
  if ( (*clp).ol_gr  > 0 ) GRP    = (*clp).ol_gr;
  if ( (*clp).ol_num > 0 ) Nitera = (*clp).ol_num;

  if ( (*clp).ol_out > 0 ) ol_out = (*clp).ol_out;
  if ( (*clp).sca > 0 )    scale  = (*clp).sca;

  //
  // Get Radar Data
  //
  rdi = SetRDInfo();

  //
  // Set Grid Info
  //
  //
  // Get space to hold the structures.
  //
  if((gridi =(struct gridInfo*)malloc(sizeof(struct gridInfo)*(2)))==NULL){
    fprintf(stderr,"gridInfo: cannot get enough memory\n"); exit(1);}
  (*gridi).Nd = NGD;
  (*gridi).Nc = NFC;
  (*gridi).Nx = (*rdi).Nx ;
  (*gridi).Ny = (*rdi).Ny ;
  (*gridi).Nz = (*rdi).Nz ;
  (*gridi).Np = (*gridi).Nx * (*gridi).Ny * (*gridi).Nz ;
  (*gridi).Ne = (*gridi).Np * (*gridi).Nc;

  (*gridi).sc = (*rdi).sc;
  (*gridi).Dx = (*rdi).Dx ;
  (*gridi).Dy = (*rdi).Dy ;
  (*gridi).Dz = (*rdi).Dz ;

  (*gridi).x0 = (*rdi).lon0 ;
  (*gridi).y0 = (*rdi).lat0 ;
  (*gridi).z0 = (*rdi).alt0 ;


  //
  // Set Up bgf structure if required
  //
  if (PBG > 0. ) bgf =  SetBGFdata(clp, gridi );

  //For testing Only write a file with the interpolated Model Velocities
  //fun_writeModel(gridi,bgf); exit(1);

  // 
  // SetUp cgInfo structure
  //
  if((cgi =(struct cgInfo*)malloc(sizeof(struct cgInfo)*(2)))==NULL){
    fprintf(stderr,"cgInfo: cannot get enough memory\n"); exit(1);}
  (*cgi).MXFUN = cg_nfe;
  (*cgi).eps = sqrt(FLT_EPSILON);  // Nash & Soffer pg.375;
  (*cgi).acc = FLT_EPSILON*100.0;  // Arithmetic is actually double.
                                   // tests in conmin are against acc
                                   // So in multiplying by 100, I'm 
                                   // relaxing the criteria.;
  (*cgi).WA  =  Dgetbuff(6*((*gridi).Ne)+2); //Auxiliar array for CG.
  //(*cgi).grid = gridi;

  //
  // SetUp Storage for the target field and its gradient.
  //
  Nx = (*gridi).Ne;
  xx = Dgetbuff(Nx+2);
  xg = Dgetbuff(Nx+2);
  x0 = Dgetbuff(Nx+2);

  // 
  // SetUp funInfo structure
  // 
  ifg = SetfunInfo(clp,rdi,bgf,gridi,cgi,PRA,PMC,PBC,PHS,PVS,PBG,PXC); 

  fprintf(stderr,"\t\t CG--BeginOfInitialization--- \n");


  //
  //Initialize Target Field to zero unless there is an input file
  // bringing the initial solution. 
  //
  SetInitalSol(ifg,x0);

  //
  // Now we look for a field that is closer the observations. 
  // We'll use this field as our initial point in the search for
  // the minimun of the Target Function
  //
  (*ifg).PRA = 1.0;
  (*ifg).PBC = 0.0;
  (*ifg).PMC = 0.0;
  (*ifg).PHS = 0.0;
  (*ifg).PVS = 0.0;
  (*ifg).PXC = 0.0;
  (*ifg).PBG = 0.0;

  cg_CONMINc(Nx,x0,&xf,xg,ifg,(int (*)(void*,int, void*,void*,void*))calcfg);

  //
  // Reset Vertical velocity at the boundaries to zero
  //
  Nxg = (*gridi).Nx;
  Nyg = (*gridi).Ny;
  Nzg = (*gridi).Nz;
  for (i = 0; i < Nxg; i++) {
    for (j = 0; j < Nyg; j++) {
      x0[gpe(Nxg,Nyg,Nzg,2,i,j,0)] = 0.0; 
      x0[gpe(Nxg,Nyg,Nzg,2,i,j,Nzg-1)] = 0.0; 
    }
  }


  fprintf(stderr,"CG-- EndOfInitialization--------------- \n");


  // 
  // Transfer Initial Sol to current Sol
  //
  for (i= 0; i< Nx; i++) {xx[i] = x0[i];}

  if ( ini_out > 0 ) {
    sprintf(Ofile,"initial.d");
    fprintf(stderr,"<%s>\n",Ofile);
    fun_ws2f(clp,ifg,xx,Ofile);
  }

  //
  // The Initialization is done. Now we honor the selection
  // of the terms to be included in the Target Function.
  //

  (*ifg).PRA = (*ifg).TFT0.PRA;
  (*ifg).PMC = (*ifg).TFT0.PMC;
  (*ifg).PHS = (*ifg).TFT0.PHS;
  (*ifg).PVS = (*ifg).TFT0.PVS;
  (*ifg).PXC = (*ifg).TFT0.PXC;
  (*ifg).PBC = (*ifg).TFT0.PBC;
  (*ifg).PBG = (*ifg).TFT0.PBG;

  // 
  // Set the Penalty parameter to its initial value and its Grow Rate.
  //
  //(*ifg).PP = PP0;

  //
  // Before solving, Scale Penalty terms
  // based on Function@currentSol. 
  //
  //HERE
  //if (scale > 0) {
  //  fprintf(stderr,"The scaling is not working yet\n");exit(1);
  //}

  // Scale only at the beginning 
  //if (scale > 0) ScaleTF(ifg,Nx,xx,&xf,xg); 

  //
  // Main Loop -- increases penalty 
  //
  for ( nn = 1; nn <= Nitera;  nn++) {
    // Scale every iteration
    if (scale > 0) ScaleTF(ifg,Nx,xx,&xf,xg);
     
    cg_CONMINc(Nx,xx,&xf,xg,ifg,(int (*)(void*,int, void*,void*,void*))calcfg);
  
    // Check convergence
    DXQ = 0.0;
    XSQ = 0.0;
    for (i= 0; i< Nx; i++) {
      XSQ += xx[i]*xx[i]; 
      DX   = (xx[i] -x0[i]);
      DXQ += (DX*DX);
    }
     
    CC1 = sqrt(DXQ)/(1.0 + sqrt(XSQ));

    // NOTE: The value of the following assesments is not scaled,
    // they are obtained assuming a weight of 1 for the term of
    // interest and zero for the others. 

    //
    //Get Penalty assestment of the observations.
    //
    RadMissFit = RadMissFit_L2norm(ifg,Nx,xx,xg);
    BGFit      = BGFit_L2norm(ifg,Nx,xx,xg);

    //
    //Get assestment of Mass Continuity. 
    //
    if ( (*ifg).sc > 0 ) MCMax = scmc_L1norm(ifg,xx);
    else MCMax = mc_L1norm(ifg,xx);

    fprintf(stderr,
	    "MainLoop: iter= %3i PP= %e CC1= %e MCMax= %e Dfit= %e BGfit= %e\n",
	    nn,(*ifg).PMC,CC1,MCMax,RadMissFit,BGFit);

    // 
    // Check Stopping Criteria
    //
    if ( fabs(MCMax) < MCtol ) {

      fprintf(stderr,
	      "MainLoopEndedAt: iter= %3i PP= %e CC1= %e MCMax= %e Dfit= %e BGfit= %e\n",
	      nn,(*ifg).PMC,CC1,MCMax,RadMissFit,BGFit);

      break;
    } else {
      // Save Current Solution
      for (i= 0; i< Nx; i++) {x0[i] = xx[i];}

      // OutPut Current solution
      if ( ol_out > 0 ) {
	sprintf(Ofile,"initial-%i.d",nn);
	fprintf(stderr,"<%s>\n",Ofile);
	fun_ws2f(clp,ifg,xx,Ofile);
      }

      //Grow Penalty
      (*ifg).PMC *= GRP;
   
      //GO try again
    }
  }

  fprintf(stderr,"TFE=%i\n",(*ifg).tfe); //exit(1);

  //
  // WriteSolution2StandarOutput 
  //
  fun_ws2f(clp,ifg,xx,"stdout");

  exit(0);
}
//-------------------------------------------------------------------
