// ng3dvar_tf_bgf.c -- This file contains the necessary functions
// to generate the contribution from the background field data to the 
// target function. (The target function is considered as a summation 
// of terms that measure the misfit of the current velocity field to 
// several conditions.) The externally visible function is
//
// double if_bgf(struct funInfo *finfo,int N,double *vel,double *g);
//
// if_bgf 
// is just an interface function that calls the internal
// function fun_bgf, which do the actual work. 
//
// fun_bgf 
// returns the contribution to the target function due to the misfit 
// between the current velocity field and the background field data. 
// It also calculates the contribution to the gradient of the target 
// function with respect to the velocity components due to the misfit 
// with the background field. These contributions are weighted by the
// current value of info.PBG (penalty associated with the misfit with 
// background field. Data from the model is obtained by SetBGFdata
//
//
// SetBGFdata 
// read the model data and store them in a structure.
// The background field data are model data obtained from the file 
// specified in the command line parameters. Those data are assumed to
// be interpolated into the analysis grid by the companion python scrip
// ng3dvar_PrepModle.py. So, the dimensions of the model grid axis 
// are supposed to be the same as those of the analysis grid and their
// names assumed to be lon,lat, z. The file should have the horizontal
// velocities in m/s, with names vx and vy and positive to the East and
// North. The velocity data is read only once and stored in a BGFdata 
// structure (see ng3dvar.h). 
//
// Carlos Lopez Carrillo
// Geophysical Research Center, NMT.
//

#include "ng3dvar.h"
#include "cdfhdr.h"
#define USAGE "should be called by function calcfg inside calcfg_if.c"


//
// Externally Visble Function
//
double Fbgf(struct funInfo *finfo,int N,double *vel,double *g);
//double if_bgf(struct funInfo *finfo,int N,double *vel,double *g);
struct BGFdata *SetBGFdata(struct clpInfo *clp, struct gridInfo *gridi);
//void fun_writeModel(struct gridInfo *gridi, struct BGFdata *bgf);

//
// External Functions
//
char *strcpy(char *dest, const char *src);
float *Fgetbuff(long nelem);
FILE *OpenFileW(char *FileName);

//
// Internal Functions.
//
double fun_bgf(struct funInfo *finfo, double *vel,double *cgf);
struct BGFdata *SetBGFdata(struct clpInfo *clp, struct gridInfo *gridi );


//
//
// Define this module interface
//
double Fbgf(struct funInfo *finfo,int N,double *vel,double *g){
//double if_bgf(struct funInfo *finfo,int N,double *vel,double *cgf)
//
// Return value -- contribution to the function from background data
// funInfo -- structure  defined in 
// f -- 1D vector that holds the velocity field on the grid.
// g -- 1D vector that holds the contribution from radar data to 
//      the gradient with respect to the velocity field components
//      of the function being minimized. 
// N -- is the size of the vectors f and g
//

double cf;


  cf = fun_bgf(finfo,vel,g);

  return(cf); 
}
//-------------------------------------------------------------------
double fun_bgf(struct funInfo *finfo, double *vel,double *cgf){

	int ig,jg,kg,pg,pgpNp,pgp2Np;
  int Nxg,Nyg,Nzg,Np;

  float bad,badlim;

	double Vmx,Vmy,Vmz;
	double Vx,Vy,Vz;
  double M1,M2,M3;
	double missfit,sum,sigma;
	double ZIP;
	double PBG;
	float a_x,a_y,a_z;
	double VerticalConstraint = 0; // 0=No; 1=Yes;

  Nxg = (*finfo).grid.Nx; 
  Nyg = (*finfo).grid.Ny;
  Nzg = (*finfo).grid.Nz;
	Np  = (*finfo).grid.Np;

  bad = (*finfo).bgf.bad;
  badlim = (*finfo).bgf.badlim;

	PBG = (*finfo).PBG; //This is actually a scale factor.
	                    //It is important only with the -s option.
											//Otherwise it is set to 1 when including 
											//the background term and to -1 when it is not 
	sigma = (*finfo).TFT0.BGsigma; //This is the real penalty

	sum   = 0.0;

  // loop on Data Grid
  for (kg = 0; kg < Nzg; kg++) {

   ZIP = 1.0; if ( (kg == 0) || (kg == (Nzg -1)) ) ZIP = 0.0; //HCB

   for (jg = 0; jg < Nyg; jg++) {
    for (ig = 0; ig < Nxg; ig++) {
       
      pg =  kg + (jg + ig * Nyg) * Nzg;
			pgpNp = pg+Np;
			pgp2Np = pg+2*Np;

      Vmx = (*finfo).bgf.vx[pg];
      Vmy = (*finfo).bgf.vy[pg];
      Vmz = (*finfo).bgf.vz[pg];

			//Misffit
			//The 3 components are totally independent
			M1 = M2 = M3= 0;
      if ( fabs(Vmx) < badlim ) { M1 = (vel[pg]     - Vmx)/sigma; }
      if ( fabs(Vmy) < badlim ) { M2 = (vel[pgpNp]  - Vmy)/sigma; }
      if ( fabs(Vmz) < badlim ) { M3 = (vel[pgp2Np] - Vmz)/sigma; }
			M3 *= VerticalConstraint;
               
      missfit = (M1*M1 + M2*M2 + M3*M3);

      if (missfit < 0 ) {
          fprintf(stderr,"calcfg_ngrad@fun_ngradX:\n");
          fprintf(stderr,"missfit = %g ByE!!\n",missfit);
          exit(1);
      }

        //
        // make appropiate contributions to the gradient. 
        //
        //a_x =  PBG * M1; 
        //a_y =  PBG * M2;
        //a_z =  PBG * M3;
        //a_x =  M1/sigma;
        //a_y =  M2/sigma;
        //a_z =  M3/sigma;
        a_x =  PBG * M1/sigma;
        a_y =  PBG * M2/sigma;
        a_z =  PBG * M3/sigma;

        cgf[pg]     += a_x;
        cgf[pgpNp]  += a_y;
        cgf[pgp2Np] += (ZIP * a_z);


        // contribution to functional.
        //sum   += (PBG * missfit);
        sum   += (missfit);

    } //i-loop
   } //j-loop
  } //k-loop

  return((double) 0.5 * PBG * sum); 
  
}
//
//
//
//-------------------------------------------------------------------
//
//
//
struct BGFdata *SetBGFdata(struct clpInfo *clp, struct gridInfo *gridi ){
// 
//
// Carlos Lopez Carrillo
// Geophysical Research Center, NMT.
//
#define LOND "lon"     //longitude degrees [0,359]
#define LATD "lat"     //latitude degrees [-90,90]
#define ALTD "z"       //altitude km 
#define U1   "vx"      //zonal velocity (m/s)
#define U2   "vy"      //meridional velocity (m/s)

FILE *fh;
static char hbin[MAXCHAR][MAXLINES]; //Header Buffer
char pline[MAXCHAR];
long nsbin,nvbin;
static float *sbin,*vbin;
float bad, badlim;
int Nmx,Nmy,Nmz;            // axis sizes and total # of points
float *xm,*ym,*zm;
float xm0,ym0,zm0;
float dmx,dmy,dmz;
float *vmx, *vmy;        // velocities (m/s,m/s,Pa/s)
static  struct field *fpx,*fpy,*fpz,*fpaux;
int ii,Np,stopit;

struct BGFdata *bgf;
char *bgf_fname;

// Get space to hold background field
//
  if((bgf =(struct BGFdata*)malloc(sizeof(struct BGFdata)*(2))) ==NULL){
  	 fprintf(stderr,"SetBGFdata: cannot get enough memory\n"); exit(1);}

	Np = (*gridi).Np;
  (*bgf).vx = Fgetbuff(Np);
  (*bgf).vy = Fgetbuff(Np);
  (*bgf).vz = Fgetbuff(Np);
	
	//Get the Background velocity field 
	bgf_fname = (*clp).bgf_fname;

  // Open GFS_FA File
  //
  if ((fh = fopen(bgf_fname,"r")) == NULL) {
    fprintf(stderr,"can't open BackgroudData file %s\n",bgf_fname);
    exit(1);
  }

  // 
  // get header 
  // 
  gethdr(fh,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"make_header:format expected\n");
    exit(1);
  }

  //
  // Get Some Header Info.
  //
  if (seekpar(hbin,HBMAX,"bad",pline) == OK) bad = atof(pline);
  else                                       bad = BAD; 
	if (seekpar(hbin,HBMAX,"badlim",pline) == OK) badlim = atof(pline);
  else                                          badlim = BADLIM;

  // 
  // Get Buffer Sizes
  //
  nsbin  = elemcnt(hbin,HBMAX,'s');
  nvbin  = elemcnt(hbin,HBMAX,'v');

  //
  // Get Memory to hold data.
  //
  sbin = Fgetbuff(nsbin);
  vbin = Fgetbuff(nvbin);

	//
  // Read in static buffer.
  // 
  getslice(fh,nsbin,sbin);

  xm = getptr2(hbin,HBMAX,sbin,'s','d',LOND,&fpx);
	Nmx = fpx->dsize1;
	xm0 = xm[0];
	dmx = xm[1] - xm[0];

  ym = getptr2(hbin,HBMAX,sbin,'s','d',LATD,&fpy);
	Nmy = fpy->dsize1;
	ym0 = ym[0];
	dmy = ym[1] - ym[0];

  zm = getptr2(hbin,HBMAX,sbin,'s','d',ALTD,&fpz);
	Nmz = fpz->dsize1;
	zm0 = zm[0];
	dmz = zm[1] - zm[0];

	//
	//check dimensions for compatibility with our analysis grid
	//
	if ( Nmx != (*gridi).Nx ) { stopit = 1;
     fprintf(stderr,"NxModel = %i != NxGrid = %i\n",Nmx,(*gridi).Nx); }
	if ( Nmy != (*gridi).Ny ) { stopit = 1;
     fprintf(stderr,"NyModel = %i != NyGrid = %i\n",Nmy,(*gridi).Ny); }
	if ( Nmz != (*gridi).Nz ) { stopit = 1;
     fprintf(stderr,"NzModel = %i != NzGrid = %i\n",Nmz,(*gridi).Nz); }

	if ( fabs(dmx - (*gridi).Dx)  > 1e-4) { stopit = 1;
     fprintf(stderr,"dxModel = %f != dxGrid = %f\n",dmx,(*gridi).Dx); }
	if ( fabs(dmy - (*gridi).Dy) > 1e-4) { stopit = 1;
     fprintf(stderr,"dyModel = %f != dyGrid = %f\n",dmy,(*gridi).Dy); }
	if ( fabs(dmz - (*gridi).Dz) > 1e-4) { stopit = 1;
     fprintf(stderr,"dzModel = %f != dzGrid = %f\n",dmz,(*gridi).Dz); }

	if ( fabs(xm0 - (*gridi).x0) > 1e-4) { stopit = 1;
     fprintf(stderr,"x0Model = %f != x0Grid = %f\n",xm0,(*gridi).x0); }
	if ( fabs(ym0 - (*gridi).y0) > 1e-4) { stopit = 1;
     fprintf(stderr,"y0Model = %f != y0Grid = %f\n",ym0,(*gridi).y0); }
	if ( fabs(zm0 - (*gridi).z0) > 1e-4) { stopit = 1;
     fprintf(stderr,"z0Model = %f != z0Grid = %f\n",zm0,(*gridi).z0); }

  if (stopit == 1) exit(1);

  //
  // Data is assumed to come all in one slice, get it
  //
  getslice(fh,nvbin,vbin);

  // 
  //Get Relevant Pointers from the Var slice
  //
  vmx = getptr(hbin,HBMAX,vbin,'v','d',U1);
  vmy = getptr(hbin,HBMAX,vbin,'v','d',U2);

  //
  // Close Model File
  //
  // close(fh);

	//set bgfmd 
	(*bgf).bad = bad;
	(*bgf).badlim = badlim;
	//transfer horizontal velocities
	for (ii = 0; ii < Np; ii++){ 
					(*bgf).vx[ii] = vmx[ii];
					(*bgf).vy[ii] = vmy[ii];
					(*bgf).vz[ii] = bad; //for now we donot include vz;
	}
  //
  //CleanUp.
  //
  // getptr2 allocate space for the structures. Since we do not need
  // them any more, we'll dispose of them.
  //
  Dfreebuff(fpx); 
  Dfreebuff(fpy); 
  Dfreebuff(fpz); 
  Dfreebuff(sbin); 
  Dfreebuff(vbin); 

	return(bgf);
}
//
//---------------------------------------------------------------------
//
//
// For testing Purposes
//
//-----------------------------------------------------------------
//
void fun_writeModel(struct gridInfo *gridi, struct BGFdata *bgf){
int Nx,Ny,Nz,Np;        /* grid dimensions */
int gloop;
float Dx,Dy,Dz;
float x0,y0,z0;
float *xg,*yg,*zg;
float *vx,*vy,*vz;

//Candis Stuff
char hbout[HBMAX][LINE];
float *sbout,*vbout;
long nsbout,nvbout;
char pline[LINE];
FILE *ofh;
char Ofile[] = "ng3dvar_modelItest.cdf";

ofh = OpenFileW(Ofile);

  Np = (*gridi).Np;
  Nx = (*gridi).Nx;
  Ny = (*gridi).Ny;
  Nz = (*gridi).Nz;
  Dx = (*gridi).Dx;
  Dy = (*gridi).Dy;
  Dz = (*gridi).Dz;
  x0 = (*gridi).x0;
  y0 = (*gridi).y0;
  z0 = (*gridi).z0;

  vx = (*bgf).vx;
  vy = (*bgf).vy;
  vz = (*bgf).vz;

//Make Header
	nullhdr(hbout,HBMAX,"float");

  /* comments */
	sprintf(pline,"TestInterpolation\n");
	addcline(hbout,HBMAX,pline);

  //parameters
	sprintf(pline,"%g",BAD);
	addpar(hbout,HBMAX,"bad",pline);
	sprintf(pline,"%g",BADLIM);
	addpar(hbout,HBMAX,"badlim",pline);
	sprintf(pline,"%g",x0);
	addpar(hbout,HBMAX,"lon0",pline);
	sprintf(pline,"%g",y0);
	addpar(hbout,HBMAX,"lat0",pline);
	sprintf(pline,"%g",z0);
	addpar(hbout,HBMAX,"z0",pline);
	sprintf(pline,"%g",Dx);
	addpar(hbout,HBMAX,"dlon",pline);
	sprintf(pline,"%g",Dy);
	addpar(hbout,HBMAX,"dlat",pline);
	sprintf(pline,"%g",Dz);
	addpar(hbout,HBMAX,"dz",pline);
	sprintf(pline,"%i",Nx);
	addpar(hbout,HBMAX,"Nxn",pline);
	sprintf(pline,"%i",Ny);
	addpar(hbout,HBMAX,"Nyn",pline);
	sprintf(pline,"%i",Nz);
	addpar(hbout,HBMAX,"Nzn",pline);

  //add index fields 
	addfld(hbout,HBMAX,'s',"lon",1000.,0.,'l',1,"lon",Nx);
	addfld(hbout,HBMAX,'s',"lat",1000.,0.,'l',1,"lat",Ny);
	addfld(hbout,HBMAX,'s',"z",1000.,0.,'l',1,"lat",Nz);
				  
	//add new fields 
  addfld(hbout,HBMAX,'v',"vmx",10.,0.,'s',3,"lon",Nx,"lat",Ny,"z",Nz);
  addfld(hbout,HBMAX,'v',"vmy",10.,0.,'s',3,"lon",Nx,"lat",Ny,"z",Nz);
  addfld(hbout,HBMAX,'v',"vmz",10.,0.,'s',3,"lon",Nx,"lat",Ny,"z",Nz);
										  
  //write output header
	puthdr(ofh,hbout,HBMAX);


// Static Slice
//
//
//allocate memory for static fields 
 nsbout = elemcnt(hbout,HBMAX,'s');
 sbout  = Fgetbuff(nsbout);
					  
 //Get Pointer to index fields and fill them up
 xg = getptr(hbout,HBMAX,sbout,'s','d',"lon");
 yg = getptr(hbout,HBMAX,sbout,'s','d',"lat");
 zg = getptr(hbout,HBMAX,sbout,'s','d',"z");
 for (gloop = 0; gloop < Nx; gloop++) xg[gloop] = x0 + (float)gloop * Dx;
 for (gloop = 0; gloop < Ny; gloop++) yg[gloop] = y0 + (float)gloop * Dy;
 for (gloop = 0; gloop < Nz; gloop++) zg[gloop] = z0 + (float)gloop * Dz;

 //write static slice 
 putslice(ofh,nsbout,sbout);


//Variable Slice 
//
//allocate memory for static & variable fields 
  nvbout = elemcnt(hbout,HBMAX,'v');
  vbout  = Fgetbuff(nvbout);

  //Get Pointer to var fields and fill them up
  vx = getptr(hbout,HBMAX,vbout,'v','d',"vmx");
	vy = getptr(hbout,HBMAX,vbout,'v','d',"vmy");
	vz = getptr(hbout,HBMAX,vbout,'v','d',"vmz");

  for (gloop = 0; gloop < Np; gloop++){ 
        vx[gloop] = (*bgf).vx[gloop];
        vy[gloop] = (*bgf).vy[gloop];
        vz[gloop] = (*bgf).vz[gloop];
  }

  //write variable slice 
	putslice(ofh,nvbout,vbout);

  // CleanUp
  if (Ofile) fclose(ofh);

  Dfreebuff(sbout);
  Dfreebuff(vbout);

}
