// calcfg_ngrad -- generates the contribution from radar data to the 
// target funtion. It works with p3 and eldora?
//
// The target function is considered as a a sumation of terms that
// measure the misfit of the current velocity field to several conditions.
//
// calcfg_ngrad returns the contribution to the target function due
// to the misfit between the current velocity field and the radar data.
// It also calculates the contribution to the gradient of the target
// function with respect to the velocity components due to the misfit
// with the radar data. 
//
// The Input radar data is actually the systhesis data reduced by
// ngradcart. It is expected satisfy some conditions:
// 1) The velocity information is expected on a regular grid
//    defined by 3 indexes (x,y,z).
// 2) The velocities are actually along the eigen-directions
//    at each grid point, and vectors defining the eigen-directions
//    are part of the velocity information at each grid point.
// 3) Also the correct weighting factor for each velocity component 
//    in the eigen frame is expected to be part of the info at each grid.
// 5) velocity units are expected in m/s.
//
//
// Carlos Lopez Carrillo
// Geophysical Research Center, NMT.
//

#include "ng3dvar.h"
#include "cdfhdr.h"
#define USAGE "should be called by function calcfg inside calcfg_if.c"

/* assumed names for Data Info */
#define LOND "lon0" 
#define LATD "lat0"
#define ALTD "alt0"

#define U1 "u1" 
#define U2 "u2" 
#define U3 "u3"
#define A1 "a1"
#define A2 "a2"
#define A3 "a3"
#define E1x "e1x"
#define E1y "e1y"
#define E1z "e1z"
#define E2x "e2x"
#define E2y "e2y"
#define E2z "e2z"
#define E3x "e3x"
#define E3y "e3y"
#define E3z "e3z"
#define DBZ "dbz"

//
// Externally Visble Function
//
double if_ngrad(struct funInfo *finfo,int N, 
                double *ndat,double *vel,double *g);

struct RadarDataInfo *SetRDInfo();

void fun_wpTVZ(struct funInfo *finfo);
void fun_wpTVZx(struct funInfo *finfo);
void fun_rho(struct funInfo *finfo);
void CheckRHOS(struct funInfo *finfo);

//
// External Functions
//
char *strcpy(char *dest, const char *src);
float *Fgetbuff(long nelem);
FILE *OpenFileW(char *FileName);

//
// Internal Functions.
//
double fun_ngradX(struct funInfo *finfo,
                 double *vel,double *cgf,double *ndat);

static void fun_weight(struct funInfo *finfo,int ix,int iy, int iz,
                float xloc,float yloc,float zloc,int *np,float *we);
//
//
// Define this module interface
//
double if_ngrad(struct funInfo *finfo,int N, 
                double *ndat, double *vel,double *cgf){
//
// Return value -- contribution to the function from radar gridded data
// funInfo -- structure  defined in mgm.h and populated in mgm_user_if.c
// f -- 1D vector that holds the velocity field on the grid.
// g -- 1D vector that holds the contribution from radar data to 
//      the gradient with respect to the velocity field components
//      of the function being minimized. 
// N -- is the size of the vectors f and g
// ndat -- on output will have the number of radar grid points
//         that contain good velocities contributing to the f and g.
//

double cf;

  cf = fun_ngradX(finfo,vel,cgf,ndat);

  return(cf); 
}
//-------------------------------------------------------------------
double fun_ngradX(struct funInfo *finfo, 
                 double *vel,double *cgf,double *ndat){

	int ig,jg,kg,pg,pgpNp,pgp2Np;
  int Nxg,Nyg,Nzg,Np;

  float bad,badlim;
  float e1x,e1y,e1z;
  float e2x,e2y,e2z;
  float e3x,e3y,e3z;
  float u1,u2,u3;
  float a1,a2,a3;
  float wpTV;

	double Vpx,Vpy,Vpz;
	double Vp1,Vp2,Vp3;
  double a_x,a_y,a_z;
  double M1,M2,M3;
	double missfit,sum;
	double ZIP;

  Nxg = (*finfo).grid.Nx; 
  Nyg = (*finfo).grid.Ny;
  Nzg = (*finfo).grid.Nz;
	Np  = (*finfo).grid.Np;

  bad = (*finfo).rd.bad;
  badlim = (*finfo).rd.badlim;

	sum   = 0.0;
	*ndat = 0.0;

  // loop on Data Grid
  for (kg = 0; kg < Nzg; kg++) {

   ZIP = 1.0; if ( (kg == 0) || (kg == (Nzg -1)) ) ZIP = 0.0; //HCB

   for (jg = 0; jg < Nyg; jg++) {
    for (ig = 0; ig < Nxg; ig++) {
       
      pg =  kg + (jg + ig * Nyg) * Nzg;
			pgpNp = pg+Np;
			pgp2Np = pg+2*Np;

      wpTV = (*finfo).wtZ[pg]; 

      u1 = (*finfo).rd.u1[pg];
      u2 = (*finfo).rd.u2[pg];
      u3 = (*finfo).rd.u3[pg];

      a1 = (*finfo).rd.a1[pg];
      a2 = (*finfo).rd.a2[pg];
      a3 = (*finfo).rd.a3[pg];

      e1x = (*finfo).rd.e1x[pg];
      e1y = (*finfo).rd.e1y[pg];
      e1z = (*finfo).rd.e1z[pg];

      e2x = (*finfo).rd.e2x[pg];
      e2y = (*finfo).rd.e2y[pg];
      e2z = (*finfo).rd.e2z[pg];

      e3x = (*finfo).rd.e3x[pg];
      e3y = (*finfo).rd.e3y[pg];
      e3z = (*finfo).rd.e3z[pg];

      //if ( ( fabs(wpTV) < badlim ) &&
      if ( ( fabs(u1) < badlim ) &&
           ( fabs(u2) < badlim ) &&
           ( fabs(u3) < badlim ) &&
           ( fabs(a1) < badlim ) &&
           ( fabs(a2) < badlim ) &&
           ( fabs(a3) < badlim ) &&
           ( fabs(e1x) < badlim ) &&
           ( fabs(e1y) < badlim ) &&
           ( fabs(e1z) < badlim ) &&
           ( fabs(e2x) < badlim ) &&
           ( fabs(e2y) < badlim ) &&
           ( fabs(e2z) < badlim ) &&
           ( fabs(e3x) < badlim ) &&
           ( fabs(e3y) < badlim ) &&
           ( fabs(e3z) < badlim ) 
         ) {
               
        Vpx = vel[pg];
        Vpy = vel[pgpNp];
        Vpz = vel[pgp2Np];
        Vpz += wpTV; 

				//Projection over EigenDirections.
        Vp1 = Vpx * e1x + Vpy * e1y + Vpz * e1z;
        Vp2 = Vpx * e2x + Vpy * e2y + Vpz * e2z;
        Vp3 = Vpx * e3x + Vpy * e3y + Vpz * e3z;

        //
        // weighted contribution to missfit.
        //
        M1 = Vp1 - u1; 
        M2 = Vp2 - u2;
        M3 = Vp3 - u3;

        missfit = a1 * M1*M1 + a2 * M2*M2 + a3 * M3*M3;

        if (missfit < 0 ) {
          fprintf(stderr,"calcfg_ngrad@fun_ngradX:\n");
          fprintf(stderr,"missfit = %g ByE!!\n",missfit);
          exit(1);
        }

        //
        // make ppropiate contributions to the gradient. 
        //
        a_x = a1 * e1x * M1 + a2 * e2x * M2 + a3 * e3x * M3;
        a_y = a1 * e1y * M1 + a2 * e2y * M2 + a3 * e3y * M3;
        a_z = a1 * e1z * M1 + a2 * e2z * M2 + a3 * e3z * M3;

        cgf[pg]     += a_x;
        cgf[pgpNp]  += a_y;
        cgf[pgp2Np] += (ZIP * a_z);


        // contribution to functional.
        sum   += missfit;
        *ndat += 1.0;

      } //good data
    } //i-loop
   } //j-loop
  } //k-loop

  return((double) 0.5*sum); 
  
} 
//-------------------------------------------------------------------

struct RadarDataInfo *SetRDInfo(){
// 
// SetRDInfo -- gets all Data info required to proccess the data. This info
// is incorporated in the RadarDataInfo structure, which will later be 
// attached to the funInf structure, so it will be available for 
// the function proccessing. 
//
// Expects that funinfo already has allocated the space to hold the
// header.
//
// This functions will handel the allocation of buffers required to 
// read in the radar data. Then it reads in the data just once.
//
// It Assumes there is only one variable slice in the input candis file.
//
// Arguments
// funInfo -- structure  defined in mgm.h and populated in mgm_user_if.c
//
// Carlos Lopez Carrillo
// Geophysical Research Center, NMT.
//--------------------------------------------------------------------
#define LON_NAME "lon" //name of x axis in radar data when lon,lat are used

  //  extern FILE *stdin;

  struct RadarDataInfo *rd;

  FILE *fh;
  char fn[MAXCHAR];
  char pline[MAXCHAR];
  char aline[MAXCHAR];
  char *hbin;
  long nsbin,nvbin;
  float *sbin,*vbin;
  float lon0,lat0,alt0;
  float bad, badlim,Dx,Dy,Dz;
  int Nx,Ny,Nz;
  float *xg,*yg,*zg;
static  struct field *fpx,*fpy,*fpz,*fpaux;

  int ig,jg,kg,pg,Nobs;
  

  float lon0AG,lat0AG,alt0AG;

  // 
  // Get space to hold the structures.
  //
  if((rd =(struct RadarDataInfo*)malloc(sizeof(struct RadarDataInfo)*(2)))==NULL){
     fprintf(stderr,"SetRDInfo: cannot get enough memory\n"); exit(1);}


  //sprintf(fn,"%s",(*finfo).rd.fn);
  hbin = (*rd).hbin[0];
  //
  // Open Radar File
  //
//  if ((fh = fopen((*rd).fn,"r")) == NULL) {
//   fprintf(stderr,"can't open aircraftRadarData file %s\n",fn);
//   exit(1);
//  }
 fh = stdin;

  // 
  // get header 
  // 
  gethdr(fh,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"make_header:format expected\n");
    exit(1);
  }

  //
  // Get Some Header Info.
  //

  if (seekpar(hbin,HBMAX,"bad",pline) == OK)
      bad = atof(pline);
  else                                       
      bad = BAD; 
	if (seekpar(hbin,HBMAX,"badlim",pline) == OK) 
     badlim = atof(pline);
  else                                       
    badlim = BADLIM;
	if (seekpar(hbin,HBMAX,LOND,pline) == OK) lon0 = atof(pline);
  else {
	  fprintf(stderr,"Could not find %s in %s parameters ..Quiting",LOND,fn);
		exit(1);
	}
	if (seekpar(hbin,HBMAX,LATD,pline) == OK) lat0 = atof(pline);
  else {
	  fprintf(stderr,"Could not find %s in %s parameters ..Quiting",LATD,fn);
		exit(1);
	}
	if (seekpar(hbin,HBMAX,ALTD,pline) == OK) alt0 = atof(pline);
  else {
	  fprintf(stderr,"Could not find %s in %s parameters ..Quiting",ALTD,fn);
		exit(1);
	}

  (*rd).badlim =  badlim;
  (*rd).bad  = bad;
  (*rd).lon0 = lon0;
  (*rd).lat0 = lat0;
  (*rd).alt0 = alt0;


  // 
  // Get Buffer Sizes
  //
  nsbin  = elemcnt(hbin,HBMAX,'s');
  nvbin  = elemcnt(hbin,HBMAX,'v');
  
  (*rd).nsbin  = nsbin;
  (*rd).nvbin  = nvbin;

  //
  // Get Memory to hold data.
  //
  (*rd).sbin = Fgetbuff(nsbin);
  (*rd).vbin = Fgetbuff(nvbin);

  sbin = (*rd).sbin;
  vbin = (*rd).vbin;
  
  //
  (*rd).dbz = getptr2(hbin,HBMAX,vbin,'v','d',DBZ,&fpaux);
  if (fpaux->dim != 3 ) {
		fprintf(stderr,"dbz is expected to be a 3dimensional field!\n");
    exit(1);
  } else {
		strcpy((*rd).xname,fpaux->dname1);
		(*rd).Nx = fpaux->dsize1;
		strcpy((*rd).yname,fpaux->dname2);
		(*rd).Ny = fpaux->dsize2;
		strcpy((*rd).zname,fpaux->dname3);
		(*rd).Nz = fpaux->dsize3;
  }
  //
  //
  // Read in static buffer.
  // 
  getslice(fh,nsbin,(*rd).sbin);

  (*rd).xg = getptr2(hbin,HBMAX,sbin,'s','d',(*rd).xname,&fpx);
  if (fpx->dim != 1 || strcmp(fpx->fname,fpx->dname1) != 0) {
    fprintf(stderr,"@alloc: %s is not an index field\n",fpx->fname);
    exit(1);
  } else {
		(*rd).x0 = (*rd).xg[0];
		(*rd).Dx = (*rd).xg[1] - (*rd).xg[0];
		(*rd).Nx = fpx->dsize1;
  
	}
  (*rd).yg = getptr2(hbin,HBMAX,sbin,'s','d',(*rd).yname,&fpy);
  if (fpy->dim != 1 || strcmp(fpy->fname,fpy->dname1) != 0) {
    fprintf(stderr,"@alloc: %s is not an index field\n",fpy->fname);
    exit(1);
  } else {
		(*rd).y0 = (*rd).yg[0];
		(*rd).Dy = (*rd).yg[1] - (*rd).yg[0];
		(*rd).Ny = fpy->dsize1;

	}
  (*rd).zg = getptr2(hbin,HBMAX,sbin,'s','d',(*rd).zname,&fpz);
  if (fpz->dim != 1 || strcmp(fpz->fname,fpz->dname1) != 0) {
    fprintf(stderr,"@alloc: %s is not an index field\n",fpz->fname);
    exit(1);
  } else {
		(*rd).z0 = (*rd).zg[0];
		(*rd).Dz = (*rd).zg[1] - (*rd).zg[0];
		(*rd).Nz = fpz->dsize1;

	}
  //
  if ( strcmp((*rd).xname,LON_NAME) == 0 ) {
    //Spherical coordinates are been used, change degrees to
    // radians -- for internal usage. Donot Forget to convert
    // them back to radians at for output.
     (*rd).sc = 1;
		 
   } else (*rd).sc = -1;

  //
  // Data is assumed to come all in one slice, get it
  //
  getslice(fh,nvbin,(*rd).vbin);

  // 
  //Get Relevant Pointers from the Var slice
  //
  (*rd).u1 = getptr(hbin,HBMAX,vbin,'v','d',U1);
  (*rd).u2 = getptr(hbin,HBMAX,vbin,'v','d',U2);
  (*rd).u3 = getptr(hbin,HBMAX,vbin,'v','d',U3);

  (*rd).a1 = getptr(hbin,HBMAX,vbin,'v','d',A1);
  (*rd).a2 = getptr(hbin,HBMAX,vbin,'v','d',A2);
  (*rd).a3 = getptr(hbin,HBMAX,vbin,'v','d',A3);

  (*rd).e1x = getptr(hbin,HBMAX,vbin,'v','d',E1x);
  (*rd).e1y = getptr(hbin,HBMAX,vbin,'v','d',E1y);
  (*rd).e1z = getptr(hbin,HBMAX,vbin,'v','d',E1z);
  
  (*rd).e2x = getptr(hbin,HBMAX,vbin,'v','d',E2x);
  (*rd).e2y = getptr(hbin,HBMAX,vbin,'v','d',E2y);
  (*rd).e2z = getptr(hbin,HBMAX,vbin,'v','d',E2z);

  (*rd).e3x = getptr(hbin,HBMAX,vbin,'v','d',E3x);
  (*rd).e3y = getptr(hbin,HBMAX,vbin,'v','d',E3y);
  (*rd).e3z = getptr(hbin,HBMAX,vbin,'v','d',E3z);

  (*rd).dbz = getptr(hbin,HBMAX,vbin,'v','d',DBZ);

  // Find Out How Many good Obs there are on the grid.
  // loop on Data Grid
  Nx = (*rd).Nx;
  Ny = (*rd).Ny;
  Nz = (*rd).Nz;
  Nobs = 0;
  for (kg = 0; kg < Nz; kg++) {
   for (jg = 0; jg < Ny; jg++) {
    for (ig = 0; ig < Nx; ig++) {
     pg =  kg + (jg + ig * Ny) * Nz;
     if ( ( fabs((*rd).u1[pg]) < badlim ) &&
          ( fabs((*rd).u2[pg]) < badlim ) &&
          ( fabs((*rd).u3[pg]) < badlim ) &&
          ( fabs((*rd).a1[pg]) < badlim ) &&
          ( fabs((*rd).a2[pg]) < badlim ) &&
          ( fabs((*rd).a3[pg]) < badlim ) &&
          ( fabs((*rd).e1x[pg]) < badlim ) &&
          ( fabs((*rd).e1y[pg]) < badlim ) &&
          ( fabs((*rd).e1z[pg]) < badlim ) &&
          ( fabs((*rd).e2x[pg]) < badlim ) &&
          ( fabs((*rd).e2y[pg]) < badlim ) &&
          ( fabs((*rd).e2z[pg]) < badlim ) &&
          ( fabs((*rd).e3x[pg]) < badlim ) &&
          ( fabs((*rd).e3y[pg]) < badlim ) &&
          ( fabs((*rd).e3z[pg]) < badlim )
       ) { Nobs += 1.0; } //good data
    } //i-loop
   } //j-loop
  } //k-loop
  (*rd).Nobs = Nobs;


  //
  // getptr2 allocate space for the structures. Since we do not need
  // them any more, we'll dispose of them.
  //
  Dfreebuff(fpaux); 
  Dfreebuff(fpx); 
  Dfreebuff(fpy); 
  Dfreebuff(fpz); 

  //
  // Close Radar File
  //
// close(fh);

  (*rd).xxd = 0.0;
  (*rd).yyd = 0.0;
  (*rd).zzd = 0.0;

  return(rd);
} 
//
//---------------------------------------------------------------------
//
void fun_wpTVZ(struct funInfo *finfo) {
//
int iloop,jloop,kloop,lloop,pd;
double *dbzi,*wtZi,*wtZ,*rhoF;
double vts,vtw,Zw,Zs,refl;
double R,S,rhos;
float Hd, dbz;

/* names for Data info */
float *in_dbz;
float Dzgd;
int   Nxgd,Nygd,Nzgd;
float bad,badlim;
double wtcalc();

  // Get Radar Data
  Nxgd = (*finfo).rd.Nx; 
  Nygd = (*finfo).rd.Ny;
  Nzgd = (*finfo).rd.Nz;
  Dzgd = (*finfo).rd.Dz;
  in_dbz = (*finfo).rd.dbz;
  bad    = (*finfo).rd.bad;
  badlim = (*finfo).rd.badlim;

  //
  // Output Variables that are interpolated to the analysis grid
  //
  dbzi = (*finfo).dbzi;
  wtZi = (*finfo).wtZi;

  // 
  // Terminal Vertical velocity from reflectivity
  // it is defined on the data grid.
  // (as required by calcfg_ng)
  //
  wtZ = (*finfo).wtZ;

  //
  //  Proccess Data 
  //

  // loop on Data Grid
  for (kloop = 0; kloop < Nzgd; kloop++) {
     //
     // Ok work on the current horizontal plane.
     //
     for (jloop = 0; jloop < Nygd; jloop++) {
        for (iloop = 0; iloop < Nxgd; iloop++) {

           //Location in data vectors
           pd =  kloop + ( jloop + iloop * Nygd) * Nzgd;

           //
           //Proccess only good data 
           //
           if ( fabs(in_dbz[pd]) < badlim ) {

	     // get height and reflectivity factor
	      Hd = (float) kloop * Dzgd;
	      dbz = in_dbz[pd];

              // compute terminal velocity
              wtZ[pd] = wtcalc(Hd, dbz);

              wtZi[pd] = wtZ[pd];
              dbzi[pd] = (double) in_dbz[pd];

            } //good data
            //
            // Zero-out wtZ instead of making bad when it can not be 
            // determined.
            //
            else {
                 wtZ[pd] = 0.0;
                 wtZi[pd] = bad; 
                 dbzi[pd] = bad;
           }

         } //i-loop
       } //j-loop
     } //k-loop
}
//
//---------------------------------------------------------------------
//
void fun_rho(struct funInfo *finfo){
//
// Uses the following formula(used on radvert) 
//
//          'rhomsl 1 z z1 / - mum1 pow * rho ='" 
//
//to come up with an density profile when one is 
//not available from the input data.
//
//This function is used used by the routine that sets up the 
// initialization. It is here because, the density profile is part 
// of the definition of the target function and it only affects the 
// mass continiuty term.
//
// Furthermore while in here, we'll get the density Factor that 
// will be needed to obtain vertical terminal velocity from 
// reflectivity data.
// 
double TMSL=300.0; //Kelvin
double PMSL=100.0; //KiloPascals
double GAMMA=7.0;  //temperature lapse rate (K/km)
double RGAS=287.6; //gas constant for air (J/kg/K)
double GRAV=9.8;   //acceleration of gravity (m/s^2)
double z,z1,mu,rhomsl,mum1,dzg,temp,term1;
double *rho,*rhoF;
int k,Nzg;

  
  rho    = (*finfo).rho;
  rhoF   = (*finfo).rhoF;
  Nzg    = (*finfo).grid.Nz;
  dzg    = (*finfo).grid.Dz;
  z1     = TMSL/GAMMA;
  mu     = GRAV/(RGAS*GAMMA*0.001);
  rhomsl = 1000.0 * PMSL/(RGAS*TMSL);
  mum1   = mu -1.0;

  for (k = 0; k<Nzg; k++) {
     z = dzg * (double) k;
     rho[k] = rhomsl*pow(1.0 - (z/z1),mum1);

     temp = TMSL - GAMMA * z;
     if (rho[k] != 0.0) {
        term1 = pow(10.0,(0.43*log10(rhomsl/rho[k]) -
                          0.4*pow(log10(rhomsl/rho[k]),2.5)));
          rhoF[k] = (1.+0.0023*(1.1-(rho[k]/rhomsl))*(TMSL-temp))*term1;
     } else rhoF[k] = 0.0; // a consistent choice with radvert4.

   //rho[k] = 1.0; //Hack for the Updraft Problem.

  }
}

//
//-------------------------------------------------------------------
//
void CheckRHOS(struct funInfo *finfo){
int k,cx,cy,sy,Nz0;
float z,H;
FILE *fhandle;
double *rho,rhoz;
char Pfile[MAXCHAR] = "rho.gri";
char Ofile[MAXCHAR] = "rho.txt";


  cx=1; cy =2; sy=0;
  fhandle = OpenFileW(Pfile);
    fprintf(fhandle,"set line width rapidograph 3x0 \n");
    fprintf(fhandle,"set symbol size 0.2\n");
    fprintf(fhandle,"open %s \n",Ofile); 
    fprintf(fhandle,"   read columns x=%i y=%i \n",cx,cy);
    //fprintf(fhandle,"   draw curve\n");
    fprintf(fhandle,"		draw symbol %i\n",sy);
    fprintf(fhandle,"close %s\n",Ofile);
  fclose(fhandle);

  fhandle = OpenFileW(Ofile);
  Nz0 = (*finfo).grid.Nz;
  H = ((float) Nz0) * (*finfo).grid.Dz;
  for (k = 0; k< (*finfo).grid.Nz ; k++) {
       z = ((float) k) * (*finfo).grid.Dz;
       if ( z <= H ) {
         rho = (*finfo).rho;
         rhoz = rho[k];
         fprintf(fhandle,"%f %f ",z,rhoz);
       }
     fprintf(fhandle,"\n");
  }
  fclose(fhandle);
  //Be carefull not to mess up the stdout.
	system("gri rho.gri >&- && gv rho.ps >&- ");

}
//
//-------------------------------------------------------------------
//
