/* ngauxcart -- Create a cartesian grid with sounding and/or satwinds
 * data represented in it for input into ngradcart.

 20101117-CarlosLopezCarrillo: the program was modified to also 
 accept satellite winds. The default behavior is exactly the same
 as the previous version.
 The satellite data are brought in by an optional argument ( see
 --satel option below.) When only satellite data are to be processed,
 the option -n must be used to indicate that no data is expected on the
 standard input. Processing of the data was move from main() to two
 separated functions: one for soundings and one for satellite.
 Satellite data is expected in a candis file in which each  variable
 slice contains the following information: 
 jday,lon,lat,pres,u_wind,v_wind,Q. All of these variable are scalars and 
 The file is put together by the companion python scrip satwinds.py
 When both satellite and sounding data are to be combined, they are
 processed as follows: first averages at each grid point from each
 kind of data are obtained independently. Then the velocity at
 a grid point is set to the sounding average if it is available, 
 otherwise the point is set to the satellite average if available;
 The velocity is set to BAD none if no data was available at the grid
 point.
           
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include "quickcandis.h"
#define _GNU_SOURCE
#include <getopt.h>

/* index macro */
#define I(ix,iy,iz) ((iz) + nz*((iy) + ny*(ix)))

/* assumed field names */
#define PRES "pres"
#define JDAY "jday"
#define LONM "lon.m"
#define LATM "lat.m"
#define LON "lon"
#define LAT "lat"
#define UWIND "u_wind"
#define VWIND "v_wind"

/* some constants */
#define D2R 0.017453    /* degrees to radians conversion */
#define EARTH_R 6370.0  /* radius of earth (km) */

/* candis stuff */
static char hbin[HBMAX][LINE], hbout[HBMAX][LINE];
static char cline[LINE], pline[LINE];
//static float bad, badlim;
static float *sbin, *vbin, *sbout, *vbout;
static long nsbin, nvbin, nsbout, nvbout;
static struct field *fp;

/* pressure to height conversion ++++++++++++++++++++++++++++++ */
/* this assumes a surface temperature of 300 K and a lapse rate of 5.5
 * K/km -- surface pressure is 1020 hPa -- incoming pressure is also
 * in hPa and the returned height is in km
 */
float zcalc(float pres)
{
  float gamma = 5.5;
  float gee = 9800.;
  float t0 = 300;
  float rd = 287;
  float psurf = 1020.;
  float zscale, power, z;

  zscale = t0/gamma;
  power = rd*gamma/gee;
  z = zscale*(1 - pow(pres/psurf, power));
  return(z);
}

static  int nx, ny, nz, num;
static  float svx, svy, reftime, tstart, tend, sigmult;
static  float jday0, lon0, dx, lat0, dy, alt0, dz;
static  int lonlatflag, propagation;

static void fun_getSondeWinds(float *vax, float *vay);
static void fun_getSatelWinds(char *fsatel, float *vax,float *vay);

/* main routine +++++++++++++++++++++++++++++++++++++++++++++++ */
main(int argc, char **argv)
{
  int c,i;

  char fsonde[LINE], fsatel[LINE];
  float sigmult_sonde, sigmult_satel;
  int fsonde_flag = 0;
  int fsatel_flag = 0;
  int sonde_flag;

  int npres;
  float *jday, *pres, *lon, *lat, *uwind, *vwind;
  float *vax, *vay, *sigma;
  float *vax_so, *vay_so;
  float *vax_sa, *vay_sa;
  int handle, dh1, dh2, dh3;

  /* initialize input parameters */
  sonde_flag = 1;
  lonlatflag = 0;
  propagation = 0;
  svx = svy = reftime = 0.;
  tstart = -1000.;
  tend = 1000.;
  sigmult = 1.;

  /* process command line options */
  optind = 1;
  int option_index = 0;
  static struct option long_options[] = {
    {"satel", 1, 0, 1},
    {0, 0, 0, 0}
  };

  while ((c = getopt_long(argc,argv,"ls:t:m:n",long_options, &option_index)) != -1) {
    switch (c) {
    case 1: fsatel_flag = 1;
      if (sscanf(optarg,"%f:%s",&sigmult_satel,fsatel) != 2) {
	fprintf(stderr,"ngauxcart: incorrect format for satel\n");
	exit(1); } 

      break;
    case 'n': // no standard input (means no sonde data)
              // by default sonde data is expected on stdin
      sonde_flag = 0;
      break;
    case 'l':
      lonlatflag = 1;
      break;
    case 's':
      propagation = 1;
      if (sscanf(optarg,"%f:%f:%f",&svx,&svy,&reftime) != 3) {
	fprintf(stderr,"ngauxcart: incorrect format for propagation\n");
	exit(1);
      }
      break;
    case 'm':
      if (sscanf(optarg,"%f",&sigmult) != 1) {
	fprintf(stderr,"ngauxcart: incorrect format for sigmult\n");
	exit(1);
      }
      break;
    case 't':
      if (sscanf(optarg,"%f:%f",&tstart,&tend) != 2) {
	fprintf(stderr,"ngauxcart: incorrect format for time limits\n");
	exit(1);
      }
      break;
    default:
      fprintf(stderr,"ngauxcart: unrecognized option -%c\n", c);
      exit(1);
      break;
    }
  }

  /* print usage statement */
  if (argc - optind != 10) {
    fprintf(stderr,"Usage: ngauxcart [-n] [-l] [-s svx,svy,reftime]\n");
    fprintf(stderr,"  [-t tstart:tend] [-m sigmult] \n"); 
    fprintf(stderr,"  [--satel sigmult:fname] --\n"); 
    fprintf(stderr,"  jday0 lon0 dx nx lat0 dy ny alt0 dz nz\n");
    exit(1);
  }

  /* get positional command line arguments */
  jday0 = atof(argv[optind++]);
  lon0 = atof(argv[optind++]);
  dx = atof(argv[optind++]);
  nx = atoi(argv[optind++]);
  lat0 = atof(argv[optind++]);
  dy = atof(argv[optind++]);
  ny = atoi(argv[optind++]);
  alt0 = atof(argv[optind++]);
  dz = atof(argv[optind++]);
  nz = atoi(argv[optind++]);

  /* get sizes of gridded fields */
  num = nx*ny*nz;

  /* allocate grid buffers and fill with zeros */
  vax = getbuff(num);
  vay = getbuff(num);
  vax_so = getbuff(num);
  vay_so = getbuff(num);
  vax_sa = getbuff(num);
  vay_sa = getbuff(num);
  sigma = getbuff(num);
  for (i = 0; i < num; i++) {
    vax[i]=vay[i]=sigma[i]=vax_so[i]=vay_so[i]=vax_sa[i]=vay_sa[i]=0.0; 
  }

  /* read in soundings from standard input, assumed to be all in
   * one file with a variable slice for each sounding
   */

  // Get average velocity at each grid point from soundings
  if ( sonde_flag  == 1 ) fun_getSondeWinds(vax_so,vay_so);
  else { for (i = 0; i < num; i++) { vax_so[i]=vay_so[i]=BAD;}}

  // Get average velocity at each grid point from satellite
  if ( fsatel_flag == 1 ) fun_getSatelWinds(fsatel,vax_sa,vay_sa);
  else { for (i = 0; i < num; i++) { vax_sa[i]=vay_sa[i]=BAD;}}


  //The two averages calculated above are now combined as follows:
  //At each grid point the velocity is set to:
  //the sounding average, it it exists at the point
  //otherwise the satellite value is assigned to the point.
  // Note that averages are set to bad if no data was available
  // when they were calculated.
  //
  for (i = 0; i < num; i++) {
    sigma[i] = BAD; // BugFix:20110712; so to Keep test at ngradcart 
    if ( vax_so[i] < BADLIM && vay_so[i] < BADLIM ){
      vax[i] = vax_so[i];
      vay[i] = vay_so[i];
      sigma[i] = sigmult;
    } else {
      if ( vax_sa[i] < BADLIM && vay_sa[i] < BADLIM ){
      	vax[i] = vax_sa[i];
      	vay[i] = vay_sa[i];
      	sigma[i] = sigmult_satel;
      } 
    }
  }

  /* write a candis file with the results */
  handle = candiscreate("-", "");
  candiscomment(handle, "ngauxcart:");
  candisparam(handle, "svx", svx);
  candisparam(handle, "svy", svy);
  candisparam(handle, "reftime", reftime);
  candisparam(handle, "tstart", tstart);
  candisparam(handle, "tend", tend);
  candisparam(handle, "sigmult", sigmult);
  if (lonlatflag) {
    dh1 = candisdim(handle, "lon", nx, lon0, dx);
    dh2 = candisdim(handle, "lat", ny, lat0, dy);
  }
  else {
    dh1 = candisdim(handle, "x", nx, 0., dx);
    dh2 = candisdim(handle, "y", ny, 0., dy);
  }
  dh3 = candisdim(handle, "z", nz, 0., dz);
  candisfld(handle, "vax", 3, dh1, dh2, dh3, 0, vax);
  candisfld(handle, "vay", 3, dh1, dh2, dh3, 0, vay);
  candisfld(handle, "sigma", 3, dh1, dh2, dh3, 0, sigma);
  candiswrite(handle);

  /* done! */
  exit(0);
}
//
//---------------------------------------------------------------
//

void fun_getSatelWinds(char *fsatel, float *vax,float *vay){
  float *sigma;
  float *jday, *lon, *lat, *pres, *uwind, *vwind;
  float bad_sa,badlim_sa;
  FILE *fo;
  float monotimein, deltatime, deltax, deltay, xloc, yloc, zloc;
  int i, ix, iy, iz, igrid;

  if ((fo = fopen(fsatel,"r")) == NULL) {
    fprintf(stderr,"cdfmerge: can't open %s\n",fsatel);
    exit(1);
  }

  gethdr(fo,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"ngauxcart: float format expected\n");
    exit(1);
  }
  
  /* bad data values */
  if (seekpar(hbin,HBMAX,"bad",pline) == OK) bad_sa = atof(pline);
  else {
    bad_sa = BAD;
  }
  if (seekpar(hbin,HBMAX,"badlim",pline) == OK) badlim_sa = atof(pline);
  else {
    badlim_sa = BADLIM;
  }

  /* allocate input buffers */
  nsbin = elemcnt(hbin,HBMAX,'s');
  nvbin = elemcnt(hbin,HBMAX,'v');
  sbin = getbuff(nsbin);
  vbin = getbuff(nvbin);

  /* get pointers to needed fields */
  jday = getptr(hbin,HBMAX,vbin,'v','d',JDAY);
  lon = getptr(hbin,HBMAX,vbin,'v','d',LON);
  lat = getptr(hbin,HBMAX,vbin,'v','d',LAT);
  pres = getptr(hbin,HBMAX,vbin,'v','d',PRES);
  uwind = getptr(hbin,HBMAX,vbin,'v','d',UWIND);
  vwind = getptr(hbin,HBMAX,vbin,'v','d',VWIND);

  /* get static slice */
  getslice(fo,nsbin,sbin);

  sigma = getbuff(num);
  for (i = 0; i < num; i++) { sigma[i] = 0.; }

  /* loop on variable slices */
  while (getslice(fo,nvbin,vbin) != EOF) {

    /* compute monotime for this sounding and skip if out of time range */
    monotimein = 86.4*(jday[0] - jday0);
    if (monotimein < tstart || monotimein > tend) continue;

    /* navigate the location of the measurement */
    deltatime = monotimein - reftime;
    deltax = svx*deltatime;
    deltay = svy*deltatime;

    i = 0;
    if (fabs(pres[i]) > badlim_sa  ||\
	fabs(uwind[i]) > badlim_sa ||\
	fabs(vwind[i]) > badlim_sa) continue;

    if (lonlatflag) {
      xloc = lon[i] - lon0 - deltax/(EARTH_R*D2R*cos(lat[i]*D2R));
      yloc = lat[i] - lat0 - deltay/(EARTH_R*D2R);
    }
    else {
      xloc = (EARTH_R)*cos(lat[i]*D2R)*((lon[i] - lon0)*D2R) - deltax;
      yloc = (EARTH_R)*((lat[i] - lat0)*D2R) - deltay;
    }

    /* avoid possible negative value truncation problems */
    ix = xloc/dx + 1.5;
    ix -= 1;
    iy = yloc/dy + 1.5;
    iy -= 1;

    /* skip if out of space range or on very edge */
    if (ix < 1 || ix >= nx - 1 || iy < 1 || iy >= ny - 1) continue;

    /* navigate vertically continuing if bad values or out of z range */
    zloc = zcalc(pres[i]);
    iz = zloc/dz + 1.5;
    iz -= 1;
    if (iz < 0 || iz >= nz) continue; 

    /* good coordinates -- increment values for average calculation */
    igrid = I(ix,iy,iz);
    sigma[igrid] += 1.;
    vax[igrid] += uwind[i];
    vay[igrid] += vwind[i];
  }


  /* take average where data are good, set things to bad otherwise */
  for (i = 0; i < num; i++) {
    if (sigma[i] > 0.) {
      vax[i] /= sigma[i];
      vay[i] /= sigma[i];
      sigma[i] = sigmult;
    }
    else {
      vax[i] = vay[i] = sigma[i] = BAD;
    }
  }
  free(sigma);
  free(sbin);
  free(vbin);
}
//
//-----------------------------------------------------------
//
void fun_getSondeWinds(float *vax, float *vay){
  float *sigma;
  float bad_so,badlim_so;
  float *jday, *lon, *lat, *pres, *uwind, *vwind;
  float monotimein, deltatime, deltax, deltay, xloc, yloc, zloc;
  int i, ix, iy, iz, igrid;
  int npres;

  /* read in soundings from standard input, assumed to be all in
   * one file with a variable slice for each sounding
   */
  gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"ngauxcart: float format expected\n");
    exit(1);
  }
  
  /* bad data values */
  if (seekpar(hbin,HBMAX,"bad",pline) == OK) bad_so = atof(pline);
  else {
    bad_so = BAD;
  }
  if (seekpar(hbin,HBMAX,"badlim",pline) == OK) badlim_so = atof(pline);
  else {
    badlim_so = BADLIM;
  }
  
  /* allocate input buffers */
  nsbin = elemcnt(hbin,HBMAX,'s');
  nvbin = elemcnt(hbin,HBMAX,'v');
  sbin = getbuff(nsbin);
  vbin = getbuff(nvbin);

  /* get pointer to pressure field, either in static or variable slice */
  if ((pres = getptr2(hbin,HBMAX,sbin,'s','r',PRES,&fp)) == (float*)NULL) {
    pres = getptr2(hbin,HBMAX,vbin,'v','d',PRES,&fp);
  }
  if (fp->dim != 1) {
    fprintf(stderr,"ngauxcart: %s is not a 1-D field\n",fp->fname);
    exit(1);
  }

  /* get the number of pressure levels */
  npres = fp->dsize1;

  /* get pointers to needed fields */
  jday = getptr(hbin,HBMAX,vbin,'v','d',JDAY);
  lon = getptr(hbin,HBMAX,vbin,'v','d',LONM);
  lat = getptr(hbin,HBMAX,vbin,'v','d',LATM);
  uwind = getptr(hbin,HBMAX,vbin,'v','d',UWIND);
  vwind = getptr(hbin,HBMAX,vbin,'v','d',VWIND);

  /* get static slice */
  getslice(stdin,nsbin,sbin);

  sigma = getbuff(num);
  for (i = 0; i < num; i++) { sigma[i] = 0.; }

  /* loop on variable slices */
  while (getslice(stdin,nvbin,vbin) != EOF) {

    /* compute monotime for this sounding and skip if out of time range */
    monotimein = 86.4*(jday[0] - jday0);
    if (monotimein < tstart || monotimein > tend) continue;

    /* navigate the location of the sounding */
    deltatime = monotimein - reftime;
    deltax = svx*deltatime;
    deltay = svy*deltatime;
    if (lonlatflag) {
      xloc = lon[0] - lon0 - deltax/(EARTH_R*D2R*cos(lat[0]*D2R));
      yloc = lat[0] - lat0 - deltay/(EARTH_R*D2R);
    }
    else {
      xloc = (EARTH_R)*cos(lat[0]*D2R)*((lon[0] - lon0)*D2R) - deltax;
      yloc = (EARTH_R)*((lat[0] - lat0)*D2R) - deltay;
    }

    /* avoid possible negative value truncation problems */
    ix = xloc/dx + 1.5;
    ix -= 1;
    iy = yloc/dy + 1.5;
    iy -= 1;

    /* skip if out of space range or on very edge */
    if (ix < 1 || ix >= nx - 1 || iy < 1 || iy >= ny - 1) continue;

    /* loop over pressure levels */
    for (i = 0; i < npres; i++) {
      
      /* navigate vertically continuing if bad values or out of z range */
      if (fabs(pres[i]) > badlim_so || fabs(uwind[i]) > badlim_so ||
	  fabs(vwind[i]) > badlim_so) continue;
      zloc = zcalc(pres[i]);
      iz = zloc/dz + 1.5;
      iz -= 1;
      if (iz < 0 || iz >= nz) continue;

      /* good coordinates -- increment values for average calculation */
      igrid = I(ix,iy,iz);
      sigma[igrid] += 1.;
      vax[igrid] += uwind[i];
      vay[igrid] += vwind[i];
    }
  }

  /* take average where data are good, set things to bad otherwise */
  for (i = 0; i < num; i++) {
    if (sigma[i] > 0.) {
      vax[i] /= sigma[i];
      vay[i] /= sigma[i];
      sigma[i] = sigmult;
    }
    else {
      vax[i] = vay[i] = sigma[i] = BAD;
    }
  }

  free(sigma);
  free(sbin);
  free(vbin);
}
