/* ngdoppler.h */
/* $Id: ngdoppler.c,v 1.9 2009/07/11 02:09:00 raymond Exp raymond $ */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "quickcandis.h"
#include "ngradcart.h"
#include "ngutils.h"

#define MINDET (1.e-10)
#define AMIN (0.0001)

/* ngdoppler +++++++++++++++++++++++++++++++++++++++++++++++++++++ */
void ngdoppler(int nx, int ny, int nz, struct param_list plist,
	       struct field_list flist)
{
  int handle, dhx, dhy, dhz, i, num, eigerr;
  char cline[80];
  double e1[3], e2[3], e3[3];
  double *a1, *a2, *a3, *b1, *b2, *b3, *u1, *u2, *u3;
  double *e1x, *e1y, *e1z, *e2x, *e2y, *e2z, *e3x, *e3y, *e3z;
  double *chisq, *avtime, *deltime;

  /* number of grid points */
  num = nx*ny*nz;

  /* allocate space for doppler analysis */
  a1 = getdbuff(num);
  a2 = getdbuff(num);
  a3 = getdbuff(num);
  b1 = getdbuff(num);
  b2 = getdbuff(num);
  b3 = getdbuff(num);
  u1 = getdbuff(num);
  u2 = getdbuff(num);
  u3 = getdbuff(num);
  e1x = getdbuff(num);
  e1y = getdbuff(num);
  e1z = getdbuff(num);
  e2x = getdbuff(num);
  e2y = getdbuff(num);
  e2z = getdbuff(num);
  e3x = getdbuff(num);
  e3y = getdbuff(num);
  e3z = getdbuff(num);
  chisq = getdbuff(num);
  avtime = getdbuff(num);
  deltime = getdbuff(num);

  /* loop over grid points */
  for (i = 0; i < num; i++) {
    
    /* process grid point if data good, enough data, sum of weights is
     * sufficient, and we can calculate eigenvalues and eigenvectors
     * of a matrix
     */
    if (flist.chisqbase[i] < BADLIM && flist.pcount[i] > 2.5) {
      eigerr = eigencalc(flist.axx[i],flist.axy[i],flist.ayy[i],
			 flist.axz[i],flist.ayz[i],flist.azz[i],
			 &a1[i],&a2[i],&a3[i],e1,e2,e3);

      /* compute velocities along principal axes */
      b1[i] = e1[0]*flist.bx[i] + e1[1]*flist.by[i] + e1[2]*flist.bz[i];
      b2[i] = e2[0]*flist.bx[i] + e2[1]*flist.by[i] + e2[2]*flist.bz[i];
      b3[i] = e3[0]*flist.bx[i] + e3[1]*flist.by[i] + e3[2]*flist.bz[i];

      /* compute u's -- eliminate results for very small a's */
      if (a1[i] < AMIN)	u1[i] = a1[i] = 0.;
      else u1[i] = b1[i]/a1[i];
      if (a2[i] < AMIN)	u2[i] = a2[i] = 0.;
      else u2[i] = b2[i]/a2[i];
      if (a3[i] < AMIN)	u3[i] = a3[i] = 0.;
      else u3[i] = b3[i]/a3[i];

      /* compute chisq */
			// Bug fix 20110621
			// when there are no radar data, chisqbase is set to zero inside
			// ngmunchrays.c, therefore we should make the following test.
			// Otherwise, we may get negative values for chisq. Furthermore,
			// in the case of no radar data we want to keep chisq = 0. 
			//
			if (flist.chisqbase[i] > 1e-6) {
      	chisq[i] = flist.chisqbase[i]
				- 2.*(u1[i]*b1[i] + u2[i]*b2[i] + u3[i]*b3[i])
				+ u1[i]*a1[i]*u1[i] + u2[i]*a2[i]*u2[i] + u3[i]*a3[i]*u3[i];
			}

      /* transfer eigenvectors */
      e1x[i] = e1[0]; e1y[i] = e1[1]; e1z[i] = e1[2];
      e2x[i] = e2[0]; e2y[i] = e2[1]; e2z[i] = e2[2];
      e3x[i] = e3[0]; e3y[i] = e3[1]; e3z[i] = e3[2];

      /* compute time variables */
      avtime[i] = flist.time[i];
      deltime[i] = pow(flist.timesq[i] - flist.time[i]*flist.time[i], 0.5);
    }
    
    /* otherwise, set everything to bad unless wzero option set -- in
     * this case, set axes to align with cartesian system, but with e2
     * rotated to the vertical, and make u2 = 0, dbz = BAD, chisq = 0,
     * and pcount = 1000.  Make a2 = a2val and a1 = a3 = 0 -- the idea
     * is to force the vertical air velocity (which equals the
     * vertical particle velocity for low reflectivity) to zero
     * outside regions with data, without constraining the horizontal
     * velocity components
     */
    else if (plist.wzero) {
      a1[i] = a3[i] = 0.;
      a2[i] = plist.a2val;
      b1[i] = b3[i] = u1[i] = u3[i] = 0.;
      b2[i] = u2[i] = 0.;
      e2x[i] = e3x[i] = 0.;
      e1y[i] = e2y[i] = 0.;
      e1z[i] = e3z[i] = 0.;
      e1x[i] = e2z[i] = e3y[i] = 1.;
      chisq[i] = 0.;
      avtime[i] = deltime[i] = 0.;
      flist.dbz[i] = BAD;
      flist.pcount[i] = 1000.;
    }
    else {
      a1[i] = a2[i] = a3[i] = BAD;
      b1[i] = b2[i] = b3[i] = BAD;
      u1[i] = u2[i] = u3[i] = BAD;
      e1x[i] = e2x[i] = e3x[i] = BAD;
      e1y[i] = e2y[i] = e3y[i] = BAD;
      e1z[i] = e2z[i] = e3z[i] = BAD;
      chisq[i] = BAD;
      avtime[i] = deltime[i] = BAD;
    }
  }

  /* create and write a candis file */
  handle = candiscreate("-", 0);
  candiscomment(handle, "ngradcart:");
  candiscomment(handle, "  new generation doppler calculation");
  if (plist.wzero) {
    candiscomment(handle, "  wzero param set");
  }
  if (!plist.radar) {
    candiscomment(handle, "  no radar data used");
  }
  if (plist.aux) {
    sprintf(cline,"  sounding file %s used", plist.auxfilename);
    candiscomment(handle, cline);
  }
  candisparam(handle, "bad", BAD);
  candisparam(handle, "badlim", BADLIM);
  candisparam(handle, "zmin0", plist.zmin0);
  candisparam(handle, "zmincoef", plist.zmincoef);
  candisparam(handle, "svx", plist.svx);
  candisparam(handle, "svy", plist.svy);
  candisparam(handle, "reftime", plist.reftime);
  candisparam(handle, "minrange", plist.minrange);
  candisparam(handle, "a2val", plist.a2val);
  candisparam(handle, "lon0", plist.lon0);
  candisparam(handle, "lat0", plist.lat0);
  candisparam(handle, "alt0", plist.alt0);
  candisparam(handle, "dx", plist.dx);
  candisparam(handle, "dy", plist.dy);
  candisparam(handle, "dz", plist.dz);
  if (plist.sphere) {
    dhx = candisdim(handle, "lon", nx, plist.lon0, plist.dx);
    dhy = candisdim(handle, "lat", ny, plist.lat0, plist.dy);
  }
  else {
    dhx = candisdim(handle, "x", nx, 0., plist.dx);
    dhy = candisdim(handle, "y", ny, 0., plist.dy);
  }
  //dhz = candisdim(handle, "z", nz, 0., plist.dz);
  dhz = candisdim(handle, "z", nz, plist.alt0, plist.dz);
  dcandisfld(handle, "dbz", 3, dhx, dhy, dhz, 0, flist.dbz);
  dcandisfld(handle, "a1", 3, dhx, dhy, dhz, 0, a1);
  dcandisfld(handle, "a2", 3, dhx, dhy, dhz, 0, a2);
  dcandisfld(handle, "a3", 3, dhx, dhy, dhz, 0, a3);
  dcandisfld(handle, "b1", 3, dhx, dhy, dhz, 0, b1);
  dcandisfld(handle, "b2", 3, dhx, dhy, dhz, 0, b2);
  dcandisfld(handle, "b3", 3, dhx, dhy, dhz, 0, b3);
  dcandisfld(handle, "u1", 3, dhx, dhy, dhz, 0, u1);
  dcandisfld(handle, "u2", 3, dhx, dhy, dhz, 0, u2);
  dcandisfld(handle, "u3", 3, dhx, dhy, dhz, 0, u3);
  dcandisfld(handle, "e1x", 3, dhx, dhy, dhz, 0, e1x);
  dcandisfld(handle, "e1y", 3, dhx, dhy, dhz, 0, e1y);
  dcandisfld(handle, "e1z", 3, dhx, dhy, dhz, 0, e1z);
  dcandisfld(handle, "e2x", 3, dhx, dhy, dhz, 0, e2x);
  dcandisfld(handle, "e2y", 3, dhx, dhy, dhz, 0, e2y);
  dcandisfld(handle, "e2z", 3, dhx, dhy, dhz, 0, e2z);
  dcandisfld(handle, "e3x", 3, dhx, dhy, dhz, 0, e3x);
  dcandisfld(handle, "e3y", 3, dhx, dhy, dhz, 0, e3y);
  dcandisfld(handle, "e3z", 3, dhx, dhy, dhz, 0, e3z);
  dcandisfld(handle, "chisq", 3, dhx, dhy, dhz, 0, chisq);
  dcandisfld(handle, "avtime", 3, dhx, dhy, dhz, 0, avtime);
  dcandisfld(handle, "deltime", 3, dhx, dhy, dhz, 0, deltime);
  dcandisfld(handle, "pcount", 3, dhx, dhy, dhz, 0, flist.pcount);
  candiswrite(handle);
}
