/* ngmunchrays.c */
/* $Id: ngmunchrays.c,v 1.4 2009/06/15 15:59:52 raymond Exp $ */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#include "ngradcart.h"

/* assumed names of radial velocity and reflectivity fields */
#define VR "vr"
#define DBZ "dbz"

/* other assumed names defining positioning */
#define RANGE "range"   /* range */
#define LON "lon"	/* longitude */
#define LAT "lat"	/* latitude */
#define ALT "alt"	/* altitude */
#define TIME "monotime" /* monotonically guaranteed time */
#define ALTTIME "time"	/* alternate time if monotime doesn't exist */
#define AZ "az"		/* azimuth */
#define EL "el"		/* elevation */

/* some constants */
#define D2R 0.017453    /* degrees to radians conversion */
#define EARTH_R 6370.0  /* radius of earth (km) */
#define RANGE_TOL 0.5   /* tolerance defining gap to omit range = alt */

/* candis stuff */
static char hbin[HBMAX][LINE];
static char cline[LINE],pline[LINE];
static float bad,badlim,auxbad,auxbadlim;
static float *sbin,*vbin;
static long nsbin,nvbin;
static struct field *fp;

/* names for pointers to input fields */
static float *range;		/* --> RANGE (index field) */
static float *time;		/* --> TIME */
static float *lon;		/* --> LON */	
static float *lat;		/* --> LAT */
static float *alt;		/* --> ALT */
static float *az;		/* --> AZ */
static float *el;		/* --> EL */

/* ray data */
static float *in_dbz;           /* raflectivity factor --> DBZ */ 
static float *in_vr;            /* radial velocity --> VR */

/* pointloc +++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
void pointloc(int sphere, float range, float lat, float xl, float yl,
	      float zl, float orax, float oray, float oraz, float deltax,
	      float deltay, float *xloc, float *yloc, float *zloc)
{
  float delx,dely,delz;
  
  /* get locations of range point relative to radar */
  delx = range*orax;
  dely = range*oray;
  delz = range*oraz;
  
  /* do corrections for earth's curvature */
  delz += (delx*delx + dely*dely)/(2.*EARTH_R);
  
  /* translate all points to their location at the common time of reftime */
  delx -= deltax;
  dely -= deltay;

  /* convert to degrees if lat-lon grid is used -- note that this is
   * an approximation which is valid for short ranges, such as
   * those typical of the position relative to the aircraft */
  if (sphere) {
    delx = delx/((EARTH_R)*D2R*cos(lat*D2R));
    dely = dely/((EARTH_R)*D2R);
  }
  
  /* get location relative to coordinate system origin */
  *xloc = xl + delx;
  *yloc = yl + dely;
  *zloc = zl + delz;
}

/* calcweights +++++++++++++++++++++++++++++++++++++++++++++++++++++ */
void calcweights(int nx, int ny, int nz, int ix, int iy, int iz,
		float cx, float cy, float cz, int *np, float *we)
{
  int jx, jy, jz;
  float ccx, ccy, ccz;

  /* calculate relative position complements and indices */
  ccx = 1. - cx; ccy = 1. - cy; ccz = 1. - cz;
  jx = ix + 1; jy = iy + 1; jz = iz + 1;
  
  np[0] = I(ix,iy,iz); we[0] = ccx*ccy*ccz;
  np[1] = I(jx,iy,iz); we[1] = cx*ccy*ccz;
  np[2] = I(ix,jy,iz); we[2] = ccx*cy*ccz;
  np[3] = I(ix,iy,jz); we[3] = ccx*ccy*cz;
  np[4] = I(jx,jy,iz); we[4] = cx*cy*ccz;
  np[5] = I(jx,iy,jz); we[5] = cx*ccy*cz;
  np[6] = I(ix,jy,jz); we[6] = ccx*cy*cz;
  np[7] = I(jx,jy,jz); we[7] = cx*cy*cz;
  
}

/* ngmunchrays +++++++++++++++++++++++++++++++++++++++++++++++++++++ */
void ngmunchrays(int nx, int ny, int nz, struct param_list plist,
		 struct field_list flist)
{
  int ix, iy, iz, i, num, nrange, irange, corner, np[8];
  float xl, yl, zl, orax, oray, oraz, deltatime, refl;
  float deltax, deltay, xloc, yloc, zloc, mingatez, alt1, alt2;
  float cx, cy, cz, weight, we[8];
  float *vax, *vay, *sigma;
  int auxlonlat;
  FILE *auxfile;

  /* zero gridded fields */
  num = nx*ny*nz;
  for (i = 0; i < num; i++) {
    flist.axx[i] = flist.axy[i] = flist.ayy[i] = 0.;
    flist.axz[i] = flist.ayz[i] = flist.azz[i] = 0.;
    flist.bx[i] = flist.by[i] = flist.bz[i] = 0.;
    flist.chisqbase[i] = flist.weightsum[i] = 0.;
    flist.pcount[i] = 0.;
  }

  /* expect radar data on the standard input if set -- otherwise no radar */
  if (plist.radar) {
    
    /* set up for reading rays from standard input */
    /* get header of input file */
    gethdr(stdin,hbin,HBMAX);
    if (getfmt(hbin,HBMAX) != 'f') {
      fprintf(stderr,"ngradcart: float format expected\n");
      exit(1);
    }
    
    /* bad data values */
    if (seekpar(hbin,HBMAX,"bad",pline) == OK) bad = atof(pline);
    else {
      bad = BAD;
      sprintf(pline,"%g",bad);
    }
    if (seekpar(hbin,HBMAX,"badlim",pline) == OK) badlim = atof(pline);
    else {
      badlim = BADLIM;
      sprintf(pline,"%g",badlim);
    }
    
    /* allocate input buffers */
    nsbin  = elemcnt(hbin,HBMAX,'s');
    nvbin  = elemcnt(hbin,HBMAX,'v');
    sbin   = getbuff(nsbin);
    vbin   = getbuff(nvbin);
    
    /* get pointer for range and number of range gates */
    range = getptr2(hbin,HBMAX,sbin,'s','d',RANGE,&fp);
    if (fp->dim != 1 || strcmp(fp->fname,fp->dname1) != 0) {
      fprintf(stderr,"ngradcart: %s is not an index field\n",fp->fname);
      exit(1);
    }
    nrange = fp->dsize1;
    
    /* get pointers for ray data */
    in_dbz = getptr(hbin,HBMAX,vbin,'v','d',DBZ);
    in_vr  = getptr(hbin,HBMAX,vbin,'v','d',VR);
    
    /* get pointer for time */
    if((time  = getptr(hbin,HBMAX,vbin,'v','r',TIME)) == NULL) {
      fprintf(stderr,"ngradcart: warning: %s not available, %s used\n",
	      TIME,ALTTIME);
      time = getptr(hbin,HBMAX,vbin,'v','r',ALTTIME);
    }
    
    /* get pointers for other non-ray variables of interest */
    lon = getptr(hbin,HBMAX,vbin,'v','d',LON);
    lat = getptr(hbin,HBMAX,vbin,'v','d',LAT);
    alt = getptr(hbin,HBMAX,vbin,'v','d',ALT);
    az = getptr(hbin,HBMAX,vbin,'v','d',AZ);
    el = getptr(hbin,HBMAX,vbin,'v','d',EL);
    
    /* get input static slice */
    getslice(stdin,nsbin,sbin);
    
    /* loop through rays */
    while (getslice(stdin,nvbin,vbin) != EOF) {
      
      /* compute radar location relative to specified origin */
      if (plist.sphere) {   /* degrees */
	xl = *lon - plist.lon0;
	yl = *lat - plist.lat0;
      }
      else {   /* kilometers */
	xl = (EARTH_R)*cos(*lat*D2R)*((*lon - plist.lon0)*D2R);
	yl = (EARTH_R)*((*lat - plist.lat0)*D2R);
      }
      
      /* height is the same for both */
      zl = (*alt - plist.alt0);
      
      /* compute the direction cosines for this ray */
      orax = cos(*el*D2R);
      oray = cos(*az*D2R)*orax;
      orax = sin(*az*D2R)*orax;
      oraz = sin(*el*D2R);
      
      /* get the time elapsed since the reference time */
      deltatime = *time - plist.reftime;
      
      /* get the position correction due to storm velocity */
      deltax = plist.svx*deltatime;
      deltay = plist.svy*deltatime;

      /* compute range of range which should be omitted to
       * avoid contamination by sidebands looking at surface specular
       * reflection, assumed to occur at zero altitude
       */
      alt1 = *alt - RANGE_TOL;
      alt2 = *alt + RANGE_TOL;
      
      /* loop on range */
      for (irange = 0; irange < nrange; irange++) {
	
	/* use data only if both velocity and reflectivity good */
	if (in_dbz[irange] < badlim && fabs(in_vr[irange]) < badlim) {
	  
	  /* Get location of point with respect to grid origin */
	  pointloc(plist.sphere,range[irange],*lat,xl,yl,zl,orax,oray,oraz,
		   deltax,deltay,&xloc,&yloc,&zloc);
	  
	  /* check against surface clutter and range thresholds */
	  mingatez = plist.zmin0 + plist.zmincoef*range[irange];
	  if ((zloc > mingatez || *el > 0.)
	      && range[irange] >= plist.minrange
	      && (range[irange] < alt1 || range[irange] > alt2)) {
	    
	    /* find grid location -- 10000 added to make int conv unambig */
	    ix = (int)(xloc/plist.dx + 10000.) - 10000;
	    iy = (int)(yloc/plist.dy + 10000.) - 10000;
	    iz = (int)(zloc/plist.dz + 10000.) - 10000;
	    
	    /* find interpolation parameters */
	    cx = xloc/plist.dx - ix;
	    cy = yloc/plist.dy - iy;
	    cz = zloc/plist.dz - iz;
	    
	    /* check that the data point is in the grid space */
	    if (ix >= 0 && ix < nx - 1 && 
		iy >= 0 && iy < ny - 1 && 
		iz >= 0 && iz < nz - 1) { 
	      
	      /* define neighbourhood points np[8] & their weights we[8] */
	      calcweights(nx,ny,nz,ix,iy,iz,cx,cy,cz,np,we);
	      
	      /* fill the eight neighbourhood points around */
	      for(corner = 0; corner < 8 ; corner++) {
		i = np[corner];
		weight = we[corner];
		
		/* get reflectivity from reflectivity factor */
		refl = pow(10.,(in_dbz[irange]/10.));
		flist.dbz[i] += weight*refl;  /* convert back to dbz later */
		flist.axx[i] += weight*orax*orax;
		flist.axy[i] += weight*orax*oray;
		flist.ayy[i] += weight*oray*oray;
		flist.axz[i] += weight*orax*oraz;
		flist.ayz[i] += weight*oray*oraz;
		flist.azz[i] += weight*oraz*oraz;
		flist.bx[i] += weight*orax*in_vr[irange];
		flist.by[i] += weight*oray*in_vr[irange];
		flist.bz[i] += weight*oraz*in_vr[irange];
		flist.chisqbase[i] += weight*in_vr[irange]*in_vr[irange];
		flist.time[i] += weight*(*time);
		flist.timesq[i] += weight*(*time)*(*time);
		flist.pcount[i] += 1.0;
		flist.weightsum[i] += weight;
	      }
	    }
	  }
	}
      }
    }
  }

  /* normalize accumulated grid values and set bad data to bad */
  for (i = 0; i < num; i++) {
    if (flist.pcount[i] > 0. && flist.weightsum[i] > 1.e-5) {
      flist.dbz[i] /= flist.weightsum[i];
      flist.axx[i] /= flist.weightsum[i];
      flist.axy[i] /= flist.weightsum[i];
      flist.ayy[i] /= flist.weightsum[i];
      flist.axz[i] /= flist.weightsum[i];
      flist.ayz[i] /= flist.weightsum[i];
      flist.azz[i] /= flist.weightsum[i];
      flist.bx[i] /= flist.weightsum[i];
      flist.by[i] /= flist.weightsum[i];
      flist.bz[i] /= flist.weightsum[i];
      flist.chisqbase[i] /= flist.weightsum[i];
      flist.time[i] /= flist.weightsum[i];
      flist.timesq[i] /= flist.weightsum[i];

      /* turn reflectivity back into dbz */
      flist.dbz[i] = 10.*log10(flist.dbz[i]);
    }
    else {
      flist.dbz[i] = BAD;
      flist.axx[i] = BAD;
      flist.axy[i] = BAD;
      flist.ayy[i] = BAD;
      flist.axz[i] = BAD;
      flist.ayz[i] = BAD;
      flist.azz[i] = BAD;
      flist.bx[i] = BAD;
      flist.by[i] = BAD;
      flist.bz[i] = BAD;
      flist.chisqbase[i] = BAD;
      flist.time[i] = BAD;
      flist.timesq[i] = BAD;
    }
  }

  /* if we have an auxiliary wind file, get it and incorporate */
  if (plist.aux) {
    
    /* open the auxiliary file for input */
    if ((auxfile = fopen(plist.auxfilename, "r")) == (FILE*)NULL) {
      fprintf(stderr,"ngradcart: auxiliary file %s doesn't exist\n",
	      plist.auxfilename);
    }

    /* get the header */
    gethdr(auxfile, hbin, HBMAX);

    /* check format */
    if (getfmt(hbin, HBMAX) != 'f') {
      fprintf(stderr,"ngradcart: aux file has non-float format\n");
      exit(1);
    }
  
    /* bad data values */
    if (seekpar(hbin,HBMAX,"bad",pline) == OK) auxbad = atof(pline);
    else {
      auxbad = BAD;
    }
    if (seekpar(hbin,HBMAX,"badlim",pline) == OK) auxbadlim = atof(pline);
    else {
      auxbadlim = BADLIM;
    }

    /* check for fields of interest */
    if ((fp = seekfld(hbin, HBMAX, 's', "x")) == NULL) {
      if ((fp = seekfld(hbin, HBMAX, 's', "lon")) == NULL) {
	fprintf(stderr,"ngradcart: input aux field has wrong index fields\n");
	exit(1);
      }
      auxlonlat = 1;
    }
    else {
      auxlonlat = 0;
    }
    
    /* check to be sure the aux field grid choice matches command line */
    if ((auxlonlat && !plist.sphere) || (!auxlonlat && plist.sphere)) {
      fprintf(stderr,"ngradcart: aux grid type doesn't match command line\n");
      exit(1);
    }

    /* get data buffers */
    nsbin = elemcnt(hbin, HBMAX, 's');
    nvbin = elemcnt(hbin, HBMAX, 'v');
    sbin = getbuff(nsbin);
    vbin = getbuff(nvbin);
    getslice(auxfile, nsbin, sbin);
    getslice(auxfile, nvbin, vbin);
      
    /* get pointers to the fields of interest */
    vax = getptr(hbin, HBMAX, vbin, 'v', 'd', "vax");
    vay = getptr(hbin, HBMAX, vbin, 'v', 'd', "vay");
    sigma = getptr2(hbin, HBMAX, vbin, 'v', 'd', "sigma", &fp);

    /* make sanity check on field shape and size if radar used */
    if (fp->dim != 3) {
      fprintf(stderr, "ngradcart: number of dimensions wrong in aux file\n");
      exit(1);
    }
    if (fp->dsize1 != nx || fp->dsize2 != ny || fp->dsize3 != nz) {
      fprintf(stderr, "ngradcart: field sizes in aux file incorrect\n");
      fprintf(stderr, "dsize1,2,3: %d %d %d; nx,ny,nz: %d %d %d\n",
	      fp->dsize1,fp->dsize2,fp->dsize3,nx,ny,nz);
      exit(1);
    }

    /* loop through the auxiliary file and add on needed stuff */
    for (i = 0; i < num; i++) {

      /* only do this where aux data are good */
      if (fabs(sigma[i]) < auxbadlim) {

	/* if radar data are good, add to radar values */
	if (fabs(flist.axx[i]) < badlim) {
	  flist.axx[i] += sigma[i];
	  flist.ayy[i] += sigma[i];
	  flist.bx[i] += sigma[i]*vax[i];
	  flist.by[i] += sigma[i]*vay[i];
	}

	/* otherwise, simply set values */
	else {
	  flist.axx[i] = sigma[i];
	  flist.ayy[i] = sigma[i];
	  flist.azz[i] = flist.axy[i] = flist.axz[i] = flist.ayz[i] = 0.;
	  flist.bx[i] = sigma[i]*vax[i];
	  flist.by[i] = sigma[i]*vay[i];
	  flist.bz[i] = 0.;
	  flist.chisqbase[i] = 0.;
	  flist.time[i] = flist.timesq[i] = 0.;
	  flist.pcount[i] = 1000.;  /* keep thresholding happy downstream */
	}
      }
    }
  }
}
