/* ngradcart.c -- First of a new generation of radcart programs
 * which analyzes Doppler radar rays to a Cartesian grid.  Unlike
 * the earlier versions of radcart, no limitation is put on the
 * orientation of the rays -- they can take on any elevation
 * angle.  The price paid for this is that the actual Cartesian
 * velocities must be obtained via a more complex global analysis
 * program than the old generation of radvert programs.  Typically
 * a three-dimensional variational scheme would be used for this
 * purpose.
 * $Id: ngradcart.c,v 1.5 2009/07/11 02:09:49 raymond Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include "ngradcart.h"
#include "ngutils.h"
#include "ngmunchrays.h"
#include "ogdoppler.h"
#include "ngdoppler.h"

/* main routine +++++++++++++++++++++++++++++++++++++++++++++++ */
main(int argc, char **argv)
{
  int c;
  int nx, ny, nz, num;
  struct param_list plist;
  struct field_list flist;

  /* initialize parameter list to default values */
  plist.aux = 0;
  strcpy(plist.auxfilename,"\n");
  plist.clutterthresh = 0;
  plist.zmin0 = 0.;
  plist.zmincoef = 0.;
  plist.propagation = 0;
  plist.svx = 0.;
  plist.svy = 0.;
  plist.reftime = 0.;
  plist.rangethresh = 0;
  plist.minrange = 0.;
  plist.oldgeneration = 0;
  plist.sphere = 0;
  plist.wzero = 0;
  plist.radar = 1;
  plist.a2val = 0.;

  /* process command line arguments */
  while ((c = getopt(argc,argv,"a:c:s:r:olz:x")) != -1) {
    switch (c) {
    case 'a':
      plist.aux = 1;
      if (sscanf(optarg," %s",&plist.auxfilename) != 1) {
	fprintf(stderr,"ngradcart: incorrect format for aux filename\n");
	exit(1);
      }
      break;
    case 'c':
      plist.clutterthresh = 1;
      if (sscanf(optarg," %f:%f",&plist.zmin0,&plist.zmincoef) != 2) {
	fprintf(stderr,"ngradcart: incorrect format for clutter_thresh\n");
	exit(1);
      }
      break; 
    case 's':
      plist.propagation = 1;
      if (sscanf(optarg,"%f:%f:%f",&plist.svx,
		 &plist.svy,&plist.reftime) != 3) {
	fprintf(stderr,"ngradcart: incorrect format for propagation\n");
	exit(1);
      }
      break; 
    case 'r':
      plist.rangethresh = 1;
      if (sscanf(optarg," %f",&plist.minrange) != 1) {
	fprintf(stderr,"ngradcart: incorrect format for minimum range\n");
	exit(1);
      }
      break;
    case 'o':
      plist.oldgeneration = 1;
      break;
    case 'l':
      plist.sphere = 1;
      break;
    case 'x':
      plist.radar = 0;
      break;
    case 'z':
      plist.wzero = 1;
      if (sscanf(optarg," %f",&plist.a2val) != 1) {
	fprintf(stderr,"incorrect format for a2val");
	exit(1);
      }
      break;
    default:
      fprintf(stderr,"ngradcart: unrecognized option -%s\n", c);
      exit(1);
      break;
    } 
  }

  /* print a usage statement */
  if (argc - optind != 9) {
    fprintf(stderr,"Usage: ngradcart [-o] [-l] [-x]\n");
    fprintf(stderr,"  [-c zmin0:zmincoef] [-a auxfilename]\n");
    fprintf(stderr,"  [-s svx:svy:reftime] [-r minrange] [-z a2val] --\n");
    fprintf(stderr,"  lon0 dx nx lat0 dy ny alt0 dz nz\n");
    exit(1);
  }

  /* -a and -x are inconsistent with -o, which just deals with radar alone */
  if ((!plist.radar || plist.aux) && plist.oldgeneration) {
    fprintf(stderr, "ngradcart: -x and -a are both inconsistent with -o\n");
    exit(1);
  }

  /* finish positional command line arguments */
  plist.lon0 = atof(argv[optind++]);
  plist.dx = atof(argv[optind++]);
  nx = plist.nx = atoi(argv[optind++]);
  plist.lat0 = atof(argv[optind++]);
  plist.dy = atof(argv[optind++]);
  ny = plist.ny = atoi(argv[optind++]);
  plist.alt0 = atof(argv[optind++]);
  plist.dz = atof(argv[optind++]);
  nz = plist.nz = atoi(argv[optind++]);

  /* get sizes of gridded fields */
  num = nx*ny*nz;

  /* allocate fields needed for scan through rays */
  flist.axx = getdbuff(num);
  flist.axy = getdbuff(num);
  flist.ayy = getdbuff(num);
  flist.axz = getdbuff(num);
  flist.ayz = getdbuff(num);
  flist.azz = getdbuff(num);
  flist.bx = getdbuff(num);
  flist.by = getdbuff(num);
  flist.bz = getdbuff(num);
  flist.time = getdbuff(num);
  flist.timesq = getdbuff(num);
  flist.chisqbase = getdbuff(num);
  flist.pcount = getdbuff(num);
  flist.dbz = getdbuff(num);
  flist.weightsum = getdbuff(num);

  /* read through rays from standard input and fill analysis fields */
  ngmunchrays(nx, ny, nz, plist, flist);

  /* perform doppler analysis and send to stdout */
  if (plist.oldgeneration) {
    ogdoppler(nx, ny, nz, plist, flist);
  }
  else {
    ngdoppler(nx, ny, nz, plist, flist);
  }

  /* done */
  exit(0);
}
