/* ngradvert.c -- This program computes the vertical particle velocity,
 * the vertical wind, the particle terminal velocity, and the
 * horizontal wind components.  The command line arguments are the
 * desired names of these fields.  A column divergence correction is
 * always done.
 * $Id: ngradvert.c,v 1.7 2009/06/27 18:03:19 raymond Exp $
 */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "cdfhdr.h"

#define PMSL 1000.0      /* sea level pressure (hPa) */
#define TMSL 298.0      /* sea level temperature (K) */
#define GAMMA 6.25      /* temperature lapse rate (K/km) */
#define GEE 9.8         /* acceleration of gravity (m/s^2) */
#define RGAS 287.6      /* gas constant for air (J/kg/K) */
#define SCALE 32.0      /* integer packing scale factor */
#define WCPA 31.8       /* constants for reflectivity-terminal velocity */
#define WCPB 0.918      /* relationship                                 */
#define WTMIN 2.0       /* assumed terminal velocity for undefined refl */
#define WTVAR1 1.0      /* top terminal velocity for no divergence */
#define WTVAR2 5.0      /* max terminal velocity for max divergence */
#define BOTT_SNOW 7.    /* height of bottom of snow layer */
#define TOP_RAIN 4.3    /* height of top of rain layer */
#define KMPDEG 111.177  /* kilometers per degree at equator */
#define DEGPRAD 57.296  /* degrees per radian */
#define A 6370.         /* radius of earth in km */
#define I(ix,iy,iz) (ix)*ny*nz + (iy)*nz + (iz)  /* array index macro */
#define J(ix,iy) (ix)*ny + (iy)  /* array index macro */
#define U "u"           /* name of x wind */
#define V "v"           /* name of y wind */
#define W "w"           /* name of z wind */
#define MFLUX "mflux"   /* name of vertical mass flux */
#define DFLUX "dflux"   /* name of horizontal mass divergence */
#define DCORR "dcorr"   /* name of divergence correction term */
#define DIV "div"       /* name of horizontal volume divergence */
#define WP "wp"         /* name of vertical particle velocity */
#define WT "wt"         /* name of particle terminal velocity */

static char hbin[HBMAX][LINE],hbout[HBMAX][LINE];  /* header buffers */
static int nhbin;              /* number of lines in input header buffer */
static float *sbuff,*vbuff;    /* slice buffers */
static long nsbuffin,nsbuffot,nvbuffin,nvbuffot;   /* sizes of slice buffers */
static float *dbz,*ux0,*uy0;   /* field pointers */
static float *gx,*gy;
static float *u,*v,*w,*dflux,*dcorr,*mflux,*div2,*wp,*wt;
static float *rho,*pres,*x,*y,*z,*zs;
static float z1,mu,rhomsl;     /* intermediates for pres and density calcs */
static int i,ix,iy,iz,nx,ny,nz;   /* looping variables and array limits */
static int xygrid;             /* 1 if x-y, 0 if lon-lat */
static float dx,dy,dz;         /* grid intervals */
static char alpha[LINE];       /* line buffer */
static char d1[LINE],d2[LINE],d3[LINE];  /* dimension names */
static struct field *fp;       /* field structure pointer */
static float bad,badlim;       /* bad data stuff */

/* compute horizontal wind -------------------------------------------- */
hwindcalc(int ix, int iy, int iz, float *rho, float *ux0, float *uy0, float *gx,
	  float *gy, float *mflux, float *wt, float *u, float *v)
{
  float wprov;

  /* compute provisional u and v if data good (gx reps all data from
   * ngradcart) -- if wbove is bad, set it to zero in the horizontal
   * wind calculation on the philosphy that non-active regions should
   * have small vertical velocity
   */
  if (fabs(gx[I(ix,iy,iz)]) < badlim) {
    
    /* wprov calculation */
    if (fabs(mflux[I(ix,iy,iz)]) < badlim) {
      wprov = mflux[I(ix,iy,iz)]/rho[iz];
    }
    else {
      wprov = 0.;
    }
    
    /* horizontal wind calculation */
    u[I(ix,iy,iz)] = ux0[I(ix,iy,iz)] +
      gx[I(ix,iy,iz)]*(wprov + wt[I(ix,iy,iz)]);
    v[I(ix,iy,iz)] = uy0[I(ix,iy,iz)] +
      gy[I(ix,iy,iz)]*(wprov + wt[I(ix,iy,iz)]);
  }
  
  /* if ux0, uy0, gx, gy, dbz (and hence wt) bad, then winds must be
   * bad -- note that in ngradcart output, all are good or bad
   * simulataneously
   */
  else {
    u[I(ix,iy,iz)] = v[I(ix,iy,iz)] = bad;
  }
}

/* compute 2-D mass divergence on x-y grid ------------------------------ */
static float xydivcalc(int ix, int iy, int iz, float *u, float *v,
	  float dx, float dy)
{
  int jx,kx,jy,ky;
  float delx,dely,div;
  
  /* set up for centered differences except at boundaries */
  jx = ix + 1;
  if (jx >= nx) jx = nx - 1;
  kx = ix - 1;
  if (kx < 0) kx = 0;
  jy = iy + 1;
  if (jy >= ny) jy = ny - 1;
  ky = iy - 1;
  if (ky < 0) ky = 0;
  
  /* only do the calculation if data are good */
  if (fabs(u[I(ix,iy,iz)]) < badlim && fabs(v[I(ix,iy,iz)]) < badlim &&
      fabs(u[I(jx,iy,iz)]) < badlim && fabs(v[I(jx,iy,iz)]) < badlim &&
      fabs(u[I(kx,iy,iz)]) < badlim && fabs(v[I(kx,iy,iz)]) < badlim &&
      fabs(u[I(ix,jy,iz)]) < badlim && fabs(v[I(ix,jy,iz)]) < badlim &&
      fabs(u[I(ix,ky,iz)]) < badlim && fabs(v[I(ix,ky,iz)]) < badlim) {
    
    /* check to see whether we are dividing by zero */
    if ((jx > kx) || (jy > ky)) {
      delx = dx*(jx - kx);
      dely = dy*(jy - ky);
      
      /* everything appears ok -- define the mass divergence at this point */
      div = rho[iz]*((u[I(jx,iy,iz)] - u[I(kx,iy,iz)])/delx +
		     (v[I(ix,jy,iz)] - v[I(ix,ky,iz)])/dely);
    }
    else {
      div = bad;
    }
  }
  else {
    div = bad;
  }

  /* return value */
  return(div);
}

/* compute 2-D mass divergence on lon-lat grid -------------------------- */
static float lldivcalc(int ix, int iy, int iz, float *u, float *v,
		       float dx, float dy, float lat)
{
  int jx,kx,jy,ky;
  float delx,dely,div;
  
  /* set up for centered differences except at boundaries */
  jx = ix + 1;
  if (jx >= nx) jx = nx - 1;
  kx = ix - 1;
  if (kx < 0) kx = 0;
  jy = iy + 1;
  if (jy >= ny) jy = ny - 1;
  ky = iy - 1;
  if (ky < 0) ky = 0;
  
  /* only do the calculation if data are good */
  if (fabs(u[I(ix,iy,iz)]) < badlim && fabs(v[I(ix,iy,iz)]) < badlim &&
      fabs(u[I(jx,iy,iz)]) < badlim && fabs(v[I(jx,iy,iz)]) < badlim &&
      fabs(u[I(kx,iy,iz)]) < badlim && fabs(v[I(kx,iy,iz)]) < badlim &&
      fabs(u[I(ix,jy,iz)]) < badlim && fabs(v[I(ix,jy,iz)]) < badlim &&
      fabs(u[I(ix,ky,iz)]) < badlim && fabs(v[I(ix,ky,iz)]) < badlim) {
    
    /* check to see whether we are dividing by zero */
    if ((jx > kx) || (jy > ky)) {
      delx = dx*(jx - kx)*KMPDEG*cos(lat/DEGPRAD);
      dely = dy*(jy - ky)*KMPDEG;
      
      /* everything appears ok -- define the mass divergence at this point */
      div = rho[iz]*((u[I(jx,iy,iz)] - u[I(kx,iy,iz)])/delx +
		     (v[I(ix,jy,iz)] - v[I(ix,ky,iz)])/dely -
		     (v[I(ix,jy,iz)] + v[I(ix,ky,iz)])*tan(lat/DEGPRAD)/(2.*A));
    }
    else {
      div = bad;
    }
  }
  else {
    div = bad;
  }

  /* return value */
  return(div);
}

/* compute particle terminal velocity --------------------------------- */
wtcalc(float *dbz, float *wt)
{
  float vts,vtw,Zw,Zs,rhos;
  float temp,term1,gamma;
  for (iz = 0; iz < nz; iz++)  {
    for ( ix = 0; ix < nx; ix++)  {
      for ( iy = 0; iy < ny; iy++)  {
	if (fabs(dbz[I(ix,iy,iz)]) < badlim) {

	  /* compute Z's for rain and snow */
	  Zw = pow(10.,(dbz[I(ix,iy,iz)]/10.));
	  Zs = 4.72*pow(10.,(dbz[I(ix,iy,iz)]/10.));
	
	  /* calculate the density factor */
	  temp = TMSL - GAMMA*z[iz];
	  term1 = pow(10.0,(0.43*log10(rhomsl/rho[iz]) -
			    0.4*pow(log10(rhomsl/rho[iz]),2.5)));
	  rhos = (1.+0.0023*(1.1-(rho[iz]/rhomsl))*(TMSL-temp))*term1;

	  /* calculate the terminal velocity for snow */
	  vts = -rhos*0.817*pow(Zs,0.063);
	  
	  /* calculate the terminal velocity for rain
	   * Joss and Waldvolgel (1970) formula
	   */
	  vtw = -rhos*2.6*pow(Zw,0.107);

	  /* linearly interpolate snow and rain terminal velocities between
	   * TOP_RAIN and BOTT_SNOW
	   */
	  if (z[iz] >= BOTT_SNOW) {
	    wt[I(ix,iy,iz)] = vts;
	  }
	  else if (z[iz] <= TOP_RAIN) {
	    wt[I(ix,iy,iz)] = vtw;
	  }
	  else {
	    wt[I(ix,iy,iz)] = (BOTT_SNOW - z[iz])*vtw/(BOTT_SNOW - TOP_RAIN)
	      + (z[iz] - TOP_RAIN)*vts/(BOTT_SNOW - TOP_RAIN);
	  }
	}
	else {  /* no reflectivity available */

	  /* set to bad */
	  wt[I(ix,iy,iz)] = bad;
	}
      }
    }
  }
}

/* main routine -------------------------------------------------------- */
main(argc,argv)
int argc;
char *argv[];
{
  float rdmean, divmean, divprev, divcurr, mfabove, mfbelow, mfcurr, lat;
  int loop, topdown, divcorr, nrdmean;

  /* check arguments */
  if (argc != 2) {
    fprintf(stderr,"Usage: ngradvert -[t|b|c] < ngradcart_file > outfile\n");
    exit(1);
  }
  else if (strcmp(argv[1],"-t") == 0) {
    topdown = 1;
    divcorr = 0;
  }
  else if (strcmp(argv[1],"-b") == 0) {
    topdown = 0;
    divcorr = 0;
  }
  else if (strcmp(argv[1],"-c") == 0) {
    topdown = 1;
    divcorr = 1;
  }
  else {
    fprintf(stderr,"Usage: ngradvert -[t|b|c] < ngradcart_file > outfile\n");
    exit(1);
  }

  /* get header from standard input */
  nhbin = gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"ngradvert: input file must be float format\n");
    exit(1);
  }

  /* extract bad and badlim from parameter section of header */
  if (seekpar(hbin,HBMAX,"bad",alpha) == FAIL) {
    fprintf(stderr,"ngradvert: parameter bad missing\n");
    exit(1);
  }
  bad = atof(alpha);
  if (seekpar(hbin,HBMAX,"badlim",alpha) == FAIL) {
    fprintf(stderr,"ngradvert: parameter badlim missing\n");
    exit(1);
  }
  badlim = atof(alpha);

  /* check for existing and non-existent fields */
  /* x-velocity */
  if ((fp = seekfld(hbin,HBMAX,'v',"ux0")) == NULL) {
    fprintf(stderr,"ngradvert: no ux0 field\n");
    exit(1);
  }

  /* get dimension names and sizes */
  if (fp->dim != 3) {
    fprintf(stderr,"ngradvert: incorrect field dimensions = %d\n",fp->dim);
    exit(1);
  }

  /* we don't do extensive error checking */
  strcpy(d1,fp->dname1);
  strcpy(d2,fp->dname2);
  strcpy(d3,fp->dname3);
  nx = fp->dsize1;
  ny = fp->dsize2;
  nz = fp->dsize3;

  /* check for grid in kilometers or degrees */
  if (strcmp(d1,"x") == 0) {
    xygrid = 1;
    fprintf(stderr,"ngradvert: x-y grid\n");
  }
  else if (strcmp(d1,"lon") == 0) {
    xygrid = 0;
    fprintf(stderr,"ngradvert: lon-lat grid\n");
  }
  else {
    fprintf(stderr,"ngradvert: x grid name %s not recognized\n",d1);
    exit(1);
  }

  /* y-velocity */
  if (seekfld(hbin,HBMAX,'v',"uy0") == NULL) {
    fprintf(stderr,"ngradvert: no uy0 field\n");
    exit(1);
  }

  /* reflectivity */
  if (seekfld(hbin,HBMAX,'v',"dbz") == NULL) {
    fprintf(stderr,"ngradvert: no dbz field\n");
    exit(1);
  }

  /* get deltas */
  if (seekpar(hbin,HBMAX,"dx",alpha) == FAIL) {
    fprintf(stderr,"ngradvert: can't find a needed parameter\n");
    exit(1);
  }
  dx = atof(alpha);
  if (seekpar(hbin,HBMAX,"dy",alpha) == FAIL) {
    fprintf(stderr,"ngradvert: can't find a needed parameter\n");
    exit(1);
  }
  dy = atof(alpha);
  if (seekpar(hbin,HBMAX,"dz",alpha) == FAIL) {
    fprintf(stderr,"ngradvert: can't find a needed parameter\n");
    exit(1);
  }
  dz = atof(alpha);

  /* make new header */
  /* start by copying over old header */
  for (i = 0; i < nhbin; i++) strcpy(hbout[i],hbin[i]);

  /* add comment */
  sprintf(alpha,"ngradvert: %s\n",argv[1]);
  addcline(hbout,HBMAX,alpha);

  /* add horizontal wind components */
  addfld(hbout,HBMAX,'v',U,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
  addfld(hbout,HBMAX,'v',V,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add divergences and vertical mass flux */
  addfld(hbout,HBMAX,'v',DFLUX,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
  addfld(hbout,HBMAX,'v',DCORR,SCALE,0.,'s',2,d1,nx,d2,ny);
  addfld(hbout,HBMAX,'v',MFLUX,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
  addfld(hbout,HBMAX,'v',DIV,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add dual doppler continuity-generated vertical velocity */
  addfld(hbout,HBMAX,'v',W,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add vertical particle velocity */
  addfld(hbout,HBMAX,'v',WP,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add terminal velocity of hydrometeors */
  addfld(hbout,HBMAX,'v',WT,SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add density to static slice */
  if (seekfld(hbout,HBMAX,'s',"rho") == NULL)
    addfld(hbout,HBMAX,'s',"rho",10000.,0.,'s',1,d3,nz);

  /* add pressure to static slice */
  if (seekfld(hbout,HBMAX,'s',"pres") == NULL)
    addfld(hbout,HBMAX,'s',"pres",10.,0.,'s',1,d3,nz);

  /* add density scale height to static slice */
  if (seekfld(hbout,HBMAX,'s',"zs") == NULL)
    addfld(hbout,HBMAX,'s',"zs",100.,0.,'s',1,d3,nz);

  /* allocate space for slices */
  nsbuffin = elemcnt(hbin,HBMAX,'s');
  nsbuffot = elemcnt(hbout,HBMAX,'s');
  sbuff = getbuff(nsbuffot);
  nvbuffot = elemcnt(hbout,HBMAX,'v');
  nvbuffin = elemcnt(hbin,HBMAX,'v');
  vbuff = getbuff(nvbuffot);

  /* set pointers for fields -- NULL if it doesn't exist */
  x = getptr(hbout,HBMAX,sbuff,'s','d',d1);
  y = getptr(hbout,HBMAX,sbuff,'s','d',d2);
  dbz = getptr(hbout,HBMAX,vbuff,'v','d',"dbz");
  ux0 = getptr(hbout,HBMAX,vbuff,'v','d',"ux0");
  uy0 = getptr(hbout,HBMAX,vbuff,'v','d',"uy0");
  gx = getptr(hbout,HBMAX,vbuff,'v','d',"gx");
  gy = getptr(hbout,HBMAX,vbuff,'v','d',"gy");
  u = getptr(hbout,HBMAX,vbuff,'v','d',U);
  v = getptr(hbout,HBMAX,vbuff,'v','d',V);
  dflux = getptr(hbout,HBMAX,vbuff,'v','d',DFLUX);
  dcorr = getptr(hbout,HBMAX,vbuff,'v','d',DCORR);
  mflux = getptr(hbout,HBMAX,vbuff,'v','d',MFLUX);
  div2 = getptr(hbout,HBMAX,vbuff,'v','d',DIV);
  w = getptr(hbout,HBMAX,vbuff,'v','d',W);
  wp = getptr(hbout,HBMAX,vbuff,'v','d',WP);
  wt = getptr(hbout,HBMAX,vbuff,'v','d',WT);
  rho = getptr(hbout,HBMAX,sbuff,'s','d',"rho");
  pres = getptr(hbout,HBMAX,sbuff,'s','d',"pres");
  z = getptr(hbout,HBMAX,sbuff,'s','d',"z");
  zs = getptr(hbout,HBMAX,sbuff,'s','d',"zs");

  /* put new header */
  puthdr(stdout,hbout,HBMAX);

  /* get static slice */
  getslice(stdin,nsbuffin,sbuff);
  
  /* fill density, pressure, and scale height fields if they exist */
  if ((rho != NULL) && (pres != NULL) && (zs != NULL)) {
    z1 = TMSL/GAMMA;
    mu = GEE/(RGAS*GAMMA*0.001);
    rhomsl = 100.0*PMSL/(RGAS*TMSL);
    for (iz = 0; iz < nz; iz++) {
      pres[iz] = PMSL*pow(1. - z[iz]/z1,mu);
      rho[iz] = rhomsl*pow(1. - z[iz]/z1,mu - 1.);
      zs[iz] = (z1 - z[iz])/(mu - 1.);
    }
  }
  else {
    fprintf(stderr,"ngradvert: panic -- rho, pres, or zs doesn't exist!\n");
    exit(1);
  }

  /* put static slice and do computations */
  putslice(stdout,nsbuffot,sbuff);

  /* get variable slice */
  while (getslice(stdin,nvbuffin,vbuff) != EOF) {

    /* initialize fields */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	for (iz = 0; iz < nz; iz++) {
	  
	  /* all input fields set to bad if dbz is bad */
	  if (fabs(dbz[I(ix,iy,iz)]) > badlim) {
	    ux0[I(ix,iy,iz)] = bad;
	    uy0[I(ix,iy,iz)] = bad;
	    gx[I(ix,iy,iz)] = bad;
	    gy[I(ix,iy,iz)] = bad;
	  }

	  /* initialize output fields to bad */
	  w[I(ix,iy,iz)] = bad; 
	  mflux[I(ix,iy,iz)] = bad; 
	  dflux[I(ix,iy,iz)] = bad; 
	  div2[I(ix,iy,iz)] = bad; 
	}

	/* initialize 2-D divergence correction field to zero */
	dcorr[J(ix,iy)] = 0.;
      }
    }
    
    /* compute particle terminal velocity from wt-Z relationship  */
    wtcalc(dbz, wt);
    
    /* top down integration option */
    if (topdown) {
      
      /* integrate down from the top to the surface */
      for (iz = nz - 2; iz >= 0; iz--) {
	
	/* now get the vertical mass flux at the current level */
	for (ix = 0; ix < nx; ix++) {
	  for (iy = 0; iy < ny; iy++) {
	    
	    /* if the vertical mass flux above is bad, set to zero */
	    if (fabs(mflux[I(ix,iy,iz + 1)]) > badlim) {
	      mfabove = 0.;
	    }
	    
	    /* if it is not, set it to its actual value */
	    else {
	      mfabove = mflux[I(ix,iy,iz + 1)];
	    }
	    
	    /* get the horizontal divergence from the level above */
	    divmean = dflux[I(ix,iy,iz + 1)];
	    
	    /* if the divergence is bad, the vertical mass flux is bad */
	    if (fabs(divmean) > badlim) {
	      mflux[I(ix,iy,iz)] = bad;
	    }
	    
	    /* if the divergence is good, we can integrate down */
	    else {
	      mflux[I(ix,iy,iz)] = mfabove + divmean*dz;
	    }
	    
	    /* compute the horizontal winds at the current level */
	    hwindcalc(ix,iy,iz,rho,ux0,uy0,gx,gy,mflux,wt,u,v);
	  }
	}
	
	/* now compute the div and recompute mflux at the current level */
	for (ix = 0; ix < nx; ix++) {
	  for (iy = 0; iy < ny; iy++) {
	    
	    /* get the mean of the divergence at the current level */
	    if (xygrid) {
	      dflux[I(ix,iy,iz)] = xydivcalc(ix,iy,iz,u,v,dx,dy);
	    }
	    else {
	      lat = y[iy];
	      dflux[I(ix,iy,iz)] = lldivcalc(ix,iy,iz,u,v,dx,dy,lat);
	    }
	  }
	}
      }

      /* do column divergence correction if requested */
      if (divcorr) {
	for (ix = 0; ix < nx; ix++) {
	  for (iy = 0; iy < ny; iy++) {
	    
	    /* get the mean value of rho*div */
	    rdmean = 0.;
	    nrdmean = 0;
	    for (iz = 0; iz < nz; iz++) {
	      if (fabs(dflux[I(ix,iy,iz)]) < badlim) {
		rdmean += dflux[I(ix,iy,iz)];
		nrdmean++;
	      }
	    }
	    
	    /* if there are valid data, correct the divergence */
	    if (nrdmean > 0) {
	      rdmean = rdmean/nrdmean;
	      for (iz = 0; iz < nz; iz++) {
		dflux[I(ix,iy,iz)] -= rdmean;
	      }
	      
	      /* store divergence correction if non-zero */
	      dcorr[J(ix,iy)] = rdmean;
	    }
	  }
	}
      }

      /* now that we have a good divergence, recompute the vertical velocity */
      for (iz = nz - 2; iz >= 0; iz--) {
	for (ix = 0; ix < nx; ix++) {
	  for (iy = 0; iy < ny; iy++) {
	    
	    /* compute the mean divergence */
	    divprev = dflux[I(ix,iy,iz + 1)];
	    divcurr = dflux[I(ix,iy,iz)];
	    
	    /* can get mean over current and above levels */
	    if (fabs(divprev) < badlim && fabs(divcurr) < badlim) {
	      divmean = 0.5*(divprev + divcurr);
	    }
	    
	    /* only above level is available */
	    else if (fabs(divprev) < badlim && fabs(divcurr) > badlim) {
	      divmean = divprev;
	    }
	    
	    /* only current level is available */
	    else if (fabs(divprev) > badlim && fabs(divcurr) < badlim) {
	      divmean = divcurr;
	    }
	    
	    /* sorry, nothing available -- divergence bad */
	    else {
	      divmean = bad;
	    }
	    
	    /* if the vertical mass flux above is bad, set it to zero */
	    if (fabs(mflux[I(ix,iy,iz + 1)]) > badlim) {
	      mfabove = 0.;
	    }
	    
	    /* if it is not, set it to its actual value */
	    else {
	      mfabove = mflux[I(ix,iy,iz + 1)];
	    }
	    
	    /* compute the vertical mass flux with more accurate divergence */
	    if (fabs(divmean) > badlim) {
	      mflux[I(ix,iy,iz)] = bad;
	    }
	    else {
	      
	      /* recompute with everything 2nd order */
	      mflux[I(ix,iy,iz)] = mfabove + divmean*dz;
	      
	      /* force vertical velocity to be zero at surface */
	      if (iz == 0) {
		mflux[I(ix,iy,iz)] = 0.;
	      }
	      
	      /* if divergence is good and it is the first integration
	       * step (evidenced by zero mfabove), set the bad mflux at the
	       * above level to zero -- this and the surface zero
	       * condition force mass balance on the vertical mass flux
	       */
	      if (mfabove == 0.) {
		mflux[I(ix,iy,iz + 1)] = mfabove;
	      }
	    }
	    
	    /* recompute the horizontal winds at the current level */
	    hwindcalc(ix,iy,iz,rho,ux0,uy0,gx,gy,mflux,wt,u,v);
	    
	    /* compute the vertical particle velocity */
	    if (fabs(mflux[I(ix,iy,iz)]) < badlim &&
		fabs(wt[I(ix,iy,iz)]) < badlim) {
	      wp[I(ix,iy,iz)] = mflux[I(ix,iy,iz)]/rho[iz] + wt[I(ix,iy,iz)];
	    }
	    else {
	      wp[I(ix,iy,iz)] = bad;
	    }
	  }
	}
      }
    }

    /* bottom up integration */
    else {
      
      /* zero mflux at surface and compute horizontal wind */
      for (ix = 0; ix < nx; ix++) {
	for (iy = 0; iy < ny; iy++) {
	  mflux[I(ix,iy,0)] = 0.;
	  hwindcalc(ix,iy,0,rho,ux0,uy0,gx,gy,mflux,wt,u,v);
	}
      }
	
      /* compute the mass divergence at the surface */
      for (ix = 0; ix < nx; ix++) {
	for (iy = 0; iy < ny; iy++) {
	  
	  /* get the mean of the divergence at the current level */
	  if (xygrid) {
	    dflux[I(ix,iy,0)] = xydivcalc(ix,iy,0,u,v,dx,dy);
	  }
	  else {
	    lat = y[iy];
	    dflux[I(ix,iy,0)] = lldivcalc(ix,iy,0,u,v,dx,dy,lat);
	  }
	}
      }
      
      /* integrate up from the surface */
      for (iz = 1; iz < nz; iz++) {
	
	/* get the vertical mass flux at the current level */
	for (ix = 0; ix < nx; ix++) {
	  for (iy = 0; iy < ny; iy++) {
	    
	    /* if the vertical mass flux below is bad, set to zero */
	    if (fabs(mflux[I(ix,iy,iz - 1)]) > badlim) {
	      mfbelow = 0.;
	    }
	    
	    /* if it is not, set it to its actual value */
	    else {
	      mfbelow = mflux[I(ix,iy,iz - 1)];
	    }
	    
	    /* get the horizontal divergence from the level below */
	    divmean = dflux[I(ix,iy,iz - 1)];
	    
	    /* if the divergence is bad, the vertical mass flux is bad */
	    if (fabs(divmean) > badlim) {
	      mflux[I(ix,iy,iz)] = bad;
	    }
	    
	    /* if the divergence is good, we can integrate up */
	    else {
	      mflux[I(ix,iy,iz)] = mfbelow - divmean*dz;
	    }
	    
	    /* compute the horizontal winds at the current level */
	    hwindcalc(ix,iy,iz,rho,ux0,uy0,gx,gy,mflux,wt,u,v);
	  }
	}
	
	/* now compute the div and recompute mflux at the current level */
	for (ix = 0; ix < nx; ix++) {
	  for (iy = 0; iy < ny; iy++) {
	    
	    /* get the mean of the divergence at the current level */
	    if (xygrid) {
	      dflux[I(ix,iy,iz)] = xydivcalc(ix,iy,iz,u,v,dx,dy);
	    }
	    else {
	      lat = y[iy];
	      dflux[I(ix,iy,iz)] = lldivcalc(ix,iy,iz,u,v,dx,dy,lat);
	    }
	    
	    /* compute the mean divergence */
	    divprev = dflux[I(ix,iy,iz - 1)];
	    divcurr = dflux[I(ix,iy,iz)];
	    
	    /* can get mean over current and above levels */
	    if (fabs(divprev) < badlim && fabs(divcurr) < badlim) {
	      divmean = 0.5*(divprev + divcurr);
	    }
	    
	    /* only below level is available */
	    else if (fabs(divprev) < badlim && fabs(divcurr) > badlim) {
	      divmean = divprev;
	    }
	    
	    /* only current level is available */
	    else if (fabs(divprev) > badlim && fabs(divcurr) < badlim) {
	      divmean = divcurr;
	    }
	    
	    /* sorry, nothing available -- divergence bad */
	    else {
	      divmean = bad;
	    }
	    
	    /* if the vertical mass flux below is bad, set it to zero */
	    if (fabs(mflux[I(ix,iy,iz - 1)]) > badlim) {
	      mfbelow = 0.;
	    }
	    
	    /* if it is not, set it to its actual value */
	    else {
	      mfbelow = mflux[I(ix,iy,iz - 1)];
	    }
	    
	    /* recompute the vertical mass flux */
	    if (fabs(divmean) > badlim) {
	      mflux[I(ix,iy,iz)] = bad;
	    }
	    
	    /* compute mflux with more accurate divergence */
	    else {
	      mflux[I(ix,iy,iz)] = mfbelow - divmean*dz;
	    }
	    
	    /* recompute the horizontal winds at the current level */
	    hwindcalc(ix,iy,iz,rho,ux0,uy0,gx,gy,mflux,wt,u,v);
	    
	    /* compute the vertical particle velocity */
	    if (fabs(mflux[I(ix,iy,iz)]) < badlim &&
		fabs(wt[I(ix,iy,iz)]) < badlim) {
	      wp[I(ix,iy,iz)] = mflux[I(ix,iy,iz)]/rho[iz] + wt[I(ix,iy,iz)];
	    }
	    else {
	      wp[I(ix,iy,iz)] = bad;
	    }
	  }
	}
      }

      /* cap with a zero vertical mass flux */
      for (ix = 0; ix < nx; ix++) {
	for (iy = 0; iy < ny; iy++) {
	  for (iz = nz - 1; iz > 0; iz--) {
	    if (fabs(mflux[I(ix,iy,iz - 1)]) < badlim) {
	      mflux[I(ix,iy,iz)] = 0.;
	      break;
	    }
	  }
	}
      }
    }

    /* compute vertical velocity and volume divergence */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
	for (iz = 0; iz < nz; iz++) {
	  if (fabs(mflux[I(ix,iy,iz)]) < badlim) {
	    w[I(ix,iy,iz)] = mflux[I(ix,iy,iz)];
	  }
	  if (fabs(dflux[I(ix,iy,iz)]) < badlim) {
	    div2[I(ix,iy,iz)] = dflux[I(ix,iy,iz)]/rho[iz];
	  }
	}
      }
    }

    /* write output */
    putslice(stdout,nvbuffot,vbuff);
  }

  /* done! */
  exit(0);
}
