/* ogdoppler.h -- old style doppler calculation like in radcart5 */
/* $Id: ogdoppler.c,v 1.7 2009/06/06 23:32:39 raymond Exp $ */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "quickcandis.h"
#include "ngradcart.h"
#include "ngutils.h"

#define MINDET (1.e-10)

/* ogdoppler +++++++++++++++++++++++++++++++++++++++++++++++++++++ */
void ogdoppler(int nx, int ny, int nz, struct param_list plist,
	       struct field_list flist)
{
  int handle, dhx, dhy, dhz, i, num;
  double det, atemp1, atemp2, atemp3, dualtemp, chisqincr;
  double *dual1, *dual2, *ux0, *uy0, *gx, *gy, *eup, *evp,
    *chisq, *avtime, *deltime;

  /* number of grid points */
  num = nx*ny*nz;

  /* allocate space for doppler analysis */
  dual1 = getdbuff(num);
  dual2 = getdbuff(num);
  ux0 = getdbuff(num);
  uy0 = getdbuff(num);
  gx = getdbuff(num);
  gy = getdbuff(num);
  eup = getdbuff(num);
  evp = getdbuff(num);
  chisq = getdbuff(num);
  avtime = getdbuff(num);
  deltime = getdbuff(num);

  /* loop over grid points */
  for (i = 0; i < num; i++) {
  
    /* compute the determinant of the system */
    det = flist.axx[i]*flist.ayy[i] - flist.axy[i]*flist.axy[i];
    
    /* if data are good, det is ok, and more than two wind measurements,
     * make the analysis
     */
    if (flist.chisqbase[i] < BADLIM && fabs(det) > MINDET &&
	flist.pcount[i] > 2.5) {
      
      /* calculate the smaller error covariance eigenvalue */
      atemp1 = flist.axx[i] - flist.ayy[i];
      atemp2 = flist.axy[i];
      dualtemp = pow(atemp1*atemp1 + 4.*atemp2*atemp2,0.5);
      atemp3 = flist.axx[i] + flist.ayy[i];
      dual1[i] = (atemp3 - dualtemp)/2.;
      dual2[i] = (atemp3 + dualtemp)/2.;
      
      /* Dual Doppler velocities */
      ux0[i] = (flist.bx[i]*flist.ayy[i] - flist.axy[i]*flist.by[i])/det;
      uy0[i] = (flist.axx[i]*flist.by[i] - flist.bx[i]*flist.axy[i])/det;
      
      /* now calculate the gx and gy */
      gx[i] = (flist.axy[i]*flist.ayz[i] - flist.ayy[i]*flist.axz[i])/det;
      gy[i] = (flist.axy[i]*flist.axz[i] - flist.axx[i]*flist.ayz[i])/det;
      
      /* next, assuming error in vr is zero */
      eup[i] = gx[i]*gx[i];
      evp[i] = gy[i]*gy[i];
      
      /* Update the reduced chi squared */
      chisqincr = ux0[i]*ux0[i]*flist.axx[i] + uy0[i]*uy0[i]*flist.ayy[i]
	+ 2.*(ux0[i]*uy0[i]*flist.axy[i]
	      - ux0[i]*flist.bx[i] - uy0[i]*flist.by[i]);
      chisq[i] = flist.chisqbase[i] + chisqincr;

      /* compute time variables */
      avtime[i] = flist.time[i];
      deltime[i] = pow(flist.timesq[i] - flist.time[i]*flist.time[i], 0.5);
    }
    
    /* otherwise, set everything to bad */
    else {
      ux0[i] = uy0[i] = gx[i] = gy[i] = BAD;
      eup[i] = evp[i] = BAD;
      dual1[i] = dual2[i] = BAD;
      chisq[i] = BAD;
      avtime[i] = deltime[i] = BAD;
    }
  }

  /* create and write a candis file */
  handle = candiscreate("-", 0);
  candiscomment(handle, "ngradcart:");
  candiscomment(handle, "  radcart5 style doppler calculation");
  candisparam(handle, "bad", BAD);
  candisparam(handle, "badlim", BADLIM);
  candisparam(handle, "zmin0", plist.zmin0);
  candisparam(handle, "zmincoef", plist.zmincoef);
  candisparam(handle, "svx", plist.svx);
  candisparam(handle, "svy", plist.svy);
  candisparam(handle, "reftime", plist.reftime);
  candisparam(handle, "minrange", plist.minrange);
  candisparam(handle, "lon0", plist.lon0);
  candisparam(handle, "lat0", plist.lat0);
  candisparam(handle, "alt0", plist.alt0);
  candisparam(handle, "dx", plist.dx);
  candisparam(handle, "dy", plist.dy);
  candisparam(handle, "dz", plist.dz);
  if (plist.sphere) {
    dhx = candisdim(handle, "lon", nx, plist.lon0, plist.dx);
    dhy = candisdim(handle, "lat", ny, plist.lat0, plist.dy);
  }
  else {
    dhx = candisdim(handle, "x", nx, 0., plist.dx);
    dhy = candisdim(handle, "y", ny, 0., plist.dy);
  }
  dhz = candisdim(handle, "z", nz, 0., plist.dz);
  dcandisfld(handle, "dbz", 3, dhx, dhy, dhz, 0, flist.dbz);
  dcandisfld(handle, "dual1", 3, dhx, dhy, dhz, 0, dual1);
  dcandisfld(handle, "dual2", 3, dhx, dhy, dhz, 0, dual2);
  dcandisfld(handle, "ux0", 3, dhx, dhy, dhz, 0, ux0);
  dcandisfld(handle, "uy0", 3, dhx, dhy, dhz, 0, uy0);
  dcandisfld(handle, "gx", 3, dhx, dhy, dhz, 0, gx);
  dcandisfld(handle, "gy", 3, dhx, dhy, dhz, 0, gy);
  dcandisfld(handle, "eup", 3, dhx, dhy, dhz, 0, eup);
  dcandisfld(handle, "evp", 3, dhx, dhy, dhz, 0, evp);
  dcandisfld(handle, "avtime", 3, dhx, dhy, dhz, 0, avtime);
  dcandisfld(handle, "deltime", 3, dhx, dhy, dhz, 0, deltime);
  dcandisfld(handle, "chisq", 3, dhx, dhy, dhz, 0, chisq);
  dcandisfld(handle, "pcount", 3, dhx, dhy, dhz, 0, flist.pcount);
  candiswrite(handle);
}
