/*
cgm_conmi -- a C version of the Congugate Gradient Minimization
             algorithm by Shanno and Phua.

author:Carlos Lopez Carrillo GRC,NMT

calling convention: 

int cg_CONMINc(int N, double *X, double *F, double *G,
               struct funInfo *finfo,
               int (*CALCFG)(void *,int,void *,void *,void *));

where CALCFG(finfo,N,X,F,G);

N,X,F,G have the same meanings as in the original code(see below), but the
rest of the required parameters are passed using an structure, finfo. 
Furthermnore, the function being minimized is now part of the argument's 
list. The structure, finfo, should be defined in the header file and 
should be the media to carry all extra information required to evaluate 
the target function and the settings of the for the conjugate gradient 
minimization.

Convergence Criteria: The original convergence criteria has been changed
and it is implemented below in function fun_ccc.

Exit Status: It is the same as in the original function, see NFLAG below.

Solution: As in the original code, if the code suceeds, X will be set
to the point where the function has its minimun value. Then F and G are
the values of the function and gradient at X.

--------------------------------------------------------------------------
C REMARK ON ALGORITHM 500
C
C MINIMIZATION OF UNCONSTRAINED MULTIVARIATE FUNCTIONS
C
C BY D.F. SHANNO AND K.H. PHUA
C
C ACM TRANSACTIONS ON MATHEMATICAL SOFTWARE 6 (DECEMBER 1980),
C 618-622
C
      SUBROUTINE CONMIN(N,X,F,G,IFUN,ITER,EPS,NFLAG,MXFUN,W,            CNMN0078
     1IOUT,MDIM,IDEV,ACC,NMETH)
C
C PURPOSE:    SUBROUTINE CONMIN MINIMIZES AN UNCONSTRAINED NONLINEAR
C             SCALAR VALUED FUNCTION OF A VECTOR VARIABLE X
C             EITHER BY THE BFGS VARIABLE METRIC ALGORITHM OR BY A
C             BEALE RESTARTED CONJUGATE GRADIENT ALGORITHM.
C
C USAGE:      CALL CONMIN(N,X,F,G,IFUN,ITER,EPS,NFLAG,MXFUN,W,
C             IOUT,MDIM,IDEV,ACC,NMETH)
C
C PARAMETERS: N      THE NUMBER OF VARIABLES IN THE FUNCTION TO
C                    BE MINIMIZED.
C             X      THE VECTOR CONTAINING THE CURRENT ESTIMATE TO
C                    THE MINIMIZER. ON ENTRY TO CONMIN,X MUST CONTAIN
C                    AN INITIAL ESTIMATE SUPPLIED BY THE USER.
C                    ON EXITING,X WILL HOLD THE BEST ESTIMATE TO THE
C                    MINIMIZER OBTAINED BY CONMIN. X MUST BE DOUBLE
C                    PRECISIONED AND DIMENSIONED N.
C             F      ON EXITING FROM CONMIN,F WILL CONTAIN THE LOWEST
C                    VALUE OF THE OBJECT FUNCTION OBTAINED.
C                    F IS DOUBLE PRECISIONED.
C             G      ON EXITING FROM CONMIN,G WILL CONTAIN THE
C                    ELEMENTS OF THE GRADIENT OF F EVALUATED AT THE
C                    POINT CONTAINED IN X. G MUST BE DOUBLE
C                    PRECISIONED AND DIMENSIONED N.
C             IFUN   UPON EXITING FROM CONMIN,IFUN CONTAINS THE
C                    NUMBER OF TIMES THE FUNCTION AND GRADIENT
C                    HAVE BEEN EVALUATED.
C             ITER   UPON EXITING FROM CONMIN,ITER CONTAINS THE
C                    TOTAL NUMBER OF SEARCH DIRECTIONS CALCULATED
C                    TO OBTAIN THE CURRENT ESTIMATE TO THE MINIZER.
C             EPS    EPS IS THE USER SUPPLIED CONVERGENCE PARAMETER.
C                    CONVERGENCE OCCURS WHEN THE NORM OF THE GRADIENT
C                    IS LESS THAN OR EQUAL TO EPS TIMES THE MAXIMUM
C                    OF ONE AND THE NORM OF THE VECTOR X. EPS
C                    MUST BE DOUBLE PRECISIONED.
C             NFLAG  UPON EXITING FROM CONMIN,NFLAG STATES WHICH
C                    CONDITION CAUSED THE EXIT.
C                    IF NFLAG=0, THE ALGORITHM HAS CONVERGED.
C                    IF NFLAG=1, THE MAXIMUM NUMBER OF FUNCTION
C                       EVALUATIONS HAVE BEEN USED.
C                    IF NFLAG=2, THE LINEAR SEARCH HAS FAILED TO
C                       IMPROVE THE FUNCTION VALUE. THIS IS THE
C                       USUAL EXIT IF EITHER THE FUNCTION OR THE
C                       GRADIENT IS INCORRECTLY CODED.
C                    IF NFLAG=3, THE SEARCH VECTOR WAS NOT
C                       A DESCENT DIRECTION. THIS CAN ONLY BE CAUSED
C                       BY ROUNDOFF,AND MAY SUGGEST THAT THE
C                       CONVERGENCE CRITERION IS TOO STRICT.
C             MXFUN  MXFUN IS THE USER SUPPLIED MAXIMUM NUMBER OF
C                    FUNCTION AND GRADIENT CALLS THAT CONMIN WILL
C                    BE ALLOWED TO MAKE.
C             W      W IS A VECTOR OF WORKING STORAGE.IF NMETH=0,
C                    W MUST BE DIMENSIONED 5*N+2. IF NMETH=1,
C                    W MUST BE DIMENSIONED N*(N+7)/2. IN BOTH CASES,
C                    W MUST BE DOUBLE PRECISIONED.
C             IOUT   IOUT IS A USER  SUPPLIED OUTPUT PARAMETER.
C                    IF IOUT = 0, THERE IS NO PRINTED OUTPUT FROM
C                    CONMIN. IF IOUT > 0,THE VALUE OF F AND THE
C                    NORM OF THE GRADIENT SQUARED,AS WELL AS ITER
C                    AND IFUN,ARE WRITTEN EVERY IOUT ITERATIONS.
C             MDIM   MDIM IS THE USER SUPPLIED DIMENSION OF THE
C                    VECTOR W. IF NMETH=0,MDIM=5*N+2. IF NMETH=1,
C                    MDIM=N*(N+7)/2.
C             IDEV   IDEV IS THE USER SUPPLIED NUMBER OF THE OUTPUT
C                    DEVICE ON WHICH OUTPUT IS TO BE WRITTEN WHEN
C                    IOUT>0.
C             ACC    ACC IS A USER SUPPLIED ESTIMATE OF MACHINE
C                    ACCURACY. A LINEAR SEARCH IS UNSUCCESSFULLY
C                    TERMINATED WHEN THE NORM OF THE STEP SIZE
C                    BECOMES SMALLER THAN ACC. IN PRACTICE,
C                    ACC=10.D-20 HAS PROVED SATISFACTORY. ACC IS
C                    DOUBLE PRECISIONED.
C             NMETH  NMETH IS THE USER SUPPLIED VARIABLE WHICH
C                    CHOOSES THE METHOD OF OPTIMIZATION. IF
C                    NMETH=0,A CONJUGATE GRADIENT METHOD IS
C                    USED. IF NMETH=1, THE BFGS METHOD IS USED.
C
C REMARKS:    IN ADDITION TO THE SPECIFIED VALUES IN THE ABOVE
C             ARGUMENT LIST, THE USER MUST SUPPLY A SUBROUTINE
C             CALCFG WHICH CALCULATES THE FUNCTION AND GRADIENT AT
C             X AND PLACES THEM IN F AND G(1),...,G(N) RESPECTIVELY.
C             THE SUBROUTINE MUST HAVE THE FORM:
C                    SUBROUTINE CALCFG(N,X,F,G)
C                    DOUBLE PRECISION X(N),G(N),F
C
*/

#include "cgm_conmin.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define min(a,b) (a < b ? a : b)
#define max(a,b) (a > b ? a : b)

// The number of times to output during a function
#define IOUT_NUM 100

//
// Internal Functions
//
static void conmninExitStatus(int nflag);

static void fun_ccc(struct funInfo *finfo, double EPS, int Nx,
              double Fprev, double Fcurrent,
              double *XP, double *XC, double *GC);

static void CONMINc(int N, double *X, double *F, double *G,
                  int *ITER, double EPS, int *NFLAG, int MXFUN, double *W,
                  int IOUT, int MDIM,int IDEV, double ACC, int NMETH,
                  struct funInfo *finfo,
                  int (*CALCFG)(struct funInfo *,int,double *,double *,double *));


void cg_CONMINc(int N, double *X, double *F, double *G,
               struct funInfo *finfo,
               int (*CALCFG)(funInfo *,int,double *,double *,double *)) {
//
// This is an interface to CONMINc 
//
int ITER,NFLAG,MXFUN,IOUT,MDIM,IDEV,NMETH;
double EPS,ACC,*W;

  EPS   = (*finfo).cgi.eps;  
  ACC   = (*finfo).cgi.acc; 
  MXFUN = (*finfo).cgi.MXFUN;
  
  NMETH = 0;
  MDIM  = 5*N+2;
  //W should just be MDIM * sizeof(double) but for some reason the algorithm was
  //writing past the end of W, MDIM^2 should be way more than enough room
  //W     = malloc(max(MDIM, N * (N + 5) / 2) * sizeof(double));
  W     = malloc((2*MDIM) * sizeof(double)); //less conservative;
	if (!W) {
   fprintf(stderr, "cgm_conmin: Unable to get memory for Working array \n");
   exit(1);
  }
  IOUT  = finfo->cgi.MXFUN / IOUT_NUM;
  IDEV  = -1;
  
  CONMINc(N,X,F,G, &ITER, EPS, &NFLAG, MXFUN, W, IOUT,MDIM,IDEV,ACC,NMETH,
         finfo, CALCFG);
  free(W);
  conmninExitStatus(NFLAG);
}
//---------------------------------------------------------------------

void CONMINc(int N, double *X, double *F, double *G,
           int *ITER, double EPS, int *NFLAG, int MXFUN, double *W,
           int IOUT, int MDIM,int IDEV, double ACC, int NMETH,
           struct funInfo *finfo,
           int (*CALCFG)(struct funInfo *,int,double *,double *,double *)) {

// CONMINc  -- a minimization subrroutine originally written in fortran
// by Shanno. It is documented in 
//
// Shanno, D.F., and K.H. Phua, 1980:  Remark on the algorithm 500 -- 
// a variable method subrroutine for unconstrained nonlinear 
// minimization. ACM Trans. on Mathematical Software,6,618-622.
//
// The original fortran code (500.gz) was downloded from 
// www.acm.org/pubs/calgo
//
// Notes: 
//
// The version in this file is a rough translation to C
// by Carlos Lopez Carrillo. Only the part that implements the 
// ConjugateGradient algorithm has been translated so far, but the Variable
// lenght part is not hard to do. If the user request the non-implemented
// algorithm, this functions dies with a usage message.
//
// I've try to make a literal translation as much as possible.
// The only real changes where to the begining and end of do loops.
// With that done, the location of the different elements in the
// arrays do not have to be modified at all. The calling convention is 
// also different because, I needed more flexibility regarding the target
// function. So, I have extended the list to include an structure with 
// information required to evaluate the target function, which now is also
//  part of the argument list.
//



double FP,FMIN,ALPHA = 0,AT,AP,GSQ,DG = 0,DG1;
double DP,STEP,DAL,U1,U2,U3,U4;
double XSQ,RTST;  //double XSQ,RTST,DSQRT,DMIN1,DMAX1,DABS;
int RSW;
int IFUN,IOUTK,NX,NG,NRY,NRD,NCONS,NCONS1,NCONS2,I,NCALLS,NXPI,NGPI,NRST,NRDPI,NRYPI;
double DXQ,deltaX;
int PROB = -1;


//C
//C INITIALIZE ITER,IFUN,NFLAG,AND IOUTK,WHICH COUNTS OUTPUT ITERATIONS.
//C
      *ITER=0;
      IFUN=0;
      IOUTK=0;
      *NFLAG=0;
//C
//C SET PARAMETERS TO EXTRACT VECTORS FROM W.
//C W(I) HOLDS THE SEARCH VECTOR,W(NX+I) HOLDS THE BEST CURRENT
//C ESTIMATE TO THE MINIMIZER,AND W(NG+I) HOLDS THE GRADIENT
//C AT THE BEST CURRENT ESTIMATE.
//C
      NX=N;
      NG=NX+N;
//C
//C TEST WHICH METHOD IS BEING USED.
//C IF NMETH=0, W(NRY+I) HOLDS THE RESTART Y VECTOR AND
//C W(NRD+I) HOLDS THE RESTART SEARCH VECTOR.
//C
     if (NMETH == 1) goto L10;

      NRY=NG+N;
      NRD=NRY+N;
      NCONS=5*N;
      NCONS1=NCONS+1;
      NCONS2=NCONS+2;

      goto L20;
//C
//C IF NMETH=1,W(NCONS+I) HOLDS THE APPROXIMATE INVERSE HESSIAN.
//C

L10:
      NCONS=3*N; goto L330; // this goto does not belong here --it is temporal
                            // so, take it off when the segment at L330 is ready.

//C
//C  CALCULATETHE FUNCTION AND GRADIENT AT THE INITIAL
//C POINT AND INITIALIZE NRST,WHICH IS USED TO DETERMINE
//C WHETHER A BEALE RESTART IS BEING DONE. NRST=N MEANS THAT THIS
//C ITERATION IS A RESTART ITERATION. INITIALIZE RSW,WHICH INDICATES
//C THAT THE CURRENT SEARCH DIRECTION IS A GRADIENT DIRECTION.
//C

L20:  
      PROB = (*CALCFG)(finfo,N,X,F,G); 
      //
      // PROB = -1 -- Unknown Function.
      // PROB = 0 -- Auxiliar Problem (*F may be negative);
      // PROB = 1 -- Target Function  (*F must be positive);
      //
      fprintf(stderr,"CONMIN:Solving Problem %i\n",PROB);
      fprintf(stderr,"CONMIN:MXFUN = %i\n",MXFUN);

      IFUN=IFUN+1;
      NRST=N;
      RSW=1;
//C
//C CALCULATE THE INITIAL SEARCH DIRECTION , THE NORM OF X SQUARED,
//C AND THE NORM OF G SQUARED. DG1 IS THE CURRENT DIRECTIONAL
//C DERIVATIVE,WHILE XSQ AND GSQ ARE THE SQUARED NORMS.
//C
      DG1=0.0;
      XSQ=0.0;
      for (I=0; I<N; I++) {
        W[I]=-G[I];
        XSQ=XSQ+X[I]*X[I];
        DG1=DG1-G[I]*G[I];
      }
      GSQ=-DG1;

//C
//C TEST IF THE INITIAL POINT IS THE MINIMIZER.
//C
//      if (GSQ.LE.EPS*EPS*DMAX1(1.0D0,XSQ))RETURN
//
    // 
    // Test only if the initial point is not the trivial solution.
    //
    //if ( XSQ > ACC) {
    //  fun_ccc(finfo,ACC, NX,*F,*F,X,X,G);
    //  if ( (*finfo).CC.c == 1 ) {
    //     fprintf(stderr,"Initial Point is a Minimizer\n"); 
    //     FMIN = *F;
    //     fprintf(stderr,"%i\t%i\t%e\t%e\t\tCONMIN\n",*ITER,IFUN,FMIN,GSQ);
    //     return; 
    //  }
    //}

//C
//C BEGIN THE MAJOR ITERATION LOOP. NCALLS IS USED TO GUARANTEE THAT
//C AT LEAST TWO POINTS HAVE BEEN TRIED WHEN NMETH=0. FMIN IS THE
//C CURRENT FUNCTION VALUE.
//C

L40:
      FMIN=*F;
      NCALLS=IFUN;

//C
//C IF OUTPUT IS DESIRED,TEST IF THIS IS THE CORRECT ITERATION
//C AND IF SO, WRITE OUTPUT.
//C
      if (IOUT > 0 && IOUTK == 0 ) {
        fprintf(stderr,"%i\t%i\t%e\t%e\t\tCONMIN\n",*ITER,IFUN,FMIN,GSQ);
        //fprintf(stderr,"ccG=%0.4e\tccX=%0.4e\tccF=%0.4e\tCONMIN=%i\n",
        //              (*finfo).CC.g,(*finfo).CC.x,(*finfo).CC.f,ITER);
      }
      IOUTK=IOUTK+1;
      if (IOUTK == IOUT)IOUTK=0;

//C
//C BEGIN LINEAR SEARCH. ALPHA IS THE STEPLENGTH.
//C SET ALPHA TO THE NONRESTART CONJUGATE GRADIENT ALPHA.
//C
//60 
//C IF NMETH=1 OR A RESTART HAS BEEN PERFORMED, SET ALPHA=1.0.
//C IF A GRADIENT DIRECTION IS USED, SET ALPHA=1.0/DSQRT(GSQ),
//C WHICH SCALES THE INITIAL SEARCH VECTOR TO UNITY.
//

    //WatchOut for the initial point beeing the trivial sol.
    if ( (*ITER == 0) && (XSQ < ACC) && (DG1 < ACC) ) ALPHA = 1.0;
    else {
       ALPHA = ALPHA*DG/DG1;
       if (NRST == 1 || NMETH == 1)ALPHA=1.0;
       if (RSW == 1) ALPHA=1.0/sqrt(GSQ);
    }

//C
//C THE LINEAR SEARCH FITS A CUBIC TO F AND DAL, THE FUNCTION AND ITS
//C DERIVATIVE AT ALPHA, AND TO FP AND DP,THE FUNCTION
//C AND DERIVATIVE AT THE PREVIOUS TRIAL POINT AP.
//C INITIALIZE AP ,FP,AND DP.
//C
      AP=0.0;
      FP=FMIN;
      DP=DG1;
//C
//C SAVE THE CURRENT DERIVATIVE TO SCALE THE NEXT SEARCH VECTOR.
//C
      DG=DG1;
//C
//C UPDATE THE ITERATION.
//C
      *ITER= *ITER + 1;
//C
//C CALCULATE THE CURRENT STEPLENGTH  AND STORE THE CURRENT X AND G.
//C
      STEP=0.0;
      for (I=0; I<N; I++) {
         STEP = STEP+W[I]*W[I];
         NXPI = NX+I;
         NGPI = NG+I;
         W[NXPI]=X[I];
         W[NGPI]=G[I];
      }
      STEP=sqrt(STEP);

//C
//C BEGIN THE LINEAR SEARCH ITERATION.
//C TEST FOR FAILURE OF THE LINEAR SEARCH.
//C
L80:

      if (ALPHA*STEP > ACC) goto L90;
//C
//C TEST IF DIRECTION IS A GRADIENT DIRECTION.
//C
      if (RSW == 0) goto L20;
      *NFLAG=2;
      fprintf(stderr,"%i\t%i\t%e\t%e\t\tCONMIN\n",*ITER,IFUN,FMIN,GSQ);
      return;
//C
//C CALCULATE THE TRIAL POINT.
//C
L90:
      for( I = 0; I < N; I++ ) {
         NXPI=NX+I;
         X[I]=W[NXPI]+ALPHA*W[I];
      }

//C
//C EVALUATE THE FUNCTION AT THE TRIAL POINT.
//C
      (*CALCFG)(finfo,N,X,F,G);

//C
//C TEST IF THE MAXIMUM NUMBER OF FUNCTION CALLS HAVE BEEN USED.
//C
      IFUN=IFUN+1;
      if (IFUN <= MXFUN) goto L110;
      *NFLAG=1;
      fprintf(stderr,"%i\t%i\t%e\t%e\t\tCONMIN\n",*ITER,IFUN,FMIN,GSQ);
      return;
//C
//C COMPUTE THE DERIVATIVE OF F AT ALPHA.
//C
L110:
      DAL=0.0;
      for (I=0; I <N; I++ ) {
          DAL=DAL+G[I]*W[I];
      }

//C
//C TEST WHETHER THE NEW POINT HAS A NEGATIVE SLOPE BUT A HIGHER
//C FUNCTION VALUE THAN ALPHA=0. IF THIS IS THE CASE,THE SEARCH
//C HAS PASSED THROUGH A LOCAL MAX AND IS HEADING FOR A DISTANT LOCAL
//C MINIMUM.
//C
      if ( *F > FMIN && DAL < 0.0) goto L160;
//C
//C IF NOT, TEST WHETHER THE STEPLENGTH CRITERIA HAVE BEEN MET.
//C
      if (*F > (FMIN+.0001*ALPHA*DG) || fabs(DAL/DG) > (.9)) goto L130;
//C
//C IF THEY HAVE BEEN MET, TEST IF TWO POINTS HAVE BEEN TRIED
//C IF NMETH=0 AND IF THE TRUE LINE MINIMUM HAS NOT BEEN FOUND.
//C
      if( ((IFUN-NCALLS) <= 1) && 
          (fabs(DAL/DG) > EPS) && 
          (NMETH == 0) ) goto L130;
      goto L170;
//C
//C A NEW POINT MUST BE TRIED. USE CUBIC INTERPOLATION TO FIND
//C THE TRIAL POINT AT.
//C
L130: 
      U1=DP+DAL-3.0*(FP- (*F))/(AP-ALPHA);
      U2=U1*U1-DP*DAL;
      if (U2 < 0.0) U2=0.0;
      U2=sqrt(U2);
      AT=ALPHA-(ALPHA-AP)*(DAL+U2-U1)/(DAL-DP+2.*U2);
//C
//C TEST WHETHER THE LINE MINIMUM HAS BEEN BRACKETED.
//C
      if((DAL/DP) > 0.0) goto L140;
//C
//C THE MINIMUM HAS BEEN BRACKETED. TEST WHETHER THE TRIAL POINT LIES
//C SUFFICIENTLY WITHIN THE BRACKETED INTERVAL.
//C IF IT DOES NOT, CHOOSE AT AS THE MIDPOINT OF THE INTERVAL.
//C
      if ( (AT < (1.01*min(ALPHA,AP))) || 
           (AT > (.99*max(ALPHA,AP))) ) AT=(ALPHA+AP)/2.0;
      goto L150;
//C
//C THE MINIMUM HAS NOT BEEN BRACKETED. TEST IF BOTH POINTS ARE
//C GREATER THAN THE MINIMUM AND THE TRIAL POINT IS SUFFICIENTLY
//C SMALLER THAN EITHER.
//C
L140:
      if ( (DAL > 0.0) && 
         (0.0 < AT) && 
         (AT < (.99*min(AP,ALPHA)))) goto L150;
//C
//C TEST IF BOTH POINTS ARE LESS THAN THE MINIMUM AND THE TRIAL POINT
//C IS SUFFICIENTLY LARGE.
//C
      if (DAL <= 0.0 && AT > (1.01*max(AP,ALPHA))) goto L150;
//C
//C IF THE TRIAL POINT IS TOO SMALL,DOUBLE THE LARGEST PRIOR POINT.
//C
      if (DAL <= 0.0) AT=2.0*max(AP,ALPHA);
//C
//C IF THE TRIAL POINT IS TOO LARGE, HALVE THE SMALLEST PRIOR POINT.
//C
      if (DAL > 0.0) AT=min(AP,ALPHA)/2.0;
//C
//C SET AP=ALPHA, ALPHA=AT,AND CONTINUE SEARCH.
//C
L150:
      AP=ALPHA;
      FP=*F;
      DP=DAL;
      ALPHA=AT;
      goto L80;
//C
//C A RELATIVE MAX HAS BEEN PASSED.REDUCE ALPHA AND RESTART THE SEARCH.
//C
L160:
      ALPHA=ALPHA/3.0;
      AP=0.0;
      FP=FMIN;
      DP=DG;
      goto L80;
//C
//C THE LINE SEARCH HAS CONVERGED. TEST FOR CONVERGENCE OF THE ALGORITHM.
//C
L170:
      GSQ=0.0;
      XSQ=0.0;
      DXQ  = 0.0;               
      for (I=0; I<N; I++) {
         GSQ=GSQ+G[I]*G[I];
         XSQ=XSQ+X[I]*X[I];
         deltaX = X[I] - W[NX+I];
         DXQ += deltaX*deltaX;
      }

      //IF(GSQ.LE.EPS*EPS*DMAX1(1.0D0,XSQ))RETURN
      fun_ccc(finfo,ACC, NX,FMIN,*F,&W[NX],X,G);
      if ( (*finfo).CC.c == 1 ) {
        fprintf(stderr,"%i\t%i\t%e\t%e\t\tCONMIN\n",*ITER,IFUN,FMIN,GSQ);
        return; //goto L1000;
      }
      //if (GSQ <= EPS*EPS*max(1.0,XSQ)) return; //BackUp Criteria.


//C
//C SEARCH CONTINUES. SET W(I)=ALPHA*W(I),THE FULL STEP VECTOR.
//C
      for (I=0; I<N; I++) { W[I]=ALPHA*W[I]; }

//C
//C COMPUTE THE NEW SEARCH VECTOR. FIRST TEST WHETHER A
//C CONJUGATE GRADIENT OR A VARIABLE METRIC VECTOR IS USED.
//C
      if (NMETH == 1) goto L330;
//C
//C CONJUGATE GRADIENT UPDATE SECTION.
//C TEST IF A POWELL RESTART IS INDICATED.
//C
      RTST=0.0;
      for (I=0; I<N; I++){
         NGPI = NG+I;
         RTST = RTST+G[I]*W[NGPI];
      }

      if (fabs(RTST/GSQ) > 0.2) NRST=N;
//C
//C IF A RESTART IS INDICATED, SAVE THE CURRENT D AND Y
//C AS THE BEALE RESTART VECTORS AND SAVE D'Y AND Y'Y
//C IN W(NCONS+1) AND W(NCONS+2).
//C
      if (NRST != N) goto L220;

      W[NCONS+1]=0.0;
      W[NCONS+2]=0.0;
      for (I=0; I< N; I++){
         NRDPI = NRD+I;
         NRYPI = NRY+I;
         NGPI = NG+I;
         W[NRYPI] =G[I]-W[NGPI];
         W[NRDPI] =W[I];
         W[NCONS1] =W[NCONS1]+W[NRYPI]*W[NRYPI];
         W[NCONS2] =W[NCONS2]+W[I]*W[NRYPI];
      }

//C
//C CALCULATE  THE RESTART HESSIAN TIMES THE CURRENT GRADIENT.
//C
L220:
      U1=0.0;
      U2=0.0;
      for (I=0; I<N; I++) {
         NRDPI = NRD+I;
         NRYPI = NRY+I;
         U1 = U1 -W[NRDPI]*G[I]/W[NCONS1];
         U2 = U2 +W[NRDPI]*G[I]*2./W[NCONS2]-W[NRYPI]*G[I]/W[NCONS1];
      }

      U3=W[NCONS2]/W[NCONS1];
      for (I=0; I<N; I++) {
         NXPI = NX+I;
         NRDPI = NRD+I;
         NRYPI = NRY+I;
         W[NXPI] = -U3*G[I] -U1*W[NRYPI] -U2*W[NRDPI];
      }

//C
//C IF THIS IS A RESTART ITERATION,W(NX+I) CONTAINS THE NEW SEARCH
//C VECTOR.
//C
      if (NRST == N) goto L300;

//C
//C NOT A RESTART ITERATION. CALCULATE THE RESTART HESSIAN
//C TIMES THE CURRENT Y.
//C
//L250: This label is not used
      U1=0.0;
      U2=0.0;
      U3=0.0;
      U4=0.0;
      for (I=0; I<N; I++) {
         NGPI = NG+I;
         NRDPI = NRD+I;
         NRYPI = NRY+I;
         U1 = U1 - (G[I]-W[NGPI])*W[NRDPI]/W[NCONS1];
         U2 = U2 - (G[I]-W[NGPI])*W[NRYPI]/W[NCONS1] + 
                   2.0*W[NRDPI]*(G[I]-W[NGPI])/W[NCONS2];
         U3 = U3 + W[I]*(G[I]-W[NGPI]);
      }

      STEP = 0.0;
      for ( I=0; I<N; I++){
         NGPI = NG+I;
         NRDPI = NRD+I;
         NRYPI = NRY+I;
         STEP = (W[NCONS2]/W[NCONS1])*(G[I]-W[NGPI]) + 
                U1*W[NRYPI]+U2*W[NRDPI];
         U4 = U4 + STEP*(G[I]-W[NGPI]);
         W[NGPI]=STEP;
      }


//C
//C CALCULATE THE DOUBLY UPDATED HESSIAN TIMES THE CURRENT
//C GRADIENT TO OBTAIN THE SEARCH VECTOR.
//C
      U1=0.0;
      U2=0.0;
      for ( I=0; I < N; I++) {
         U1 = U1-W[I]*G[I]/U3;
         NGPI = NG+I;
         U2 = U2+(1.0+U4/U3)*W[I]*G[I]/U3-W[NGPI]*G[I]/U3;
      }

      for (I=0; I<N; I++) {
        NGPI = NG+I;
        NXPI = NX+I;
        W[NXPI] = W[NXPI]-U1*W[NGPI]-U2*W[I];
      }


//C
//C CALCULATE THE DERIVATIVE ALONG THE NEW SEARCH VECTOR.
//C
L300:   
      DG1=0.0;
      for (I=0; I< N; I++) {
         NXPI = NX+I;
         W[I] = W[NXPI];
         DG1 = DG1+W[I]*G[I];
       }

//C
//C IF THE NEW DIRECTION IS NOT A DESCENT DIRECTION,STOP.
//C
      if (DG1 > 0.0) goto L320;
//C
//C UPDATE NRST TO ASSURE AT LEAST ONE RESTART EVERY N ITERATIONS.
//C
      if (NRST == N)NRST=0;
      NRST=NRST+1;
      RSW=0;
      goto L40;
//C
//C ROUNDOFF HAS PRODUCED A BAD DIRECTION.
//C
L320: 
      *NFLAG=3;
      fprintf(stderr,"%i\t%i\t%e\t%e\t\tCONMIN\n",*ITER,IFUN,FMIN,GSQ);
      return;
//C
//C A VARIABLE METRIC ALGORITM IS BEING USED. CALCULATE Y AND D'Y.
//C
L330: 
      fprintf(stderr,"\n !! NMETH = 1 (VARIABLE METRIC ALGORITM) \n");
      fprintf(stderr,"is not implemented yet. set NMETH to zero!! \n\n");
      exit(1);
/*
      U1=0.0
      DO 340 I=1,N
        NGPI=NG+I
        W(NGPI)=G(I)-W(NGPI)
340   U1=U1+W(I)*W(NGPI)
C
C IF RSW=.TRUE.,SET UP THE INITIAL SCALED APPROXIMATE HESSIAN.
C
      IF(.NOT.RSW)GO TO 380
C
C CALCULATE Y'Y.
C
      U2=0.
      DO 350 I=1,N
        NGPI=NG+I
350   U2=U2+W(NGPI)*W(NGPI)
C
C CALCULATE THE INITIAL HESSIAN AS H=(P'Y/Y'Y)*I
C AND THE INITIAL U2=Y'HY AND W(NX+I)=HY.
C
      IJ=1
      U3=U1/U2
      DO 370 I=1,N
        DO 360 J=I,N
          NCONS1=NCONS+IJ
          W(NCONS1)=0.0
          IF(I.EQ.J)W(NCONS1)=U3
360     IJ=IJ+1
        NXPI=NX+I
        NGPI=NG+I
370   W(NXPI)=U3*W(NGPI)
      U2=U3*U2
      GO TO 430
C
C CALCULATE W(NX+I)=HY AND U2=Y'HY.
C
380   U2=0.0
      DO 420 I=1,N
        U3=0.0
        IJ=I
        IF(I.EQ.1)GO TO 400
        II=I-1
        DO 390 J=1,II
          NGPJ=NG+J
          NCONS1=NCONS+IJ
          U3=U3+W(NCONS1)*W(NGPJ)
390     IJ=IJ+N-J
400     DO 410 J=I,N
          NCONS1=NCONS+IJ
          NGPJ=NG+J
          U3=U3+W(NCONS1)*W(NGPJ)
410     IJ=IJ+1
        NGPI=NG+I
        U2=U2+U3*W(NGPI)
        NXPI=NX+I
420   W(NXPI)=U3
C
C CALCULATE THE UPDATED APPROXIMATE HESSIAN.
C
430   U4=1.0+U2/U1
      DO 440 I=1,N
        NXPI=NX+I
        NGPI=NG+I
440   W(NGPI)=U4*W(I)-W(NXPI)
      IJ=1
      DO 450 I=1,N
        NXPI=NX+I
        U3=W(I)/U1
        U4=W(NXPI)/U1
        DO 450 J=I,N
          NCONS1=NCONS+IJ
          NGPJ=NG+J
          W(NCONS1)=W(NCONS1)+U3*W(NGPJ)-U4*W(J)
450   IJ=IJ+1
C
C CALCULATE THE NEW SEARCH DIRECTION W(I)=-HG AND ITS DERIVATIVE.
C
      DG1=0.0
      DO 490 I=1,N
        U3=0.0
        IJ=I
        IF(I.EQ.1)GO TO 470
        II=I-1
        DO 460 J=1,II
          NCONS1=NCONS+IJ
          U3=U3-W(NCONS1)*G(J)
460     IJ=IJ+N-J
470     DO 480 J=I,N
          NCONS1=NCONS+IJ
          U3=U3-W(NCONS1)*G(J)
480     IJ=IJ+1
        DG1=DG1+U3*G(I)
490   W(I)=U3
C
C TEST FOR A DOWNHILL DIRECTION.
C
      IF(DG1.GT.0.)GO TO 320
      RSW=.FALSE.
      GO TO 40
*/
//500   FORMAT(10H ITERATION,I5,20H      FUNCTION CALLS,I6/5H F = ,
//    1D15.8,13H G-SQUARED = ,D15.8/)
//     END
}

//
//-------------------------------------------------------------
//

void conmninExitStatus(int nflag) {
//
// NFLAG=0, THE ALGORITHM HAS CONVERGED.
// NFLAG=1, THE MAXIMUM NUMBER OF FUNCTION EVALUATIONS HAVE BEEN USED.
// NFLAG=2, THE LINEAR SEARCH HAS FAILED TO IMPROVE THE FUNCTION VALUE.
//          THIS IS THE USUAL EXIT IF EITHER THE FUNCTION OR THE GRADIENT 
//          IS INCORRECTLY CODED.
// NFLAG=3, THE SEARCH VECTOR WAS NOT A DESCENT DIRECTION. THIS CAN ONLY 
//          BE CAUSED BY ROUNDOFF,AND MAY SUGGEST THAT THE CONVERGENCE 
//          CRITERION IS TOO STRICT.

  while(1){
    if ( nflag == 0 ) {
      fprintf(stderr,"CONMIN: NFLAG=0\n");
      fprintf(stderr,"THE ALGORITHM HAS CONVERGED.\n\n");
      break;
    }
    if ( nflag == 1 ) {
      fprintf(stderr,"CONMIN:  NFLAG=1\n");
      fprintf(stderr,"THE MAXIMUM NUMBER OF FUNCTION EVALUATIONS HAVE BEEN USED.\n\n");
      break;
    }
    if ( nflag == 2 ) {
      fprintf(stderr,"CONMIN: NFLAG=2\n");
      fprintf(stderr,"THE LINEAR SEARCH HAS FAILED TO IMPROVE THE FUNCTION VALUE.\n\n");
      fprintf(stderr,"THIS IS THE USUAL EXIT IF EITHER THE FUNCTION OR THE GRADIENT\n");
      fprintf(stderr,"IS INCORRECTLY CODED.\n\n");
      break;
    }
    if ( nflag == 3 ) {
      fprintf(stderr,"CONMIN: NFLAG=3\n"); 
      fprintf(stderr,"THE SEARCH VECTOR WAS NOT A DESCENT DIRECTION. THIS CAN ONLY\n");
      fprintf(stderr,"BE CAUSED BY ROUNDOFF,AND MAY SUGGEST THAT THE CONVERGENCE\n"); 
      fprintf(stderr,"CRITERION IS TOO STRICT.\n\n");
      break;
    }
    fprintf(stderr,"CONMIN: UnknowExitStatus.. Quiting!!\n\n");exit(1);
  }
}
//
//-------------------------------------------------------------
//
// Various convergence criteria
//
//-------------------------------------------------------------------
//
void fun_ccc(struct funInfo *finfo, double EPS, int Nx,
              double Fprev, double Fcurrent,
              double *XP, double *XC, double *GC){
//
// CCC--ConvergenceCriteriaCheck
//
// Nash and Soffer Convergence Test
// 
// no special meaning is attached to the function or its gradient. 
// This is a general L2norm convergence criteria used during the 
// minimization of a general function
//
//
//
int i;
double GSQ,XSQ,DXQ,DeltaX,T,DT;
double eps4Fseq,eps4Xseq,eps4grad;
double N1G; //Norm1 of G.

  eps4Fseq = EPS; // most accurate
  eps4Xseq = eps4grad = sqrt(EPS); // less accurate

  GSQ=0.0;
  XSQ=0.0;
  DXQ=0.0;
  N1G = 0;
  for (i=0; i<Nx; i++){
    DeltaX = XC[i] - XP[i];
    XSQ +=(XC[i]*XC[i]);
    GSQ +=(GC[i]*GC[i]);
    DXQ += (DeltaX*DeltaX);
    if ( fabs(GC[i]) > N1G ) N1G = fabs(GC[i]);
  }
  T = 1.0 + fabs(Fcurrent);
  DT = (Fprev - Fcurrent);

  //Normalize to GSQ the number of variables --critical for large problems.
  //if (Nx > 1000) { 
  //   GSQ /= ((double) (Nx*Nx)); // GSQ= N1G*N1G;
  //   T   /= ((double) Nx);
  //   DT  /= ((double) Nx);
  //}

  (*finfo).CC.g =  sqrt(GSQ)/T ;
  (*finfo).CC.x =  sqrt(DXQ)/(1.0 + sqrt(XSQ));
  (*finfo).CC.f =  DT/T ;

  (*finfo).CC.c = 0; //first check that trivial solution is not accepted.
  if ( XSQ > EPS && \
       ((*finfo).CC.g <= eps4grad) && \
       ((*finfo).CC.f <  eps4Fseq) && \
       ((*finfo).CC.x <  eps4Xseq) ) (*finfo).CC.c=1;

}
//--------------------------------------------------------------
//
//int fun_ncc(double EPS, int NX, double Fprev,double Fcurrent,
//           double XSQ,double GSQ, double DXQ);
//int fun_ncc(double EPS, int NX, double Fprev,double Fcurrent,
//          double XSQ,double GSQ, double DXQ);
//
//------------------------------------------------------------
//
int fun_scc(double EPS, int NX, double Fprev,double Fcurrent,
            double XSQ,double GSQ, double DXQ){
//
// Shanno Convergence Test, for simplicity we have the
// same interface as in fun_ncc
//
   if (GSQ <= EPS*EPS*max(1.0,XSQ)) return(1);
   else return(0);
}
//
//------------------------------------------------------------
//
int fun_ncc(double EPS, int NX, double Fprev,double Fcurrent,
            double XSQ,double GSQ, double DXQ){
//
// New Convergence Criteria
// EPS is supossed to the accurecy on the function values -- tyupically
// EPSmachine;
//
// Following Nash and Sofer (pg 375), the function values are more
// accurate than the vector X and the gradient. i.e. if the function
// is known to an accuracy of EPSmach, then X and the gradient will only
// be accurate to sqrt(EPSmachine). Hence, I'll set the eps values as
//
//           
//           eps4Fsequence = EPS; 
//           eps4Xsequence = eps4grad = sqrt(EPS);
//
// Also, we have to scale the norm-two by the number of variables in
// the problem as their number is quite large and the value of 
// the norm-two of a vector scales with the number of variables.
//
double T;
double cc4grad, cc4Fseq, cc4Xseq;
double eps4grad,eps4Fseq,eps4Xseq;

  
  //Normalize to the number of variables --critical for large problems.
  GSQ /= (double) NX;
  // XSQ /= (double) NX;
  // DXQ /= (double) NX;
  
  //Define criterias
  T = 1.0 + fabs(Fcurrent);

  cc4grad = sqrt(GSQ)/T;                 // Criteria to identify a Min
  cc4Fseq = (Fprev - Fcurrent)/T;        //Convergence of sequence {Fi}
  cc4Xseq = sqrt(DXQ)/(1.0 + sqrt(XSQ)); //Convergence of sequence {Xi}

  eps4Fseq = EPS;
	eps4Xseq = eps4grad = sqrt(EPS); 

fprintf(stderr,"fun_ncc: ccG=%0.4e\tccX=%0.4e\tccF=%0.4e\n",
                cc4grad,cc4Xseq,cc4Fseq);

  //
  // Nash and Soffer Convergence Test
  //
     if ( (cc4grad <= eps4grad) && \
          (cc4Fseq <  eps4Fseq) && \
          (cc4Xseq <  eps4Xseq) ) return(1);
	   else return(0);
}
