#include "nglfit.h"
#include "ngutil.h"
#include <cdfhdr.h>
#include <assert.h>
#include <math.h>


/* define functions that should be defined in cdfhdr.h */
int seekpar(char buffer[][LINE], int max, char pname[], char value[]); /* from seekpar.c */
int gethdr(FILE *stream, char buffer[][LINE], int max); /* from cdfio.c */

/* local functions */

/**
 * Gets a pointer from either the static or variable slice.
 * Note that field must be freed by caller.
 *
 * Exits on failure.
 */
float* ngGetPtr(char hbuff[HBMAX][LINE], float *sbuff, float *vbuff, char *name,
		struct field **field)
{
  float *ret;
  
  /* first check the static slice */
  ret = getptr2(hbuff, HBMAX, sbuff, 's', 'r', name, field);
  
  if(!ret)
    {
      /* next check the variable slice */
      ret = getptr2(hbuff, HBMAX, vbuff, 'v', 'r', name, field);
    }
  
  if(!ret)
    {
      /* we have failed */
      fprintf(stderr, "ngsonde: Failed to find field named: '%s'", name);
      exit(1);
    }
  
  return ret;
}

/**
 * Check to see if a value is bad.
 *
 * @return True if the value is good, false if bad.
 */
int ngCheck(float value, float bad, float badlim)
{
  return value != bad && value < badlim;
}

/* functions defined in nglfit.h */
void ngFreeSonde(ngSonde *sonde)
{
  size_t i;
  
  free(sonde->x);
  sonde->x = NULL;
  
  free(sonde->y);
  sonde->y = NULL;
  
  
  free(sonde->z);
  sonde->z = NULL;
  
  
  if(sonde->data)
    {
      for(i = 0; i < sonde->count; i++)
	{
	  if(sonde->data[i])
	    {
	      free(sonde->data[i]);
	      sonde->data[i] = NULL;
	    }
	}
      
      free(sonde->data);
      sonde->data = NULL;
    }
  
  sonde->count = 0;
}

float ngZcalc(float pres)
{
  float gamma = 5.5;
  float gee = 9800.;
  float t0 = 300;
  float rd = 287;
  float psurf = 1020.;
  float zscale, power, z;
  
  zscale = t0 / gamma;
  power = rd * gamma / gee;
  z = zscale * (1 - pow(pres / psurf, power));
  
  return z;
}

void ngReadData(FILE *fin, ngCMDArgs *args, ngSonde **dest, size_t *destCount,  float *BaD, float *BadLim)
{
  /*constants*/
  const float D2R = (float)(acos(-1) / 180.0); /* degrees to radians conversion (acos -1 = PI) */
  const float EARTH_R = 6370.0;  /* radius of earth (km) */

  /* stuff for D. J. Raymond's mean longitude-latitude hack */
  float meanlon, meanlat;
  int nlon, nlat;
  
  /* sonde variables */
  ngSonde *sonde;
  float *gph, *lat, *lon, *time, *pres, *idx; /* temp = tempature */
  float **data;
  int timeDim, warned = 0;
  size_t nIdx, index;
  //size_t nPres;
  size_t i, j, k;
  float dt, dx, dy;
  
  /* variables for reading the input file */
  char hbuff[HBMAX][LINE];
  char c1[LINE];
  float *sbuff, *vbuff;
  long nsbuff, nvbuff;
  float bad, badlim;
  struct field *fp;

  /* read in the candis file */
  gethdr(fin, hbuff, HBMAX);
  if(getfmt(hbuff, HBMAX) != 'f')
    {
      fprintf(stderr, "ngsonde: File format must be f");
      exit(1);
    }
  
  if(seekpar(hbuff, HBMAX, "bad", c1) == OK)
    {
      bad = atof(c1); 
    } else
    {
      bad = BAD; 
    }
  *BaD = bad;
  
  if(seekpar(hbuff, HBMAX, "badlim", c1) == OK)
    {
      badlim = atof(c1);
    } else 
    {
      badlim = BADLIM;
    }
  *BadLim = badlim;
  
  nsbuff = elemcnt(hbuff, HBMAX, 's');
  sbuff = getbuff(nsbuff);
  nvbuff = elemcnt(hbuff, HBMAX, 'v');
  vbuff = getbuff(nvbuff);
  getslice(fin, nsbuff, sbuff);
  
  /* we don't free dest so we need to make sure we don't have a memory leak */
  /* we also use realloc so dest can't be a wild pointer */
  assert(*dest == NULL);
  assert(*destCount == 0);
  
  while(getslice(fin, nvbuff, vbuff) != EOF)
    {
      (*destCount)++;
      *dest = realloc(*dest, (*destCount) * sizeof(ngSonde));
      sonde = *dest + ((*destCount) - 1);
      sonde->x = NULL;
      sonde->y = NULL;
      sonde->z = NULL;
      sonde->data = NULL;
      sonde->count = 0;
      
      /* retrieve the pointers need to calculate lon, lat, z, and t */
      
      idx = ngGetPtr(hbuff, sbuff, vbuff, args->idx, &fp);
      if(fp->dim != 1 || strcmp(fp->dname1, args->idx) != 0) {
	fprintf(stderr, "ngsonde: index must be One-Dimensional.\n");
	exit(1);
      }
      nIdx = fp->dsize1;
      free(fp);
      
      // check if pressure is required to calculate altitude.
      if ( args->pres != NULL) { // we need the pressure
	
	// check if the index is already  the pressure --default case
    	if ( strcmp(args->idx,args->pres) == 0 ) {
	  pres = idx; //press points to the same place as the index.
	  //nPres = nIdx;
	} else {	
	  //presure is not the index, so we need to get a pointer to it.
	  pres = ngGetPtr(hbuff, sbuff, vbuff, args->pres, &fp);
	  if(fp->dim != 1 || strcmp(fp->dname1, args->idx) != 0) {
	    fprintf(stderr, "ngsonde: pres -- the index for pressure must be %s\n", args->idx);
	    exit(1);
	  }
	  //nPres = fp->dsize1; // should be the same as nIdx
	  free(fp);
	}
      } else { pres = NULL;} // we don't need the pressure
      
      if (args->gph != NULL ) {
	gph = ngGetPtr(hbuff, sbuff, vbuff, args->gph, &fp);
	if(fp->dim != 1 || strcmp(fp->dname1, args->idx) != 0) {
	  fprintf(stderr, "ngsonde: The geopotential height field must be a 1-d filed whose index name is %s  \n", args->idx);
	  exit(1);
	}
	free(fp);
      } else gph = NULL;
      
      lon = ngGetPtr(hbuff, sbuff, vbuff, args->lon, &fp);
      if(fp->dim != 1 || strcmp(fp->dname1, args->idx) != 0)
	{
	  fprintf(stderr, "ngsonde: The longitude field must be a 1d filed whose index name is %s\n", args->idx);
	  exit(1);
	}
      free(fp);
      
      lat = ngGetPtr(hbuff, sbuff, vbuff, args->lat, &fp);
      if(fp->dim != 1 || strcmp(fp->dname1, args->idx) != 0)
	{
	  fprintf(stderr, "ngsonde: The latitude field must be a 1d field whose index name is %s\n", args->idx);
	  exit(1);
	}
      free(fp);
      
      time = ngGetPtr(hbuff, sbuff, vbuff, args->time, &fp);
      timeDim = fp->dim;
      if(timeDim == 0)
	{
	  if(!warned)
	    {
	      fprintf(stderr, "ngsonde: warning, time has only one dimension\n");
	    }
	  warned = 1;
	} else if(timeDim != 1 || strcmp(fp->dname1, args->idx) != 0)
	{
	  fprintf(stderr, "ngsonde: If the time field has a dimension it must be %s \n", args->idx);
	  exit(1);
	}
      free(fp);
      
      
      /* allocate space for lfit fields */
      data = malloc(sizeof(float*) * args->numFields);
      if(!data)
	{
	  fprintf(stderr, "ngsonde: Out of memory at line __LINE__ of __FILE__\n");
	  exit(1);
	}
      
      for(i = 0; i < args->numFields; i++)
	{
	  data[i] = ngGetPtr(hbuff, sbuff, vbuff, args->fields[i].inName, &fp);
	  if(fp->dim != 1 || strcmp(fp->dname1, args->idx) != 0)
	    {
	      fprintf(stderr, "ngsonde: %s must be indexed by %s \n", args->fields[i].inName,args->idx);
	      exit(1);
	    }
	  free(fp);
	}
      
      /* Hack by D. J. Raymond to put mean value of lon and lat into lon and lat fields
       * first we have to compute these means
       */
      meanlon = 0.;
      meanlat = 0.;
      nlon = 0;
      nlat = 0;
      for (i = 0; i < nIdx; i++) {
	if (fabs(lon[i]) < badlim) {
	  meanlon += lon[i];
	  nlon++;
	}
	if (fabs(lat[i]) < badlim) {
	  meanlat += lat[i];
	  nlat++;
	}
      }
      if (nlon > 0 && nlat > 0) {
	meanlon = meanlon/nlon;
	meanlat = meanlat/nlat;
      }
      else {
	meanlon = meanlat = bad;
      }
      for (i = 0; i < nIdx; i++) {
	lon[i] = meanlon;
	lat[i] = meanlat;
      }
	
      
      /* now copy the data into the ngSonde structure */
      if(timeDim > 0 || ngCheck(*time, bad, badlim))
	{
	  sonde->count = 0;
	  //for(j = 0; j < nPres; j++)
	  for(j = 0; j < nIdx; j++)
	    {
	      //i = nPres - j - 1;
	      i = nIdx - j - 1;
	      /* ensure there is no bad data */
	      if(ngCheck(lon[i],  bad, badlim) &&
		 ngCheck(lat[i],  bad, badlim) &&
		 ngCheck(idx[i], bad, badlim) &&
		 (pres == NULL || ngCheck(pres[i], bad, badlim)) &&
		 (gph  == NULL || ngCheck(gph[i] , bad, badlim)) &&
		 (timeDim < 1  || ngCheck(time[i], bad, badlim)))
		{
		  /* encrease the number data points */
		  sonde->count++;
		  sonde->x = realloc(sonde->x,       sizeof(float) * sonde->count);
		  sonde->y = realloc(sonde->y,       sizeof(float) * sonde->count);
		  sonde->z = realloc(sonde->z,       sizeof(float) * sonde->count);
		  sonde->data = realloc(sonde->data, sizeof(float*) * sonde->count);
		  sonde->data[sonde->count - 1] = malloc(sizeof(float) * args->numFields);
		  
		  /* copy the data into the sonde structure */
		  index = sonde->count - 1;
		  dt = timeDim < 1 ? * time : time[i];
		  dt -= args->reftime; //HERE Substract reference time.
		  dx = args->svx * dt;
		  dy = args->svy * dt;
		  sonde->x[index] = lon[i] - dx / (EARTH_R * D2R * cos(lat[i] * D2R));
		  sonde->y[index] = lat[i] - dy / (EARTH_R * D2R);
		  
		  // By default, we use the standard atmosphere to convert pressure
		  // to height. However if the user indicates that the geopotential
		  // height should be used, then we will honor the request.
		  if ( args->gph == NULL)
		    sonde->z[index] = ngZcalc(pres[i]);
		  else
		    sonde->z[index] = gph[i];
		  
		  /* now copy data fields */
		  for(k = 0; k < args->numFields; k++)
		    {
		      sonde->data[index][k] = ngCheck(data[k][i], bad, badlim) ? data[k][i] : bad;
		    }
		}
	    }
	  fprintf(stderr, "ngsonde: %u good data points for sonde %u\n", (unsigned int)sonde->count, (unsigned int)(*destCount));
	  if (sonde->count == 0) {
	    fprintf(stderr,"ngsonde: Oops, bad sonde!\n");
	    exit(1);
	  }
	}
      
      free(data); /* we do not have to free what data points to  as candis should do that */
      data = NULL;
    }
  
  
  /* clean up */
  free(sbuff);
  sbuff = NULL;
  free(vbuff);
  vbuff = NULL;
}

float* ngLocalFit(ngCMDArgs *args, unsigned int findex, ngSonde *sondes, size_t numSondes,float bad, float badlim)
{
  float *ret = malloc(sizeof(float) * args->nx * args->ny * args->nz);
  int i, j, k, ix, iy, iz;
  float x, y, z;
  ngSonde *sonde;
  float xmax = args->dx, ymax = args->dy, zmax = args->dz;
  float zmaxsonde;
  float *weights, *values;
  long double top, bottom, temp;
  int nvalues;

  fprintf(stderr, "ngsonde: (%u/%u) Processing %s\n", findex + 1, args->numFields, args->fields[findex].outName);
  for(ix = 0; ix < args->nx; ix++)
    {
      x = ((float)ix * (args->dx * (float)args->nx)) / ((float)args->nx) + args->sx;
      for(iy = 0; iy < args->ny; iy++)
	{
	  y = ((float)iy * (args->dy * (float)args->ny)) / ((float)args->ny) + args->sy;
	  for(iz = 0; iz < args->nz; iz++)
	    {
	      z = ((float)iz * (args->dz * (float)args->nz + args->sz)) / ((float)args->nz) + args->sz;
	      
	      /* now find what datapoints will effect the gridpoint at x, y, z */
	      nvalues = 0;
	      weights = NULL;
	      values = NULL;
	      for(j = 0; j < numSondes; j++)
		{
		  sonde = &sondes[j];

		  // the z range should be the sonde z step rather than the analysis z step. Otherwise
		  // the surface value of analysis target is biased to data above the surface.
		  // D. J. Raymond
		  zmaxsonde = sonde->z[1] - sonde->z[0];
		  
		  for(k = 0; k < sonde->count; k++)
		    {
		      if(fabs(z - sonde->z[k]) <= zmaxsonde && fabs(x - sonde->x[k]) <= xmax &&
			 fabs(y - sonde->y[k]) <= ymax)
			{
			  nvalues++;
			  weights = realloc(weights, sizeof(float) * nvalues);
			  values = realloc(values, sizeof(float) * nvalues);
			  weights[nvalues - 1] =
			    (x - sonde->x[k]) * (x - sonde->x[k]) +
			    (y - sonde->y[k]) * (y - sonde->y[k]) +
			    (z - sonde->z[k]) * (z - sonde->z[k]);
			  values[nvalues - 1] = sonde->data[k][findex];
			}
		    }
		}
	      
	      /* now calculate the return value */
	      i = gp(args->ny, args->nz, ix, iy, iz);
	      if(nvalues <= 0)
		{
		  ret[i] = bad; //HERE
		} else {
		if(nvalues == 1)
		  {
		    ret[i] = values[0];
		  } else
		  {
		    top = 0;
		    bottom = 0;
		    for(j = 0; j < nvalues; j++)
		      {
			temp = 1;
			for(k = 0; k < nvalues; k++)
			  {
			    if(k != j)
			      {
				temp *= weights[k];
			      }
			  }
			//HERE
			if (values[j] < badlim) {
			  top += values[j] * temp;
			  bottom += temp;
			}
		      }
		    ret[i] = bottom != 0 ? top / bottom : bad; //HERE
		  }
		
		free(values);
		free(weights);
		values = NULL;
		weights = NULL;
	      }
	    }
	}
    }
  
  return ret;
}
