// ngsonde_tf_misfit -- generates the contribution from dropsonde data
// to the target funtion. 
//
// The target function is considered as a a sumation of terms that
// measure the misfit of the analyzed field to several conditions.
//
// ngsonde_tf_misfit returns the contribution to the target function due
// to the misfit between the analyzed field and the sonde data.
// It also calculates the contribution to the gradient of the target
// function with respect to the analyzed filed due to the misfit
// with the sonde data. 
//
// The Input sonde data is actually the gridded data reduced by nglfit. 
//
// Carlos Lopez Carrillo*  and Max Brister+
// *Geophysical Research Center, NMT.
// +Undergrad at the Physics Department, NMT
//

#include "ngsonde_tf_misfit.h"
#include "ngutil.h"
#include "cdfhdr.h"
#include "nggfit.h"
#include <math.h>
#include <stdio.h>

#define USAGE "should be called by function calcfg inside calcfg_if.c"

//
// Internal Functions.
//
double fun_misfit(struct funInfo *finfo,
				  double *afld,double *cgf,double *ndat);
//-------------------------------------------------------------------

//
// Exported function
//
double if_misfit(struct funInfo *finfo,int N, 
                double *ndat,double *afld,double *g)
//
// Return value -- contribution to the function from radar gridded data
// funInfo -- structure  function information
// f -- 1D vector that holds the analyzed field on the grid.
// g -- 1D vector that holds the contribution from sonde data to 
//      the gradient with respect to the analyzed field components
//      of the function being minimized. 
// N -- is the size of the vectors f and g
// ndat -- on output will have the number of grid points that actualy
//         have good data.
//
{
	double cf;

	cf = fun_misfit(finfo,afld,g,ndat);

	return(cf); 
}
//-------------------------------------------------------------------
double fun_misfit(struct funInfo *finfo, 
				  double *afld,double *cgf,double *ndat){

	int ig,jg,kg,pg; //,pgpNp,pgp2Np;
	int Nxg,Nyg,Nzg; //,Np;

	double bad,badlim;
	double value, svalue;
	double misfit,sum;
	ngFitInfo *ginfo = (ngFitInfo*)finfo->userData;

	

	Nxg = ginfo->nx; //(*finfo).grid.Nx; 
	Nyg = ginfo->ny; //(*finfo).grid.Ny;
	Nzg = ginfo->nz; //(*finfo).grid.Nz;

	bad = (double)ginfo->bad;
	badlim = (double)ginfo->badlim;

	sum   = 0.0;
	*ndat = 0.0;

	// loop on Data Grid
	for (kg = 0; kg < Nzg; kg++) {
		for (jg = 0; jg < Nyg; jg++) {
			for (ig = 0; ig < Nxg; ig++) {
       
				pg = gp(Nyg, Nzg, ig, jg, kg);
				value = ginfo->data[pg];

				if ( fabs(value) < badlim) {
               
					svalue = afld[pg];

					//
					// weighted contribution to misfit.
					//
					misfit = (value - svalue) * (value - svalue);

					//
					// make ppropiate contributions to the gradient. 
					//
					cgf[pg] += svalue - value;

					// contribution to functional.
					sum   += misfit;
					*ndat += 1.0;

				} //good data
			} //i-loop
		} //j-loop
	} //k-loop

	return((double) 0.5*sum); 
  
}
