/* quickcandis -- This set of functions implements a higher level
 * (and somewhat restricted) method for Candis files.
 * The major restrictions are: a) only one variable slice, and
 * b) only index fields in static slice, and c) index fields
 * are required for each dimension.
 */

#include <stdio.h>
#include "cdfhdr.h"
#include "string.h"

#define HANDLE 1
#define MAXNAMLEN 30
#define MAXIFLDS 4
#define SMUL 1.e10
#define DEBUG 0

static int open_flag = 0;           /* this is set to 1 when a file is open */
static FILE *candis_file;           /* file pointer */
static int not_stdout;              /* flag for opening and closing */
static char hbuff[HBMAX][LINE];     /* header buffer */
static char cline[LINE],pline[LINE];  /* character line buffers */
static float *vfields[HBMAX];       /* pointers to potential variable fields */
static long vsizes[HBMAX];          /* sizes of variable fields */
static int nfields;                 /* number of variable fields */
static float *ifields[MAXIFLDS];    /* pointers to potential index fields */
static int nindex;                  /* number of index fields (dimensions) */
static long isizes[MAXIFLDS];       /* dimension sizes */
static char inames[MAXIFLDS][LINE]; /* names of dimensions */
static long stat_elements;     /* total number of elements in static slice */
static long var_elements;      /* total number of elements in variable slice */

static int checkstring();

/* candiscreate -- This routine opens a new Candis file for output. */
/* It returns a file handle that needs to be used for all other calls. */

int candiscreate(file_name,instance)
char *file_name;                    /* name of Candis file to be created */
int instance;                       /* qualifier for file name */
{
  char q_file_name[256];

#if DEBUG
  fprintf(stderr,"candiscreate: %s.%4d\n",file_name,instance);
  fflush(stderr);
#endif

/* if a file is open already, croak */
  if (open_flag) {
    fprintf(stderr,"candiscreate: a Candis file is already open for output\n");
    exit(1);
  }

/* check file name */
  if (checkstring(file_name) > LINE) {
    fprintf(stderr,"candiscreate: file name %s too long\n",file_name);
    exit(1);
  }

/* if file name is "-", treat as stdout and don't open or close */
  if (strcmp(file_name,"-") == 0) not_stdout = 0;
  else not_stdout = 1;

/* see if we can open the desired file for write */
  if (not_stdout) {

/* first compute fully qualified file name */
    sprintf(q_file_name,"%s.%04d",file_name,instance);
    if ((candis_file = fopen(q_file_name,"w")) == NULL) {
      fprintf(stderr,"candiscreate: cannot open file %s\n",file_name);
      exit(1);
    }
  }
  else {
    candis_file = stdout;
  }

/* if so, create a null Candis header buffer */
  nullhdr(hbuff,HBMAX,"float");

/* initialize some things */
  nfields = nindex = 0;
  stat_elements = var_elements = 0;

/* set lock and return */
  open_flag = 1;
  return(HANDLE);
}

/* candiscomment -- Add a comment to Candis header. */

int candiscomment(handle,comment)
int handle;                           /* file handle */
char *comment;                        /* comment to be inserted */
{

#if DEBUG
  fprintf(stderr,"candiscomment: %d %s\n",handle,comment);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candiscomment: Candis file is not open\n");
    exit(1);
  }

/* it is, and we can add comment after checks */
  if (checkstring(comment) >= LINE - 1) {
    fprintf(stderr,"candiscomment: comment %s is too long\n",comment);
    exit(1);
  }
  sprintf(cline,"%s\n",comment);
  addcline(hbuff,HBMAX,cline);
  return(handle);
}

/* candisparam -- Add a parameter to header. */

int candisparam(handle,parameter,value)
int handle;                         /* file handle */
char *parameter;                    /* parameter name */
float value;                        /* parameter value */
{

#if DEBUG
  fprintf(stderr,"candisparam: %d %s %g\n",handle,parameter,value);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candisparam: Candis file is not open\n");
    exit(1);
  }

/* add the comment after appropriate checks */
  if (checkstring(parameter) >= MAXNAMLEN) {
    fprintf(stderr,"candisparam: parameter name %s is too long\n",
      parameter);
    exit(1);
  }
  sprintf(cline,"%g",value);
  addpar(hbuff,HBMAX,parameter,cline);
  return(handle);
}

/* add a dimension */
/* returns a dimension handle */

int candisdim(handle,dim_name,size,start_val,increment)
int handle;                    /* file handle */
char *dim_name;                /* name of the dimension */
long size;                     /* number of elements in dimension direction */
float start_val;               /* starting value of dimension */
float increment;               /* distance between dimension values */
{
  long loop;

#if DEBUG
  fprintf(stderr,"candisdim: %d %s %d %g %g\n",handle,dim_name,
       size,start_val,increment);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candisdim: Candis file is not open\n");
    exit(1);
  }

/* check name size */
  if (checkstring(dim_name) > MAXNAMLEN) {
    fprintf(stderr,"candisdim: dimension name %s too long\n",dim_name);
    exit(1);
  }

/* do we have too many index fields already? */
  if (nindex >= MAXIFLDS) {
    fprintf(stderr,"candisdim: %s makes too many dimensions\n",dim_name);
    exit(1);
  }

/* add field to static slice */
  addfld(hbuff,HBMAX,'s',dim_name,SMUL,0.,'l',1,dim_name,size);

/* create the field and fill it */
  if (size < 0) {
    fprintf(stderr,"candisdim: negative dimension sizes %d not allowed\n",
         size);
    exit(1);
  }
  ifields[nindex] = getbuff(size);
  for (loop = 0; loop < size; loop++) {
    ifields[nindex][loop] = start_val + loop*increment;
  }

/* add index parameters */
  sprintf(cline,"%s0",dim_name);
  sprintf(pline,"%g",start_val);
  addpar(hbuff,HBMAX,cline,pline);
  sprintf(cline,"d%s",dim_name);
  sprintf(pline,"%g",increment);
  addpar(hbuff,HBMAX,cline,pline);

/* increment total static element counter */
  stat_elements += size;

/* save name and size information */
  isizes[nindex] = size;
  strcpy(inames[nindex],dim_name);

/* return a dimension handle */
  return(nindex++);
}

/* add a dimension with possibly irregular grid */
/* returns a dimension handle */

int candisidim(handle,dim_name,size,data)
int handle;                    /* file handle */
char *dim_name;                /* name of the dimension */
long size;                     /* number of elements in dimension direction */
float *data;                   /* dimension grid data */
{
  long loop;

#if DEBUG
  fprintf(stderr,"candisdim: %d %s %d\n",handle,dim_name,size);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candisidim: Candis file is not open\n");
    exit(1);
  }

/* check name size */
  if (checkstring(dim_name) > MAXNAMLEN) {
    fprintf(stderr,"candisidim: dimension name %s too long\n",dim_name);
    exit(1);
  }

/* do we have too many index fields already? */
  if (nindex >= MAXIFLDS) {
    fprintf(stderr,"candisidim: %s makes too many dimensions\n",dim_name);
    exit(1);
  }

/* add field to static slice */
  addfld(hbuff,HBMAX,'s',dim_name,SMUL,0.,'l',1,dim_name,size);

/* create the field and fill it */
  if (size <= 1) {
    fprintf(stderr,"candisidim: dimension sizes %d < 2 not allowed\n",size);
    exit(1);
  }
  ifields[nindex] = getbuff(size);
  for (loop = 0; loop < size; loop++) {
    ifields[nindex][loop] = data[loop];
  }

/* add index parameters */
  sprintf(cline,"%s0",dim_name);
  sprintf(pline,"%g",data[0]);
  addpar(hbuff,HBMAX,cline,pline);
  sprintf(cline,"d%s",dim_name);
  sprintf(pline,"%g",(data[size - 1] - data[0])/(size - 1));
  addpar(hbuff,HBMAX,cline,pline);

/* increment total static element counter */
  stat_elements += size;

/* save name and size information */
  isizes[nindex] = size;
  strcpy(inames[nindex],dim_name);

/* return a dimension handle */
  return(nindex++);
}

/* add a dimension with possibly irregular grid */
/* returns a dimension handle -- double size data version of candisidim */

int dcandisidim(handle,dim_name,size,data)
int handle;                    /* file handle */
char *dim_name;                /* name of the dimension */
long size;                     /* number of elements in dimension direction */
double *data;                  /* dimension grid data */
{
  long loop;

#if DEBUG
  fprintf(stderr,"candisdim: %d %s %d\n",handle,dim_name,size);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candisidim: Candis file is not open\n");
    exit(1);
  }

/* check name size */
  if (checkstring(dim_name) > MAXNAMLEN) {
    fprintf(stderr,"candisidim: dimension name %s too long\n",dim_name);
    exit(1);
  }

/* do we have too many index fields already? */
  if (nindex >= MAXIFLDS) {
    fprintf(stderr,"candisidim: %s makes too many dimensions\n",dim_name);
    exit(1);
  }

/* add field to static slice */
  addfld(hbuff,HBMAX,'s',dim_name,SMUL,0.,'l',1,dim_name,size);

/* create the field and fill it */
  if (size <= 1) {
    fprintf(stderr,"candisidim: dimension sizes %d < 2 not allowed\n",size);
    exit(1);
  }
  ifields[nindex] = getbuff(size);
  for (loop = 0; loop < size; loop++) {
    ifields[nindex][loop] = data[loop];
  }

/* add index parameters */
  sprintf(cline,"%s0",dim_name);
  sprintf(pline,"%g",data[0]);
  addpar(hbuff,HBMAX,cline,pline);
  sprintf(cline,"d%s",dim_name);
  sprintf(pline,"%g",(data[size - 1] - data[0])/(size - 1));
  addpar(hbuff,HBMAX,cline,pline);

/* increment total static element counter */
  stat_elements += size;

/* save name and size information */
  isizes[nindex] = size;
  strcpy(inames[nindex],dim_name);

/* return a dimension handle */
  return(nindex++);
}

/* candisfld -- This routine creates a Candis field and transmits the data */

int candisfld(handle,field_name,num_dims,
     dhand1,dhand2,dhand3,dhand4,data_pointer)
int handle;                     /* file handle */
char *field_name;               /* name of field */
int num_dims;                   /* number of dimensions */
int dhand1;                     /* dimension handles */
int dhand2;
int dhand3;
int dhand4;
float *data_pointer;            /* pointer to data */
{
  long loop;
  long tsize,dsize1,dsize2,dsize3,dsize4;

#if DEBUG
  fprintf(stderr,"candisfld: %d %s %d %d %d %d %d\n",handle,field_name,
       num_dims,dhand1,dhand2,dhand3,dhand4);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candisfld: Candis file is not open\n");
    exit(1);
  }

/* check names */
  if (checkstring(field_name) > MAXNAMLEN) {
    fprintf(stderr,"candisfld: field name %s too long\n",field_name);
    exit(1);
  }

/* find dimension sizes from dimension data base */
  tsize = 1;
  if (num_dims > 0) {
    if (dhand1 < 0 || dhand1 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand1);
      exit(1);
    }
    dsize1 = isizes[dhand1];
    tsize *= dsize1;
  }
  if (num_dims > 1) {
    if (dhand2 < 0 || dhand2 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand2);
      exit(1);
    }
    dsize2 = isizes[dhand2];
    tsize *= dsize2;
  }
  if (num_dims > 2) {
    if (dhand3 < 0 || dhand3 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand3);
      exit(1);
    }
    dsize3 = isizes[dhand3];
    tsize *= dsize3;
  }
  if (num_dims > 3) {
    if (dhand4 < 0 || dhand4 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand4);
      exit(1);
    }
    dsize4 = isizes[dhand4];
    tsize *= dsize4;
  }

/* add the field to the header */
  addfld(hbuff,HBMAX,'v',field_name,SMUL,0.,'l',num_dims,
       inames[dhand1],isizes[dhand1],
       inames[dhand2],isizes[dhand2],
       inames[dhand3],isizes[dhand3],
       inames[dhand4],isizes[dhand4]);

/* increment total number of variable slice elements */
  var_elements += tsize;
  vsizes[nfields] = tsize;

/* make a buffer and transfer the data to it */
  vfields[nfields] = getbuff(tsize);
  for (loop = 0; loop < tsize; loop++) {
    vfields[nfields][loop] = data_pointer[loop];
  }

/* finish */
  nfields++;
  return(handle);
}

/* dcandisfld -- Just like candisfld, except handles double arrays */

int dcandisfld(handle,field_name,num_dims,
     dhand1,dhand2,dhand3,dhand4,data_pointer)
int handle;                     /* file handle */
char *field_name;               /* name of field */
int num_dims;                   /* number of dimensions */
int dhand1;                     /* dimension handles */
int dhand2;
int dhand3;
int dhand4;
double *data_pointer;           /* pointer to data */
{
  long loop;
  long tsize,dsize1,dsize2,dsize3,dsize4;

#if DEBUG
  fprintf(stderr,"candisfld: %d %s %d %d %d %d %d\n",handle,field_name,
       num_dims,dhand1,dhand2,dhand3,dhand4);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candisfld: Candis file is not open\n");
    exit(1);
  }

/* check names */
  if (checkstring(field_name) > MAXNAMLEN) {
    fprintf(stderr,"candisfld: field name %s too long\n",field_name);
    exit(1);
  }

/* find dimension sizes from dimension data base */
  tsize = 1;
  if (num_dims > 0) {
    if (dhand1 < 0 || dhand1 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand1);
      exit(1);
    }
    dsize1 = isizes[dhand1];
    tsize *= dsize1;
  }
  if (num_dims > 1) {
    if (dhand2 < 0 || dhand2 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand2);
      exit(1);
    }
    dsize2 = isizes[dhand2];
    tsize *= dsize2;
  }
  if (num_dims > 2) {
    if (dhand3 < 0 || dhand3 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand3);
      exit(1);
    }
    dsize3 = isizes[dhand3];
    tsize *= dsize3;
  }
  if (num_dims > 3) {
    if (dhand4 < 0 || dhand4 >= nindex) {
      fprintf(stderr,"candisfld: dimension handle %d out of range\n",dhand4);
      exit(1);
    }
    dsize4 = isizes[dhand4];
    tsize *= dsize4;
  }

/* add the field to the header */
  addfld(hbuff,HBMAX,'v',field_name,SMUL,0.,'l',num_dims,
       inames[dhand1],isizes[dhand1],
       inames[dhand2],isizes[dhand2],
       inames[dhand3],isizes[dhand3],
       inames[dhand4],isizes[dhand4]);

/* increment total number of variable slice elements */
  var_elements += tsize;
  vsizes[nfields] = tsize;

/* make a buffer and transfer the data to it */
  vfields[nfields] = getbuff(tsize);
  for (loop = 0; loop < tsize; loop++) {
    vfields[nfields][loop] = data_pointer[loop];
  }

/* finish */
  nfields++;
  return(handle);
}

/* candiswrite -- This routine writes out everything to the file */

int candiswrite(handle)
int handle;                    /* file handle */
{
  int loop;
  char header[17];

#if DEBUG
  fprintf(stderr,"candiswrite: %d\n",handle);
  fflush(stderr);
#endif

/* check to see if file is open */
  if (!open_flag) {
    fprintf(stderr,"candiswrite: Candis file is not open\n");
    exit(1);
  }

/* write the header */
  puthdr(candis_file,hbuff,HBMAX);

/* non-standard writing of slices -- free buffers as we go along */

/* static slice first */
  sprintf(header,"@%15ld",stat_elements);
  if (fwrite(header,sizeof(char),16,candis_file) != 16) {
    fprintf(stderr,"candiswrite: write error\n");
    fflush(candis_file);
    exit(1);
  }
  for (loop = 0; loop < nindex; loop++) {
    if (FLOATSWAP) floatswap(isizes[loop],ifields[loop]);
    if (fwrite(ifields[loop],sizeof(float),isizes[loop],candis_file)
         != isizes[loop]) {
      fprintf(stderr,"candiswrite: write error\n");
      exit(1);
    }
    free(ifields[loop]);
  }

/* then variable slice */
  sprintf(header,"@%15ld",var_elements);
  if (fwrite(header,sizeof(char),16,candis_file) != 16) {
    fprintf(stderr,"candiswrite: write error\n");
    fflush(candis_file);
    exit(1);
  }
  for (loop = 0; loop < nfields; loop++) {
    if (FLOATSWAP) floatswap(vsizes[loop],vfields[loop]);
    if (fwrite(vfields[loop],sizeof(float),vsizes[loop],candis_file)
         != vsizes[loop]) {
      fprintf(stderr,"candiswrite: write error\n");
      exit(1);
    }
    free(vfields[loop]);
  }

/* done */
  if (not_stdout) fclose(candis_file);
  open_flag = 0;
  return(handle);
}

/* stringcheck -- This routine checks a string for reasonableness */
static int checkstring(st)
char *st;
{
  int sln,i;

/* check length */
  sln = strlen(st);
  if (sln < 0 || sln > 256) {
    fprintf(stderr,"checkstring: string is too long: %s?\n",st);
    exit(1);
  }

/* check content */
  for (i = 0; i < sln; i++ ) {
    if (st[i] < ' ' || st[i] > '~') {
      fprintf(stderr,"checkstring: string %s contains weird character %d\n",
           st,st[i]);
      exit(1);
    }
  }

/* return length */
  return(sln);
}

/* candisread -- This routine reads in an entire candis file, assumed
   to contain only one variable slice -- the stuff is left in static
   variables accessible by other routines */

#define STANDIN "-"

static int in_flag = 0;             /* this is set to 1 when a file has
				       been read */
static FILE *in_file;               /* pointer to input file */
static char inbuff[HBMAX][LINE];    /* input header buffer */
static long instat_els;             /* number of static slice elements */
static long invar_els;              /* number of variable slice elements */
static float *insslice,*invslice;   /* pointers to static and
				       variable slices */

static int incmtlines;              /* number of input comment lines */
static char incmts[HBMAX][LINE];    /* pointers to each input comment line */

static int inparams;                /* number of input parameters */
static char inpnames[HBMAX][LINE];  /* input parameter names */
static float inpvals[HBMAX];        /* input parameter values */

static int innindex;                /* number of input index fields */
static float *inifields[MAXIFLDS];  /* pointers to input index fields */
static long inisizes[MAXIFLDS];     /* dimension sizes */
static char ininames[MAXIFLDS][LINE]; /* names of dimensions */

static int innfields;               /* number of input variable fields */
static float *invfields[HBMAX];     /* pointers to input variable fields */
static long invsizes[HBMAX];        /* sizes of input variable fields */
static int inndims[HBMAX];          /* number of dimensions of each field */
static int invdims[HBMAX][MAXIFLDS];    /* dimensions of input fields */
static long invdsizes[HBMAX][MAXIFLDS];  /* dimension sizes of input fields */
static char invnames[HBMAX][LINE];  /* names of fields */

int candisread(filename)
char *filename;
{
  long i,j;
  struct field *fb;

  /* zero arrays */
  for (j = 0; j < LINE; j++) {
    for (i = 0; i < HBMAX; i++) {
      incmts[i][j] = 0;
      inpnames[i][j] = 0;
      invnames[i][j] = 0;
    }
    for (i = 0; i < MAXIFLDS; i++) {
      ininames[i][j] = 0;
    }
  }
  
  /* open input file and get header */
  if (strcmp(filename,STANDIN) == 0) {
    in_file = stdin;
  }
  else {
    if ((in_file = fopen(filename,"r")) == NULL) {
      fprintf(stderr,"candisread: file %s does not exist\n",filename);
      exit(1);
    }
  }
  gethdr(in_file,inbuff,HBMAX);
  if (getfmt(inbuff,HBMAX) != 'f') {
    fprintf(stderr,"candisread: float format expected\n");
    exit(2);
  }

  /* compute the expected number of static and variable slice elements */
  instat_els = elemcnt(inbuff,HBMAX,'s');
  invar_els = elemcnt(inbuff,HBMAX,'v');

  /* allocate slice space */
  if ((insslice = (float *)malloc(sizeof(float)*instat_els)) == NULL) {
    fprintf(stderr,"candisread: can't allocate static slice buffer\n");
    exit(3);
  }
  if ((invslice = (float *)malloc(sizeof(float)*invar_els)) == NULL) {
    fprintf(stderr,"candisread: can't allocate variable slice buffer\n");
    exit(3);
  }

  /* read static and variable slices */
  if (getslice(in_file,instat_els,insslice) == EOF) {
    fprintf(stderr,
	    "candisread: unexpected end of file reading static slice\n");
    exit(4);
  }
  if (getslice(in_file,invar_els,invslice) == EOF) {
    fprintf(stderr,
	    "candisread: unexpected end of file reading variable slice\n");
    exit(4);
  }

  /* get the comment section */
  incmtlines = getcmt(inbuff,HBMAX,incmts);

  /* squash newlines in comment section */
  for (i = 0; i < incmtlines; i++) {
    for (j = 0; j < LINE; j++) {
      if (incmts[i][j] == '\n') incmts[i][j] = '\0';
    }
  }

  /* get the parameters section */
  i = 0;
  while (getpar(inbuff,HBMAX,i,cline,pline) == OK) {
    strcpy(inpnames[i],cline);
    inpvals[i] = atof(pline);
    i++;
  }
  inparams = i;

  /* get the static field dimension information -- non-index fields
     are ignored */
  i = 0;
  innindex = 0;
  while ((fb = getfld(inbuff,HBMAX,'s',i)) != NULL) {
    if (fb->dim == 1 && strcmp(fb->fname,fb->dname1) == 0) {
      strcpy(ininames[innindex],fb->fname);
      inisizes[innindex] = fb->tsize;
      inifields[innindex] = getptr(inbuff,HBMAX,insslice,'s','d',fb->fname);
      innindex++;
    }
    i++;
  }

  /* get the variable field information */
  innfields = 0;
  while ((fb = getfld(inbuff,HBMAX,'v',innfields)) != NULL) {
    strcpy(invnames[innfields],fb->fname);
    inndims[innfields] = fb->dim;
    invfields[innfields] = getptr(inbuff,HBMAX,invslice,'v','d',fb->fname);
    invsizes[innfields] = fb->tsize;
    
    /* be sure each dimension matches an index field */
    if (fb->dim > 0) {
      j = 0;
      for (i = 0; i < innindex; i++) {
	if (strcmp(fb->dname1,ininames[i]) == 0) {
	  invdims[innfields][0] = i;
	  invdsizes[innfields][0] = fb->dsize1;
	  j++;
	  break;
	}
      }
      if (j == 0) {
	fprintf(stderr,"candisread: no index field for dimension %s in %s\n",
		fb->dname2,fb->fname);
	exit(10);
      }
    }
    j = 0;
    if (fb->dim > 1) {
      for (i = 0; i < innindex; i++) {
	if (strcmp(fb->dname2,ininames[i]) == 0) {
	  invdims[innfields][1] = i;
	  invdsizes[innfields][1] = fb->dsize2;
	  j++;
	  break;
	}
      }
      if (j == 0) {
	fprintf(stderr,"candisread: no index field for dimension %s in %s\n",
		fb->dname2,fb->fname);
	exit(10);
      }
    }
    j = 0;
    if (fb->dim > 2) {
      for (i = 0; i < innindex; i++) {
	if (strcmp(fb->dname3,ininames[i]) == 0) {
	  invdims[innfields][2] = i;
	  invdsizes[innfields][2] = fb->dsize3;
	  j++;
	  break;
	}
      }
      if (j == 0) {
	fprintf(stderr,"candisread: no index field for dimension %s in %s\n",
		fb->dname3,fb->fname);
	exit(10);
      }
    }
    j = 0;
    if (fb->dim > 3) {
      for (i = 0; i < innindex; i++) {
	if (strcmp(fb->dname4,ininames[i]) == 0) {
	  invdims[innfields][3] = i;
	  invdsizes[innfields][3] = fb->dsize4;
	  j++;
	  break;
	}
      }
      if (j == 0) {
	fprintf(stderr,"candisread: no index field for dimension %s in %s\n",
		fb->dname4,fb->fname);
	exit(10);
      }
    }
    innfields++;
  }

#if DEBUG
  /* do a debugging printout */
  fprintf(stderr,"Comments:\n");
  for (i = 0; i < incmtlines; i++) {
    fprintf(stderr,"  %s",incmts[i]);
  }
  fprintf(stderr,"\nParameters:\n");
  for (i = 0; i < inparams; i++) {
    fprintf(stderr,"  %s: %g\n",inpnames[i],inpvals[i]);
  }
  fprintf(stderr,"\nDimensions:\n");
  for (i = 0; i < innindex; i++) {
    fprintf(stderr,"  %d: %s: %d\n",i,ininames[i],inisizes[i]);
  }
  fprintf(stderr,"\nFields:\n");
  for (i = 0; i < innfields; i++) {
    fprintf(stderr,"  %s: %d: %d\n",invnames[i],inndims[i],invsizes[i]);
    for (j = 0; j < inndims[i]; j++) {
      fprintf(stderr,"    %d\n",invdims[i][j]);
    }
  }
#endif

  /* return a zero indicating success */
  if (strcmp(filename,STANDIN) != 0) fclose(in_file);
  return(0);
}

/* candiscinfo -- returns number of comment lines */
int candiscinfo(handle)
     int handle;
{
  return(incmtlines);
}

/* candispinfo -- returns number of parameters */
int candispinfo(handle)
     int handle;
{
  return(inparams);
}

/* candisdinfo -- returns number of dimensions */
int candisdinfo(handle)
     int handle;
{
  return(innindex);
}

/* candisfinfo -- returns number of variable fields */
int candisfinfo(handle)
     int handle;
{
  return(innfields);
}

/* candiscget -- returns char array containing a comment line */
char *candiscget(handle,i)
     int handle;                    /* file handle */
     int i;                         /* line number */
{
  if (i < 0 || i >= incmtlines) {
    fprintf(stderr,"candiscget: comment line %d doesn't exist\n",i);
    exit(1);
  }
  return(incmts[i]);
}

/* candispnameget -- returns null-termed char array containing param name */
char *candispnameget(handle,i)
     int handle;                    /* file handle */
     int i;                         /* parameter number */
{
  if (i < 0 || i >= inparams) {
    fprintf(stderr,"candispnameget: parameter %d doesn't exist\n",i);
    exit(1);
  }
  return(inpnames[i]);
}

/* candispvalget -- returns float containing param value */
float candispvalget(handle,i)
     int handle;                    /* file handle */
     int i;                         /* parameter number */
{
  if (i < 0 || i >= inparams) {
    fprintf(stderr,"candispvalget: parameter %d doesn't exist\n",i);
    exit(1);
  }
  return(inpvals[i]);
}

/* candisdnameget -- returns char array containing dimension name */
char *candisdnameget(handle,i)
     int handle;                    /* file handle */
     int i;                         /* dimension number */
{
  if (i < 0 || i >= innindex) {
    fprintf(stderr,"candisdnameget: dimension %d doesn't exist\n",i);
    exit(1);
  }
  return(ininames[i]);
}

/* candisdsizeget -- returns long containing dimension size */
long candisdsizeget(handle,i)
     int handle;                     /* file handle */
     int i;                          /* dimension number */
{
  if (i < 0 || i >= innindex) {
    fprintf(stderr,"candisdsizeget: dimension %d doesn't exist\n",i);
    exit(1);
  }
  return(inisizes[i]);
}

/* candisdvalsget -- returns float array containing dimension values */
float *candisdvalsget(handle,i)
     int handle;                     /* file handle */
     int i;                          /* dimension number */
{
  if (i < 0 || i >= innindex) {
    fprintf(stderr,"candisdvalsget: dimension %d doesn't exist\n",i);
    exit(1);
  }
  return(inifields[i]);
}

/* candisfnameget -- returns char array containing field name */
char *candisfnameget(handle,i)
     int handle;                     /* file handle */
     int i;                          /* field number */
{
  if (i < 0 || i >= innfields) {
    fprintf(stderr,"candisfnameget: field %d doesn't exist\n",i);
    exit(1);
  }
  return(invnames[i]);
}

/* candisfsizeget -- returns long containing size of field */
long candisfsizeget(handle,i)
     int handle;                     /* file handle */
     int i;                          /* field number */
{
  if (i < 0 || i >= innfields) {
    fprintf(stderr,"candisfsizeget: field %d doesn't exist\n",i);
    exit(1);
  }
  return(invsizes[i]);
}

/* candisfndimget -- returns number of dimensions of field */
int candisfndimget(handle,i)
     int handle;                     /* file handle */
     int i;                          /* field number */
{
  if (i < 0 || i >= innfields) {
    fprintf(stderr,"candisfndimsget: field %d doesn't exist\n",i);
    exit(1);
  }
  return(inndims[i]);
}

/* candisfdimsget -- returns list of dimension handles for field */
int *candisfdimsget(handle,i)
     int handle;                     /* file handle */
     int i;                          /* field number */
{
  if (i < 0 || i >= innfields) {
    fprintf(stderr,"candisfdimsget: field %d doesn't exist\n",i);
    exit(1);
  }
  return(invdims[i]);
}

/* candisfvalsget -- returns float array containing field data */
float *candisfvalsget(handle,i)
     int handle;                     /* file handle */
     int i;                          /* field number */
{
  if (i < 0 || i >= innfields) {
    fprintf(stderr,"candisfvalsget: field %d doesn't exist\n",i);
    exit(1);
  }
  return(invfields[i]);
}

/* The following undocumented routines pretend to be fortran and are
   for the Sisal input interface -- they are wrappers for the above
   functions returning arrays and are defined because arrays can only
   be returned to Sisal via fortran subroutines, not C */

int cdfcget_(handle,i,ad,cline)
     int *handle;
     int *i;
     int *ad;
     char *cline;
{
  char *clptr;
  long j;
  clptr = candiscget(*handle,*i);
  for (j = 0; j < ad[4]; j++) {
    cline[j] = clptr[j];
  }
}

int cdfpnameget_(handle,i,ad,pname)
     int *handle;
     int *i;
     int *ad;
     char *pname;
{
  char *pnptr;
  long j;
  pnptr = candispnameget(*handle,*i);
  for (j = 0; j < ad[4]; j++) {
    pname[j] = pnptr[j];
  }
}

int cdfdnameget_(handle,i,ad,dname)
     int *handle;
     int *i;
     int *ad;
     char *dname;
{
  char *dnptr;
  long j;
  dnptr = candisdnameget(*handle,*i);
  for (j = 0; j < ad[4]; j++) {
    dname[j] = dnptr[j];
  }
}

int cdffnameget_(handle,i,ad,fname)
     int *handle;
     int *i;
     int *ad;
     char *fname;
{
  char *fnptr;
  long j;
  fnptr = candisfnameget(*handle,*i);
  for (j = 0; j < ad[4]; j++) {
    fname[j] = fnptr[j];
  }
}

int cdfdvalsget_(handle,i,ad,dvals)
     int *handle;
     int *i;
     int *ad;
     float *dvals;
{
  float *dptr;
  long j;
  dptr = candisdvalsget(*handle,*i);
  for (j = 0; j < ad[4]; j++) {
    dvals[j] = dptr[j];
  }
}

int cdffvalsget_(handle,i,ad,fvals)
     int *handle;
     int *i;
     int *ad;
     float *fvals;
{
  float *fptr;
  long j;
  fptr = candisfvalsget(*handle,*i);
  for (j = 0; j < ad[4]; j++) {
    fvals[j] = fptr[j];
  }
}

int cdffdimsget_(handle,i,ad,fdims)
     int *handle;
     int *i;
     int *ad;
     int *fdims;
{
  int *fdptr;
  int j;
  fdptr = candisfdimsget(*handle,*i);
  for (j = 0; j < ad[4]; j++) {
    fdims[j] = fdptr[j];
  }
}
