/* subspace1 -- This function locates a one-dimensional sub-space from
 * an arbitrarily dimensioned field, the sub-space being defined by
 * holding all dimensions except the specified dimension constant.
 * There are tsize/dsize instances of this subspace.  Particular
 * instances are defined by an instance number, which takes the range
 * 0 to tsize/dsize - 1.  Each instance is located by a starting point,
 * an increment, and an element count for that sub-space.
 * Theory:
 * Starting points come in groups of elements, with starting points
 * sequential in each group.  The group size equals the product of
 * all dimension sizes to the right, while the number of groups equals
 * the product of all dimension sizes to the left.  The increment
 * for a particular instance equals the group size.  The spacing
 * between groups equals the group size times the dimension size
 * for the dimension of interest.
 * Arguments and return values:
 *     fp -- pointer to field structure for desired field
 *     dname -- string containing name of desired dimension
 *     instance -- long integer containing number of desired instance
 *     start -- points to where starting element number should be returned
 *     incr -- points to where increment should be returned
 *     size -- points to where size of sub-space should be returned
 * Subspace1 returns OK if successful, FAIL otherwise.
 */

#include <stdio.h>
#include "cdfhdr.h"

int subspace1(fp,dname,instance,start,incr,size)
struct field *fp;
char *dname;
long instance;
long *start,*incr,*size;
{
  long membership,position,ngroups,groupsize,index,dloc,dsize,ds[4];
/* see if we have the requested dimension */
  dloc = 0;
  ds[0] = fp->dsize1;
  ds[1] = fp->dsize2;
  ds[2] = fp->dsize3;
  ds[3] = fp->dsize4;
  if ((fp->dim > 0) && (strcmp(fp->dname1,dname) == 0)) {
    dloc = 1;
    dsize = fp->dsize1;
  }
  else if ((fp->dim > 1) && (strcmp(fp->dname2,dname) == 0)) {
    dloc = 2;
    dsize = fp->dsize2;
  }
  else if ((fp->dim > 2) && (strcmp(fp->dname3,dname) == 0)) {
    dloc = 3;
    dsize = fp->dsize3;
  }
  else if ((fp->dim > 3) && (strcmp(fp->dname4,dname) == 0)) {
    dloc = 4;
    dsize = fp->dsize4;
  }
/* nope -- fail */
  if (dloc == 0) return(FAIL);
/* fail if instance out of range */
  if ((instance < 0) || (instance >= fp->tsize/dsize)) return(FAIL);
/* compute number of groups and group size */
  index = dloc;
  groupsize = 1;
  while (++index <= fp->dim) groupsize *= ds[index - 1];
  index = dloc;
  ngroups = 1;
  while (--index >= 1) ngroups *= ds[index - 1];
/* compute group membership and position of instance */
  position = instance % groupsize;
  membership = instance/groupsize;
/* compute starting point of instance */
  *start = position + membership*groupsize*dsize;
/* increment equals groupsize */
  *incr = groupsize;
/* total number equals dsize */
  *size = dsize;
/* successful return */
  return(OK);
}
