// gocat.go -- A Go version of cdfcat, written for test purposes.
package main

import (
	"fmt"
	"strconv"
	"os"
	"gocandis"
)

func main() {

	// command line arguments are testvar start end maxslices
	args := os.Args
	if len(args) != 5 {
		fmt.Fprintln(os.Stderr,"Usage: gocat testvar start end maxslices")
		os.Exit(1)
	}

	// convert the command line arguments as needed
	testvar := args[1]
	start, err := strconv.ParseFloat(args[2],64)
	if err != nil {
		fmt.Fprintln(os.Stderr,"Gocat: malformed command line argument 2")
		os.Exit(1)
	}
	end, err := strconv.ParseFloat(args[3],64)
	if err != nil {
		fmt.Fprintln(os.Stderr,"Gocat: malformed command line argument 3")
		os.Exit(1)
	}
	maxslices, err := strconv.ParseInt(args[4],0,0)
	if err != nil {
		fmt.Fprintln(os.Stderr,"Gocat: malformed command line argument 4")
		os.Exit(1)
	}

	// get Candis structure from standard input
	a := gocandis.GetCandis("-")

	// get list of dimension names and sizes
	inputdnames := a.GetDimNames()

	// get the dimension data for each dimension
	inputdimdata := make([][]float64,0,10)
	inputdimsizes := make([]int,0,10)
	for _, dname := range inputdnames {
		mydata := a.GetDimData(dname)
		inputdimdata = append(inputdimdata,mydata)
		inputdimsizes = append(inputdimsizes,len(mydata))
	}

	// get variable field names, dimensions, and sizes and initialize
	// field data buffers -- also identify new index field
	newifld := -1
	inputfldnames := a.GetVFieldNames()
	inputflddims := make([][]string,0,100)
	inputfldsizes := make([]int,0,100)
	infldbufs := make([][]float64,0,100)
	for iv, fname := range inputfldnames {
		mydims := a.GetFieldDims(fname)
		inputflddims = append(inputflddims,mydims)
		mysize := 1

		// find the size of the field from the dimension data
		for _, mydim := range mydims {
			for i, dname := range inputdnames {
				if mydim == dname {
					mysize = mysize*inputdimsizes[i]
				}
			}
		}
		inputfldsizes = append(inputfldsizes,mysize)

		// make buffer for all variable field slices
		mybuffer := make([]float64,mysize*int(maxslices))
		infldbufs = append(infldbufs,mybuffer)

		// new index field by order in variable fields
		if fname == testvar {
			if mysize != 1 {
				fmt.Fprintln(os.Stderr,"Gocat: Testvar is not a scalar field")
				os.Exit(1)
			}
			newifld = iv
		}
	}

	// check to see whether testvar is a variable field name
	if newifld == -1 {
		fmt.Fprintln(os.Stderr,"Gocat: The testvar is not a variable slice field name")
		os.Exit(1)
	}

	// Loop over variable slices
	nvslices := 0
	for {

		// determine whether this slice is within specified range
		mydata := a.GetFieldData(testvar)
		inrange := mydata[0] >= start && mydata[0] <= end
		
		// read in all variable field data
		if inrange {
			for iv, fname := range inputfldnames {
				mydata := a.GetFieldData(fname)
				mylen := len(mydata)
				if mylen != inputfldsizes[iv] {
					fmt.Fprintln(os.Stderr,"Gocat: field size mismatch")
					os.Exit(1)
				}
				
				// transfer data to buffer
				for i := 0; i < mylen; i++ {
					infldbufs[iv][i + nvslices*mylen] = mydata[i]
				}
			}
			nvslices++
		}

		// get next variable slice -- break if it doesn't exist
		ssize := a.GetNextVSlice()
		if ssize == 0 {
			break
		} else {

			// check to see if maxslices is big enough
			if int(nvslices) >= int(maxslices) {
				fmt.Fprintln(os.Stderr,"Gocat: maxslices is too small")
				os.Exit(1)
			}
		}
	}

	// eliminate trivial case
	if nvslices <= 1 {
		fmt.Fprintln(os.Stderr,"Gocat: only one variable slice in input")
		os.Exit(1)
	}

	// trim unneeded space off of end of data buffers
	for iv, _ := range infldbufs {
		vfsize := inputfldsizes[iv]
		infldbufs[iv] = infldbufs[iv][0:nvslices*vfsize]
	}

	// create an output Candis structure
	b := gocandis.NullCandis()

	// transfer comments
	comments := a.GetComments()
	for _, comment := range comments {
		b.AddComment(comment)
	}

	// add comment
	b.AddComment("gocat:")
	b.AddComment("  "+args[1]+" "+args[2]+" "+args[3]+" "+args[4])

	// transfer parameters
	paramnames := a.GetParamNames()
	for _, pname := range paramnames {
		pvalue := a.GetParamValue(pname)
		b.AddParam(pname,pvalue)
	}

	// add index parameters
	x0 := infldbufs[newifld][0]
	s0 := strconv.FormatFloat(x0,'g',8,64)
	dx := infldbufs[newifld][1] - x0
	ds := strconv.FormatFloat(dx,'g',8,64)
	b.AddParam(testvar+"0",s0)
	b.AddParam("d"+testvar,ds)

	// add the new dimension
	b.AddDim(testvar,infldbufs[newifld])

	// transfer the rest of the dimensions
	dimnames := a.GetDimNames()
	for _, dname := range dimnames {
		dimdata := a.GetDimData(dname)
		b.AddDim(dname,dimdata)
	}

	// add the variable fields
	for i, fname := range inputfldnames {

		// skip the new index field
		if i != newifld {

			// construct the new dimension list, testvar first
			olddims := a.GetFieldDims(fname)
			newdims := make([]string,0,len(olddims) + 1)
			newdims = append(newdims,testvar)
			for _, dname := range olddims {
				newdims = append(newdims,dname)
			}
			b.AddVField(fname,newdims,infldbufs[i])
		}
	}

	// write the new candis file to standard error and close files
	a.CloseCandis()
	b.PutCandis("-")
	b.CloseCandis()
}
