// gomask.go -- Generate a 2D convex mask with 1.0 in the interior
// and BAD in the exterior.  Traverse counterclockwise and don't
// repeat the last point.
package main

import (
	"fmt"
	"os"
	"gocandis"
)

func main() {

	// command line arguments
	args := os.Args
	if len(args) != 10 {
		fmt.Fprintln(os.Stderr,
			"Usage: gomask maskfile xname yname x0 y0 dx dy nx ny")
		os.Exit(1)
	}
	maskfile := args[1]
	xname := args[2]
	yname := args[3]
	x0 := 0.0
	y0 := 0.0
	dx := 0.0
	dy := 0.0
	fmt.Sscanf(args[4],"%f",&x0)
	fmt.Sscanf(args[5],"%f",&y0)
	fmt.Sscanf(args[6],"%f",&dx)
	fmt.Sscanf(args[7],"%f",&dy)
	nx := 0
	ny := 0
	fmt.Sscanf(args[8],"%d",&nx)
	fmt.Sscanf(args[9],"%d",&ny)

	// make the mask array and fill with ones
	mask := make([]float64, nx*ny)
	x := make([]float64, nx)
	y := make([]float64, ny)
	for i := 0; i < nx*ny; i++ {
		mask[i] = 1.0
	}

	// initialize index fields
	for ix := 0; ix < nx; ix++ {
		x[ix] = x0 + dx*float64(ix)
	}
	for iy := 0; iy < ny; iy++ {
		y[iy] = y0 + dy*float64(iy)
	}

	// read mask file
	maxpts := 100
	xloc := make([]float64, maxpts)
	yloc := make([]float64, maxpts)
	numpts := 0
	maskhandle, _ := os.Open(maskfile)
	for {
		n, _ := fmt.Fscanf(maskhandle, "%f %f\n",
			&xloc[numpts],&yloc[numpts])
		if n != 2 {
			break
		}
		numpts++
		if numpts >= maxpts - 1 {
			break
		}
	}
	maskhandle.Close()

	// repeat the first point
	xloc[numpts] = xloc[0]
	yloc[numpts] = yloc[0]
	numpts++

	// calculate mask
	for j := 0; j < numpts - 1; j++ {

		// compute vector between successive points
		xmask := xloc[j + 1] - xloc[j]
		ymask := yloc[j + 1] - yloc[j]
		for ix := 0; ix < nx; ix++ {
			for iy := 0; iy < ny; iy++ {
				i := ix*ny + iy

				// compute location vector rel to mask pt
				xrel := x[ix] - xloc[j]
				yrel := y[iy] - yloc[j]

				// sign of cross product gives in vs out
				// assumes counterclockwise traverse
				outside := xmask*yrel - ymask*xrel < 0.0
				if outside {
					mask[i] = 1.e30
				}
			}
		}
	}

	// write Candis file with mask as sole field
	c := gocandis.NullCandis()
	c.AddComment("Mask file")
	c.AddDim(xname, x)
	c.AddDim(yname, y)
	c.AddVField("mask", []string{xname,yname}, mask)
	c.PutCandis(maskfile + ".cdf")
}
