// goorder.go -- Program to change the order of dimensions in
// a Candis file.
package main

import (
	"fmt"
	"os"
	"gocandis"
)

func main() {

	// command line arguments
	args := os.Args
	numargs := len(args)
	if numargs < 3 || numargs > 5 {
		fmt.Fprintln(os.Stderr,"Usage: goorder dim1 dim2 [dim3 [dim4]]")
		os.Exit(1)
	}

	// get the dimensions to be used and number of dimensions
	newdims := args[1:]
	ndims := len(newdims)

	// check that two dimensions are not identical
	for i := 1; i < ndims; i++ {
		for j := 0; j < i; j++ {
			if newdims[i] == newdims[j] {
				fmt.Fprintln(os.Stderr,"goorder: repeated dimensions not allowed")
				os.Exit(1)
			}
		}
	}

	// get input Candis file from standard input
	a := gocandis.GetCandis("-")

	// be sure that the number of dimensions matches
	olddims := a.GetDimNames()
	if len(olddims) != ndims {
		fmt.Fprintln(os.Stderr,"goorder: number of dims doesn't match")
		os.Exit(1)
	}

	// determine ordering
	iold := make([]int, ndims)

	// loop over new dims
	for i := 0; i < ndims; i++ {
		iold[i] = -1

		// loop over old dims to seek a match
		for j := 0; j < ndims; j++ {
			if olddims[j] == newdims[i] {
				iold[i] = j
			}
		}

		// if no match, quit
		if iold[i] == -1 {
			fmt.Fprintln(os.Stderr,"goorder: dim names don't match")
			os.Exit(1)
		}
	}

	// start creation of output Candis file
	c := gocandis.NullCandis()

	// transfer comments and add our comment
	incomments := a.GetComments()
	for _, comment := range incomments {
		c.AddComment(comment)
	}
	newcomment := ""
	for j, x := range args {
		if j == 0 {
			newcomment = "goorder:"
		} else {
			newcomment = newcomment + " " + x
		}
	}

	// add new comment to new file
	c.AddComment(newcomment)

	// get parameters and extract bad and badlim if they exist
	inparams := a.GetParamNames()
	bad := ""
	badlim := ""
	for _, x := range inparams {
		if x == "bad" {
			bad = a.GetParamValue("bad")
		}
		if x == "badlim" {
			badlim = a.GetParamValue("badlim")
		}
	}

	// add bad and badlim to output file if they exist
	if bad != "" {
		c.AddParam("bad", bad)
	}
	if badlim != "" {
		c.AddParam("badlim", badlim)
	}

	// add the new dimensions
	for i := 0; i < ndims; i++ {
		dimname := newdims[i]
		dimvals := a.GetFieldData(dimname)
		c.AddDim(dimname, dimvals)
	}

	// transfer reordered variable fields
	vnames := a.GetVFieldNames()
	for _, vname := range vnames {

		// get the dimensions
		vdims := a.GetFieldDims(vname)

		// if less than 2 dimensions, nothing to be done -- just transfer
		if len(vdims) < 2 {
			vdata := a.GetFieldData(vname)
			c.AddVField(vname, vdims, vdata)
			continue
		}

		// find the dims of the vfield and order them properly
		newvdims := make([]string,0,4)
		for _, newdim := range newdims {
			for _, vdim := range vdims {
				if vdim == newdim {
					newvdims = append(newvdims,newdim)
					continue
				}
			}
		}
		if len(newvdims) != len(vdims) {
			fmt.Fprintln(os.Stderr,"goorder: malformed dims for",
				vname)
			os.Exit(1)
		}

		// define the mapping from old to new dim order and get sizes
		nvdims := len(vdims)
		iv := make([]int, nvdims)
		is := make([]int, nvdims)
		for i := 0; i < nvdims; i++ {
			is[i] = len(a.GetDimData(vdims[i]))
			for j := 0; j < nvdims; j++ {
				if vdims[i] == newvdims[j] {
					iv[j] = i
				}
			}
		}

		// transfer the data from old to new field
		vdata := a.GetFieldData(vname)
		nvdata := make([]float64,len(vdata))
		idx := make([]int,nvdims)
		for i := 0; i < len(vdata); i++ {

			// get indices associated with each old dimension
			imod := i
			for j := nvdims - 1; j > 0; j-- {
				jmod := imod/is[j]
				idx[j] = imod - jmod*is[j]
				imod = jmod
			}
			idx[0] = imod

			// construct the new flat index
			k := idx[iv[0]]
			for j := 1; j < nvdims; j++ {
				k = k*is[iv[j]] + idx[iv[j]]
			}

			// transfer data
			nvdata[k] = vdata[i]
		}

		// write the field
		c.AddVField(vname, newvdims, nvdata)
	}

	// write the new file to standard output
	c.PutCandis("-")

	// clean up
	c.CloseCandis()
	a.CloseCandis()
}
