package main
import (
	"fmt"
	"strings"
	"strconv"
	"os"
	"os/exec"
	"io/ioutil"
)

// define GRI file pointer
var grifile *os.File

/* Escape plotstring formatting characters */
func escape(raw string) string{
	raw1 := strings.Replace(raw, "~,", "~comma", -1)
	raw2 := strings.Replace(raw1, "~/", "~slash", -1)
	raw3 := strings.Replace(raw2, "~~", "~tilde", -1)
	return raw3
}

/* Unescape plotstring formatting characters */
func unescape(pstr string) string {
	pstr1 := strings.Replace(pstr, "~comma", ",", -1)
	pstr2 := strings.Replace(pstr1, "~slash", "/", -1)
	pstr3 := strings.Replace(pstr2, "~tilde", "~", -1)
	return pstr3
}

/* Replaces color short names with the actual names used in GRI */
func colormap(c string) string{
	colors := make(map[string]string)
	colors["r"] = "red"
	colors["g"] = "green"
	colors["b"] = "blue"
	colors["c"] = "cyan"
	colors["m"] = "magenta"
	colors["y"] = "yellow"
	colors["w"] = "white"
	colors["n"] = "brown"
	colors["k"] = "black"

	name, ok := colors[c] 

	if !ok {
		name = "black"
	}

	return name
}

/* Replaces symbol short names with the actual names used in GRI */
func symbolmap(s string) string {
	symbols := make(map[string]string)

	symbols["p"] = "plus"
	symbols["m"] = "times"
	symbols["b"] = "box"
	symbols["c"] = "circ"
	symbols["d"] = "diamond"
	symbols["t"] = "triangleup"
	symbols["a"] = "asterisk"
	symbols["fb"] = "filledbox"
	symbols["fc"] = "bullet"
	symbols["fd"] = "filleddiamond"
	symbols["ft"] = "filledtriangleup"

	name, ok := symbols[s] 

	if !ok {
		fmt.Println("Error: Invalid symbol.")
		os.Exit(1)
	}

	return name

}

/* Convert a string to a float64 */
func s2f64(x string) (q float64) {
	q, err := strconv.ParseFloat(x, 64)
	if err != nil {
		fmt.Println("goplot: bad string to float64 conversion %s\n", x)
		os.Exit(1)
	}
	return
}
		

/* Check whether the increment in filled map is zero */
func checkinc(dval string) {
	if dval != "0" {
		fmt.Fprintln(grifile,"increment", dval)
	} else {
		fmt.Fprintf(grifile,"\n")
	}
}

/* creates the color map based on our short names and range values */
func cmap(in string, val1 string, val2 string, dval string) {
	if in == "1" {
		fmt.Fprintf(grifile,"set image colorscale red %v cyan %v ", val1, val2)
		checkinc(dval)
	} else if in == "2" {
		fmt.Fprintf(grifile,"set image colorscale blue %v yellow %v ", val1, val2)
		checkinc(dval)
	} else if in == "3" {
		fmt.Fprintf(grifile,"set image colorscale black %v white %v ", val1, val2)
		checkinc(dval)		
	} else {
		fmt.Println("Error: unrecognized color map")
		os.Exit(1)
	}
}


/* Labels the axes */
func labelAxis(axis string, label string) {
	if axis == "x" {
		fmt.Fprintf(grifile,"set x name \"%s\"\n", label)
	} else if axis == "y" {
		fmt.Fprintf(grifile,"set y name \"%s\"\n", label)
	}
}


func main() {

	// If no arguments, print usage and exit
	prefix := ""
	defaultcase := true
	if len(os.Args) < 2 || len(os.Args) > 3 {
		fmt.Println ("Usage: goplot 'plot_string' [prefix_name]")
		fmt.Println ("  Goplot expects a Candis file on the standard input.")
		fmt.Println ("  Plot_string is of the form command/command/.../command")
		fmt.Println ("  Escape ',' '/' '~' with '~' e. g. '~,' --> ',' etc.")
		fmt.Println ("  A plotting command is mandatory, others optional.")
		fmt.Println ("  If you would like to a subplot, use the fg command to")
		fmt.Println ("  set a location for the new plot and then put all of the")
		fmt.Println ("  commands for the new plot. Subplots must use data from")
		fmt.Println ("  the same input file. If you wish to plot multiple lines")
		fmt.Println ("  on one plot, simply use multiple plotting commands without")
		fmt.Println ("  using the fg command between them.")
		fmt.Println ("  The optional prefix name is given to the output")
		fmt.Println ("  gri and postscript files, which are deleted otherwise")
		fmt.Println ("")
		fmt.Println ("pre-plotting commands:")
		fmt.Println ("  /sx,sy,lm,bm,fg/ - define figure size")
		fmt.Println ("                    sx, sy give figure dimensions in cm,")
		fmt.Println ("                    lm, bm give left, bottom margins in cm")
		fmt.Println ("                    defaults are 10, 10, 6, 6 cm")
		fmt.Println ("  /missing_val,mv/ - set missing value (default 1e30)")
		fmt.Println ("  /size,fs/ - set font size (points, 12 default) - optional,")
		fmt.Println ("  /xmin,xmax,dx,ymin,ymax,dy,ax/ - set up the axes")
		fmt.Println ("                    set ranges of axes with min and max")
		fmt.Println ("                    dx and dy are the intervals at which labelled")
		fmt.Println ("                    will be located")		
		fmt.Println ("  /axis,format,nf/ - set axis number format")
		fmt.Println ("                    axis is x or y")
		fmt.Println ("                    format is o (off) or d (default)")
		fmt.Println ("  /xtype,ytype,ty/ - set axis types (linear or log)")	
		fmt.Println ("  /label,lx/ - label x axis. Default label will be x variable name")
		fmt.Println ("  /label,ly/ - label y axis. Default label will be y variable name")
		fmt.Println ("")
		fmt.Println ("plotting commands:")
		fmt.Println ("  /x,y,var[,cmap,val1,val2,dval[,resolution]],f/ - make a shaded filled plot")
		fmt.Println ("                    from gridded data, optional cmap =(1,2,3), val1, val2 define")
		fmt.Println ("                    range of data, dval is increment for stepped fills")
		fmt.Println ("  /x,y,var[,c,lw,dash[,val1[,val2,incr]]],c/ - make a contour plot")
		fmt.Println ("                    line color c with legal colors: (r,g,b,c,m,y,w,n,k)")
		fmt.Println ("                    line width lw select from range (0.1-5.0)")
		fmt.Println ("                    dash = 0: solid; = 1, 2, ...: dashed")
		fmt.Println ("                    val1, val2, incr: range, increment of contours")
		fmt.Println ("                    just val1 => single contour")
		fmt.Println ("                    comes after a fill plot if used")
		fmt.Println ("  /x,y,u,v,uscale,vscale,xdecim,ydecim,v/ - make a vector plot with")
		fmt.Println ("                    (u, v): vector components")
		fmt.Println ("                    (uscale,vscale): axis units per vector unit")
		fmt.Println ("                    (xdecim,ydecim): decimation factors in x, y directions")
		fmt.Println ("  /x,y,u,v,uscale,vscale,s/ - make a vectors at specified station point")
		fmt.Println ("                    (x, y): list of vector positions")
		fmt.Println ("                    (u, v): list of vector component values")
		fmt.Println ("                    (uscale,vscale): axis units per vector unit")
		fmt.Println ("  /x,y[,c[,lw[,dash,[legendLabel,xloc,yloc]]]],p/ - make a line plot of y vs x")
		fmt.Println ("                    color, linewidth, and dash as above ")		
		fmt.Println ("                    legend creates a legend at xloc,yloc with label")
		fmt.Println ("  /x,y[c,[,size[,symbol,[legendLabel,xloc,yloc]]]],m/ - make a symbol plot of y vs x")
		fmt.Println ("                    c - color as above")
		fmt.Println ("                    symbol size in cm, legal symbols (p,m,b,c,d,t,a,fb,fc,fd,ft)")
		fmt.Println ("                    legend creates a legend at xloc,yloc with label")
		fmt.Println ("  /[location,[leftval,rightval,incr,]]fp/ - make a fill palette")
		fmt.Println ("                    location is either t or r for top or right side of plot")
		fmt.Println ("                    leftval, rightval, and incr give range of palette")
		fmt.Println ("")
		fmt.Println ("beautification commands:")
		fmt.Println ("  /xlo,ylo,xhi,yhi[,c[,lw[,dash]]],dl/ - draw line from (xlo,ylo) to (xhi,yhi)")
		fmt.Println ("                    c, lw, and dash as in line plotting")
		fmt.Println ("  /label,xloc,yloc,ll/ - label at (x,y)")
		fmt.Println ("  /title,tl/ - make title")
		os.Exit(0)
	} else if len(os.Args) == 2 {
		prefix = "default" + strconv.FormatInt(int64(os.Getpid()),10)
		defaultcase = true
	} else {
		prefix = os.Args[2]
		defaultcase = false
	}

	// initialize vector case index
	vcase := 0

	// write a temporary file from candis file on standard input
	ifilename := "interfile.cdf"
	cmd1 := exec.Command("cat")
	cmd1.Stdin = os.Stdin
	ifile, _ := os.Create(ifilename)
	cmd1.Stdout = ifile
	cmd1.Run()

	// open output gri file
	grifilename := prefix + ".gri"
	grifile, _ = os.Create(grifilename)

	//================================================================
	
	rawplotstring := os.Args[1]
	
	/* Escape formating characters in plotstring */
	plotstring := escape(rawplotstring)

	// ================================================================
	/* split plot string into commands with their arguments */
	commandlist := strings.Split(plotstring, "/")
	
	/* Get a list of all of the commands in the plot sting and 
		   order the commands in the proper order for gri*/
	var cmdCheck = make([]string,len(commandlist))
	
	
	copy(cmdCheck[:], commandlist)
	
	
	for i := 0; i < len(cmdCheck); i++ {
		long := len(cmdCheck[i])
		cmdIndex := strings.LastIndex(cmdCheck[i], ",")
		cmdCheck[i] = cmdCheck[i][(cmdIndex + 1): long]
	}
	
	/* Checking to see if label and subplot commands were given*/	
	lxc := false
	lyc := false
	plot_done := 0

	/* do some initializations */
	boxxsize := 10.0
	boxysize := 10.0
	boxxmargin := 6.0
	boxymargin := 6.0
	fpleft := 6.0
	fpright := 16.0
	fpbot := 18.0
	fptop := 19.0
	pal1 := 0.0
	pal2 := 0.0
	dpal := 0.0

	/* Go through each of the commands and print the proper gri commands */
	for i := 0; i < len(commandlist); i++ {
		
		/* Skip over null commands */
		if commandlist[i] == "" {
			continue
		}
		
		/* Get the arguments for the commands */
		parts := strings.Split(commandlist[i], ",")
		
		
		nargs := len(parts)
		
		if nargs > 1 {
			/* Unescape formatting charecters */
			for j := 0; j < nargs - 1; j++ {
				parts[j] = unescape(parts[j])
			}
		}
		
		cmd := parts[nargs - 1]
		
		
		// Pre-plotting Commands
		// ================================================================
		// ================================================================
		// Open netcdf file before anything else
	 	if i == 0 {
			fmt.Fprintln(grifile,"open default.nc netcdf")
			fmt.Fprintln(grifile,"set missing value 1.0e30")
			fmt.Fprintln(grifile,"set font size 8")
		}
		
		// ================================================================
		// Set missing value
		if cmd == "mv" {
			
			if nargs != 2 {
				fmt.Printf("goplot: expect /missing_val,mv/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}
			
			fmt.Fprintln(grifile,"set missing value", parts[0])	
		}
		
		// ================================================================
		// Set font size
		if cmd == "fs" {
			if nargs != 2 {
				fmt.Printf("goplot: expect /size,fs/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}
			
	 		fmt.Fprintln(grifile,"set font size", parts[0])
		}

		// ================================================================
		// Set figure size and position
		if cmd == "fg" {
			
		 	if nargs != 5 {
				fmt.Printf("goplot: expect /sx,sy,lm,rm,fg/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			if plot_done != 0 {
				lxc = false
				lyc = false
			}


			boxxsize = s2f64(parts[0])
			boxysize = s2f64(parts[1])
			boxxmargin = s2f64(parts[2])
			boxymargin = s2f64(parts[3])

			fmt.Fprintln(grifile,"set x size", parts[0])
			fmt.Fprintln(grifile,"set y size", parts[1])
			fmt.Fprintln(grifile,"set x margin", parts[2])
			fmt.Fprintln(grifile,"set y margin", parts[3])
			fmt.Fprintln(grifile,"set x axis unknown")
			fmt.Fprintln(grifile,"set y axis unknown")

		}

		// ================================================================
		// Set the axis types
		if cmd == "ty" {

			if nargs != 3 {
				fmt.Printf("goplot: expect /xtype,ytype,ty/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			fmt.Fprintf(grifile,"set x type \"%s\"", parts[0])
			fmt.Fprintf(grifile,"set y type \"%s\"", parts[1])
		}
		// ================================================================
		// Set axes
		if cmd == "ax" {

			if nargs != 7 {
				fmt.Printf("goplot: expect /xmin,xmax,dx,ymin,ymax,dy,ax/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			fmt.Fprintln(grifile,"set x axis", parts[0], parts[1], parts[2])
			fmt.Fprintln(grifile,"set y axis", parts[3], parts[4], parts[5])
		}

		// ================================================================
		// Format axis numbers
		if cmd == "nf" {
			if nargs != 3 {
				fmt.Printf("goplot: expect /axis,format,fx/ got /%s/\n",commandlist[i])
				os.Exit(1)
			}
			afmt := ""
			if parts[1] == "d" {
				afmt = "default"
			} else if parts[1] == "o" {
				afmt = "off"
			} else {
				fmt.Printf("goplot: expect axis format d or o, got %s\n",parts[1])
				os.Exit(1)
			}
			fmt.Fprintln(grifile,"set",parts[0],"format",afmt)
		}

		// ================================================================
		// Label axes
		if cmd == "lx" {

			lxc = true
			if nargs != 2 {
				fmt.Printf("goplot: expect /label,lx/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}			

			labelAxis("x", parts[0]) 
		}

		if cmd == "ly" {

			lyc = true
			if nargs != 2 {
				fmt.Printf("goplot: expect /label,ly/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			labelAxis("y", parts[0])
		}		

		// Plotting Commands
		// ================================================================
		// ================================================================
		// Line Plot
		if cmd == "p" {

			if !(nargs == 3 || nargs == 4 || nargs == 5 || nargs == 6 || nargs == 9) {
				fmt.Printf("goplot: expect /x,y[,c[,lw[,dash,[legend,xloc,yloc]]]],p/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}
			
			// Start with black fonts
			fmt.Fprintln(grifile,"set font color black")
			
			/* Label the axes with variable names if no labels were given*/
			if !lxc {
				labelAxis("x", parts[0])
			}

			if !lyc {
				labelAxis("y", parts[1])
			}

			fmt.Fprintln(grifile,"set clip on")

			fmt.Fprintf(grifile,"read columns x = \"%s\" y = \"%s\"\n", parts[0], parts[1])

			fmt.Fprintln(grifile,"draw axes if needed")

			if nargs >= 4 {
				fmt.Fprintln(grifile,"set color", colormap(parts[2]))
			}

			if nargs >= 5 {
				fmt.Fprintln(grifile,"set line width", parts[3])
			}		

			if nargs >= 6 {
				fmt.Fprintln(grifile,"set dash", parts[4])
			}

			if nargs == 9 {
				fmt.Fprintf(grifile,"draw line legend \"%s\" at ", parts[5])
				fmt.Fprintln(grifile,parts[6], parts[7])
			}

			fmt.Fprintln(grifile,"set font color black")
			fmt.Fprintln(grifile,"draw curve")
			fmt.Fprintln(grifile,"set color black")
			fmt.Fprintln(grifile,"set line width default")
			fmt.Fprintln(grifile,"set dash 0")

			plot_done++

		}
		// ================================================================
		// Point Plot
		if cmd == "m" {

			if !(nargs == 3 || nargs == 4 || nargs == 5 || nargs == 6 || nargs == 9) {
				fmt.Printf("goplot: expect /x,y[c,[,size[,symbol,[legend,xloc,yloc]]]],m/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}
			
			// Start with black fonts
			fmt.Fprintln(grifile,"set font color black")

			/* Label the axes with variable names if no labels were given*/
			if !lxc {
				labelAxis("x", parts[0])
			}

			if !lyc {
				labelAxis("y", parts[1])
			}

			fmt.Fprintln(grifile,"set clip on")
			fmt.Fprintf(grifile,"read columns x = \"%s\" y = \"%s\"\n", parts[0], parts[1])

			fmt.Fprintln(grifile,"draw axes if needed")

			if nargs >= 5 {
				fmt.Fprintln(grifile,"set symbol size", parts[3])
			} else {
				fmt.Fprintln(grifile,"set symbol size", "0.2")
			}

			if nargs >= 4 {
				fmt.Fprintln(grifile,"set color", colormap(parts[2]))
			} else {
				fmt.Fprintln(grifile,"set color", colormap("r"))
			}

			if nargs == 9 {
				fmt.Fprintf(grifile,"draw symbol legend %s ", symbolmap(parts[4]))
				fmt.Fprintf(grifile,"\"%s\" at %s %s\n", parts[5], parts[6], parts[7])
			}

			if nargs >= 6 {
				fmt.Fprintln(grifile,"draw symbol", symbolmap(parts[4]))
			} else {
				fmt.Fprintln(grifile,"draw symbol", symbolmap("fc"))
			}


			fmt.Fprintln(grifile,"set color black")
			fmt.Fprintln(grifile,"set symbol size default")

			plot_done++
		}

		// ================================================================
		// Contour Plot
		if cmd == "c" {

			if !(nargs == 4 || nargs == 7 || nargs == 8 || nargs == 10) {
				fmt.Printf("goplot: expect /x,y,var[,c,lw,dash[,val1[,val2,incr]]],c/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			/* Label the axes with variable names if no labels were given*/
			if !lxc {
				labelAxis("x", parts[0])
			}

			if !lyc {
				labelAxis("y", parts[1])
			}


			fmt.Fprintln(grifile,"set clip on")
			fmt.Fprintf(grifile,"read grid x = \"%s\"\n", parts[0])
			fmt.Fprintf(grifile,"read grid y = \"%s\"\n", parts[1])
			fmt.Fprintf(grifile,"read grid data = \"%s\"\n", parts[2])
			
			fmt.Fprintln(grifile,"draw axes if needed")
			
			if nargs >= 7 {
				fmt.Fprintln(grifile,"set color", colormap(parts[3]))
				fmt.Fprintln(grifile,"set line width", parts[4])
				fmt.Fprintln(grifile,"set dash", parts[5])

			}



			if nargs == 4 || nargs == 7 {
				fmt.Fprintln(grifile,"draw contour")
			}
			if nargs == 8 {
				fmt.Fprintln(grifile,"draw contour", parts[6], "unlabelled")
			}
			if nargs == 10 {
				fmt.Fprintln(grifile,"draw contour", parts[6], parts[7], parts[8], "unlabelled")
			}

			if nargs >= 7 {
				fmt.Fprintln(grifile,"set color black")
				fmt.Fprintln(grifile,"set line width default")
				fmt.Fprintln(grifile,"set dash 0")
			}		

			plot_done++
		}


		// ================================================================
		// Vector Plot
		if cmd == "v" {

			if nargs != 9 {
				fmt.Printf("goplot: expect /x,y,u,v,uscale,vscale,xdecim,ydecim,v/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			/* increment vcase and convert to ascii */
			vcase = vcase + 1
			suf := fmt.Sprintf(".l%d",vcase)

			/* Label the axes with variable names if no labels were given */
			if !lxc {
				labelAxis("x", parts[0])
			}

			if !lyc {
				labelAxis("y", parts[1])
			}

			/* construct the gri call, assuming cdfgriv has been used to serialize fields */
			fmt.Fprintf(grifile,"read columns x = \"%s\" y = \"%s\" u = \"%s\" v = \"%s\"\n",
				parts[0] + suf, parts[1] + suf, parts[2] + suf, parts[3] + suf)
			fmt.Fprintln(grifile,"set u scale as x")
			fmt.Fprintln(grifile,"set v scale as y")
			fmt.Fprintln(grifile,"u *=", parts[4])
			fmt.Fprintln(grifile,"v *=", parts[5])
			fmt.Fprintln(grifile,"set arrow size as 15 percent of length")
			fmt.Fprintln(grifile,"draw arrows")

			/* run cdfgriv on the candis file */
			tmpname := ifilename + ".new"
			cmd := exec.Command("cdfgriv",parts[0],parts[6],
				parts[1],parts[7],parts[2],parts[3])
			vinfile, _ := os.Open(ifilename)
			cmd.Stdin = vinfile
			voutfile, _ := os.Create(tmpname)
			cmd.Stdout = voutfile
			cmd.Run()
			exec.Command("mv",tmpname,ifilename).Run()
			
			plot_done++
		}

		// ================================================================
		// Plot 1-D list of vectors at specified station points
		if cmd == "s" {
			if nargs != 7 {
				fmt.Printf("goplot: expect /x,y,u,v,uscale,vscale,s/ got /%s/\n", commandlist[i])
			}

			/* Label the axes with variable names if no labels were given */
			if !lxc {
				labelAxis("x", parts[0])
			}

			if !lyc {
				labelAxis("y", parts[1])
			}

			/* construct the gri call */
			fmt.Fprintf(grifile,"read columns x = \"%s\" y = \"%s\" u = \"%s\" v = \"%s\"\n",
				parts[0], parts[1], parts[2], parts[3])
			fmt.Fprintln(grifile,"set u scale as x")
			fmt.Fprintln(grifile,"set v scale as y")
			fmt.Fprintln(grifile,"u *=", parts[4])
			fmt.Fprintln(grifile,"v *=", parts[5])
			fmt.Fprintln(grifile,"set arrow size as 15 percent of length")
			fmt.Fprintln(grifile,"draw arrows")
		}

		// ================================================================
		// Filled Plot from Gridded Data
		if cmd == "f" {

			if !(nargs == 4 || nargs == 8 || nargs == 9) {
				fmt.Printf("goplot: expect /x,y,var[,cmap,val1,val2,dval[,resolution]],f/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			/* Label the axes with variable names if no labels were given*/
			if !lxc {
				labelAxis("x", parts[0])
			}

			if !lyc {
				labelAxis("y", parts[1])
			}

			fmt.Fprintln(grifile,"set clip on")
			fmt.Fprintf(grifile,"read grid x = \"%s\"\n", parts[0])
			fmt.Fprintf(grifile,"read grid y = \"%s\"\n", parts[1])
			fmt.Fprintf(grifile,"read grid data = \"%s\"\n", parts[2])

			if nargs >= 8 {
				pal1 = s2f64(parts[4])
				pal2 = s2f64(parts[5])
				dpal = s2f64(parts[6])
				fmt.Fprintln(grifile,"set image range", parts[4], parts[5])
				cmap(parts[3], parts[4], parts[5], parts[6])
				
			} else {
				fmt.Fprintf(grifile,"set image range from grid\n")
				pal1 = 0.0
				pal2 = 0.0
				dpal = 0.0
			}

			if nargs == 9 {
				fmt.Fprintln(grifile,"convert grid to image size", parts[7], parts[7])
			} else {
				fmt.Fprintln(grifile,"convert grid to image")
			}

			fmt.Fprintln(grifile,"draw image")

			/* do this last so that the image doesn't mask the axes */
			fmt.Fprintln(grifile,"draw axes if needed")
			
			plot_done++
		}

		// ================================================================
		// VFill Palette		
		if cmd == "fp" {

			if !(nargs == 1 || nargs == 2 || nargs == 5) {
				fmt.Printf("goplot: expect /[location,[left, right, incr,]]fp/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			location := ""
			palettewidth := 0.7

			if nargs == 1 {
				fpleft = boxxmargin
				fpbot = boxymargin + boxysize + 2.0
				fpright = boxxmargin + boxxsize
				fptop = fpbot + palettewidth
				location = "axisbottom"
			}

			if nargs > 1 {
				if parts[0] == "t" {
					fpleft = boxxmargin
					fpbot = boxymargin + boxysize + 2.0
					fpright = boxxmargin + boxxsize
					fptop = fpbot + palettewidth
					location = "axisbottom"
				} else if parts[0] == "r" {
					fpleft = boxxmargin + boxxsize + 1.0
					fpbot = boxymargin
					fpright = fpleft + palettewidth
					fptop = boxymargin + boxysize
					location = "axisright"
				} else {
					fmt.Printf("goplot: bad palette location\n")
					os.Exit(1)
				}
			}

			if nargs <= 2 {
				if pal1 == 0.0 && pal2 == 0.0 {
					fmt.Fprintf(grifile,"draw image palette %s box %g %g %g %g\n", location, fpleft, fpbot, fpright, fptop)
				} else {
					fmt.Fprintf(grifile,"draw image palette %s left %g right %g increment %g box %g %g %g %g\n",
						location, pal1, pal2, dpal, fpleft, fpbot, fpright, fptop)
				}
			} else {
				fmt.Fprintf(grifile,"draw image palette %s left %s right %s increment %s box %g %g %g %g\n",
					location, parts[1], parts[2], parts[3], fpleft, fpbot, fpright, fptop)
			}
			plot_done++
		}

		// Beautification Commands
		// ================================================================ 
		// ================================================================
		// Label at (x,y)
		if cmd == "ll"{

			if nargs != 4 {
				fmt.Printf("goplot: expect /label,xloc,yloc,ll/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}
			fmt.Fprintf(grifile,"draw label \"%s\" at ",parts[0])
			fmt.Fprintln(grifile,parts[1], parts[2])
		}
		// ================================================================
		// Draw line
		if cmd == "dl"{

			if nargs != 5 && nargs != 6 && nargs != 7 && nargs != 8 {
				fmt.Printf("goplot: expect /xlo,ylo,xhi,yhi[,c[,lw[,dash]]],dl/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}
			
			if nargs >= 6 {
				fmt.Fprintln(grifile,"set color", colormap(parts[4]))
			}

			if nargs >= 7 {
				fmt.Fprintln(grifile,"set line width", parts[5])
			}		

			if nargs >= 8 {
				fmt.Fprintln(grifile,"set dash", parts[6])
			}

			fmt.Fprintln(grifile,"draw line from", parts[0], parts[1], "to", parts[2], parts[3])

			fmt.Fprintln(grifile,"set color black")
			fmt.Fprintln(grifile,"set line width default")
			fmt.Fprintln(grifile,"set dash 0")
		}

		// ================================================================
		// Make title
		if cmd == "tl" {

			if nargs != 2 {
				fmt.Printf("goplot: expect /title,tl/ got /%s/\n", commandlist[i])
				os.Exit(1)
			}

			fmt.Fprintf(grifile,"draw title \"%s\"\n", parts[0])
		}

		// ================================================================

	}

	// swap the order of dimensions to be consistent with gri
	tmpname := ifilename + ".new"
	cmd := exec.Command("cdforder")
	infile, _ := os.Open(ifilename)
	cmd.Stdin = infile
	outfile, _ := os.Create(tmpname)
	cmd.Stdout = outfile
	cmd.Run()
	exec.Command("mv",tmpname,ifilename).Run()

	// write the netcdf file and remove the temporary file
	ncfile := "default.nc"
	cmd2 := exec.Command("uniput","-r",ncfile)
	infile, _ = os.Open(ifilename)
	cmd2.Stdin = infile
	cmd2.Run()
	exec.Command("rm","-f",ifilename).Run()

	// close the gri file and run gri on it -- delete the netcdf file
	grifile.Close()
	cmd = exec.Command("gri",grifilename)
	stdout, _ := cmd.StdoutPipe()
	cmd.Start()
	slurp, _ := ioutil.ReadAll(stdout)
	fmt.Printf("%s\n",slurp)
	cmd.Wait()
	
	exec.Command("rm","-f",ncfile).Run()

	// display the gri output in default case
	psfilename := prefix + ".ps"
	if defaultcase {
		exec.Command("gv","--scale=1.0",psfilename).Run()
	}

	// remove gri and postscript files in default case
	if len(os.Args) == 2 {
		exec.Command("rm","-f",grifilename,psfilename).Run()
	}
}
