// goregress.go -- Make a linear regression over multiple variables

package main

import (
	"fmt"
	"matrix"
	"gocandis"
	"strconv"
	"math"
	"os"
)

const (
	aname = "avals"
	fitsuffix = "-fit"
	bname = "bval"
	rsqname = "rsq"
	fparamname = "F"
	xcorrsname = "xcorrs"
)

func main() {

	// get command line arguments
	args := os.Args
	nargs := len(args)
	if nargs < 3 {
		fmt.Fprintln(os.Stderr, "Usage: goregress f x1 ...")
		os.Exit(1)
	}
	fname := args[1]
	xnames := make([]string, nargs - 2)
	for iv, argname := range args[2:] {
		xnames[iv] = argname
	}

	// number of predictor variables
	nvars := len(xnames)

	// get candis file from the standard input and close
	c := gocandis.GetCandis("-")
	c.CloseCandis()

	// get badlim from parameters
	params := c.GetParamNames()
	badlim := float64(9.99e29)
	for _, pname := range params {
		if pname == "badlim" {
			blstring := c.GetParamValue(pname)
			badlim, _ = strconv.ParseFloat(blstring,64)
			break
		}
	}
	
	
	// get the predicted field
	fdims := c.GetFieldDims(fname)
	f := c.GetFieldData(fname)
	if len(fdims) != 1 {
		fmt.Fprintln(os.Stderr,"Goregress: Incorrect number of dimensions for" + fname)
		os.Exit(1)
	}
	sdimname := fdims[0]

	// get the predictor fields and put them in a 2-D array
	x := make([][]float64, nvars)
	for iv, xname := range xnames {
		xdims := c.GetFieldDims(xname)
		if len(xdims) != 1 {
			fmt.Fprintln(os.Stderr,"Goregress: Field " + xname + " not one-dimensional")
			os.Exit(1)
		}
		if fdims[0] != xdims[0] {
			fmt.Fprintln(os.Stderr,"Goregress: dimension of " + xname + " inconsistent")
			os.Exit(1)
		}
		x[iv] = c.GetFieldData(xname)
	}

	// do the regression
	as, b, rsq, fparam, fits, xcorrs := regress(x, f, badlim)

	// add needed dimensions to the candis structure
	ivdata := make([]float64, nvars)
	for iv := range xnames {
		ivdata[iv] = float64(iv)
	}
	c.AddDim("iv1", ivdata)
	c.AddDim("iv2", ivdata)

	// add the fields
	c.AddVField(aname, []string{"iv1"}, as)
	c.AddVField(bname, make([]string, 0), []float64{b})
	c.AddVField(rsqname, make([]string, 0), []float64{rsq})
	c.AddVField(fparamname, make([]string, 0), []float64{fparam})
	c.AddVField(fname + fitsuffix, []string{sdimname}, fits)
	c.AddVField(xcorrsname, []string{"iv1", "iv2"},
		gocandis.Flatten2(xcorrs))

	// write the modified candis structure to the standard output
	c.PutCandis("-")
	c.CloseCandis()
}

// Do least squares fit to f = a*x + b
// x: first index is variable; second index is samples
// f: index is samples
// a: index is variable
func regress(x [][]float64, f []float64, badlim float64) (as []float64, b, rsq,
	fparam float64, fits []float64, xcorrs [][]float64) {

	// get dimensions
	nvars := len(x)
	nsamples := len(x[0])
	if nsamples != len(f) {
		fmt.Fprintln(os.Stderr,"Goregress: inconsistent sizes between x and f")
		os.Exit(1)
	}

	// compute x and f means over samples
	xmean := make([]float64, nvars)
	for iv := range xmean {
		if len(x[iv]) != nsamples {
			fmt.Fprintln(os.Stderr,"Goregress: x is ragged array")
			os.Exit(1)
		}
		xmean[iv] = 0.0
		goodsamples := 0
		for is := 0; is < nsamples; is++ {
			if math.Abs(x[iv][is]) < badlim {
				xmean[iv] += x[iv][is]
				goodsamples += 1
			}
		}
		if goodsamples > 0 {
			xmean[iv] /= float64(goodsamples)
		} else {
			fmt.Fprintln(os.Stderr,"Goregress: no good data")
			os.Exit(1)
		}
	}
	fmean := float64(0)
	goodsamples := 0
	for is := 0; is < nsamples; is++ {
		if math.Abs(f[is]) < badlim {
			fmean += f[is]
			goodsamples += 1
		}
	}
	if goodsamples > 0 {
		fmean /= float64(goodsamples)
	} else {
		fmt.Fprintln(os.Stderr,"Goregress: no good data")
		os.Exit(1)
	}

	// compute the correlations
	As := make([]float64, nvars*nvars)
	Bs := make([]float64, nvars)
	for iv := range xmean {
		for jv := range xmean {
			goodsamples := 0
			for is := 0; is < nsamples; is++ {
				if math.Abs(x[iv][is]) < badlim && math.Abs(x[jv][is]) < badlim {
					As[iv*nvars + jv] += (x[iv][is] - xmean[jv])*
						(x[jv][is] - xmean[jv])
					goodsamples += 1
				}
			}
			if goodsamples > 0 {
				As[iv*nvars + jv] = As[iv*nvars + jv]/float64(goodsamples)
			} else {
				fmt.Fprintln(os.Stderr,"Goregress: no good data")
				os.Exit(1)
			}
		}
		goodsamples := 0
		for is := 0; is < nsamples; is++ {
			if math.Abs(f[is]) < badlim && math.Abs(x[iv][is]) < badlim {
				Bs[iv] += f[is]*(x[iv][is] - xmean[iv])
				goodsamples += 1
			}
		}
		if goodsamples > 0 {
			Bs[iv] = Bs[iv]/float64(goodsamples)
		} else {
			fmt.Fprintln(os.Stderr,"Goregress: no good data")
			os.Exit(1)
		}
	}

	// solve the linear equation
	A := matrix.MakeDenseMatrix(As,nvars,nvars)
	if math.Abs(A.Det()) < 1.e-10 {
		fmt.Fprintln(os.Stderr,"Goregress: problem is ill-posed")
		os.Exit(1)
	}
	B := matrix.MakeDenseMatrix(Bs,nvars,1)
	a, err := A.Solve(B)
	if err != nil {
		fmt.Fprintln(os.Stderr,"Goregress: Couldn't solve the linear equation")
		os.Exit(1)
	}
	as = a.Array()

	// compute b
	b = fmean
	for iv := range as {
		b -= xmean[iv]*as[iv]
	}

	// compute the regression statistics
	sa := float64(0)
	sb := float64(0)
	fits = make([]float64, nsamples)
	goodsamples = 0
	for is := 0; is < nsamples; is++ {
		fji := b
		for jv := 0; jv < nvars; jv++ {
			if math.Abs(x[jv][is]) < badlim {
				fji += as[jv]*x[jv][is]
			}
		}
		fits[is] = fji
		if math.Abs(f[is]) < badlim {
			temp := fji - fmean
			sa += temp*temp
			temp = f[is] - fmean
			sb += temp*temp
			goodsamples += 1
		}
	}

	// coefficient of determination
	rsq = sa/sb

	// F parameter
	fparam = (rsq/float64(nvars))/((1.0 - rsq)/float64(goodsamples - nvars - 1))

	// compute cross correlations between dependent variables
	xmag := make([]float64, nvars)
	for iv := range xmag {
		xmag[iv] = 0
		for is := 0; is < nsamples; is++ {
			if math.Abs(x[iv][is]) < badlim {
				temp := x[iv][is] - xmean[iv]
				xmag[iv] += temp*temp
			}
		}
		xmag[iv] = math.Pow(xmag[iv], 0.5)
	}
	xcorrsflat := make([]float64, nvars*nvars)
	xcorrs = gocandis.Unflatten2(nvars, nvars, xcorrsflat)
	for iv := 0; iv < nvars; iv++ {
		for jv := 0; jv < nvars; jv++ {
			xcorrs[iv][jv] = 0.0
			for is := 0; is < nsamples; is++ {
				if math.Abs(x[iv][is]) < badlim && math.Abs(x[jv][is]) < badlim {
					xpi := x[iv][is] - xmean[iv]
					xpj := x[jv][is] - xmean[jv]
					xcorrs[iv][jv] += xpi*xpj
				}
			}
			xcorrs[iv][jv] = xcorrs[iv][jv]/(xmag[iv]*xmag[jv])
		}
	}
	
	return
}
