// gosndextrap.go -- Program to extrapolate dropsonde data upward
// intelligently.
package main

import (
	"math"
	"gocandis"
)

func main() {

	// constants
	ztrop := 14.0         // km
	tlapse := 5.5         // K/km
	mrscale := 3.0        // km
	presscale := 8.0      // km

	// variables
	zname := "z"          // km
	tempname := "temp"    // C
	presname := "pres"    // hPa
	mrname := "mr"        // g/kg
	uname := "u_wind"     // m/s
	vname := "v_wind"     // m/s

	// conservative bad data limit
	badlim := 10000.0

	// get sounding from standard input
	c := gocandis.GetCandis("-")

	// get index field data, assumed to be height in km
	z := c.GetDimData(zname)
	nlevels := len(z)

	// access sounding data
	temp := c.GetFieldData(tempname)
	pres := c.GetFieldData(presname)
	mr := c.GetFieldData(mrname)
	u_wind := c.GetFieldData(uname)
	v_wind := c.GetFieldData(vname)

	// do interpolations above launch altitude -- code depends
	// on height decreasing with increasing loop index

	// decrease temp at constant rate below tropopause.  Above
	//tropopause maintain constant value.
	lgood := 0
	for l := 0; l < nlevels; l++ {
		if temp[l] < badlim {
			lgood = l
			break
		}
	}
	for l := lgood - 1; l >= 0; l-- {
		if z[l] < ztrop {
			temp[l] = temp[l + 1] - (z[l] - z[l + 1])*tlapse
		} else {
			temp[l] = temp[l + 1]
		}
	}

	// decrease mr at a fixed scale height rate
	lgood = 0
	for l := 0; l < nlevels; l++ {
		if mr[l] < badlim {
			lgood = l
			break
		}
	}
	for l := lgood - 1; l >= 0; l-- {
		mr[l] = mr[l + 1]*math.Exp((z[l + 1] - z[l])/mrscale)
	}

	// decrease pres at a fixed scale height rate
	lgood = 0
	for l := 0; l < nlevels; l++ {
		if pres[l] < badlim {
			lgood = l
			break
		}
	}
	for l := lgood - 1; l >= 0; l-- {
		pres[l] = pres[l + 1]*math.Exp((z[l + 1] - z[l])/presscale)
	}
			
	// decrease u and v wind to zero at top
	// u_wind
	lgood = 0
	for l := 0; l < nlevels; l++ {
		if u_wind[l] < badlim {
			lgood = l
			break
		}
	}
	for l := lgood - 1; l >= 0; l-- {
		u_wind[l] = u_wind[lgood]*(float64(l)/float64(lgood))
	}

	// v_wind
		lgood = 0
	for l := 0; l < nlevels; l++ {
		if v_wind[l] < badlim {
			lgood = l
			break
		}
	}
	for l := lgood - 1; l >= 0; l-- {
		v_wind[l] = v_wind[lgood]*(float64(l)/float64(lgood))
	}

	// output modified sounding data
	c.PutCandis("-")
	c.CloseCandis()
}
