// Simple input/output package
// This is mostly a thin wrapper around I/O packages in os
// enhanced with a few convenience functions.
// Use the fmt package for formatted I/O.
package sio

import(
	"fmt"
	"os"
	"errors"
)

// Create a new file for writing
// filenames "stdout" and "stderr" don't create
// a new file pointer but just used the pre-opened
// ones of the same name
func Create(filename string) (fileptr *os.File, err error) {
	err = nil
	if filename == "stdout" {
		fileptr = os.Stdout
		return
	} else if filename == "stderr" {
		fileptr = os.Stderr
		return
	} else {
		fileptr, err = os.Create(filename)
		return
	}
}

// Append to an existing file or if it doesn't exist, create
// a new one for writing -- can't be stdout or stderr
func Append(filename string) (fileptr *os.File, err error) {

	// no pipes
	if filename == "stdout" || filename == "stderr" {
		fileptr = nil
		err = errors.New("No pipes with Append")
		return
	}

	// open the file for append
	fileptr, err = os.OpenFile(filename,
		os.O_APPEND|os.O_CREATE|os.O_WRONLY, 0666)
	return
}


// Open an existing file for reading
// filename "stdin" doesn't create a new file pointer
// but just uses the existing pre-opened pointer
func Open(filename string) (fileptr *os.File, err error) {
	err = nil
	if filename == "stdin" {
		fileptr = os.Stdin
		return
	} else {
		fileptr, err = os.Open(filename)
		return
	}
}

// Close an open file
func Close(fileptr *os.File) (err error) {
	err = fileptr.Close()
	return
}

// Reliable byte slice read
// reads from pipes might not get entire desired block
// so may have to do multiple Reads
func Get(fileptr *os.File, nbytes int) (inbuff []byte, err error) {

	// definite number of bytes to be attempted
	if nbytes > 0 {
		inbuff = make([]byte,nbytes)
		totalread := 0
		
		// loop on read -- terminates when got enough bytes or EOF
		for {
			nread := 0
			nread, err = fileptr.Read(inbuff[totalread:nbytes])
			totalread += nread
			if totalread == nbytes || nread == 0 {
				inbuff = inbuff[:totalread]
				return
			}
		}
	} else { // read the full file
		bytebuffer := make([][]byte, 0)
		buffsize := 1000
		loop := 0
		totalread := 0
		for {
			newbuff := make([]byte, buffsize)
			nread := 0
			nread, err = fileptr.Read(newbuff)
			bytebuffer = append(bytebuffer, newbuff[:nread])
			loop++
			totalread += nread

			// reduce buffer size if short read so
			// we don't needlessly allocate a big buffer
			if nread < buffsize {
				buffsize = 500
			}
			if err != nil {
				break
			}

			// use successively bigger buffers
			buffsize = buffsize*2
		}

		// transfer to a single slice
		inbuff = make([]byte, totalread)
		i := 0
		for _, buff := range bytebuffer {
			for _, value := range buff {
				inbuff[i] = value
				i++
			}
		}
		return
	}
}

// Read a text file line omitting the trailing newline
// carriage returns are silently omitted
// reads a file byte by byte until it hits a newline or an EOF
func GetLine(fileptr *os.File) (s string, err error) {
	
	// we will let inbuff grow as needed
	inbuff := make([]byte, 0)
	bytebuff := make([]byte, 0)

	// loop, reading a byte at a time
	for {

		// get a byte
		bytebuff, err = Get(fileptr, 1)
		if err != nil {
			break
		}
		
		// newline
		if bytebuff[0] == '\n' {
			break
		}
		
		// carriage returns shouldn't exist except
		// in text files from Windows as part of end-of-line
		if bytebuff[0] != '\r' {
			
			// got a valid character
			inbuff = append(inbuff,bytebuff[0])
		}
	}

	// convert to string
	s = string(inbuff)
	
	// done
	return
}

// Write a byte slice to an open file
func Put(fileptr *os.File, b []byte) (err error) {
	_, err = fileptr.Write(b)
	return
}	

// Write a text line to an open file
// no newline is needed at the end, as Fprintln provides it
// this is just a wrapper around fmt.Fprintln
func PutLine(fileptr *os.File, s string) (err error) {
	_, err = fmt.Fprintln(fileptr, s)
	return
}
