/* ngutils.c -- utility functions for ngradcart */
/* $Id: ngutils.c,v 1.3 2009/06/26 20:11:55 raymond Exp $ */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define EPS 1.e-12

/* newton ++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* find an eigenvalue using Newton's method */
double newton(double A, double B, double C, double astart, double frac)
{
  int i;
  double dela, a, numer, denom;
  a = astart;
  i = 0;
  while (1) {
    numer =  -(a*a*a + A*a*a + B*a + C);
    denom = 3.*a*a +2*A*a + B;
    dela = numer/denom;
    a = a + frac*dela;
    if (fabs(dela) < EPS) break;
    if (++i > 50) {
      fprintf(stderr,"newton: no conv, A,B,C,astart,frac: %g %g %g %g %g\n",
	      A,B,C,astart,frac);
      return(-1.);
    }
  }
  return(a);
}

/* dot +++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* compute a dot product */
double dot(double v1[3], double v2[3])
{
  return(v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]);
}

/* eigenvector +++++++++++++++++++++++++++++++++++++++++++++++ */
/* exploit fact that eigenvector is normal to rows of c_kl */
void eigenvector(double cxx, double cyy, double czz, double cxy,
		 double cyz, double cxz, double e[3])
{
  double ea[3], eb[3], ec[3], lensqa, lensqb, lensqc, len, cyx, czy, czx;

  /* make it easy on ourselves! */
  cyx = cxy; czy = cyz; czx = cxz;

  /* compute cross products for all possible row combinations */
  ea[0] = cyy*czz - czy*cyz;
  ea[1] = cyz*czx - czz*cyx;
  ea[2] = cyx*czy - czx*cyy;
  lensqa = dot(ea, ea);
  eb[0] = czy*cxz - cxy*czz;
  eb[1] = czz*cxx - cxz*czx;
  eb[2] = czx*cxy - cxx*czy;
  lensqb = dot(eb, eb);
  ec[0] = cxy*cyz - cyy*cxz;
  ec[1] = cxz*cyx - cyz*cxx;
  ec[2] = cxx*cyy - cyx*cxy;
  lensqc = dot(ec, ec);

  /* pick the one with the greatest length */
  if (lensqa >= lensqb && lensqa >= lensqc) {
    len = pow(lensqa, 0.5);
    e[0] = ea[0]/len; e[1] = ea[1]/len; e[2] = ea[2]/len;
  }
  else if (lensqb >= lensqc && lensqb >= lensqa) {
    len = pow(lensqb, 0.5);
    e[0] = eb[0]/len; e[1] = eb[1]/len; e[2] = eb[2]/len;
  }
  else if (lensqc >= lensqa && lensqc >= lensqb) {
    len = pow(lensqc, 0.5);
    e[0] = ec[0]/len; e[1] = ec[1]/len; e[2] = ec[2]/len;
  }
  else {
    fprintf(stderr,"eigenvector: shouldn't get here! %g %g %g\n",
	    lensqa, lensqb, lensqc);
    exit(1);
  }
}

/* eigencalc +++++++++++++++++++++++++++++++++++++++++++++++++ */
/* calculate eigenvalues and eigenvectors of a matrix */
int eigencalc(double axx, double axy, double ayy, double axz,
	      double ayz, double azz, double *a1, double *a2, double *a3,
	      double e1[3], double e2[3], double e3[3])
{
  double A, B, C, D;

  /* calculate the coefficients of the cubic eigenvalue equation */
  A = -(axx + ayy + azz);
  B = axx*ayy + axx*azz + ayy*azz - axz*axz - axy*axy - ayz*ayz;
  C = axx*ayz*ayz + ayy*axz*axz + azz*axy*axy - axx*ayy*azz - 2.*axy*ayz*axz;

  /* from this calculate the discriminant, checking for a small value */
  D = -4.*A*A*A*C + A*A*B*B -4.*B*B*B + 18.*A*B*C - 27.*C*C;

  /* negative discriminant shouldn't occur */
  if (D < -EPS) {
    fprintf(stderr,"ngradcart: negative discriminant %g in eigencalc\n",D);
  }

  /* check for (essentially) zero discriminant -- special case */
  if (D < EPS) {

    /* check for equation of form x(x - c)^2 = 0 with x = 0 corresponding
     * to the z direction 
     */
    if (fabs(C) < EPS && fabs(azz) < EPS) {

      /* ok, we can solve this directly -- one eigenvalue is zero,
       * the other two are degenerate -- this is most likely a pure
       * sounding point with 1 = x, 2 = y 3 = z */
      *a1 = *a2 = -A/2.;
      *a3 = 0.;
      e1[0] = e2[1] = e3[2] = 1.;
      e1[1] = e1[2] = 0.;
      e2[0] = e2[2] = 0.;
      e3[0] = e3[1] = 0.;
      return(0);
    }
    else {
      return(1);
    }
  }

  /* calculate eigenvalues -- we know enough to pick good start points */
  if ((*a1 = newton(A, B, C, 0., 1.)) < -0.5) return(3);
  if ((*a3 = newton(A, B, C, 1., 1.)) < -0.5) return(3);
  *a2 = -A - *a1 - *a3;   /* the trace is invariant! */
  if (fabs(*a1 - *a2) < EPS || fabs(*a2 - *a3) < EPS ||
      fabs(*a1 - *a3) < EPS) {
    fprintf(stderr,"newton: unexpected degeneracy!\n");
    fprintf(stderr,"  a1,a2,a3: %g %g %g\n",*a1,*a2,*a3);
    return(2);
  }

  /* calc eigenvectors */
  eigenvector(axx - *a1, ayy - *a1, azz - *a1, axy, ayz, axz, e1);
  eigenvector(axx - *a2, ayy - *a2, azz - *a2, axy, ayz, axz, e2);
  eigenvector(axx - *a3, ayy - *a3, azz - *a3, axy, ayz, axz, e3);

  /* success */
  return(0);
}

/* getdbuff +++++++++++++++++++++++++++++++++++++++++++++++++++ */
double *getdbuff(nelem)
long nelem;
{
  double *point;
  if ((point = (double*)malloc(sizeof(double)*nelem)) == NULL) {
    fprintf(stderr,"getdbuff: can't get enough memory\n");
    exit(1);
  }
  return(point);
}

/* getfbuff +++++++++++++++++++++++++++++++++++++++++++++++++++ */
float *getfbuff(nelem)
long nelem;
{
  float *point;
  if ((point = (float*)malloc(sizeof(float)*nelem)) == NULL) {
    fprintf(stderr,"getfbuff: can't get enough memory\n");
    exit(1);
  }
  return(point);
}
